/*
    File:               argonArray.h
    Description:        Array and hashtable handling classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         23.11.2003
    last updated:       05.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonArray_h
#define __argonArray_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #ifndef _MFC_VER
    // Do this include for a non-MFC project.
    #include <new.h>                // for new-operator with 2 parameters
    #endif

    #include "argonBase.h"          // for Iterator etc.
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




///<summary>
///  This is a class template to support arrays.
///   <b>T</b> can be even a class, because its constructors and destructors
///   will be called indeed. But be aware: because of the nature of a
///   dynamic array, elements (so also for objects if <b>T</b> is a class)
///   can be moved to different memory-locations. You should hold this in mind
///   when you use a class for <b>T</b>. <br/>
///  Regarding memory is to say that it only supports heap memory.
///   But if the array is very small, it can use a memory local to the class
///   to prevent allocating it on the heap and fragmenting heap too much.
///   That size is defined by <b>cmini</b>. <br/>
///   In that case, it is nevertheless possible, to use an <b>Array</b> object
///   in a moving memory handling. That means you can move the location of
///   the object to another address and the content data stays valid and accessible. <br/>
///   AND: It is possible to specify <b>cmini</b> to be zero. In that case,
///   the internal memory array will have a size of only one element, but
///   it will never be used.
///  NOTE: <br/>
///   If <b>T</b> is a class, it must have defined/overloaded the following methods: <br/>
///      - default constructor  <br/>
///      - copy-constructor     <br/>
///      - operator =()         <br/>
///      - operator ==()        <br/>
///   If <b>T</b> is a struct, you can but need not define these methods. <br/>
///   If <b>T</b> is a class or struct and it contains compound classes or structs,
///   then those will have to be handled by the <b>T</b> object
///   in its 4 methods accordingly.
///</summary>
template< class T, int cmini = 3 >
class Array
{
    friend class Array< T, cmini >;
    friend class DimSize;
    friend class DimPt;

    public :
        ///<summary> Type of the array itself </summary>
        typedef Array<T, cmini> typeArray;
        ///<summary> Type of a contained element </summary>
        typedef T               typeElement;

    protected :
        T *m_aElements;   // if this is NULL (and #m_cElements# is not zero) then #m_aElements_fixed_Mem# contains data instead.
        _size_t_ m_cElements;
        // Note: We cannot set #m_aElements# to #m_aElements_fixed_Mem#, because if #this# array
        // is in another array, and if elements are removed in the outer array, and the following
        // elements (some inner arrays) are moved in the memory, then #this#'s #m_aElements# will
        // not point onto #m_aElements_fixed_Mem# any more. It will point on wrong memory address.
        // So I decided to have the condition #m_aElements == NULL# and #m_cElements != 0# as indicator
        // to use #m_aElements_fixed_Mem# as memory content.
        // *OR* if #m_cElements# is less or equal #cmini# then #m_aElements# should be ignored,
        // because then, #m_aElements_fixed_Mem# has the data, regardless of the block size.
        _size_t_ m_cBlockSize;  // size in Ts of in how many T-item-steps to alloc on each realloc
    private :
        mutable BYTE m_aElements_fixed_Mem[ (cmini + 1) * sizeof(T) ];  // the last element (+1) will be used as dummy to be returned as reference var when an index outside the array specified
        mutable bool m_bdummy_set; // if the last element of #m_aElements_fixed_Mem# (the dummy element) holds a constructed object
        _size_t_ m_cAllocatedElements;  // number of elements available in allocated memory (including up to the last block boundary)

    public :
        Array() { Ctor(); }
        Array( const Array<T,cmini>& src ) { Ctor(); CopyFrom( src ); } // copy constructor
        Array( _size_t_            cSize ) { Ctor(); SetSize( cSize ); }
        Array( const T *pElements, _size_t_ cSize ) { Ctor(); SetData( pElements, cSize ); }
        virtual ~Array() { Empty(); DestructDummy(); }
    private :
        void Ctor() { m_aElements = NULL; m_cElements = 0; m_cBlockSize = 1; m_bdummy_set = false; m_cAllocatedElements = 0; }

    public :
        _size_t_ GetSize    () const { return m_cElements; }
        _size_t_ GetByteSize() const { return m_cElements * sizeof(T); }
        const T *Ptr( bool bNullable = true ) const;                                                        // intended for an r-value
        const T *PtrAt( _idx_ ix, bool bNullable = true ) const;                                            // intended for an r-value
        const T *PtrLast( bool bNullable = true ) const { return PtrAt( m_cElements - 1, bNullable ); }     // intended for an r-value
        const T *PtrRangeAt( _idx_ ix, _size_t_ cRange ) const;                                             // intended for an r-value
        T       *Ptr( bool bNullable = true );                                                              // intended for an l-value
        T       *PtrAt( _idx_ ix, bool bNullable = true, bool bGrow = false );                              // intended for an l-value
        T       *PtrLast( bool bNullable = true ) { return PtrAt( m_cElements - 1, bNullable, false ); }    // intended for an l-value
        T       *PtrRangeAt( _idx_ ix, _size_t_ cRange, bool bGrow = false );                               // intended for an l-value
        const T& operator[]( _idx_ ix ) const { return GetAt( ix ); } // intended for an r-value
        T& operator[]( _idx_ ix )             { return GetAt( ix ); } // intended for an l-value
        //operator T *() const { return Ptr(); }  // cannot be used or else compiler will fail
        // I do not want to implement an #IsEmpty()# method in here to
        // prevent meaning-conflicts for possibly derived classes.
        // Use #GetSize() == 0# instead to determine if the array is empty.

        virtual void SetSize( _size_t_ cSize ) { _SetSize( cSize, true, true ); } // Note: If the array is made larger, new elements will be filled with zeros.
        T *IncreaseSize( _size_t_ cIncreaseBy ) { _size_t_ c = GetSize(); SetSize( c + cIncreaseBy ); return Ptr() + c; }
        virtual void Clear();
        void Empty() { _SetSize( 0, true, false ); }
        void CopyFrom( const Array<T,cmini>& ar ) { SetData( ar.Ptr(), ar.m_cElements ); }
        Array<T,cmini>& operator =( const Array<T,cmini>& ar ) { CopyFrom( ar ); return *this; }
        void SetData( const T *pElements, _size_t_ cSize );
        bool SetAt( _idx_ ix, const T& nElement, bool bGrow = false );
        bool SetAtByVal( _idx_ ix, T nElement, bool bGrow = false ) { return SetAt( ix, nElement, bGrow ); }
        bool SetFromPtrAt( _idx_ ix, const T *pElement, bool bGrow = false );
        const T& GetAt( _idx_ ix ) const;                               // intended for an r-value
        const T& GetFirst() const { return GetAt( 0 ); }                // intended for an r-value
        const T& GetLast() const  { return GetAt( m_cElements - 1 ); }  // intended for an r-value
        T& GetAt( _idx_ ix );                                           // intended for an l-value
        T& GetFirst()             { return GetAt( 0 ); }                // intended for an l-value
        T& GetLast()              { return GetAt( m_cElements - 1 ); }  // intended for an l-value
        _idx_ Find( const T& nSearchFor ) const                 { return FindNext( nSearchFor, 0 ); }
        _idx_ FindNext( const T& nSearchFor, _idx_ iStart ) const;
        _idx_ FindByVal( T nSearchFor ) const                   { return FindNext( nSearchFor, 0 ); }
        _idx_ FindNextByVal( T nSearchFor, _idx_ iStart ) const { return FindNext( nSearchFor, iStart ); }
        const T *FindElement( const T& nSearchFor ) const                   { return PtrAt( FindNext( nSearchFor, 0      ) ); } // intended for an r-value
        const T *FindNextElement( const T& nSearchFor, _idx_ iStart ) const { return PtrAt( FindNext( nSearchFor, iStart ) ); } // intended for an r-value
        const T *FindElementByVal( T nSearchFor ) const                     { return PtrAt( FindNext( nSearchFor, 0      ) ); } // intended for an r-value
        const T *FindNextElementByVal( T nSearchFor, _idx_ iStart ) const   { return PtrAt( FindNext( nSearchFor, iStart ) ); } // intended for an r-value
        T *FindElement( const T& nSearchFor )                   { return PtrAt( FindNext( nSearchFor, 0      ) ); } // intended for an l-value
        T *FindNextElement( const T& nSearchFor, _idx_ iStart ) { return PtrAt( FindNext( nSearchFor, iStart ) ); } // intended for an l-value
        T *FindElementByVal( T nSearchFor )                     { return PtrAt( FindNext( nSearchFor, 0      ) ); } // intended for an l-value
        T *FindNextElementByVal( T nSearchFor, _idx_ iStart )   { return PtrAt( FindNext( nSearchFor, iStart ) ); } // intended for an l-value
        void Add( const T& nElement ) { Add( &nElement, 1 ); }
        void AddByVal( T nElement )   { Add( &nElement, 1 ); }
        void Add( const T *pElements, _size_t_ cSize = 1 ) { Insert( m_cElements, pElements, cSize ); }
        void Add( const Array<T,cmini>& ar ) { Add( ar.Ptr(), ar.m_cElements ); }
        T *AddNew( _size_t_ cSize = 1 ) { return InsertNew( m_cElements, cSize ); }
        Array<T,cmini>& operator +=( const T& nElement ) { Add( nElement ); return *this; }
        Array<T,cmini>& operator +=( const Array<T,cmini>& ar ) { Add( ar ); return *this; }
        bool Insert( _idx_ ix, T nElement ) { return Insert( ix, &nElement, 1 ); }
        bool Insert( _idx_ ix, const T *pElements, _size_t_ cSize = 1 );
        bool Insert( _idx_ ix, Array<T,cmini>& ar ) { return Insert( ix, ar.Ptr(), ar.m_cElements ); }
        T *InsertNew( _idx_ ix, _size_t_ cSize = 1 );
        bool Remove( _idx_ ix ) { return Remove( ix, 1 ); }
        bool Remove( _idx_ ix, _size_t_ count );
        bool IsValidIndex( _idx_ ix ) const { return ix >= 0 && (_size_t_) ix < m_cElements; }
        bool IsValidRange( _idx_ ix, _size_t_ count ) const { return ix >= 0 && ((_size_t_) ix + count <= m_cElements); }

        _size_t_ GetBlockSize() const { return m_cBlockSize; }
            // IMPORTANT: When setting a block size, it will take effect the next time the array is changed in its size.
        void     SetBlockSize( _size_t_ blocksize ) { m_cBlockSize = blocksize >= 1 ? blocksize : 1; }

        // The "Available" size and pointer is meant for the available memory area
        // which can be larger than the number of elements because of the local fixed
        // memory (cmini) or the block-allocation of memory (SetBlockSize()).
        _size_t_ GetAvailableSize    () const { return m_cElements <= (_size_t_) cmini ? (_size_t_) cmini : m_cAllocatedElements; }
        _size_t_ GetAvailableByteSize() const { return GetAvailableSize() * sizeof(T); }
        const T *GetAvailablePtr() const { return ( m_cElements <= (_size_t_) cmini )? (T*) &m_aElements_fixed_Mem : m_aElements; } // intended for an r-value
        T       *GetAvailablePtr()       { return ( m_cElements <= (_size_t_) cmini )? (T*) &m_aElements_fixed_Mem : m_aElements; } // intended for an l-value

    private :
        void _SetSize( _size_t_ cSize, bool bAllowDestructs, bool bAllowCtors );
        bool ReAllocElements( _size_t_ cSize );
        void FreeElements();
        T&   GetDummy() const;
        void DestructDummy();
};



///<param name="bNullable"> if TRUE then this method may return NULL (which
///  can happen only when cmini is 0 and the number of elements is also zero),
///  otherwise it should return a dummy object in case of empty array </param>
template< class T, int cmini >
const T *Array<T,cmini>::Ptr( bool bNullable ) const
{
    return ( m_cElements <= (_size_t_) cmini ) ?
                (m_cElements > 0 ? (T*) &m_aElements_fixed_Mem :
                                   (bNullable ? (T*) NULL : &GetDummy())) :
                m_aElements;
}

///<param name="bNullable"> if TRUE then this method may return NULL (which
///  can happen only when cmini is 0 and the number of elements is also zero),
///  otherwise it should return a dummy object in case of empty array </param>
template< class T, int cmini >
T *Array<T,cmini>::Ptr( bool bNullable )
{
    return ( m_cElements <= (_size_t_) cmini ) ?
                (m_cElements > 0 ? (T*) &m_aElements_fixed_Mem :
                                   (bNullable ? (T*) NULL : &GetDummy())) :
                m_aElements;
}

///<param name="bNullable"> if TRUE then this method may return NULL,
///  otherwise it should return a dummy object for illegal indices </param>
template< class T, int cmini >
const T *Array<T,cmini>::PtrAt( _idx_ ix, bool bNullable ) const
{
    if( ix < 0 ) return bNullable ? NULL : &GetDummy();
    if( (_size_t_) ix >= m_cElements )
        return bNullable ? NULL : &GetDummy();
    return Ptr() + ix;
}

///<param name="bNullable"> if TRUE then this method may return NULL,
///  otherwise it should return a dummy object for illegal indices </param>
template< class T, int cmini >
T *Array<T,cmini>::PtrAt( _idx_ ix, bool bNullable, bool bGrow )
{
    if( ix < 0 ) return bNullable ? NULL : &GetDummy();
    if( (_size_t_) ix >= m_cElements ) {
        if( !bGrow ) return bNullable ? NULL : &GetDummy();
        SetSize( ix + 1 );
        if( GetSize() < (_size_t_) ix + 1 ) return NULL;
    }
    return Ptr() + ix;
}

///<remarks> IMPORTANT: On error (negative index, or range less-eq zero)
///  then NULL will be returned. </remarks>
template< class T, int cmini >
const T *Array<T,cmini>::PtrRangeAt( _idx_ ix, _size_t_ cRange ) const
{
    if( ix < 0 || cRange <= 0 ) return NULL;
    if( (_size_t_) ix + cRange > m_cElements )
        return NULL;
    return Ptr() + ix;
}

///<remarks> IMPORTANT: On error (negative index, or range less-eq zero) or if not growable
///  (only happens if <b>bGrow</b> is false) then NULL will be returned. </remarks>
template< class T, int cmini >
T *Array<T,cmini>::PtrRangeAt( _idx_ ix, _size_t_ cRange, bool bGrow )
{
    if( ix < 0 || cRange <= 0 ) return NULL;
    if( (_size_t_) ix + cRange > m_cElements ) {
        if( !bGrow ) return NULL;
        SetSize( ix + cRange );
        if( GetSize() < (_size_t_) ix + cRange ) return NULL;
    }
    return Ptr() + ix;
}

///<summary> This will fill all present elements with zeros.
/// It will first destruct all the #T# objects, then fill
/// all the memory with zeros, and then again recreate the objects. </summary>
template< class T, int cmini >
void Array<T,cmini>::Clear()
{
    if( m_cElements > 0 )
    {
        T *p = Ptr() + m_cElements;
        _size_t_ c = m_cElements;
        while( c-- ) (--p)->~T();

        memset( p = Ptr(), 0, m_cElements * sizeof(T) );

        c = m_cElements;
        while( c-- ) new((void*)(p++)) T;
    }
}

template< class T, int cmini >
void Array<T,cmini>::SetData( const T *pElements, _size_t_ cSize )
{
    T *p = Ptr();
    if( p == pElements ) {
        // Note: The pointers could be equal when #this# array is
        // copied/assigned to itself ( e.g. using operator =() ).
        // In that case, do not perform the copy operation.
        _SetSize( cSize, true, true );
        return;
    }

    p += m_cElements;
    _size_t_ c = m_cElements;
    while( c-- ) (--p)->~T();

    _SetSize( cSize, false, false );

    p = Ptr(); // get ptr again since memory has been reallocated
    c = m_cElements;
    while( c-- ) new((void*)(p++)) T( *(pElements++) );
}

///<summary> Set an element at the specified index </summary>
///<param name="bGrow"> if the array should be growed if an index outside is specified.
///  If growed, the new size will be <b>ix + 1</b>. </param>
///<returns> false = not done (wrong index) </returns>
///<remarks> Note: If the array is made larger, new elements (except the last one) will
///  be filled with zeros by default (or if <b>SetSize()</b> is overloaded, the
///  filling provided in this method will be done). The last one (which *is*
///  the element at index <b>ix</b> will be set to the value <b>nElement</b> of course). </remarks>
template< class T, int cmini >
bool Array<T,cmini>::SetAt( _idx_ ix, const T& nElement, bool bGrow )
{
    if( ix < 0 ) return false;
    if( (_size_t_) ix >= m_cElements ) {
        if( !bGrow ) return false;
        _size_t_ cSize_old = m_cElements;
        _SetSize( ix + 1, true, false );
        T *p = Ptr() + cSize_old;
        if( ix > (_idx_) m_cElements ) ix = m_cElements;
        for( _size_t_ i = cSize_old ; i < (_size_t_) ix ; ++i )
            new((void*)(p++)) T;
        if( ix >= (_idx_) m_cElements ) return false;
        new((void*)p) T( nElement );
    } else
        Ptr()[ix] = nElement;
    return true;
}

///<summary> Set an element at the specified index </summary>
///<param name="bGrow"> if the array should be growed if an index outside is specified.
///  If growed, the new size will be <b>ix + 1</b>. </param>
///<returns> false = not done (wrong index) </returns>
///<remarks> Note: If the array is made larger, new elements (except the last one) will
///  be filled with zeros by default (or if <b>SetSize()</b> is overloaded, the
///  filling provided in this method will be done). The last one (which *is*
///  the element at index <b>ix</b> will be set to the value <b>nElement</b> of course). </remarks>
template< class T, int cmini >
bool Array<T,cmini>::SetFromPtrAt( _idx_ ix, const T *pElement, bool bGrow )
{
    if( ix < 0 || !pElement ) return false;
    if( (_size_t_) ix >= m_cElements ) {
        if( !bGrow ) return false;
        _size_t_ cSize_old = m_cElements;
        _SetSize( ix + 1, true, false );
        T *p = Ptr() + cSize_old;
        if( ix > (_idx_) m_cElements ) ix = m_cElements;
        for( _size_t_ i = cSize_old ; i < (_size_t_) ix ; ++i )
            new((void*)(p++)) T;
        if( ix >= m_cElements ) return false;
        new((void*)p) T( *pElement );
    } else
        Ptr()[ix] = *pElement;
    return true;
}

template< class T, int cmini >
const T& Array<T,cmini>::GetAt( _idx_ ix ) const
{
    return ( ix < 0 || (_size_t_) ix >= m_cElements )? GetDummy() : Ptr()[ix];
}

template< class T, int cmini >
T& Array<T,cmini>::GetAt( _idx_ ix )
{
    return ( ix < 0 || (_size_t_) ix >= m_cElements )? GetDummy() : Ptr()[ix];
}

///<summary> Returns the index of the first item found (beginning
///  at index <b>iStart</b>) that is equal to <b>nSearchFor</b>,
///  or -1 if not found. </summary>
template< class T, int cmini >
_idx_ Array<T,cmini>::FindNext( const T& nSearchFor, _idx_ iStart ) const
{
    const T *p0 = Ptr();
    for( _size_t_ i = ( iStart >= 0 )? (_size_t_) iStart : (_size_t_) 0 ; i < m_cElements ; ++i )
        if( p0[i] == nSearchFor ) return (_idx_) i;
    return -1;
}

///<summary> Insert elements at the specified index. </summary>
///<param name="ix"> if index points directly after the end of the array -> added to the end. <br/>
///  if index too large -> nothing done. <br/>
///  if index negative -> nothing done. </param>
///<returns> false = not done (wrong index) </returns>
template< class T, int cmini >
bool Array<T,cmini>::Insert( _idx_ ix, const T *pElements, _size_t_ cSize )
{
    if( ix < 0 || (_size_t_) ix > m_cElements ) return false;
    if( cSize == 0 || !pElements ) return true;
    _size_t_ cEl_old = m_cElements;
    _size_t_ ccopy = m_cElements - (_size_t_) ix;
    _SetSize( m_cElements + cSize, true, false );
    if( m_cElements <= cEl_old ) return false;
    T *p0 = Ptr();
    if( ccopy > 0 ) memmove( &p0[ix + cSize], &p0[ix], ccopy * sizeof(T) );
    memcpy( &p0[ix], pElements, cSize * sizeof(T) );
    T *p = &p0[ix];
    while( cSize-- ) new((void*)(p++)) T( *(pElements++) );
    return true;
}

///<summary> Insert elements at the specified index. </summary>
///<param name="ix"> if index points directly after the end of the array -> added to the end. <br/>
///  if index too large -> nothing done. <br/>
///  if index negative -> nothing done. </param>
///<returns> returns a pointer to the start of the new inserted memory area. <br/>
///  NULL = not done (wrong index). </returns>
///<remarks> NOTE: If <b>cSize</b> was zero, and <b>ix</b> a valid index, then the pointer
///  will be returned correctly (as non-NULL), but you should *not*
///  copy and memory in it because no new memory has been inserted. </remarks>
template< class T, int cmini >
T *Array<T,cmini>::InsertNew( _idx_ ix, _size_t_ cSize )
{
    if( ix < 0 || (_size_t_) ix > m_cElements ) return NULL;
    if( cSize == 0 ) return &Ptr()[ix]; // just to indicate that no error occured
    _size_t_ cEl_old = m_cElements;
    _size_t_ ccopy = m_cElements - (_size_t_) ix;
    _SetSize( m_cElements + cSize, true, false );
    if( m_cElements <= cEl_old ) return NULL;
    T *p0 = Ptr();
    if( ccopy > 0 ) memmove( &p0[ix + cSize], &p0[ix], ccopy * sizeof(T) );
    memset( &p0[ix], 0, cSize * sizeof(T) );
    T *p = &p0[ix];
    while( cSize-- ) new((void*)(p++)) T;
    return &p0[ix];
}

///<summary> Remove elements </summary>
///<returns> false = not done (wrong index) </returns>
///<remarks> Note: If <b>count</b> is so large, that it (together with <b>ix</b>) goes
///  over the end of the array -> only those parts of the array which are
///  occupied by the area definition [ix,ix+count[ will be removed
///  without error.
///  But if <b>ix</b> negative -> nothing done. </remarks>
template< class T, int cmini >
bool Array<T,cmini>::Remove( _idx_ ix, _size_t_ count )
{
    if( ix < 0 || (_size_t_) ix >= m_cElements ) return false;
    if( count == 0 ) return true;
    if( (_size_t_) ix + count > m_cElements ) count = m_cElements - (_size_t_) ix;
    T *p0 = Ptr();
    T *p; p = &p0[ix + count];  // prevent warning variable initialized but not referenced
    _size_t_ c; c = count;      // prevent warning variable initialized but not referenced
    while( c-- ) (--p)->~T();
    _size_t_ ccopy = m_cElements - (ix + count);
    if( ccopy > 0 ) memmove( &p0[ix], &p0[ix + count], ccopy * sizeof(T) );
    memset( &p0[m_cElements - count], 0, count * sizeof(T) );   // if the following resizing fails, we have at least empty objects
    _SetSize( m_cElements - count, false, true );
    return true;
}



template< class T, int cmini >
void Array<T,cmini>::_SetSize( _size_t_ cSize, bool bAllowDestructs, bool bAllowCtors )
{
    // First destruct elements if needed,
    // but only those which will be removed.
    _size_t_ cSize_old = m_cElements;
    if( bAllowDestructs && cSize < cSize_old ) {
        T *p; p = Ptr() + cSize_old;  // prevent warning variable initialized but not referenced
        for( _size_t_ i = cSize ; i < cSize_old ; ++i )
            (--p)->~T();
    }

    bool ret = ReAllocElements( cSize );

    if( !ret )
        if( bAllowDestructs && cSize < cSize_old && m_cElements == cSize_old ) {
            T *p; p = Ptr() + cSize_old;  // prevent warning variable initialized but not referenced
            memset( p, 0, (cSize_old - cSize) * sizeof(T) );   // if the resizing above failed, we have at least empty objects
        }

    // We could throw an error if reallocation did not work.

    // Now construct new elements if needed,
    // but only the newly added ones.
    if( bAllowCtors && cSize > cSize_old ) {
        T *p; p = Ptr() + cSize_old;  // prevent warning variable initialized but not referenced
        for( _size_t_ i = cSize_old ; i < cSize ; ++i )
            new((void*)(p++)) T;
    }
}

///<summary> Resizes the <b>m_aElements</b> array (using <b>m_aElements_fixed_Mem</b> for a small array to
///  prevent too many small heap memory areas, which e.g. can occur when
///  many <b>Array</b> objects exist). </summary>
///<returns> False if reallocation failed. In that case, the original memory is unchanged. </returns>
///<remarks> Note: When resizing, the old contents will remain and if the array
///  is getting larger, the newly added memory will be cleared. <br/>
///  Note: This method does not call constructors or destructors of <b>T</b>. </remarks>
template< class T, int cmini >
bool Array<T,cmini>::ReAllocElements( _size_t_ cSize )
{
    if( cSize == m_cElements ) return true;

    if( cSize == 0 ) {
        FreeElements();
        return true;
    }

    bool onheap_new = ( cSize       > (_size_t_) cmini );   // true if for the new size we need mem on the heap instead of using #m_aElements_fixed_Mem#
    bool onheap_old = ( m_cElements > (_size_t_) cmini );   // true if for the old size we need mem on the heap instead of using #m_aElements_fixed_Mem#

    _size_t_ cElements_old = m_cElements;

    if( onheap_new )
    {
        _size_t_ cAllocatedElements_new = ((cSize + m_cBlockSize - 1) / m_cBlockSize) * m_cBlockSize;

        if( !onheap_old ) m_aElements = NULL; // for #realloc()# so that it will not try to free the memory.
        else if( cAllocatedElements_new == m_cAllocatedElements ) {
            m_cElements = cSize;
            if( m_cElements > cElements_old )
                memset( &m_aElements[cElements_old], 0, (m_cElements - cElements_old) * sizeof(T) );
            return true;
        }
        _size_t_ allocsize = cAllocatedElements_new * sizeof(T);
        void *p = realloc( (void*) m_aElements, (_size_t_) allocsize );
        if( !p && allocsize ) return false;
        m_aElements = (T*) p;
        m_cElements = cSize;
        m_cAllocatedElements = cAllocatedElements_new;
        if( !onheap_old )
            memcpy( m_aElements, m_aElements_fixed_Mem, cElements_old * sizeof(T) );

        if( m_cElements > cElements_old )
            memset( &m_aElements[cElements_old], 0, (m_cElements - cElements_old) * sizeof(T) );
    }
    else
    {
        if( onheap_old ) {
            memcpy( m_aElements_fixed_Mem, m_aElements, cSize * sizeof(T) );
            free( m_aElements );
        }
        m_aElements = NULL;
        m_cElements = cSize;
        m_cAllocatedElements = 0;

        if( m_cElements > cElements_old )
            memset( &m_aElements_fixed_Mem[cElements_old * sizeof(T)], 0, (m_cElements - cElements_old) * sizeof(T) );
    }

    return true;
}

///<remarks> Note: This method does not call constructors or desctructor of <b>T</b>. </remarks>
template< class T, int cmini >
void Array<T,cmini>::FreeElements()
{
    if( m_aElements && m_cElements > (_size_t_) cmini )
        free( m_aElements );
    m_cElements = 0;
    m_aElements = NULL;
}

///</summary> Clears and constructs the dummy object and returns it.
///  If already constructed, it will first be destructed
///  so that its memory can be cleared before construction. </summary>
template< class T, int cmini >
T& Array<T,cmini>::GetDummy() const
{
    Array<T,cmini> *pArray = (Array<T,cmini>*) this; // to make it non-const
    T *pDummy = (T*) &pArray->m_aElements_fixed_Mem[cmini * sizeof(T)];
    if( pArray->m_bdummy_set ) pDummy->~T();
    memset( pDummy, 0, sizeof(T) );
    new((void*) pDummy) T;
    pArray->m_bdummy_set = true;
    return *pDummy;
}

template< class T, int cmini >
void Array<T,cmini>::DestructDummy()
{
    if( m_bdummy_set ) {
        ((T*) &m_aElements_fixed_Mem[cmini * sizeof(T)])->~T();
        m_bdummy_set = false;
    }
}




///<summary>
///  This is a class template to support arrays of pointers to objects.
///   The objects will be deleted automatically when no longer needed.
///   So type <b>T</b> must be a type that can be deleted using <b>delete</b> operator. <br/>
///  Regarding array memory is to say that it only supports heap memory.
///   But if the array is very small, it can use a memory local to the class
///   to prevent allocating it on the heap and fragmenting heap too much.
///   That size is defined by <b>cmini</b>. <br/>
///   In that case, it is nevertheless possible, to use a <b>Ptrarray</b> object
///   in a moving memory handling. That means you can move the location of
///   the object to another address and the content data stays valid and accessible. <br/>
///   AND: It is possible to specify <b>cmini</b> to be zero. In that case,
///   the internal memory array will have a size of only one element, but
///   it will never be used. <br/>
///  Warning: <br/>
///   Since the Ptrarray owns each object (because it will delete it
///   on its own), a copy-constructor and = oerator are not supported
///   for Ptrarrays.
///</summary>
template< class T, int cmini = 3 >
class Ptrarray
{
    friend class Ptrarray< T, cmini >;
    friend class DimSize;
    friend class DimPt;

    public :
        ///<summary> Type of the array itself </summary>
        typedef Ptrarray<T, cmini> typeArray;
        ///<summary> Type of a contained element in the array (so a pointer type) </summary>
        typedef T                 *typePointerElement;
        typedef T                 *typeElement;         // this is same name like in #Array# for an element (which is a pointer in Ptrarray)
        typedef T                 *Tp;                  // used mainly to avoid something like "const T **" which will be only correct in the form "const Tp *" in here.
        ///<summary> Type of a dereferenced contained element (so a object type) </summary>
        typedef T                  typePointedElement;

    protected :
        T **m_aElements;   // if this is NULL (and #m_cElements# is not zero) then #m_aElements_fixed_Mem# contains data instead.
        _size_t_ m_cElements;
        // Note: We cannot set #m_aElements# to #m_aElements_fixed_Mem#, because if #this# array
        // is in another array, and if elements are removed in the outer array, and the following
        // elements (some inner arrays) are moved in the memory, then #this#'s #m_aElements# will
        // not point onto #m_aElements_fixed_Mem# any more. It will point on wrong memory address.
        // So I decided to have the condition #m_aElements == NULL# and #m_cElements != 0# as indicator
        // to use #m_aElements_fixed_Mem# as memory content.
        // *OR* if #m_cElements# is less or equal #cmini# then #m_aElements# should be ignored,
        // because then, #m_aElements_fixed_Mem# has the data, regardless of the block size.
        _size_t_ m_cBlockSize;  // size in Ts of in how many T-item-steps to alloc on each realloc
    private :
        mutable BYTE m_aElements_fixed_Mem[ (cmini + 1) * sizeof(T*) ];  // the last element (+1) will be used as dummy to be returned as reference var when an index outside the array specified
        _size_t_ m_cAllocatedElements;  // number of elements available in allocated memory (including up to the last block boundary)

    public :
        Ptrarray() { Ctor(); }
        Ptrarray( _size_t_               cSize ) { Ctor(); SetSize( cSize ); }
        Ptrarray( const Tp *ppObjects, _size_t_ cSize ) { Ctor(); SetData( ppObjects, cSize ); }
        ~Ptrarray() { Empty(); }
    protected :
        Ptrarray( const Ptrarray<T,cmini>& src );
        Ptrarray& operator =( const Ptrarray<T,cmini>& src );
    private :
        void Ctor() { m_aElements = NULL; m_cElements = 0; m_cBlockSize = 1; m_cAllocatedElements = 0; }

    public :
        _size_t_  GetSize    () const { return m_cElements; }
        _size_t_  GetByteSize() const { return m_cElements * sizeof(T*); }
        const Tp *Ptr( bool bNullable = true ) const;                                                        // intended for an r-value
        const Tp *PtrAt( _idx_ ix, bool bNullable = true ) const;                                            // intended for an r-value
        const Tp *PtrLast( bool bNullable = true ) const { return PtrAt( m_cElements - 1, bNullable ); }     // intended for an r-value
        const Tp *PtrRangeAt( _idx_ ix, _size_t_ cRange ) const;                                             // intended for an r-value
        Tp       *Ptr( bool bNullable = true );                                                              // intended for an l-value
        Tp       *PtrAt( _idx_ ix, bool bNullable = true, bool bGrow = false );                              // intended for an l-value
        Tp       *PtrLast( bool bNullable = true ) { return PtrAt( m_cElements - 1, bNullable, false ); }    // intended for an l-value
        Tp       *PtrRangeAt( _idx_ ix, _size_t_ cRange, bool bGrow = false );                               // intended for an l-value
        const T*& operator[]( _idx_ ix ) const { return GetAt( ix ); } // intended for an r-value
        T*& operator[]( _idx_ ix )             { return GetAt( ix ); } // intended for an l-value
        // I do not want to implement an #IsEmpty()# method in here to
        // prevent meaning-conflicts for possibly derived classes.
        // Use #GetSize() == 0# instead to determine if the array is empty.

        void SetSize( _size_t_ cSize ) { _SetSize( cSize, true ); } // Note: If the array is made larger, new elements will be filled with NULL pointers.
        T **IncreaseSize( _size_t_ cIncreaseBy ) { _size_t_ c = GetSize(); SetSize( c + cIncreaseBy ); return Ptr() + c; }
        void Clear() { Clear( 0, m_cElements ); }
        void Clear( _idx_ ix, _size_t_ count = 1 );
        T   *DetachObject ( _idx_ ix ) { T *p = GetAt( ix ); return DetachObjects( ix, 1 ) ? p : NULL; }
        bool DetachObjects( _idx_ ix, _size_t_ count );
        void Empty() { _SetSize( 0, true ); }
        void SetData( const Tp *ppObjects, _size_t_ cSize );
        bool SetAt( _idx_ ix, T* pObject, bool bGrow = false );
        const T*& GetAt( _idx_ ix ) const;                               // intended for an r-value
        const T*& GetFirst() const { return GetAt( 0 ); }                // intended for an r-value
        const T*& GetLast() const  { return GetAt( m_cElements - 1 ); }  // intended for an r-value
        T*& GetAt( _idx_ ix );                                           // intended for an l-value
        T*& GetFirst()             { return GetAt( 0 ); }                // intended for an l-value
        T*& GetLast()              { return GetAt( m_cElements - 1 ); }  // intended for an l-value
        _idx_ Find( T *pSearchFor ) const { return FindNext( pSearchFor, 0 ); }
        _idx_ FindNext( T *pSearchFor, _idx_ iStart ) const;
        const Tp *FindElement( T *pSearchFor ) const                   { return PtrAt( FindNext( pSearchFor, 0      ) ); } // intended for an r-value
        const Tp *FindNextElement( T *pSearchFor, _idx_ iStart ) const { return PtrAt( FindNext( pSearchFor, iStart ) ); } // intended for an r-value
        Tp *FindElement( T *pSearchFor )                   { return PtrAt( FindNext( pSearchFor, 0      ) ); } // intended for an l-value
        Tp *FindNextElement( T *pSearchFor, _idx_ iStart ) { return PtrAt( FindNext( pSearchFor, iStart ) ); } // intended for an l-value
        void Add( T *pObject ) { Add( &pObject, 1 ); }
        void Add( const Tp *ppObjects, _size_t_ cSize = 1 ) { Insert( m_cElements, ppObjects, cSize ); }
        T **AddNew( _size_t_ cSize = 1 ) { return InsertNew( m_cElements, cSize ); }
        Ptrarray<T,cmini>& operator +=( T *pObject ) { Add( pObject ); return *this; }
        bool Insert( _idx_ ix, T *pObject ) { return Insert( ix, &pObject, 1 ); }
        bool Insert( _idx_ ix, const Tp *ppObjects, _size_t_ cSize = 1 );
        T **InsertNew( _idx_ ix, _size_t_ cSize = 1 );
        bool Remove( _idx_ ix ) { return Remove( ix, 1 ); }
        bool Remove( _idx_ ix, _size_t_ count );
        void RemoveNullPointers();
        bool IsValidIndex( _idx_ ix ) const { return ix >= 0 && (_size_t_) ix < m_cElements; }
        bool IsValidRange( _idx_ ix, _size_t_ count ) const { return ix >= 0 && ((_size_t_) ix + count <= m_cElements); }

        _size_t_ GetBlockSize() const { return m_cBlockSize; }
            // IMPORTANT: When setting a block size, it will take effect the next time the array is changed in its size.
        void     SetBlockSize( _size_t_ blocksize ) { m_cBlockSize = blocksize >= 1 ? blocksize : 1; }

        // The "Available" size and pointer is meant for the available memory area
        // which can be larger than the number of elements because of the local fixed
        // memory (cmini) or the block-allocation of memory (SetBlockSize()).
        _size_t_ GetAvailableSize    () const { return m_cElements <= (_size_t_) cmini ? (_size_t_) cmini : m_cAllocatedElements; }
        _size_t_ GetAvailableByteSize() const { return GetAvailableSize() * sizeof(T*); }
        const Tp *GetAvailablePtr() const { return ( m_cElements <= (_size_t_) cmini )? (T**) &m_aElements_fixed_Mem : m_aElements; } // intended for an r-value
        Tp       *GetAvailablePtr()       { return ( m_cElements <= (_size_t_) cmini )? (T**) &m_aElements_fixed_Mem : m_aElements; } // intended for an l-value

    private :
        void _SetSize( _size_t_ cSize, bool bAllowDeletes );
        bool ReAllocElements( _size_t_ cSize );
        void FreeElements();
        T*&  GetDummy() const;
};



///<param name="bNullable"> if TRUE then this method may return NULL (which
///  can happen only when cmini is 0 and the number of elements is also zero),
///  otherwise it should return a dummy pointer (which contains NULL) in case of empty array </param>
template< class T, int cmini >
const typename Ptrarray<T,cmini>::Tp *Ptrarray<T,cmini>::Ptr( bool bNullable ) const
{
    return ( m_cElements <= (_size_t_) cmini ) ?
                (m_cElements > 0 ? (T**) &m_aElements_fixed_Mem :
                                   (bNullable ? (T**) NULL : &GetDummy())) :
                m_aElements;
}

///<param name="bNullable"> if TRUE then this method may return NULL (which
///  can happen only when cmini is 0 and the number of elements is also zero),
///  otherwise it should return a dummy pointer (which contains NULL) in case of empty array </param>
template< class T, int cmini >
typename Ptrarray<T,cmini>::Tp *Ptrarray<T,cmini>::Ptr( bool bNullable )
{
    return ( m_cElements <= (_size_t_) cmini ) ?
                (m_cElements > 0 ? (T**) &m_aElements_fixed_Mem :
                                   (bNullable ? (T**) NULL : &GetDummy())) :
                m_aElements;
}

///<param name="bNullable"> if TRUE then this method may return NULL,
///  otherwise it should return a dummy pointer (which contains NULL) for illegal indices </param>
template< class T, int cmini >
const typename Ptrarray<T,cmini>::Tp *Ptrarray<T,cmini>::PtrAt( _idx_ ix, bool bNullable ) const
{
    if( ix < 0 ) return bNullable ? NULL : &GetDummy();
    if( (_size_t_) ix >= m_cElements )
        return bNullable ? NULL : &GetDummy();
    return Ptr() + ix;
}

///<param name="bNullable"> if TRUE then this method may return NULL,
///  otherwise it should return a dummy pointer (which contains NULL) for illegal indices </param>
template< class T, int cmini >
typename Ptrarray<T,cmini>::Tp *Ptrarray<T,cmini>::PtrAt( _idx_ ix, bool bNullable, bool bGrow )
{
    if( ix < 0 ) return bNullable ? NULL : &GetDummy();
    if( (_size_t_) ix >= m_cElements ) {
        if( !bGrow ) return bNullable ? NULL : &GetDummy();
        SetSize( ix + 1 );
        if( GetSize() < (_size_t_) ix + 1 ) return NULL;
    }
    return Ptr() + ix;
}

///<remarks> IMPORTANT: On error (negative index, or range less-eq zero)
///  then NULL will be returned. </remarks>
template< class T, int cmini >
const typename Ptrarray<T,cmini>::Tp *Ptrarray<T,cmini>::PtrRangeAt( _idx_ ix, _size_t_ cRange ) const
{
    if( ix < 0 || cRange <= 0 ) return NULL;
    if( (_size_t_) ix + cRange > m_cElements )
        return NULL;
    return Ptr() + ix;
}

///<remarks> IMPORTANT: On error (negative index, or range less-eq zero) or if not growable
///  (only happens if <b>bGrow</b> is false) then NULL will be returned. </remarks>
template< class T, int cmini >
typename Ptrarray<T,cmini>::Tp *Ptrarray<T,cmini>::PtrRangeAt( _idx_ ix, _size_t_ cRange, bool bGrow )
{
    if( ix < 0 || cRange <= 0 ) return NULL;
    if( (_size_t_) ix + cRange > m_cElements ) {
        if( !bGrow ) return NULL;
        SetSize( ix + cRange );
        if( GetSize() < (_size_t_) ix + cRange ) return NULL;
    }
    return Ptr() + ix;
}

///<summary> This will fill all present elements in the specified range with NULL pointers.
/// It will first delete all the #T# objects in that range, then fill
/// that memory with NULL pointers. </summary>
template< class T, int cmini >
void Ptrarray<T,cmini>::Clear( _idx_ ix, _size_t_ count )
{
    if( ix < 0 || (_size_t_) ix >= m_cElements ) return;
    if( count == 0 ) return;
    if( (_size_t_) ix + count > m_cElements ) count = m_cElements - (_size_t_) ix;
    T **p0 = Ptr();
    T **p = p0 + (ix + count);
    _size_t_ c = count;
    while( c-- ) if( *(--p) ) delete (*p);
    memset( &p0[ix], 0, count * sizeof(T*) );
}

///<summary> Detaches objects by just simply setting NULL pointers at their positions
///  *without* deleting the objects themselves. </summary>
///<returns> false = not done (wrong index) </returns>
///<remarks> Note: If <b>count</b> is so large, that it (together with <b>ix</b>) goes
///  over the end of the array -> only those objects of the array which are
///  occupied by the area definition [ix,ix+count[ will be detached
///  without error.
///  But if <b>ix</b> negative -> nothing done.
///  <br/><br/>
///  NOTE: <br/>
///  This method does *not* remove the elements from the array. It will just
///  replace them by NULL pointers. So the array size is *not* changed.
///  If you want the objects being detached and removed, call <b>Remove()</b>
///  after the detach call passing the same parameters. </remarks>
template< class T, int cmini >
bool Ptrarray<T,cmini>::DetachObjects( _idx_ ix, _size_t_ count )
{
    if( ix < 0 || (_size_t_) ix >= m_cElements ) return false;
    if( count == 0 ) return true;
    if( (_size_t_) ix + count > m_cElements ) count = m_cElements - (_size_t_) ix;
    T **p0 = Ptr();
    memset( &p0[ix], 0, count * sizeof(T*) );
    return true;
}

///<remarks>
///  Warning: The elements copied are pointers. So this means
///  that the Ptrarray will then own the objects. You must not
///  delete the pointed objects on your own.
///</remarks>
template< class T, int cmini >
void Ptrarray<T,cmini>::SetData( const Tp *pElements, _size_t_ cSize )
{
    T **p = Ptr();
    if( p == pElements ) {
        // Note: The pointers could be equal when #this# array is
        // copied/assigned to itself ( e.g. using operator =() ).
        // In that case, do not perform the copy operation.
        _SetSize( cSize, true );
        return;
    }

    Clear();

    _SetSize( cSize, false );

    if( cSize > m_cElements ) cSize = m_cElements;
    if( cSize > 0 )
        memcpy( Ptr(), pElements, cSize );
}

///<summary> Set a pointer element at the specified index </summary>
///<param name="bGrow"> if the array should be growed if an index outside is specified.
///  If growed, the new size will be <b>ix + 1</b>. </param>
///<returns> false = not done (wrong index) </returns>
///<remarks> Note: If the array is made larger, new elements (except the last one) will
///  be filled with zeros by default (or if <b>SetSize()</b> is overloaded, the
///  filling provided in this method will be done). The last one (which *is*
///  the element at index <b>ix</b> will be set to the value <b>nElement</b> of course). </remarks>
template< class T, int cmini >
bool Ptrarray<T,cmini>::SetAt( _idx_ ix, T* pObject, bool bGrow )
{
    if( ix < 0 ) return false;
    if( (_size_t_) ix >= m_cElements ) {
        if( !bGrow ) return false;
        _SetSize( ix + 1, true );
        if( ix >= (_idx_) m_cElements ) return false;
    }
    T **p = Ptr() + ix;
    if( *p ) delete (*p);
    *p = pObject;
    return true;
}

template< class T, int cmini >
const T*& Ptrarray<T,cmini>::GetAt( _idx_ ix ) const
{
    return ( ix < 0 || (_size_t_) ix >= m_cElements )? GetDummy() : Ptr()[ix];
}

template< class T, int cmini >
T*& Ptrarray<T,cmini>::GetAt( _idx_ ix )
{
    return ( ix < 0 || (_size_t_) ix >= m_cElements )? GetDummy() : Ptr()[ix];
}

///<summary> Returns the index of the first item found (beginning
///  at index <b>iStart</b>) that is equal to <b>nSearchFor</b>,
///  or -1 if not found. </summary>
template< class T, int cmini >
_idx_ Ptrarray<T,cmini>::FindNext( T *pSearchFor, _idx_ iStart ) const
{
    const Tp *p0 = Ptr();
    for( _size_t_ i = ( iStart >= 0 )? (_size_t_) iStart : (_size_t_) 0 ; i < m_cElements ; ++i )
        if( p0[i] == pSearchFor ) return (_idx_) i;
    return -1;
}

///<summary> Insert elements at the specified index. </summary>
///<param name="ix"> if index points directly after the end of the array -> added to the end. <br/>
///  if index too large -> nothing done. <br/>
///  if index negative -> nothing done. </param>
///<returns> false = not done (wrong index) </returns>
template< class T, int cmini >
bool Ptrarray<T,cmini>::Insert( _idx_ ix, const Tp *ppObjects, _size_t_ cSize )
{
    if( ix < 0 || (_size_t_) ix > m_cElements ) return false;
    if( cSize == 0 || !ppObjects ) return true;
    _size_t_ cEl_old = m_cElements;
    _size_t_ ccopy = m_cElements - (_size_t_) ix;
    _SetSize( m_cElements + cSize, true );
    if( m_cElements <= cEl_old ) return false;
    T **p0 = Ptr();
    if( ccopy > 0 ) memmove( &p0[ix + cSize], &p0[ix], ccopy * sizeof(T*) );
    memcpy( &p0[ix], ppObjects, cSize * sizeof(T*) );
    return true;
}

///<summary> Insert elements at the specified index. </summary>
///<param name="ix"> if index points directly after the end of the array -> added to the end. <br/>
///  if index too large -> nothing done. <br/>
///  if index negative -> nothing done. </param>
///<returns> returns a pointer to the start of the new inserted memory area. <br/>
///  NULL = not done (wrong index). </returns>
///<remarks> NOTE: If <b>cSize</b> was zero, and <b>ix</b> a valid index, then the pointer
///  will be returned correctly (as non-NULL), but you should *not*
///  copy and memory in it because no new memory has been inserted. </remarks>
template< class T, int cmini >
T **Ptrarray<T,cmini>::InsertNew( _idx_ ix, _size_t_ cSize )
{
    if( ix < 0 || (_size_t_) ix > m_cElements ) return NULL;
    if( cSize == 0 ) return &Ptr()[ix]; // just to indicate that no error occured
    _size_t_ cEl_old = m_cElements;
    _size_t_ ccopy = m_cElements - (_size_t_) ix;
    _SetSize( m_cElements + cSize, true );
    if( m_cElements <= cEl_old ) return NULL;
    T **p0 = Ptr();
    if( ccopy > 0 ) memmove( &p0[ix + cSize], &p0[ix], ccopy * sizeof(T*) );
    memset( &p0[ix], 0, cSize * sizeof(T*) );
    return &p0[ix];
}

///<summary> Remove elements </summary>
///<returns> false = not done (wrong index) </returns>
///<remarks> Note: If <b>count</b> is so large, that it (together with <b>ix</b>) goes
///  over the end of the array -> only those parts of the array which are
///  occupied by the area definition [ix,ix+count[ will be removed
///  without error.
///  But if <b>ix</b> negative -> nothing done. </remarks>
template< class T, int cmini >
bool Ptrarray<T,cmini>::Remove( _idx_ ix, _size_t_ count )
{
    if( ix < 0 || (_size_t_) ix >= m_cElements ) return false;
    if( count == 0 ) return true;
    if( (_size_t_) ix + count > m_cElements ) count = m_cElements - (_size_t_) ix;
    T **p0 = Ptr();
    T **p; p = &p0[ix + count]; // prevent warning variable initialized but not referenced
    _size_t_ c; c = count;      // prevent warning variable initialized but not referenced
    while( c-- ) if( *(--p) ) delete (*p);
    _size_t_ ccopy = m_cElements - (ix + count);
    if( ccopy > 0 ) memmove( &p0[ix], &p0[ix + count], ccopy * sizeof(T*) );
    memset( &p0[m_cElements - count], 0, count * sizeof(T*) );  // if the following resizing fails, we have at least NULL pointers
    _SetSize( m_cElements - count, false );
    return true;
}

///<summary>
///  <b>RemoveNullPointers()</b> removes all NULL pointers from the array.
///</summary>
template< class T, int cmini >
void Ptrarray<T,cmini>::RemoveNullPointers()
{
    _idx_ ix = 0;
    while( ( ix = FindNext( NULL, ix ) ) >= 0 )
    {
        if( !Remove( ix ) )
            break;
    }
}



template< class T, int cmini >
void Ptrarray<T,cmini>::_SetSize( _size_t_ cSize, bool bAllowDeletes )
{
    // First destruct elements if needed,
    // but only those which will be removed.
    _size_t_ cSize_old = m_cElements;
    if( bAllowDeletes && cSize < cSize_old ) {
        T **p; p = Ptr() + cSize_old;  // prevent warning variable initialized but not referenced
        for( _size_t_ i = cSize ; i < cSize_old ; ++i )
            if( *(--p) ) delete (*p);
    }

    bool ret = ReAllocElements( cSize );

    if( !ret )
        if( bAllowDeletes && cSize < cSize_old && m_cElements == cSize_old ) {
            T **p; p = Ptr() + cSize_old;  // prevent warning variable initialized but not referenced
            memset( p, 0, (cSize_old - cSize) * sizeof(T*) );  // if the resizing above failed, we have at least NULL pointers
        }

    // We could throw an error if reallocation did not work.
}

///<summary> Resizes the <b>m_aElements</b> array (using <b>m_aElements_fixed_Mem</b> for a small array to
///  prevent too many small heap memory areas, which e.g. can occur when
///  many <b>Array</b> objects exist). </summary>
///<returns> False if reallocation failed. In that case, the original memory is unchanged. </returns>
///<remarks> Note: When resizing, the old contents will remain and if the array
///  is getting larger, the newly added memory will be cleared. <br/>
///  Note: This method does not call constructors or destructors of <b>T</b>. </remarks>
template< class T, int cmini >
bool Ptrarray<T,cmini>::ReAllocElements( _size_t_ cSize )
{
    if( cSize == m_cElements ) return true;

    if( cSize == 0 ) {
        FreeElements();
        return true;
    }

    bool onheap_new = ( cSize       > (_size_t_) cmini );   // true if for the new size we need mem on the heap instead of using #m_aElements_fixed_Mem#
    bool onheap_old = ( m_cElements > (_size_t_) cmini );   // true if for the old size we need mem on the heap instead of using #m_aElements_fixed_Mem#

    _size_t_ cElements_old = m_cElements;

    if( onheap_new )
    {
        _size_t_ cAllocatedElements_new = ((cSize + m_cBlockSize - 1) / m_cBlockSize) * m_cBlockSize;

        if( !onheap_old ) m_aElements = NULL; // for #realloc()# so that it will not try to free the memory.
        else if( cAllocatedElements_new == m_cAllocatedElements ) {
            m_cElements = cSize;
            if( m_cElements > cElements_old )
                memset( &m_aElements[cElements_old], 0, (m_cElements - cElements_old) * sizeof(T*) );
            return true;
        }
        _size_t_ allocsize = cAllocatedElements_new * sizeof(T*);
        void *p = realloc( (void*) m_aElements, (_size_t_) allocsize );
        if( !p && allocsize ) return false;
        m_aElements = (T**) p;
        m_cElements = cSize;
        m_cAllocatedElements = cAllocatedElements_new;
        if( !onheap_old )
            memcpy( m_aElements, m_aElements_fixed_Mem, cElements_old * sizeof(T*) );

        if( m_cElements > cElements_old )
            memset( &m_aElements[cElements_old], 0, (m_cElements - cElements_old) * sizeof(T*) );
    }
    else
    {
        if( onheap_old ) {
            memcpy( m_aElements_fixed_Mem, m_aElements, cSize * sizeof(T*) );
            free( m_aElements );
        }
        m_aElements = NULL;
        m_cElements = cSize;
        m_cAllocatedElements = 0;

        if( m_cElements > cElements_old )
            memset( &m_aElements_fixed_Mem[cElements_old * sizeof(T*)], 0, (m_cElements - cElements_old) * sizeof(T*) );
    }

    return true;
}

///<remarks> Note: This method does not call constructors or desctructor of <b>T</b>. </remarks>
template< class T, int cmini >
void Ptrarray<T,cmini>::FreeElements()
{
    if( m_aElements && m_cElements > (_size_t_) cmini )
        free( m_aElements );
    m_cElements = 0;
    m_aElements = NULL;
}

///</summary> Clears and constructs the dummy object and returns it.
///  If already constructed, it will first be destructed
///  so that its memory can be cleared before construction. </summary>
template< class T, int cmini >
T*& Ptrarray<T,cmini>::GetDummy() const
{
    Ptrarray<T,cmini> *pArray = (Ptrarray<T,cmini>*) this; // to make it non-const
    T **pDummy = (T**) &pArray->m_aElements_fixed_Mem[cmini * sizeof(T*)];
    *pDummy = NULL;
    return *pDummy;
}




///<summary> A <b>MemList</b> is a list of memory units which are
///  *not* intended to be like an ordered memory array like <b>Array</b>.
///  It will allocate each memory unit not separately but packed
///  in blocks to reduce heap fragmentation. <br/>
///  Note: A memory unit has a specific size in bytes and this size
///  is equal for all memory units in the list. So this list does
///  not support units with different sizes in a specific list. <br/><br/>
///
///  The <b>MemList</b> is intended for speed allocation.
///  So there are some features, which are not supported: <br/><br/>
///  - A newly created unit (using <b>MemList::Create()</b>) will be created
///    somewhere in the list. So the position where it is inserted is
///    random. <br/><br/>
///  - Two following memory units will be placed on random addresses
///    which you cannot imagine as being one after each other. <br/><br/>
///  - Retrieving or querying memory units from the list is *not*
///    supported. Only a pointer to a memory unit will be returned
///    by <b>MemList::Create()</b> when the unit is created. <br/>
///    So the functionality of <b>MemList</b> is comparable with a simple
///    memory allocation like on a heap, but with the restriction that
///    all allocated memory units have the same size. <br/>
///    If you need the units to be linked like in a linked list
///    or double linked list (etc.), then you will have to provide
///    and manage prev- and/or next-pointers in each memory unit by
///    yourself. </summary>
class MemList
{
    protected :
        struct Unit;

        struct Block
        {
            MemList  *m_pMemList;       // to which <b>MemList</b> this memory block belongs.
            Block    *m_prev;
            Block    *m_next;
            _size_t_  m_cUsedUnits;     // how many of the units in this block really are used (maximum is <b>m_cBlock</b>)
            Unit     *m_pUnusedUnit;    // pointer to a unit in this block which *is* unused (or else NULL)
        };

        struct Unit
        {
            Block    *m_pBlock;         // to which block this unit belongs. this is valid only for a used unit
            BOOL      m_bUsed;          // if this unit is used (of course, this member variable is also valid for an unused unit)
        };

        _size_t_  m_cMemUnit;       // size of one memory unit in bytes desired by the user
        _size_t_  m_cBlock;         // number of memory units to allocate at once when necessary

        _size_t_  m_cUsedUnits;     // number of memory units currently availabe

        Block    *m_pBlock1;
        Block    *m_pBlockL;

        Block    *m_pPartiallyUsedBlock;    // is a block from which it is *known* that not all of its units are used
        _size_t_  m_cPartiallyUsedBlocks;   // how many of the blocks do not have all of their units used

    public :
        MemList( _size_t_ cMemUnit, _size_t_ cBlock = 16 );
        ~MemList() { Empty(); }

        _size_t_ GetSize() const { return m_cUsedUnits; } // returns the number of units currently available to the user
        bool IsEmpty() const { return m_cUsedUnits == 0; }
        void Empty(); // deletes all units from this list

        void *Create( bool bClear = true ); // creates a new unit but somewhere into the list (the insertion position is random)
        void Delete( void *pUnit );

        // Use <b>GetMemList()</b> to retrieve the <b>MemList</b> object that
        // handles the specified memory unit (once returned by <b>Create()</b>).
        static MemList *GetMemList( void *pUnit ) { return ((Unit*) pUnit)[-1].m_pBlock->m_pMemList; }

    protected :
        _size_t_ CUnit() const  { return sizeof( Unit  ) + m_cMemUnit; }
        _size_t_ CBlock() const { return sizeof( Block ) + m_cBlock * CUnit(); }
        Block *CreateBlock();
        void DeleteBlock( Block *block );
        Block *FindAPartiallyUsedBlock();
        Unit *FindAnUnusedUnit( Block *block );
};




typedef _size_t_  _hashval_t_;    // hashkey value type (see #HashTable#) (*must* be an unsigned type)

///<summary> The <b>HashTable</b> template class allows access to a list of
///  memory items (of type <b>TItem</b>) through a key (of type <b>TKey</b>)
///  which gives a more affective (and faster) access to items in the list when
///  seaching for one as long as the hashtable is not overfilled. The hashtable
///  does not have an upper limit of items it can hold. But if there are much
///  more items in the last than the size of the hashtable <b>cHashTable</b>,
///  or if the <b>HashKey()</b> method does not generate keys which are well
///  spread over the size of the hashtable, then the access can become much
///  slower. <br/><br/>
///
///  IMPORTANT: The <b>TKey</b> can also be a class itself (because its
///  constructors and desctructors also will be called whenever necessary). <br/>
///  Of course, <b>TItem</b> can be a class, because its constructors and
///  destructors indeed will be called whenever necessary. <br/><br/>
///
///  The hashtable can then be interpreted as mapping of values of type
///  <b>TKey</b> to values of type <b>TItem</b>. <br/><br/>
///
///  In the hashtable, keys with same value are *not* supported. If you
///  call the <b>SetAt( key, newitem )</b> method or call
///  <b>hashtable[key] = newitem</b>, then this will *not* create a new
///  key with the same value <b>key</b> if one does already exist in the
///  table with that value. Instead, the existing key's item will be
///  overwritten with the new item value <b>newitem</b>. <br/><br/>
///
///  NOTE: It is recommended, to create a hashtable with an odd size (maybe
///  even with a size value which is a prim-number). Guess this could spread
///  the hash keys a bit better. <br/><br/>
///
///  NOTE: If the provided <b>TKey</b> is a struct or class, you
///  can define a method <b>EvalHashKey()</b> which will be used to
///  calculate the hash key values instead of the default implementation.
///  In that case, the method should be declared like this: <br/>
///      <b>_hashval_t_ EvalHashKey() const;</b>     <br/><br/>
///
///  IMPORTANT: If you provide <b>TKey</b> as a struct or class, you
///  must ensure, that the following TKey's operations work correctly: <br/>
///      - copy-constructor     <br/>
///      - operator =()         <br/>
///      - operator ==()        <br/><br/>
///
///  Note: Every newly created item will be created by calling its
///  constructor (or copy-constructor). If in there, all or some members
///  of the item are *not* initialized, they will be filled with zeros
///  because internally, before any constructor is called, the memory
///  allocated for the new item will first be filled with zeros. Then
///  the constructors will be called. </summary>
template< class TKey, class TItem >
class HashTable
{
    protected :
        struct Item
        {
            Item   *m_next;     // belongs to a linked list where the first <b>Item</b> is stored in an entry of the <b>m_aHashTable</b>.
            TKey    m_key;
            TItem   m_item;
            _idx_   m_ihash;    // index into the hashtable (calculated from the <b>m_key</b>)
        };

        MemList m_mlItems;  // contains all items of the hashtable, but unsorted. The order of items depends on the <b>Item</b> struct member vars
        // Note: Instead of the <b>MemList</b> usage, each <b>Item</b> could be created on the
        // heap using <b>new</b> operator as well, but this can fragment the heap very much
        // while <b>MemList</b> allows heap allocation of blocks of memory. The size of each
        // block depends on the <b>cItemsPerBlock</b> param of the constructor in here.
        // It defines the number items to be hold in one block of heap memory.

        Array<Item*> m_aHashTable; // each entry contains a pointer to the first <b>Item</b> of a list of <b>Item</b>s.
        // The hashtable is designed like that:
        // Each entry of <b>aHashTable</b> contains a pointer to a <b>Item</b> object which
        // is the first of the list of <b>Item</b>s which all are addressed by the same
        // hashkey (generated by <b>HashKey()</b>). The hashkey returned by <b>HashKey()</b>
        // selects one of the entries of the <b>m_aHashTable</b>.
        // Each entry must hold a list of items because to the table there can
        // be added different items which have the same hashkey and they have
        // to be stored then in a list to one entry of <b>m_aHashTable</b>.

    public :
        HashTable( _size_t_ cHashTable, _size_t_ cItemsPerBlock = 16 ); // calls <b>Init()</b> implicitely
        virtual ~HashTable();

        bool Init( _size_t_ cHashTable );
        void Empty();
        bool IsEmpty() const { return m_mlItems.IsEmpty();    }
        _size_t_ Count() const { return m_mlItems.GetSize();    } // number of items in the hashtable (can be samller or *larger* than <b>Size()</b>)
        _size_t_ Size () const { return m_aHashTable.GetSize(); } // size of the hashtable. This is the value <b>cHashTable</b> as passed to <b>Init()</b> or the ctor.

        TItem& operator[]( TKey key );      // lookup and add if not there
        void SetAt( TKey key, TItem newitem ) { (*this)[key] = newitem; } // add a new (key,item) pair
        void SetFromPtrAt( TKey key, TItem *newitem ) { (*this)[key] = *newitem; } // add a new (key,item) pair
        bool RemoveAt( TKey key );          // removes an existing (key,?) pair
        TItem GetAt( TKey key ) const;      // same functionality as <b>Lookup()</b>
        TItem *GetPtrAt( TKey key ) const;  // same functionality as <b>LookupPtr()</b>
        bool Lookup( TKey key, TItem& item ) const; // same functionality as <b>GetAt()</b>
        bool LookupPtr( TKey key, TItem *& item ) const; // same functionality as <b>GetPtrAt()</b>

        // iterating all (key,item) pairs
        bool GetNext   ( Iterator& nextpos, TKey&         key, TItem&   item ) const;
        bool GetNextPtr( Iterator& nextpos, const TKey *& key, TItem *& item ) const;

    protected :
        typename Item *GetItemAt( TKey key, _idx_& iHash ) const;

        // Overridables: Routine used to user-provided hash keys
        virtual _hashval_t_ HashKey( TKey key ) const;
};




template< class TKey, class TItem >
HashTable<TKey,TItem>::HashTable( _size_t_ cHashTable, _size_t_ cItemsPerBlock )
                      :m_mlItems( sizeof(Item), cItemsPerBlock ),
                       m_aHashTable()
{
    Init( cHashTable );
}

template< class TKey, class TItem >
HashTable<TKey,TItem>::~HashTable()
{
    Empty();
}



///<returns> FALSE = Error because <b>Init()</b> called although the hashtable
///  is not empty. Call <b>Empty()</b> first and then call <b>Init()</b> again. </returns>
///<remarks> Note: A hashtable *cannot* be resized, because the calculated hashkeys
///  (from <b>HashKey()</b>) will return different indices into <b>m_aHashTable</b> when
///  its size has changed. So a new hashtable size can be set only when
///  the hashtable is empty. </remarks>
template< class TKey, class TItem >
bool HashTable<TKey,TItem>::Init( _size_t_ cHashTable )
{
    if( !m_mlItems.IsEmpty() ) return false;
    m_aHashTable.SetSize( cHashTable > 0 ? cHashTable : 1 );
    return cHashTable > 0;
}

template< class TKey, class TItem >
void HashTable<TKey,TItem>::Empty()
{
    TKey *key;
    TItem *item;
    Iterator up;

    while( GetNextPtr( up, (const TKey*&) key, item ) ) {
        key->~TKey();
        item->~TItem();
    }

    m_mlItems.Empty();
    m_aHashTable.Clear();
}



template< class TKey, class TItem >
TItem& HashTable<TKey,TItem>::operator[]( TKey key )
{
    _idx_ ihash;
    Item *_item = GetItemAt( key, ihash );
    if( !_item ) {
        // Item does not exist. So create a new one and clear its item-content.
        _item = (Item*) m_mlItems.Create( true ); // true to cause the memory to be cleared
        new((void*) &_item->m_item) TItem;
        new((void*) &_item->m_key) TKey( key );
        _item->m_ihash = ihash; // used mainly for speed (expecially for #GetNext()# and #GetNextPtr()#)
        // put item into hashtable
        _item->m_next = m_aHashTable.GetAt( ihash );
        m_aHashTable.SetAt( ihash, _item );
    }
    return _item->m_item;
}

///<returns> TRUE = item was found in the hashtable and has been removed <br/>
///          FALSE = item not found in the hashtable </returns>
template< class TKey, class TItem >
bool HashTable<TKey,TItem>::RemoveAt( TKey key )
{
    _size_t_ size = Size();
    if( size <= 0 ) return false;
    _idx_ ihash = type<_idx_>::abs( (_idx_) (HashKey( key ) % size) );
    Item **__item = m_aHashTable.PtrAt( ihash );
    Item *_item = *__item;
    while( _item ) {
        if( _item->m_key == key ) {
            *__item = _item->m_next; // remove from linked list first
            _item->m_key.~TKey();
            _item->m_item.~TItem();
            m_mlItems.Delete( _item );
            return true;
        }
        __item = &_item->m_next;
        _item = _item->m_next;
    }
    return false;
}

template< class TKey, class TItem >
TItem HashTable<TKey,TItem>::GetAt( TKey key ) const
{
    _idx_ ihash;
    Item *_item = GetItemAt( key, ihash );
    return _item ? _item->m_item : NULL;
}

template< class TKey, class TItem >
TItem *HashTable<TKey,TItem>::GetPtrAt( TKey key ) const
{
    _idx_ ihash;
    Item *_item = GetItemAt( key, ihash );
    return _item ? &_item->m_item : NULL;
}

template< class TKey, class TItem >
bool HashTable<TKey,TItem>::Lookup( TKey key, TItem& item ) const
{
    _idx_ ihash;
    Item *_item = GetItemAt( key, ihash );
    if( !_item ) return false;
    item = _item->m_item;
    return true;
}

template< class TKey, class TItem >
bool HashTable<TKey,TItem>::LookupPtr( TKey key, TItem *& item ) const
{
    _idx_ ihash;
    Item *_item = GetItemAt( key, ihash );
    if( !_item ) return false;
    item = &_item->m_item;
    return true;
}



///<returns> false = no more items </returns>
///<remarks> NOTE: <b>nextpos</b> when passed should not be NULL. <br/><br/>
/// Note: The order of the items in the hashtable is *not* the order of
///       creation by the user. <br/><br/>
/// Note: Because internally the items are (for each entry in <b>m_aHashTable</b>)
///       stored as linked list, it *is* possible to call <b>GetNext()</b> to
///       get a <i>(key,item)</i> pair and immediately remove it with <b>RemoveAt()</b>
///       and then call <b>GetNext()</b> again to iterate further. <br/>
///       So when removing an item during an iteration, you will not need
///       to start iteration again right from the start, <br/>
///       EXCEPT IN THE FOLLOWING CASE: When you retrieve an item with
///       <b>GetNext()</b>, <b>nextpos</b> will point to (as pointer) the next
///       following item. So if you also delete following items, <b>nextpos</b>
///       will indicate an invalid item and a crash would occur. <br/>
///       This will *not* happen, if you ensure that *only* the item recreived
///       by a call to <b>GetNext()</b> is removed using <b>RemoveAt()</b>. <br/><br/>
/// Note: <b>GetNext()</b> and <b>GetNextPtr()</b> are implemented to be
///  as fast as possible (no hashkey-values are calculated from the key-values
///  any more) because it is used e.g. by the <b>ExpressionEvaluator</b> class
///  which needs fast variable enumeration (the variable storage is done by using
///  a hashtable). </remarks>
template< class TKey, class TItem >
bool HashTable<TKey,TItem>::GetNext( Iterator& nextpos, TKey& key, TItem& item ) const
{
    Item *_item = (Item*) nextpos.m_pos;
    if( !_item ) return false;

    if( _item == (Item*) Iterator::before_iteration_start() )
    {
        // find first item
        if( Count() == 0 ) { nextpos.m_pos = NULL; return false; }
        Item **p = (Item**) m_aHashTable.Ptr(); // used just for speed
        _size_t_ size = Size();
        for( _size_t_ ihash = 0 ; ihash < size ; ++ihash )
            { _item = p[ihash]; if( _item ) break; }
        if( !_item ) { nextpos.m_pos = NULL; return false; }
    }

    // find next item
    Item *_next = _item->m_next;
    if( !_next )
    {
        // go to next hash index
        Item **p = (Item**) m_aHashTable.Ptr(); // used just for speed
        _size_t_ size = Size();
        for( _size_t_ ihash = _item->m_ihash + 1 ; ihash < size ; ++ihash )
            { _next = p[ihash]; if( _next ) break; }
        // If #_next# is NULL here -> no more items in the hashtable
    }

    nextpos.m_pos = (Iterator::iteration_pos) _next;
    key  = _item->m_key;
    item = _item->m_item;
    return _item != NULL;
}

///<returns> FALSE = no more items </returns>
///<remarks> See remarks to <b>GetNext()</b>. </remarks>
template< class TKey, class TItem >
bool HashTable<TKey,TItem>::GetNextPtr( Iterator& nextpos, CONST TKey *& key, TItem *& item ) const
{
    Item *_item = (Item*) nextpos.m_pos;
    if( !_item ) return false;

    if( _item == (Item*) Iterator::before_iteration_start() ) {
        // find first item
        if( Count() == 0 ) { nextpos.m_pos = NULL; return false; }
        Item **p = (Item**) m_aHashTable.Ptr(); // used just for speed
        _size_t_ size = Size();
        for( _size_t_ ihash = 0 ; ihash < size ; ++ihash )
            { _item = p[ihash]; if( _item ) break; }
        if( !_item ) { nextpos.m_pos = NULL; return false; }
    }

    // find next item
    Item *_next = _item->m_next;
    if( !_next ) {
        // go to next hash index
        Item **p = (Item**) m_aHashTable.Ptr(); // used just for speed
        _size_t_ size = Size();
        for( _size_t_ ihash = _item->m_ihash + 1 ; ihash < size ; ++ihash )
            { _next = p[ihash]; if( _next ) break; }
        // If #_next# is NULL here -> no more items in the hashtable
    }

    nextpos.m_pos = (Iterator::iteration_pos) _next;
    key  = &_item->m_key;
    item = &_item->m_item;
    return _item != NULL;
}



///<param name="iHash"> will contain the index into the <b>m_aHashTable</b>
///  where the item has/shall be found. </param>
template< class TKey, class TItem >
typename HashTable<TKey,TItem>::Item *HashTable<TKey,TItem>::GetItemAt( TKey key, _idx_& iHash ) const
{
    _size_t_ size = Size();
    if( size <= 0 ) return NULL;
    iHash = type<_idx_>::abs( (_idx_) (HashKey( key ) % size) );
    Item *_item = m_aHashTable.GetAt( iHash );
    while( _item ) {
        if( _item->m_key == key ) return _item;
        _item = _item->m_next;
    }
    return NULL;
}



template< class TKey, class TItem >
_hashval_t_ HashTable<TKey,TItem>::HashKey( TKey key ) const
{
    // If #TKey# has a method #EvalHashKey()# then this will be
    // used to calculate a hashkey value.
    // The method should be declared like this:
    //      _hashval_t_ EvalHashKey() const;
    // Otherwise a default handling will be used.

    __if_exists( TKey::EvalHashKey ) {
        return key.EvalHashKey();
    }

    __if_not_exists( TKey::EvalHashKey ) {
        // default identity hash - works for most primitive values
        //return ((_hashval_t_) (void*) (DWORD) key) >> 4;
        return (_hashval_t_) (((UINT_PTR) key) >> 4);
    }
}




}

#endif
