/*
    File:               argonColor.cpp
    Description:        Provides means for useful color calculations.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         06.08.2006
    last updated:       06.08.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonColor_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonColor.h"

namespace ArgonLibrary
{




///<summary> Scales the R G B values (but does *not* change the alpha value) </summary>
_Color_ _Color_::StretchColor( float s ) const
{
    int r = (int) (getR() * s);
    int g = (int) (getG() * s);
    int b = (int) (getB() * s);

    r = privateColor::tbound( 0, r, 255 );
    g = privateColor::tbound( 0, g, 255 );
    b = privateColor::tbound( 0, b, 255 );

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, getA() );
}

///<summary> Scales the R G B values up to the specified maximum
///  (but does <u>not</u> change the alpha value). </summary>
_Color_ _Color_::StretchColorTo( BYTE max_component ) const
{
    BYTE max_comp = getMaxComponent();
    // Do *not* set #max_comp# to type WORD or else the VC7 compiler
    // generates wrong code in Release versions.

    if( max_comp == 0 )
        return _Color_( 0, 0, 0, getA() );

    BYTE r = (BYTE) ((getR() * max_component) / max_comp);
    BYTE g = (BYTE) ((getG() * max_component) / max_comp);
    BYTE b = (BYTE) ((getB() * max_component) / max_comp);

    return _Color_( r, g, b, getA() );
}

///<summary> Adds a value to the R G B values (but does *not* change the alpha value) </summary>
_Color_ _Color_::LiftColor( int n ) const
{
    int r = (int) getR() + n;
    int g = (int) getG() + n;
    int b = (int) getB() + n;

    r = privateColor::tbound( 0, r, 255 );
    g = privateColor::tbound( 0, g, 255 );
    b = privateColor::tbound( 0, b, 255 );

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, getA() );
}

///<summary> Adds a color to this color, but the alpha channel will not be changed.
///  An alpha value of <b>rgb2</b> will be ignored. </summary>
_Color_ _Color_::AddColor( _Color_ rgb2 ) const
{
    int r = (int) getR() + (int) rgb2.getR();
    int g = (int) getG() + (int) rgb2.getG();
    int b = (int) getB() + (int) rgb2.getB();

    r = privateColor::tmin( r, 255 );
    g = privateColor::tmin( g, 255 );
    b = privateColor::tmin( b, 255 );

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, getA() );
}

///<summary> Subtracts a color from this color, but the alpha channel will not be changed.
///  An alpha value of <b>rgb2</b> will be ignored. </summary>
_Color_ _Color_::SubColor( _Color_ rgb2 ) const
{
    int r = (int) getR() - (int) rgb2.getR();
    int g = (int) getG() - (int) rgb2.getG();
    int b = (int) getB() - (int) rgb2.getB();

    r = privateColor::tmax( 0, r );
    g = privateColor::tmax( 0, g );
    b = privateColor::tmax( 0, b );

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, getA() );
}

///<summary> Interpolates between <b>rgba1</b> and <b>rgba2</b> when <b>s</b> ranges
/// from 0 to 1. If <b>s</b> == 0 -> <b>rgba1</b> will be returned,
/// if <b>s</b> == 1 -> <b>rgba2</b> will be returned. <br/>
/// Note: <b>s</b> may also be outside the 0 to 1 range. <br/>
/// Note: The alpha channel will *not* be changed. </summary>
_Color_ _Color_::InterpolateColor( _Color_ rgba1, _Color_ rgba2, float s )
{
    int r = (int) rgba2.getR() - (int) rgba1.getR();
    int g = (int) rgba2.getG() - (int) rgba1.getG();
    int b = (int) rgba2.getB() - (int) rgba1.getB();
    int a = (int) rgba2.getA() - (int) rgba1.getA();

    r = (int) (s * (float) r) + (int) rgba1.getR();
    g = (int) (s * (float) g) + (int) rgba1.getG();
    b = (int) (s * (float) b) + (int) rgba1.getB();
    a = (int) (s * (float) a) + (int) rgba1.getA();

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, (BYTE) a );
}

///<summary> Stretches the color that way, that the 3 color
///  components occupy the whole range from 0 to 255.
///  The alpha channel is not affected (will not be changed).</summary>
///<remarks> Normally this would result in that one component is 0,
///  the other is 255, the third is stretched accordingly.
///  If all 3 components are equal, then nothing will be done. </remarks>
///<param name="lower_amount"> How much to move the darkest component
///  towards black. 0 = don't move at all, 1 = make black. <br/>
///  Can be any value, even outside range 0 - 1. </param>
///<param name="upper_amount"> How much to move the lightest component
///  towards white. 0 = don't move at all, 1 = make white. <br/>
///  Can be any value, even outside range 0 - 1. </param>
_Color_ _Color_::MaximizeColor( float lower_amount, float upper_amount ) const
{
    int r = (int) getR();
    int g = (int) getG();
    int b = (int) getB();

    int min_comp = (int) getMinComponent();
    int max_comp = (int) getMaxComponent();

    float comp_range = (float) (max_comp - min_comp);
    if( comp_range == 0.0f )
        return *this;

    float norm_r = (float) (r - min_comp) / comp_range;  // range 0 - 1
    float norm_g = (float) (g - min_comp) / comp_range;  // range 0 - 1
    float norm_b = (float) (b - min_comp) / comp_range;  // range 0 - 1

    float lower_val = (float) min_comp * (1.0f - lower_amount);
    float upper_val = 255.0f - (float) (255 - max_comp) * (1.0f - upper_amount);
    comp_range = upper_val - lower_val;

    r = privateColor::tbound( 0, (int) ((norm_r * comp_range) + lower_val), 255 );
    g = privateColor::tbound( 0, (int) ((norm_g * comp_range) + lower_val), 255 );
    b = privateColor::tbound( 0, (int) ((norm_b * comp_range) + lower_val), 255 );

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, getA() );
}

///<summary> Same as <b>MaximizeColor()</b> (see remarks there), except
///  that the amount values have been replaced by the final color component
///  minimum and maximum values. The final color will be placed in between
///  of this range. (For a gray color, it will be limited to that range). <br/>
///  The lower and upper component values may also be outside the normal
///  color component range (0 - 255). </summary>
_Color_ _Color_::MaximizeColorTo( int lower_component, int upper_component ) const
{
    int r = (int) getR();
    int g = (int) getG();
    int b = (int) getB();

    int min_comp = (int) getMinComponent();
    int max_comp = (int) getMaxComponent();

    float comp_range = (float) (max_comp - min_comp);
    if( comp_range == 0.0f ) {
        if( min_comp < lower_component ) return _Color_( (BYTE) lower_component, (BYTE) lower_component, (BYTE) lower_component, getA() );
        if( max_comp > upper_component ) return _Color_( (BYTE) upper_component, (BYTE) upper_component, (BYTE) upper_component, getA() );
        return *this;
    }

    float norm_r = (float) (r - min_comp) / comp_range;  // range 0 - 1
    float norm_g = (float) (g - min_comp) / comp_range;  // range 0 - 1
    float norm_b = (float) (b - min_comp) / comp_range;  // range 0 - 1

    comp_range = (float) (upper_component - lower_component);

    r = privateColor::tbound( 0, (int) (norm_r * comp_range) + lower_component, 255 );
    g = privateColor::tbound( 0, (int) (norm_g * comp_range) + lower_component, 255 );
    b = privateColor::tbound( 0, (int) (norm_b * comp_range) + lower_component, 255 );

    return _Color_( (BYTE) r, (BYTE) g, (BYTE) b, getA() );
}

///<summary> Paints the <b>rgbPaint</b> color over the <b>rgbBkgnd</b>
///  taking the alpha channels into account. </summary>
///<remarks> The resulting alpha channel will be the alpha channel
///  of <b>rgbBkgnd</b> unchanged. </remarks>
_Color_ _Color_::AlphaPaintColor( _Color_ rgbBkgnd, _Color_ rgbPaint )
{
    BYTE a1  = rgbPaint.getA();
    BYTE r1  = (rgbPaint.getR() * a1) / 255;
    BYTE g1  = (rgbPaint.getG() * a1) / 255;
    BYTE b1  = (rgbPaint.getB() * a1) / 255;

    BYTE a1_ = 255 - a1;
    BYTE r2  = (rgbBkgnd.getR() * a1_) / 255;
    BYTE g2  = (rgbBkgnd.getG() * a1_) / 255;
    BYTE b2  = (rgbBkgnd.getB() * a1_) / 255;

    return _Color_( r1 + r2, g1 + g2, b1 + b2, rgbBkgnd.getA() );
}

// Checks if 2 colors are nearly the same (inside the specified
// 'range' which is a value between 0 and 255).
// This is done by this: The colors are converted to gray and
// its difference is compared if it is inside the 'range'.
//  : range = a value between 0 and 255.
//            It *can* be >= 256, but in that case, the function will always return FALSE.
//            0      = the function will return TRUE only if the colors are exactly equal.
//            255    = the function will almost ever return FALSE.
//            >= 256 = the function will always return FALSE.
// IMPORTANT: This function *ignores* any alpha channels.
bool _Color_::SimilarColors( _Color_ rgb1, _Color_ rgb2, UINT range )
{
    UINT gray1 = (UINT) rgb1.getGray();
    UINT gray2 = (UINT) rgb2.getGray();
    return abs( (int) gray1 - (int) gray2 ) <= (int) range;
}

///<summary> Gets the minimum of each of the 4 components of the
///  passed colors. </summary>
_Color_ _Color_::GetMinColor( _Color_ rgba1, _Color_ rgba2 )
{
    return _Color_( privateColor::tmax(rgba1.getR(), rgba2.getR()),
                    privateColor::tmax(rgba1.getG(), rgba2.getG()),
                    privateColor::tmax(rgba1.getB(), rgba2.getB()),
                    privateColor::tmax(rgba1.getA(), rgba2.getA()) );
}

///<summary> Gets the maximum of each of the 4 components of the
///  passed colors. </summary>
_Color_ _Color_::GetMaxColor( _Color_ rgba1, _Color_ rgba2 )
{
    return _Color_( privateColor::tmin(rgba1.getR(), rgba2.getR()),
                    privateColor::tmin(rgba1.getG(), rgba2.getG()),
                    privateColor::tmin(rgba1.getB(), rgba2.getB()),
                    privateColor::tmin(rgba1.getA(), rgba2.getA()) );
}




}
