/*
    File:               argonColor.h
    Description:        Provides means for useful color calculations.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         06.08.2006
    last updated:       07.08.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonColor_h
#define __argonColor_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

namespace ArgonLibrary
{




// Maximum possible colors in a palette for actual display adapters
// and corresponding depth value ( bits )
#define MAXPALCOLS                  256
#define MAXPALDEPTH                 8


// Note: The next few following color macros are used for COLORREF colors

#ifndef RGBA
#define RGBA(r, g, b, a)            (((DWORD)RGB(r, g, b)) | (((DWORD)(BYTE)(a)) << 24))  // produces a COLORREF color
#endif
#define SWAP_RGBA(rgba)             ((DWORD)( (((DWORD)(BYTE)(rgba)) << 16) | ((rgba) & 0xff00ff00) | ((BYTE)(((DWORD)(rgba)) >> 16)) ))


#ifndef GetAValue
#define GetAValue(rgba)             ((BYTE)((rgba) >> 24))
#endif

// Calculates the middle color of 2 RGB colors
#ifndef GetMidColor
#define GetMidColor(rgb1,rgb2)                                           \
        RGB( ((WORD)GetRValue(rgb1) + (WORD)GetRValue(rgb2)) >> 1,       \
             ((WORD)GetGValue(rgb1) + (WORD)GetGValue(rgb2)) >> 1,       \
             ((WORD)GetBValue(rgb1) + (WORD)GetBValue(rgb2)) >> 1 )
#endif
// Calculates the middle color of 2 RGBA colors
#ifndef GetMidAlphaColor
#define GetMidAlphaColor(rgba1,rgba2)                                    \
        RGBA( ((WORD)GetRValue(rgba1) + (WORD)GetRValue(rgba2)) >> 1,    \
              ((WORD)GetGValue(rgba1) + (WORD)GetGValue(rgba2)) >> 1,    \
              ((WORD)GetBValue(rgba1) + (WORD)GetBValue(rgba2)) >> 1,    \
              ((WORD)GetAValue(rgba1) + (WORD)GetAValue(rgba2)) >> 1 )
#endif

#define ColorrefTriple(cr)          ((COLORREF)( ((DWORD)(cr)) & 0x00ffffff ))   // extracts only the R, G, B channels. Is necessary for most GDI calls.


// QUAD_TO_RGB transforms an RGBQUAD value to a DWORD-rgb (COLORREF) value.
// TRIPLE_TO_RGB transforms an RGBTRIPLE value to a DWORD-rgb (COLORREF) value.
// DEPTH15_TO_RGB transforms an RGB-color stored in only a WORD ( each color value
//  has 5 bits ) into a DWORD-rgb (COLORREF) value.
// RGB_TO_DEPTH15 transforms a DWORD-rgb (COLORREF) value into a WORD-rgb value
//  ( each color value has 5 bits ).
#define QUAD_TO_RGB(rgbquad)        RGB( (rgbquad).rgbRed,             \
                                         (rgbquad).rgbGreen,           \
                                         (rgbquad).rgbBlue )
#define QUAD_TO_RGBA(rgbquad)       RGBA( (rgbquad).rgbRed,            \
                                          (rgbquad).rgbGreen,          \
                                          (rgbquad).rgbBlue,           \
                                          (rgbquad).rgbReserved )
#define TRIPLE_TO_RGB(rgbtriple)    RGB( (rgbtriple).rgbtRed,          \
                                         (rgbtriple).rgbtGreen,        \
                                         (rgbtriple).rgbtBlue )
#define DEPTH15_TO_RGB(d15)         RGB( (((d15) >> 10) & 0x1f) << 3,  \
                                         (((d15) >> 5) & 0x1f) << 3,   \
                                         ((d15) & 0x1f) << 3 )
#define RGB_TO_DEPTH15(rgb)         ((WORD)( (( (WORD)(((BYTE)GetRValue(rgb)) >> 3) ) << 10) |   \
                                             (( (WORD)(((BYTE)GetGValue(rgb)) >> 3) ) << 5 ) |   \
                                              ( (WORD)(((BYTE)GetBValue(rgb)) >> 3) )      ))




namespace privateColor
{
    template< class T > T tbound( T _min, T x, T _max ) { return ( x < _min )? _min : ( x > _max ? _max : x ); }
    template< class T > T tmin( T a, T b ) { return ( a <= b )? a : b; }
    template< class T > T tmax( T a, T b ) { return ( a >= b )? a : b; }
}

///<summary>
///  This is a new color type.
///  It has the same value as a D3DCOLOR (with macros D3DRGB and D3DRGBA).
///  But it is contraire to COLORREF: The COLORREF has the red and blue
///  components exchanged. <br/>
///  Advantage: RgbClr uses same memory arrangement as RGBQUAD struct.
///</summary>
class _Color_
{
    public :
        DWORD m_Color;  // stores color in AARRGGBB format (Alpha as MSB, Blue as LSB)

    public :
        _Color_() { m_Color = 0; }
    protected :
        _Color_( DWORD src ) { m_Color = src; }
        _Color_( BYTE r, BYTE g, BYTE b ) { m_Color = b | (g << 8) | (r << 16); }
        _Color_( BYTE r, BYTE g, BYTE b, BYTE a ) { m_Color = b | (g << 8) | (r << 16) | (a << 24); }

    public :
        static _Color_ fromRGB ( BYTE r, BYTE g, BYTE b         ) { return _Color_( r, g, b       ); }
        static _Color_ fromRGBX( BYTE r, BYTE g, BYTE b         ) { return _Color_( r, g, b, 0xff ); }
        static _Color_ fromRGBA( BYTE r, BYTE g, BYTE b, BYTE a ) { return _Color_( r, g, b, a    ); }
        static _Color_ fromXRGB( BYTE r, BYTE g, BYTE b         ) { return _Color_( r, g, b, 0xff ); }
        static _Color_ fromARGB( BYTE a, BYTE r, BYTE g, BYTE b ) { return _Color_( r, g, b, a    ); }

        BYTE getR() const { return (BYTE) (m_Color >> 16); }
        BYTE getG() const { return (BYTE) (m_Color >>  8); }
        BYTE getB() const { return (BYTE)  m_Color       ; }
        BYTE getA() const { return (BYTE) (m_Color >> 24); }

        static _Color_ fromGray( BYTE gray, BYTE a = 0 ) { return _Color_( gray, gray, gray, a ); }
        BYTE           getGray() const { return (BYTE) (((int) getR() + (int) getG() + (int) getB()) / 3); }
        bool           isGray () const { return getB() == getR() && getB() == getG(); }
        _Color_        asGray () const { BYTE gray = getGray(); return _Color_( gray, gray, gray, getA() ); }

        BYTE getMinComponent() const { return privateColor::tmin( getR(), privateColor::tmin( getG(), getB() ) ); } // get minimum of the 3 R,G,B components
        BYTE getMaxComponent() const { return privateColor::tmax( getR(), privateColor::tmax( getG(), getB() ) ); } // get maximum of the 3 R,G,B components

        static _Color_   fromRGBQUAD  ( const RGBQUAD   *pRgbquad   ) { return *(_Color_*) pRgbquad; }
        static _Color_   fromRGBTRIPLE( const RGBTRIPLE *pRgbtriple ) { return _Color_( pRgbtriple->rgbtRed, pRgbtriple->rgbtGreen, pRgbtriple->rgbtBlue ); }
        RGBQUAD         *asRGBQUAD  ()       { return (RGBQUAD*) &m_Color; }
        const RGBQUAD   *asRGBQUAD  () const { return (const RGBQUAD*) &m_Color; }
        RGBTRIPLE       *asRGBTRIPLE()       { return (RGBTRIPLE*) &m_Color; }
        const RGBTRIPLE *asRGBTRIPLE() const { return (const RGBTRIPLE*) &m_Color; }

        static _Color_ fromColorref( COLORREF cr ) { return _Color_( SWAP_RGBA(cr) ); }
        COLORREF       asColorref() const { return (COLORREF) SWAP_RGBA( m_Color ); }

        static _Color_ fromBpp15( WORD n15 ) { return _Color_( (((n15) & 0x1f) << 3) | (((n15) & 0x3e0) << 6) | (((n15) & 0x7c00) << 9) | 0xff000000 ); }
        WORD           asBpp15() const { return (WORD)( ((WORD)((((BYTE)(m_Color)) >> 3) & 0x1f)) | ((WORD)((((WORD)(m_Color)) >> 6) & 0x3e0)) | ((WORD)(((WORD)((m_Color) >> 9)) & 0x7c00)) ); }

        _Color_ makeOpaque() const { return _Color_( m_Color | 0xff000000 ); }
        _Color_ makeTransparent() const { return _Color_( m_Color & 0x00ffffff ); }
        _Color_ setAlpha( BYTE a ) const { return _Color_( (m_Color & 0x00ffffff) | (a << 24) ); }

            ///<summary> Multiplies the alpha into the color channels. </summary>
        _Color_ PremultiplyAlpha() const
            { BYTE a = getA(); return _Color_( (getR() * a) / 255, (getG() * a) / 255, (getB() * a) / 255, a ); }

        static _Color_ ModulateColors( _Color_ clr1, _Color_ clr2 )
            { return _Color_( (clr1.getR() * clr2.getR()) / 255, (clr1.getG() * clr2.getG()) / 255, (clr1.getB() * clr2.getB()) / 255, (clr1.getA() * clr2.getA()) / 255 ); }
        _Color_        StretchColor    ( float s ) const;
        _Color_        StretchColorTo  ( BYTE max_component ) const;
        _Color_        LiftColor       ( int n ) const;
        _Color_        AddColor        ( _Color_ rgb2 ) const;
        _Color_        SubColor        ( _Color_ rgb2 ) const;
        static _Color_ InterpolateColor( _Color_ rgba1, _Color_ rgba2, float s );
        _Color_        MaximizeColor   ( float lower_amount = 1.0f, float upper_amount = 1.0f ) const;
        _Color_        MaximizeColorTo ( int lower_component = 0, int upper_component = 255 ) const;
        static _Color_ AlphaPaintColor ( _Color_ rgbBkgnd, _Color_ rgbPaint );
        static bool    SimilarColors   ( _Color_ rgb1,  _Color_ rgb2, UINT range );
        static _Color_ GetMinColor     ( _Color_ rgba1, _Color_ rgba2 );
        static _Color_ GetMaxColor     ( _Color_ rgba1, _Color_ rgba2 );
        _Color_        ContrastColor   ( float s ) const { return InterpolateColor( asGray(), *this, s ); } // 0 = gray, 1 = original color

        friend _Color_ operator *( const _Color_& clr, float s ) { return clr.StretchColor( s ); }
        friend _Color_ operator *( float s, const _Color_& clr ) { return clr.StretchColor( s ); }
        _Color_ operator *( const _Color_& clr2 ) const { return ModulateColors( *this, clr2 ); }
        _Color_ operator +( const _Color_& rgb2 ) const { return AddColor( rgb2 ); }
        _Color_ operator -( const _Color_& rgb2 ) const { return SubColor( rgb2 ); }
        friend _Color_ operator +( const _Color_& clr, int n ) { return clr.LiftColor( n ); }
        friend _Color_ operator +( int n, const _Color_& clr ) { return clr.LiftColor( n ); }
        friend _Color_ operator -( const _Color_& clr, int n ) { return clr.LiftColor( -n ); }
        friend _Color_ operator -( int n, const _Color_& clr ) { return clr.LiftColor( -n ); }

        _Color_& operator *=( float s ) { *this = StretchColor( s ); return *this; }
        _Color_& operator *=( const _Color_& clr2 ) { *this = ModulateColors( *this, clr2 ); return *this; }
        _Color_& operator +=( const _Color_& rgb2 ) { *this = AddColor( rgb2 ); return *this; }
        _Color_& operator -=( const _Color_& rgb2 ) { *this = SubColor( rgb2 ); return *this; }
        _Color_& operator +=( int n ) { *this = LiftColor( n ); return *this; }
        _Color_& operator -=( int n ) { *this = LiftColor( -n ); return *this; }

            ///<param name="nIndex"> same param as passed to <b>GetSysColor()</b>. </param>
        static _Color_ fromSysColor( int nIndex ) { return fromColorref( GetSysColor(nIndex) ); }

        static _Color_ Black ( BYTE a = 0 ) { return _Color_(   0,   0,   0, a ); }
        static _Color_ White ( BYTE a = 0 ) { return _Color_( 255, 255, 255, a ); }
        static _Color_ Gray  ( BYTE a = 0 ) { return _Color_( 128, 128, 128, a ); }
        static _Color_ LtGray( BYTE a = 0 ) { return _Color_( 192, 192, 192, a ); }
        static _Color_ Red   ( BYTE a = 0 ) { return _Color_( 255,   0,   0, a ); }
        static _Color_ Green ( BYTE a = 0 ) { return _Color_(   0, 255,   0, a ); }
        static _Color_ Blue  ( BYTE a = 0 ) { return _Color_(   0,   0, 255, a ); }
        static _Color_ Yellow( BYTE a = 0 ) { return _Color_( 255, 255,   0, a ); }
        static _Color_ Pink  ( BYTE a = 0 ) { return _Color_( 255,   0, 255, a ); }
        static _Color_ Cyan  ( BYTE a = 0 ) { return _Color_(   0, 255, 255, a ); }
};




}

#endif
