/*
    File:               argonErrors.h
    Description:        Provides easy error handling as well as easy
                        error definition creation, extension and costumization.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         27.05.2006
    last updated:       05.11.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    Switches:

    _argon_dont_use_SEH     Define _argon_dont_use_SEH if try/catch can also
                            catch asynchronous (hardware) exceptions. This is the
                            case for compiler option /EHac instead of /EHsc.
                            /EHsc usually is the default option.
                            If you don't define _argon_dont_use_SEH, then hardware
                            exceptions will be caught using __try/__except instead
                            of try/catch.
                            Hardware exceptions can be illegal pointer access, etc.

    _argon_no_Lang...       These defines allow excluding of specific languages predefined
                            in the ArgonLibrary.
                            By default, all languages available in the ArgonLibrary will
                            be included. If you define one or some of this set of defines,
                            the appropriate language strings will not be included in
                            the compilation.
                            The following languages and their excluding defines are available:
                                _argon_no_LangEnglish
                                _argon_no_LangGerman
                                _argon_no_LangItalian
*/

#define __argonErrors_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

/*
    The classes have been implemented to ease handling and
    creation of error codes.

    Let me explain 4 main chapters:
        Error codes hierarchy.
        Using error codes.
        Creating new/own error codes.
        Adding new languages/descriptions to existing error codes.


    Error codes hierarchy:
    =========================

    Error codes are structured in 2 hierarchical levels with an
    additional level for the display string representations for
    error codes.

    Highest level is the Error Class level.
    It defines how error code values have to be treated (which ones
    are success values and which ones are failure values).

    Next level is the Error Group level.
    An Error Class can contain several Error Groups where each group
    can use the full range of Error code values (the range of a
    LONG value) without interfering with other groups even of the
    same Error Class.
    Each Error Group is identified by a GUID.

    Last level is the Description Tables level where each Error Group
    can contain one or more description tables.
    These tables contain the display strings to be used for the
    error code values of an Error Group.
    The description tables support different languages, PCTSTR string
    pointers, resource string-IDs and mapping to other error codes
    of other Error Classes or Groups.


    Using error codes:
    ======================

    The main class for this is
        reterr_t

    It holds one error code of a sepcific Error Group and its Error Class.
    It is created by using one of the predefined macros
        interr
        syserr
        hreserr
    or a macro of a user-defined Error Group (see chapter "Creating new/own error codes").

    Some additional classes besides reterr_t are:
        retval_t<>
        retsiz_t
    which incorporate reterr_t with an arbitrary additional value (e.g. of type _size_t_).

    A typical usage could be like this:

        reterr_t MyFunction()
        {
            ....

            if( something_critical )
                return interr (errCritical);

            if( something_failed )
                return interr (errFailed);

            return interr (errOk);
        }

    And a calling code snippet could look like this:

        reterr_t err = MyFunction();

        // Do failure handling.
        if( err.Failed() )
        {
            // on critical error
            if( err == interr (errCritical) )
                { .... }

            // other errors
            ....
        }

    reterr_t also provides basic boolean operators for more easy handling:

        reterr_t MyFunction2()
        {
            ....

            if( something_failed )
                return false;           // same as returning interr (errFailed)

            return true;                // same as returning interr (errOk)
        }

        reterr_t err = MyFunction2();

        if( err )
        {
            // Function returned success.
            ....
        }

        if( !err )
        {
            // Function returned failure.
            ....
        }

    reterr_t also provides methods for retrieving and displaying error messages:

        reterr_t err = MyFunction2();

        if( !err )
            err.DisplayMessage();

    Note:
    When getting or displaying an error description string, it will search for the
    desired language. But if this cannot be found, the implementation will look for
    any other available language and will then return one of those (preferring
    english if available).


    Additional error code features:

    Cascaded error codes:
    ------------------------

    You can cascade error codes if you want to more general error code followed immediately
    by a more detailed error code.

        HRESULT hr = .... ;
        reterr_t err = interr (errCritical) (hreserr (hr));

    In this sample, the err is a errCritical error code, but it also contains a
    more detailed error code, the HRESULT error value.
    When retrieving the message of the err, both errors will be listed.

    It is also possible to create a (bottomless) chain of error codes.

        reterr_t err = interr (errCritical) (hreserr (hr)) (hreserr (hr2)) (syserr (another_error));

    Here a errCritical is cascaded by hr, hr2 and another_error error values.
    which all will be listed in the error message text.

    Note: Cascaded error values are the more specialized case and probably will not be
    used very often in program code. So if a reterr_t value contains also one or more
    cascaded error codes, you should not expect high execution speed. Objects
    will be created and copied.
    So the fastest way is not to use cascaded error codes.

    When you retrieve a message text from a cascaded error code, the message will be formed
    as follows:
    First the message of the main error code (in the samples this is errCritical) will
    be retrieved.
    If this one contains one or more "{m}" sequences, they will be each replaced
    by the message of the next following cascaded error code.
    If one or more "{#}" sequences are found, each one will be replaced by the
    next following cascaded error's group- or classname followed by its error code value
    in hexadecimal and decimal forms.
    If one or more "{f}" sequences are found, each one will be replaced by the
    function name the error occured in. If no function name is available,
    it is left empty.
    If none of these {...} sequences are found in the message string,
    then the cascaded error's message text, group- or classname, and its error code value
    (hex and dec) will be appended to the current message text automatically.


    Additional individual message text:
    --------------------------------------

    As you now know, an error code has a description text.
    You can add additional text to your error code object. It will be used
    only for your current error code object. So it will *not* be set permanently
    for the error code value.

    You can add additional text similar to the way to cascade error codes,
    but instead you pass string pointer:

        reterr_t err = interr (errCritical) (_T("Notice, this is really critical!"));

    Notice, the text will be appended. So if the error code object did already contain
    some additional text, the new text will be appended to the old one (with 2
    newline chars in between to separate them).

    Of course you can combine additional text definition with cascading of error codes:

        reterr_t err = interr (errCritical) (_T("Notice, this is really critical!")) (hreserr (hr)) (hreserr (hr2)) (syserr (another_error));

    Notice, the additional text you specify will be always only appended to the
    top error code object, since all the () operators always return the top error code object.

    But it is possible to add additional text to one of cascaded errors (e.g.
    in the sample above) by including it inside the brackets:

        String tst = "Test-3";
        reterr_t err = interr (errCritical) (_T("Text-1")) (hreserr (hr) (_T("Text-2"))) (hreserr (hr2)) (syserr (another_error) (tst));

    Now the text "Text-1" is, as you know already, appended to the errCritical error object,
    but "Text-2" is appended to the first hreserr object and "Text-3" to the last syserr object.


    Creating new/own error codes:
    =================================

    The creation of new error code values and definitions has been implemented
    in a way to allow easy error creation as well as flexible error creation.
    For this, there are several ways of creating new error definitions.


    The easiest is to use the
    Generic error classes:
    --------------------------

        enum eGenErrMappingType
        struct GenErrDesc
        class GenericErrorClass
        template class GenericErrorDescriptionTableT
        class GenErr

    Let's begin with a sample.
    We make this as most minimalistic as possible.

    Write in your header file myerror.h:

        enum eMy1stErrorCodes
        {
            my1_Ok          = 0,            // if bit 31 is not set, it is a success value. And 0 is interpreted as pure-success value.
            my1_AnError     = 0x80000001,   // if bit 31 is set, it is an error value
            my1_SomeOtherError,
            ....
        };

        class My1stError_Def {  };
        typedef ErrorGroupT<GenericErrorClass, My1stError_Def> My1stErrorGroup;

        #define my1err  ArgonLibrary::GenErr( &My1stErrorGroup::s_Access, __FUNCTION__ )

    Write in your cpp file myerror.cpp:

        const GenErrDesc g_My1stErrorDescs[] = {
            ....
        };

        const GUID                                     g_groupidMy1stError = { .... };    // specify your GUID here.
        My1stErrorGroup                                g_egrMy1st( g_groupidMy1stError ); // error group for our My1st-Errors
        GenericErrorDescriptionTableT<My1stErrorGroup> g_edtMy1st( g_My1stErrorDescs );   // and its description table

    Usage of our new errors:

        reterr_t MyFct()
        {
            ....
            return my1err (my1_AnError);
        }

        if( !MyFct() )
            { .... }    // error occured

    The names with "my1st" are just samples and the g_ prefixes shall show you
    that these variables are global variables.

    Let's begin in the header file:
    The enum contains your newly defined error code values.
    Instead of an enum, you can also use #defines, whatever you like.

    The My1stError_Def class is not used at all. Its only purpose is to
    cause the compiler to create a separate code base for My1stErrorGroup
    independent from other ErrorGroupT<> types.
    For each new Error Group you create, you should make a _Def class
    that is not being used for any other Error Group.

    This is necessary immediately in the following define which uses
    the s_Access static member. Don't care about this, just use at as
    shown in the sample.

    The cpp file:
    The variables in the cpp file do not need to be exported to other cpp
    files or modules.

    The GenErrDesc array contains the strings for the error code values.
    How to write that array, you can see in the array g_InternalErrorDescs
    in argonErrors.cpp.

    Then you need to define a GUID for the new Error Group.
    You can create Guids using tools like guidgen.exe or using the
    COM function CoCreateGuid().
    It is recommended to make a fixed GUID for each new Error Group.
    The sense is to allow users to add new description tables to
    existing Error Groups identified by their GUIDs.

    Next create a global object of the Error Group.
    Here you need to pass the GUID. You can also pass a name for the
    group. This name will appear in Error display messages.
    If you don't provide a name, the name of the Error Class the
    group belongs to (that's the Generic Error Class in this sample)
    will be used.

    Finally we add one Generic Error Description Table to our Error Group.
    Here we pass the g_My1stErrorDescs array.
    It is possible to add more description tables to add new languages or
    to overload existing error value definitions. For more on this, see
    chapter "Adding new languages/descriptions to existing error codes" below.


    More complicated way, with higher type safety:
    -------------------------------------------------

    This is similar to the easiest way shown above, but instead of
    using the #define with GenErr, you derive your own error value class
    from ErrorCode and provide a operator()() with your error value type
    (your enum type).

    For a sample on this, see class IntErr in here.
    It defines an operator()(internal_error_t) and a #define interr
    that expands to IntErr instead of a GenErr.
    The advantage is when using something like
        interr (errOk)
    the compiler will check the type of the passed error value.
    So calling
        interr (E_FAIL)
    will return a compiler error, because E_FAIL is not of type internal_error_t.

    The disadvantage is that you have to write a class derived from ErrorCode.

    The GenErr does not provide this type checking, because it uses
    an operator()(LONG).


    Most complicated, but most flexible way: New Error Class:
    ----------------------------------------------------------

    You can create your own Error Class and for that your own
    Error Group.
    This way have to write additional classes, but you have
    highest flexibility.
    In your Error Class, you can define your own IsSevere and
    IsPureSuccess functions.
    For samples on how to write own Error Classes, have a look
    on SystemErrorClass and SysErr as well as
    HresultErrorClass and HresErr in here.


    Adding new languages/descriptions to existing error codes:
    ============================================================

    For this, you can create new description table(s) for existing
    Error Groups.

    Let's make a sample adding a new language for an internal error code
    and overwriting an existing language of an internal error code.

    Just a cpp file is needed:

        const GenErrDesc g_MyMore_InternalErrorDescs[] = {
            { errOk,     LANG_ITALIAN, gemapt_Direct, (LONG_PTR) _T("Bravissimo!") },
            { errFailed, LANG_ENGLISH, gemapt_Direct, (LONG_PTR) _T("Oh oh. We failed!") },
            GENERRDESC_ENDMARK
        };

        GenericErrorDescriptionTableT<InternalErrorGroup> g_edtMyMoreInternal( g_MyMore_InternalErrorDescs );

    That's it.
    For errOk, we added an italian description string. (Imagine we do not have any italian
    error descriptions defined for internal error values in argonErrors.cpp).
    And for errFailed, we replaced (or at least we want to replace) the english string by a new one.

    The problem is that g_edtMyMoreInternal is a global object as well as the g_edtInternal
    (the default description table for internal errors defined in argonErrors.cpp).
    It cannot be told, which of these 2 objects is initialized first during global
    objects construction since both lie in different cpp files.

    argonErrors uses that one with higher priority, which has been initialized later.
    Well to force our g_edtMyMoreInternal to be of higher priority, we can
    call
        g_edtMyMoreInternal.MakeHighestPriority();
    in our main(), WinMain() or DllMain().

    The priority is important only when existing language description strings shall
    be replaced. If new languages are added, then the priority is unimportant.


    If desired, a description table's GetDescriptionString() method can be overridden
    to implement different algorithms of retrieving the description data (e.g.
    from databases etc.).


    A word on templates:
    =======================

    As you actually will have recognized, most of the classes in here are templates.
    The need for this comes from one of the main required features
    of the error classes:
    The ability to be initialized during global object creation time and
    complete available after that step and even functionality during
    the global object creation time.

    This would not be possible so easily when using just classes with virtual
    method overrides.
    What we need is often static member variables and methods. But we even want
    the static methods overridable. That's possible using templates.
    A good example you can see in template class ErrorClassT.
    This class has to be derived and the static methods IsSevere and IsPureSuccess
    need to be implemented.
    The advantage of static member variables and methods is, that they exist
    even before the objects have been initialized where they are lying in.
    This is necessary because it is not predictable in which order global
    objects are being initialized.
*/

#ifndef __argonErrors_h_Part1
#define __argonErrors_h_Part1

#ifndef _argon_no_includes
    #include "argonBase.h"
    #include "argonLocking.h"
    #include "argonInstance.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




// ____ Internal errors defined by the Argon Library ________________

// Defintion of the specific basic error codes.
// IMPORTANT NOTE: Many of the error codes here have set the language dependent
// message strings in "argonErrors.cpp" in the g_InternalErrorDescs[] array,
// but *NOT* all of them. Some have their strings in other cpp files (e.g.
// the ExpressionEvaluator has the error message strings defined in "argonExpressionEvaluator.cpp").
// But the enum values must be defined in here of cause.
// So if you want error message strings for those errors, you need to include the appropriate
// cpp files of the ArgonLibrary, too.
enum internal_error_t
{
    // Succeeding values

    errOk                       = 0,        // this is also the pure-success value
    suxOk                       = errOk,

    suxNoOp                     = 0x00000001,
    suxMemoryCopyProduced,

    // Failing values (severity bit = highest bit = is set)

    errFailed                   = 0x80000001,
    errCritical,
    errAborted,
    errUserAborted,
    errEndOfData,
    errIllegalMemType,
    errMemoryNotReallocatable,
    errMemoryNotResizeable,
    errMemoryDataLocked,
    errMemoryReallocLocked,
    errMemoryMustBeEmpty,
    errMemoryCouldNotAttachPointer,
    errMemoryCannotDecrease,
    errDynMemWrongDimensions,
    errDynMemNoHeader,
    errOutOfRange,
    errNoDataAvailable,
    errAccessLocked,
    errEOF,
    errFileNotOpen,
    errFileAlreadyOpen,
    errFileReadFailed,
    errFileWriteFailed,
    errFileSeekFailed,
    errFileTellFailed,
    errFileNoStream,
    errFileStreamError,
    errFileUnknownType,
    errFileReadNotEnoughData,
    errMemFileReferenceNotReallocable,
    errRegNotOpen,
    errRegAlreadyOpen,
    errRegDerivedNotOpen,
    errRegDerivedAlreadyOpen,
    errProfileNotOpen,
    errProfileAlreadyOpen,
    errEeEmptyExpression,
    errEeSyntax,
    errEeNoArgs,
    errEeWrongNumArgs,
    errEeUnbalancedBrackets,
    errEeUnknownFunction,
    errEeMissingExpressionEndChar,
};




// ____ Error handling in general ________________________

// Now we define the tryHW and catchHW exception handler macros.
// We use them to catch asynchronous (that is hardware) exceptions
// like illegal pointer access.
// Define _argon_dont_use_SEH if try/catch can also catch asynchronous
// (hardware) exceptions. This is the case for compiler option /EHac instead
// of /EHsc.
// If you don't define _argon_dont_use_SEH, then hardware
// exceptions will be caught using __try/__except instead
// of try/catch.
// Usage:
//      tryHW { .... }
//      catchHW { .... }
#ifndef tryHW
#ifdef _argon_dont_use_SEH
    #define tryHW       try
    #define catchHW     catch(...)
#else
    #define tryHW       __try
    #define catchHW     __except( EXCEPTION_EXECUTE_HANDLER )
    // The disadvantages of using SEH (structured exception handling):
    // - It is not fully designed for C++.
    // - It is not in the C++ standard (so maybe not all compilers support it).
    // - In a function with __try/__except, no local object destructors
    //   can be executed.
    // The advantage:
    // - It can catch hardware exceptions (like illegal pointer access)
    //   even if the compiler option /EHsc is specified (this is the default,
    //   it means that try/catch can catch only synchronous exceptions).
#endif
#endif
// Note: By default (so by not defining _argon_dont_use_SEH) tryHW/catchHW
// map to __try/__except because the compiler does not use SEH for
// try/catch by default (/EHsc).
// IMPORTANT: Since tryHW/catchHW can map to __try/__except, beware of the
// restrictions of __try/__except usage: No local C++ object unwinding (dtor)
// is allowed in the function where the __try/__except statements are.
// In such a case, make sub-functions where in there, the object unwindings
// are done.

template<class> class ErrorClassT;
template<class, class> class ErrorGroupT;
class ErrorCode;
class reterr_t;
class ErrorSpecializedExtras;
class MoreMessageBox;

// Now we do some declarations from "argonString.h".
}
#ifndef _argon_no_includes
    #ifndef _argon_no_CString
    #ifndef _MFC_VER
    // Do this include for a non-MFC project.
    #include <atlstr.h>
    #endif
    #endif
#endif
namespace ArgonLibrary
{
namespace privateString { template< class > class BaseStringShared; }
template< class, typename > class StringT;
#ifndef _argon_no_CString
typedef StringT<CString, TCHAR> String;
#else
typedef StringT<privateString::BaseStringShared<TCHAR>, TCHAR> String;
#endif

// Now we do some declarations from "argonLanguageTable.h".
struct LanguageTable_Access;
class LanguageTableString;

struct ErrLangTableStrStruct
{
    const LanguageTable_Access *m_pAccess;
    LONG_PTR                    m_ID;

    ErrLangTableStrStruct& Init( LONG_PTR id );
    static ErrLangTableStrStruct fromLTS( const LanguageTableString& lts );
};

#define ERRLANGTABLESTRSTRUCT_NULL      { NULL, 0 }

///<summary>
///  LanguageTableString- (Lts) IDs for <b>LanguageTableString</b>
///  and <b>LanguageTableT</b> used for our error classes in here.
///  These are strings like names or formatting strings that we need
///  to support in different languages, too (beside the normal error
///  description strings).
///  But the normal error description strings are independent from
///  <b>LanguageTableString</b> and <b>LanguageTableT</b>. For them,
///  we use a separate language handling implementation
///  (ErrorGroupT and ErrorDescriptionTableT).
///  <br/><br/>
///  Although in argonLanguageTable, it is described that any
///  ID can be used (even 0), we use 0 here is invalid ID for
///  error related language strings.
///  <br/><br/>
///  If you want to add your own IDs to the Error related
///  <b>LanguageTableT</b>, you have 2 possibilities:
///  - use new IDs beginning from <b>errltsid_USER</b>. While this
///    is nice, it is hard to prevent ID collisions with other
///    modules that probably use the same IDs as you are defining.
///  - to solve that, you can pass pointer addresses instead.
///    Just make a (any) global variable and pass its address
///    as ID value. That way, you can prevent ID conflicts within
///    process scope.
///    The disadvantage of this is that the address can change from
///    executable instance to another. So you cannot expect const
///    values in that case.
///</summary>
enum eErrorLtsIds
{
    errltsid_ErrorInformationTitle = 1,
    errltsid_InternalError,
    errltsid_SystemError,
    errltsid_COMError,
    errltsid_UnidentifiedError,
    errltsid_ErrorCode,
    errltsid_InFunction,
    errltsid_inFunction,
    errltsid_PrecedingErrors,

    errltsid_USER = 1024,
};

enum eErrorLangPriority     // note: Higher values have higher priority
{
    errlangprior_none       = 0,
    errlangprior_any,       // any language will be used, because the desired language has not been found (so language is unimportant and will be ignored)
    errlangprior_English,   // english will be searched
    errlangprior_Neutral,   // neutral language will be searched
    errlangprior_PrimaryId, // only primary lang Id will be searched
    errlangprior_LangId,    // primary lang Id and sublang Id will be searched
    errlangprior_Lcid       // primary lang Id, sublang Id and Sort Id will be searched
};

///<summary>
///  Use this template class for making a global variable
///  that will allow a <b>ErrorGroupT</b> (derived) object
///  to retrieve the description string for a specific error
///  code. <br/>
///  You will have to specify as template parameter (<b>ErrorGroup_T</b>)
///  that error class for which you want the description table to be
///  used (that is the name <b>Derived</b> that you pass as template
///  parameter to <b>ErrorClassT</b>).
///  So you can define multiple description tables for one error
///  group which will be identified by the language code.
///  <br/><br/>
///  During runtime, when a description for an error is requested,
///  the error class and group will try to find the best description table to
///  return the error string.
///  So it is not necessary for a description table in a 2nd or 3rd etc.
///  language to implement all error code values. If for one, no
///  error description string is available, the error class or group will try
///  to get it from the other description tables it has.
///</summary>
template< class ErrorGroup_T >
class ErrorDescriptionTableT
{
    template<class, class> friend class ErrorGroupT;

    private :
        ErrorDescriptionTableT *m_prev;
        ErrorDescriptionTableT *m_next;

    public :
        ErrorDescriptionTableT() { LockModuleWide lmw; AddToList(); }
        virtual ~ErrorDescriptionTableT() { LockModuleWide lmw; RemoveFromList(); }

        virtual eErrorLangPriority GetDescriptionString( String& str, LONG code, LCID language ) const = 0;

            ///<summary> Makes this description table as the one with the
            ///  highest priority for its language. </summary>
            ///<remarks> This can make sense if you have more than one
            ///  description table for the <u>same</b> Group-Id (depending
            ///  on the return value from <b>GetGroupId()</b>). </remarks>
        void MakeHighestPriority()
            { LockModuleWide lmw; RemoveFromList(); AddToList(); }

    private :
        void AddToList();
        void RemoveFromList();
};

///<summary>
///  Use this template for making a global variable
///  that will represent a specific error class. <br/>
///  An error class consists of several error groups.
///  While error groups are used to differentiate between error
///  code values, the error class is used to allow different
///  method implementations for identifying an error code as
///  success, failure or pure success so that you do not have
///  to define this for each of the groups again and again.
///  <br/><br/>
///  For your own error class type definition, derive a class from
///  the <b>ErrorClassT</b> template which is then the
///  handler for all operations related to that error class.
///  <br/><br/>
///  The error class type can have a name. Error class names
///  and error group names will be used when displaying error
///  informations and descriptions. The group name will be used
///  only. But if there is no group name defined, the class name in here
///  will be used instead.
///</summary>
template< class Derived >
class ErrorClassT
{
    template<class> friend class ErrorClassT;

    public :
        static ErrLangTableStrStruct s_ltsClassName;

    private :
        static ErrorClassT *s_pointer; // if we are using the correct object

    protected :
        ErrorClassT( const LanguageTableString& ltsClassName );
        ErrorClassT( const ErrLangTableStrStruct& ltsClassName );
        ErrorClassT( const ErrorClassT& src ); // copying shall not be allowed -> no implementation
        ErrorClassT& operator =( const ErrorClassT& src ); // copying shall not be allowed -> no implementation
    public :
        ~ErrorClassT();

    public :
        static bool GetClassName( String& str );
        static bool GetClassName( String& str, LCID language );

        // You must implement the following methods in your derived class:
        //      static bool IsSevere( LONG code );
        //      static bool IsPureSuccess( LONG code );
        // If you don't, a compiler error will occur so that you are forced to.
        // IsSevere:
        //  This method will check if the passed error code represents
        //  a severe error (return true) or a success code (return false).
        // IsPureSuccess:
        //  In here, return true only if the code represents the normal
        //  success value. In general, this will be the zero.
};

// The following struct is intended to be stored only within global
// memory (so as static member of an object).
struct ErrorGroup_Access    // for internal use only
{
    typedef bool (*tfnIsSevere)( LONG code );
    typedef bool (*tfnIsPureSuccess)( LONG code );
    typedef bool (*tfnGetDescriptionString)( String& str, LONG code, LCID language );
    typedef void (*tfnGetDisplayMessage)( String& str, PCTSTR pstrDescriptionString, LONG code, PCSTR pstrFct, LCID language );

    bool                    m_bInitialized;
    ErrLangTableStrStruct   m_ltsClassName;
    ErrLangTableStrStruct   m_ltsGroupName;
    GUID                    m_GroupId;

    tfnIsSevere             m_pfnIsSevere;
    tfnIsPureSuccess        m_pfnIsPureSuccess;
    tfnGetDescriptionString m_pfnGetDescriptionString;
    tfnGetDisplayMessage    m_pfnGetDisplayMessage;

    void Init( const ErrLangTableStrStruct& ltsClassName,
               const ErrLangTableStrStruct& ltsGroupName,
               const GUID&                  groupId,
               tfnIsSevere                  pfnIsSevere,
               tfnIsPureSuccess             pfnIsPureSuccess,
               tfnGetDescriptionString      pfnGetDescriptionString,
               tfnGetDisplayMessage         pfnGetDisplayMessage );

    // Note: Since this struct may be used only as static data,
    // the following methods may be called even *after* #this#
    // struct object has been destructed or *before* Init() has been called,
    // because the memory then will still exist and be valid,
    // EXCEPT: if this object was in a Dll's global space and the Dll
    // has been unloaded.

    bool GetGroupName( String& str ) const;
    bool GetGroupName( String& str, LCID language ) const;
    bool GetClassName( String& str ) const;
    bool GetClassName( String& str, LCID language ) const;

    bool IsSevere     ( LONG code ) const { return m_pfnIsSevere      ? m_pfnIsSevere     ( code ) : true;  }
    bool IsPureSuccess( LONG code ) const { return m_pfnIsPureSuccess ? m_pfnIsPureSuccess( code ) : false; }

    bool GetDescriptionString( String& str, LONG code, LCID language ) const
        { return m_pfnGetDescriptionString ? m_pfnGetDescriptionString( str, code, language ) : false; }

    void GetDisplayMessage( String& str, PCTSTR pstrDescriptionString, LONG code, PCSTR pstrFct, LCID language ) const
        { if( m_pfnGetDisplayMessage ) m_pfnGetDisplayMessage( str, pstrDescriptionString, code, pstrFct, language ); }
};

#define ERRORGROUP_ACCESS_INITDATA  { false, ERRLANGTABLESTRSTRUCT_NULL, ERRLANGTABLESTRSTRUCT_NULL, { 0x00000000, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } }, NULL, NULL, NULL, NULL }

///<summary>
///  Use this template for making a global variable
///  that will represent a specific error group. <br/>
///  For each error group, you can use a LONG's range of
///  error code values independently to other error groups and classes,
///  preventing error value conflicts but enabling
///  easy error value definition and usage.
///  <br/><br/>
///  For your own error group definition, pass a NewGroup class parameter
///  to differentiate between different ErrorGroupT classes.
///  <br/><br/>
///  In addition, you will have to create one or more
///  <b>ErrorDescriptionTableT</b> (derived) objects for this class
///  (pass the <b>ErrorGroupT</b> class to the <b>ErrorDescriptionTableT</b>'s
///  template parameter). From each description table template class,
///  you will then have to create one or more objects.
///  <br/><br/>
///  The error group type can have a name. Error class names
///  and error group names will be used when displaying error
///  informations and descriptions. The group name will be used
///  only. But if there is no group name defined, the class name
///  will be used instead.
///</summary>
template< class ErrorClass_T, class NewGroup >      // pass an arbitrary class as NewGroup which will cause the ErrorGroupT to differ from other ErrorGroupTs. That's the only reason and sense of NewGroup.
class ErrorGroupT
{
    template<class> friend class ErrorDescriptionTableT;

    public :
        static ErrorGroup_Access s_Access;

    private :
        static ErrorDescriptionTableT<ErrorGroupT> *s_pDescsHead;
        static ErrorDescriptionTableT<ErrorGroupT> *s_pDescsTail;   // at the tail is the one with the highest priority

        static ErrorGroupT *s_pointer; // if we are using the correct object

    public :
        ErrorGroupT( const GUID& groupId );
        ErrorGroupT( const GUID& groupId, const LanguageTableString& ltsGroupName );
        ErrorGroupT( const GUID& groupId, const ErrLangTableStrStruct& ltsGroupName );
        ~ErrorGroupT();
    protected :
        ErrorGroupT( const ErrorGroupT& src );              // copying shall not be allowed -> no implementation
        ErrorGroupT& operator =( const ErrorGroupT& src );  // copying shall not be allowed -> no implementation

    public :
        static bool GetDescriptionString( String& str, LONG code, LCID language );
        static void GetDisplayMessage   ( String& str, PCTSTR pstrDescriptionString, LONG code, PCSTR pstrFct, LCID language );

        static bool GetGroupName( String& str );
        static bool GetGroupName( String& str, LCID language );
        static bool GetClassName( String& str );
        static bool GetClassName( String& str, LCID language );

    protected :
        static void GetDefaultDisplayMessage( String& str, PCTSTR pstrDescriptionString, LONG code,
                                              PCTSTR pstrClassName, PCTSTR pstrGroupName, PCSTR pstrFct,
                                              LCID language );
};

///<summary>
///  Derive a class from the <b>ErrorCode</b> class which is then the
///   representation of error codes for a specific error class. <br/>
///  See <b>IntErr</b> or <b>SysErr</b> etc. below for an example.
///  <br/><br/>
///  IMPORTANT: <br/>
///   The derived class must not contain any additional member variables
///   because the <b>reterr_t</b> class will not be able to store them.
///   Therefore, it is not allowed to implement the copy-ctor and
///   operator= in the derived class since it will and must not be used.
///   <br/><br/>
///  IMPORTANT: <br/>
///   Do not pass reterr_t and ErrorCode (and derived) objects over module
///   boundaries unless you are sure that the passing module will not be unloaded
///   before the receiving module has used the passed error objects.
///   Otherwise the ErrorCode.m_pGroup member could point to an Error Group
///   object that was lying in the passing module's global space but does
///   not exist any more when the passing module has been unloaded already.
///   <br/>
///   Only for speed reasons we do not catch illegal pointer access in
///   an exception. Doing so would decrease execution speed dramatically.
///</summary>
class ErrorCode
{
    friend class reterr_t;
    friend class ErrorSpecializedExtras;

    protected :
        ErrorGroup_Access      *m_pGroup;
        LONG                    m_Code;
        PCSTR                   m_pstrFct;      // the function/method name in the source code where the error occured
        ErrorSpecializedExtras *m_pSpecialized; // If not NULL, points to additional stuff like more text or (a chain of) more specialized ErrorCode(s).

    protected :
            ///<param name="pGroup"> This should be the pointer to the single, global error group object for that specific error group. </param>
        ErrorCode( ErrorGroup_Access *pGroup, PCSTR pstrFct )
            { m_pGroup = pGroup; m_Code = 0; m_pstrFct = pstrFct; m_pSpecialized = NULL; }

    // the following will be used *only* in #reterr_t#.
    private :
        ErrorCode() { Ctor_as_Internal( errOk ); }
        explicit ErrorCode( internal_error_t errorcode ) { Ctor_as_Internal( errorcode ); }
            ///<summary> Create the object similar to calling #reterr_t( interr (errorcode) )#
            ///  but without the debugging infos (function). </summary>
        void Ctor_as_Internal( internal_error_t errorcode );
        void Set_as_Internal( internal_error_t errorcode );
    private :   // setting private forbids copying of derived objects
        ErrorCode( const ErrorCode& err );
        ErrorCode& operator =( const ErrorCode& err );
    public :
        ~ErrorCode();

    protected :
        bool IsEqual( const ErrorCode& err2 ) const
        {
            if( m_Code == err2.m_Code && m_pGroup == err2.m_pGroup )
                return m_pSpecialized && err2.m_pSpecialized ? AreSpecializedErrorsEqual(*m_pSpecialized, *err2.m_pSpecialized) : true;

            // If the errors are different, then we check for
            // the pure, absolute success value (the default ok value) explicitely.
            return IsPureSuccess() && err2.IsPureSuccess();
        }

        static bool AreSpecializedErrorsEqual( const ErrorSpecializedExtras& spec1, const ErrorSpecializedExtras& spec2 );

    public :
            ///<summary> <b>operator ()( const ErrorCode&amp; )</b> Appends a specizalized error code (for possibly setting deeper ones), but returns <b>this</b> object again (e.g. to be passed to a <b>reterr_t</b>). </summary>
        ErrorCode& operator() ( const ErrorCode&  ecSpecialized ) { AppendSpecializedError( ecSpecialized ); return *this; }
        ErrorCode& operator() ( const ErrorCode *pecSpecialized ) { AppendSpecializedError( pecSpecialized ); return *this; }
        ErrorCode& operator() ( const reterr_t&  retSpecialized );
        void SetSpecializedError   ( const ErrorCode&  ecSpecialized );
        void SetSpecializedError   ( const ErrorCode *pecSpecialized ) { if( pecSpecialized ) SetSpecializedError( *pecSpecialized ); }
        void AppendSpecializedError( const ErrorCode&  ecSpecialized );
        void AppendSpecializedError( const ErrorCode *pecSpecialized ) { if( pecSpecialized ) AppendSpecializedError( *pecSpecialized ); }
            ///<summary> <b>operator ()( PCTSTR )</b> Appends an additional text to the normal error description text (but only for the current ErrorCode object) </summary>
        ErrorCode& operator() ( PCTSTR pstrText ) { AppendAdditionalText( pstrText ); return *this; }
        void SetAdditionalText   ( PCTSTR pstrText );
        void AppendAdditionalText( PCTSTR pstrText );

        friend bool operator ==( const ErrorCode& err1, const ErrorCode& err2 );
        friend bool operator !=( const ErrorCode& err1, const ErrorCode& err2 );

        bool IsSevere     () const { return m_pGroup->IsSevere( m_Code ); }
        bool IsPureSuccess() const { return m_pGroup->IsPureSuccess( m_Code ); }
        //bool IsSevere     () const { tryHW { return m_pGroup->IsSevere     ( m_Code ); } catchHW { return true;  } }
        //bool IsPureSuccess() const { tryHW { return m_pGroup->IsPureSuccess( m_Code ); } catchHW { return false; } }
        //  Note: IsSevere and IsPureSuccess will be used very often in source code.
        //  Putting a try block will make the machine code much much larger and slower.
        //  So for speed, we do not use a try-block for these.
        //  So be careful to pass an error object from one Dll/Module to another in
        //  the same process when the case can happen that the passing Module is
        //  unloaded earlier than the receiving module.
        //  But if you prevent passing reterr_t and ErrorCode (and derived) objects
        //  between module boundaries, you have no problem.
        //  Otherwise the m_pGroup could point to a group object in a module that has
        //  been unloaded already which will lead to an invalid pointer access.

        bool GetDescriptionString( String& str ) const { return GetDescriptionString( str, GetThreadLocale() ); }
        bool GetDescriptionString( String& str, LCID language, bool bInclCascadedErrors = true ) const;
        void GetDisplayMessage   ( String& str ) const { GetDisplayMessage( str, GetThreadLocale() ); }
        void GetDisplayMessage   ( String& str, LCID language, bool bInclCascadedErrors = true ) const;

        LONG             GetCode       () const { return m_Code; }
        const GUID&      GetGroupId    () const
            { static const GUID guidNull = { 0x00000000, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } }; tryHW { return m_pGroup->m_GroupId; } catchHW { return guidNull; } }
        bool             GetGroupName( String& str ) const;
        bool             GetGroupName( String& str, LCID language ) const;
        bool             GetClassName( String& str ) const;
        bool             GetClassName( String& str, LCID language ) const;
        const ErrorCode *GetSpecializedError() const;       // returns NULL if no specialized error code is set
        void             GetAdditionalText  ( String& text ) const;

    private :
        bool GetSafeDescriptionString( String& str, LCID language ) const;
        void GetSafeDisplayMessage( String& str, const String& desc, LCID language ) const;
};

// A flag for param dwMmbFlags of the MoreMessageBox class taken from file "argonMoreMessageBox.h"
#define MMB_AppendExeNameInCaption  0x00000040  // if set, the MoreMessageBox caption will include the executable name of the application that opened the MessageBox.

///<summary>
///  Use class <b>reterr_t</b> as return value from a function or as error variable.
///  It can act even as a bool value where true means success and false means failure.
///   <br/><br/>
///  IMPORTANT: <br/>
///   Do not pass reterr_t and ErrorCode (and derived) objects over module
///   boundaries unless you are sure that the passing module will not be unloaded
///   before the receiving module has used the passed error objects.
///   Otherwise the ErrorCode.m_pGroup member could point to an Error Group
///   object that was lying in the passing module's global space but does
///   not exist any more when the passing module has been unloaded already.
///   <br/>
///   Only for speed reasons we do not catch illegal pointer access in
///   an exception. Doing so would decrease execution speed dramatically.
///</summary>
class reterr_t
{
    friend class ErrorCode;

    protected :
        ErrorCode m_code;

    public :
        reterr_t()                                                          {  } // def-ctor will create an #errOk# internal error value
        reterr_t( const ErrorCode& err ) : m_code( err )                    {  }
        reterr_t( bool             ok  ) : m_code( ok ? errOk : errFailed ) {  } // used just to intialize the object with an #errOk# or #errFailed# internal error code
        //  Note: The following copy operation we do not define by hand,
        //  because we want automatic memberwise copying.
        //reterr_t( const reterr_t&  err ) : m_code( err.m_code )             {  } // copy-ctor

        bool Succeeded    () const { return !m_code.IsSevere(); }
        bool Failed       () const { return  m_code.IsSevere(); }
            ///<summary> Returns true if the error code is the pure, absolue success value. </summary>
        bool IsPureSuccess() const { return  m_code.IsPureSuccess(); }

        reterr_t& operator =( const ErrorCode& err ) { m_code = err;                                     return *this; }
        reterr_t& operator =( bool             ok  ) { m_code.Set_as_Internal( ok ? errOk : errFailed ); return *this; }
        //  Note: The following copy operation we do not define by hand,
        //  because we want automatic memberwise copying.
        //reterr_t& operator =( const reterr_t&  err ) { m_code = err.m_code;                               return *this; }

        // The |= operator can be used to set an error value to #this#
        // error object only if the operand has a *failure* error value.
        reterr_t& operator |=( const ErrorCode& err ) { if( err.IsSevere() ) { m_code = err;        } return *this; }
        reterr_t& operator |=( const reterr_t&  err ) { if( err.Failed  () ) { m_code = err.m_code; } return *this; }

        bool operator !() const { return Failed(); }
        operator bool()   const { return Succeeded(); }

        friend bool operator ==( const reterr_t&  err1, const reterr_t&  err2 );
        friend bool operator ==( const reterr_t&  err1, const ErrorCode& err2 );
        friend bool operator ==( const ErrorCode& err1, const reterr_t&  err2 );
        friend bool operator !=( const reterr_t&  err1, const reterr_t&  err2 );
        friend bool operator !=( const reterr_t&  err1, const ErrorCode& err2 );
        friend bool operator !=( const ErrorCode& err1, const reterr_t&  err2 );

        reterr_t& Exchange( reterr_t& err )
            { ErrorCode code1 = m_code; m_code = err.m_code; err.m_code = code1; return *this; }

        bool GetDescriptionString( String& str ) const { return m_code.GetDescriptionString( str ); }
        bool GetDescriptionString( String& str, LCID language, bool bInclCascadedErrors = true ) const { return m_code.GetDescriptionString( str, language, bInclCascadedErrors ); }
        void GetDisplayMessage   ( String& str ) const { m_code.GetDisplayMessage( str ); }
        void GetDisplayMessage   ( String& str, LCID language, bool bInclCascadedErrors = true ) const { m_code.GetDisplayMessage( str, language, bInclCascadedErrors ); }

        const reterr_t& DisplayMessage   ( HWND hwndParent = NULL, PCTSTR pstrCaption = NULL, UINT mbType = MB_OK | MB_SETFOREGROUND, DWORD dwMmbFlags = MMB_AppendExeNameInCaption ) const { return DisplayMessage( hwndParent, pstrCaption, mbType, dwMmbFlags, GetThreadLocale() ); }
        const reterr_t& DisplayMessage   ( HWND hwndParent, PCTSTR pstrCaption, UINT mbType, DWORD dwMmbFlags, LCID language,
                                           PCTSTR pstrAdditionalText = NULL,
                                           PCTSTR pstrButton1 = NULL, PCTSTR pstrButton2 = NULL, PCTSTR pstrButton3 = NULL, PCTSTR pstrMoreButton = NULL,
                                           PCTSTR pstrCheck1 = NULL,
                                           PCTSTR pstrHeader = NULL, PCTSTR pstrExtraMsg = NULL, PCTSTR pstrExtraLink_Text = NULL, PCTSTR pstrExtraLink_Url = NULL,
                                           HMODULE hModuleToLoadIconsFrom = NULL, UINT_PTR nMsgIcon_IdOrHicon = 0, UINT_PTR nExtraIcon_IdOrHicon = 0,
                                           MoreMessageBox *pMmb = NULL ) const;
        INT_PTR         DisplayMessageRet( HWND hwndParent = NULL, PCTSTR pstrCaption = NULL, UINT mbType = MB_OK | MB_SETFOREGROUND, DWORD dwMmbFlags = MMB_AppendExeNameInCaption ) const { return DisplayMessageRet( hwndParent, pstrCaption, mbType, dwMmbFlags, GetThreadLocale() ); }
        INT_PTR         DisplayMessageRet( HWND hwndParent, PCTSTR pstrCaption, UINT mbType, DWORD dwMmbFlags, LCID language,
                                           PCTSTR pstrAdditionalText = NULL,
                                           PCTSTR pstrButton1 = NULL, PCTSTR pstrButton2 = NULL, PCTSTR pstrButton3 = NULL, PCTSTR pstrMoreButton = NULL,
                                           PCTSTR pstrCheck1 = NULL,
                                           PCTSTR pstrHeader = NULL, PCTSTR pstrExtraMsg = NULL, PCTSTR pstrExtraLink_Text = NULL, PCTSTR pstrExtraLink_Url = NULL,
                                           HMODULE hModuleToLoadIconsFrom = NULL, UINT_PTR nMsgIcon_IdOrHicon = 0, UINT_PTR nExtraIcon_IdOrHicon = 0,
                                           MoreMessageBox *pMmb = NULL ) const;

        LONG        GetCode   () const { return m_code.GetCode();    }
        const GUID& GetGroupId() const { return m_code.GetGroupId(); }
        PCSTR       GetFct    () const { return m_code.m_pstrFct;    }

        const ErrorCode& GetContent         () const { return m_code; }
        const ErrorCode *GetSpecializedError() const { return m_code.GetSpecializedError(); }   // returns NULL if no specialized error code is set
        void             GetAdditionalText  ( String& text ) const { m_code.GetAdditionalText( text ); }
};

///<summary>
///  This template can be used in the case you want to return a reterr_t error
///  additional to a value, pointer or object.
///</summary>
template< class T >
class retval_t
{
    protected :
        reterr_t m_error;
        T        m_value;

    public :
        retval_t() {  }
        retval_t( bool             b,   T val ) : m_value(val), m_error(b)   {  }
        retval_t( const ErrorCode& err, T val ) : m_value(val), m_error(err) {  }
        retval_t( const reterr_t&  err, T val ) : m_value(val), m_error(err) {  }
        retval_t( const retval_t&  src, T val ) : m_value(val), m_error(src.m_error) {  }
        //  Note: The following copy operations we do not define by hand,
        //  because we want automatic memberwise copying.
        //retval_t( const retval_t& src ) : m_value(src.m_value), m_error(src.m_error) {  }
        //retval_t& operator =( const retval_t& src ) { m_value = src.m_value; m_error = src.m_error; return *this; }

        T               getValue     () const { return m_value; }
        const reterr_t& getError     () const { return m_error; }
        bool            Succeeded    () const { return m_error.Succeeded(); }
        bool            Failed       () const { return m_error.Failed(); }
        bool            IsPureSuccess() const { return m_error.IsPureSuccess(); }

        operator T() const        { return m_value; }
        operator reterr_t() const { return m_error; }

    // to check for success and error
        operator bool() const   { return  m_error; }    // returns true if succeeded
        bool operator !() const { return !m_error; }    // returns true if failed

    // comparison with a T value to compare the contained value
        friend bool operator ==( const retval_t& r1, T               r2 ) { return r1.m_value == r2        ; }
        friend bool operator ==( T               r1, const retval_t& r2 ) { return r1         == r2.m_value; }
        friend bool operator !=( const retval_t& r1, T               r2 ) { return r1.m_value != r2        ; }
        friend bool operator !=( T               r1, const retval_t& r2 ) { return r1         != r2.m_value; }
        friend bool operator < ( const retval_t& r1, T               r2 ) { return r1.m_value <  r2        ; }
        friend bool operator < ( T               r1, const retval_t& r2 ) { return r1         <  r2.m_value; }
        friend bool operator > ( const retval_t& r1, T               r2 ) { return r1.m_value >  r2        ; }
        friend bool operator > ( T               r1, const retval_t& r2 ) { return r1         >  r2.m_value; }
        friend bool operator <=( const retval_t& r1, T               r2 ) { return r1.m_value <= r2        ; }
        friend bool operator <=( T               r1, const retval_t& r2 ) { return r1         <= r2.m_value; }
        friend bool operator >=( const retval_t& r1, T               r2 ) { return r1.m_value >= r2        ; }
        friend bool operator >=( T               r1, const retval_t& r2 ) { return r1         >= r2.m_value; }

    // comparison with error values
        friend bool operator ==( const retval_t&  err1, const reterr_t&  err2 ) { return err1.getError() == err2; }
        friend bool operator ==( const reterr_t&  err1, const retval_t&  err2 ) { return err1 == err2.getError(); }
        friend bool operator ==( const retval_t&  err1, const ErrorCode& err2 ) { return err1.getError() == err2; }
        friend bool operator ==( const ErrorCode& err1, const retval_t&  err2 ) { return err1 == err2.getError(); }
        friend bool operator !=( const retval_t&  err1, const reterr_t&  err2 ) { return err1.getError() != err2; }
        friend bool operator !=( const reterr_t&  err1, const retval_t&  err2 ) { return err1 != err2.getError(); }
        friend bool operator !=( const retval_t&  err1, const ErrorCode& err2 ) { return err1.getError() != err2; }
        friend bool operator !=( const ErrorCode& err1, const retval_t&  err2 ) { return err1 != err2.getError(); }
};

typedef retval_t<_size_t_> retsiz_t;




// ____ Generic error classes for user-defined error groups ________________

enum eGenErrMappingType {
    // 0 is used as end-of-array indicator (see GENERRDESC_ENDMARK).
    gemapt_Direct = 1,      // m_DescInfo is a PCTSTR or a String-ID
    gemapt_SysErr = 32,     // m_DescInfo is a System-Error's Error-Code.
    gemapt_HresErr,         // m_DescInfo is a HRESULT Error-Code.
};

struct GenErrDesc
{
    LONG     m_Code;        // the Error-Code value
    LCID     m_Locale;      // specifying LANG_NETURAL will indicate to use this entry for *any* current language
    LONG     m_MappingType; // a gemapt_* (eGenErrMappingType) value
    LONG_PTR m_DescInfo;
    // Note: if m_MappingType is zero -> end of array
};

#define GENERRDESC_ENDMARK      { 0, 0, 0, 0 }

class GenericErrorClass : public ErrorClassT< GenericErrorClass >
{
    public :
        GenericErrorClass()
            : ErrorClassT< GenericErrorClass > ( ErrLangTableStrStruct().Init(errltsid_InternalError) )
        {  }

        static bool IsSevere( LONG code )
        { return (LONG32) code < 0; }

        static bool IsPureSuccess( LONG code )
        { return code == 0; }
};

template< class ErrorGroup_T >
class GenericErrorDescriptionTableT : public ErrorDescriptionTableT< ErrorGroup_T >
{
    protected :
        const GenErrDesc *m_pDescs;

    public :
        GenericErrorDescriptionTableT( const GenErrDesc *pDescs )
        { m_pDescs = pDescs; }

        virtual eErrorLangPriority GetDescriptionString( String& str, LONG code, LCID language ) const;
};

class GenErr : public ErrorCode
{
    public :
        GenErr( ErrorGroup_Access *pGroup, PCSTR pstrFct )
            : ErrorCode( pGroup, pstrFct )
        {  }

        ErrorCode& operator()( LONG code )
        { m_Code = code; return *this; }
};




// ____ Internal errors defined by the Argon Library ________________

// Note: The Error code values have been defined on top of this file.

#define GROUPID_INTERNALERROR       { 0xbbca9530, 0xcd23, 0x476e, { 0x97, 0xdf, 0xff, 0x52, 0xa2, 0x95, 0x9d, 0xd3 } }  // {BBCA9530-CD23-476E-97DF-FF52A2959DD3}
#define GROUPID_INTERNALERROR_STR   "{BBCA9530-CD23-476E-97DF-FF52A2959DD3}"
extern const GUID g_groupidInternalError;

class InternalError_Def {  };
typedef ErrorGroupT<GenericErrorClass, InternalError_Def> InternalErrorGroup;

class IntErr : public ErrorCode
{
    public :
        IntErr( PCSTR pstrFct )
            : ErrorCode( &InternalErrorGroup::s_Access, pstrFct )
        {  }

        ErrorCode& operator()( internal_error_t code )
        { m_Code = (LONG) (ULONG) code; return *this; }
};

#define interr          ArgonLibrary::IntErr( __FUNCTION__ )




// ____ System (Win32) errors (ERROR_*) ____________________

#define GROUPID_SYSTEMERROR         { 0x66f20740, 0xe9fd, 0x4e4b, { 0x9e, 0xb7, 0x9b, 0xa5, 0x0b, 0xe1, 0xe6, 0xa3 } }  // {66F20740-E9FD-4E4B-9EB7-9BA50BE1E6A3}
#define GROUPID_SYSTEMERROR_STR     "{66F20740-E9FD-4E4B-9EB7-9BA50BE1E6A3}"
extern const GUID g_groupidSystemError;

class SystemErrorClass : public ErrorClassT< SystemErrorClass >
{
    public :
        SystemErrorClass()
            : ErrorClassT< SystemErrorClass > ( ErrLangTableStrStruct().Init(errltsid_SystemError) )
        {  }

        static bool IsSevere( LONG code )
        { return code != ERROR_SUCCESS; }

        static bool IsPureSuccess( LONG code )
        { return code == 0; }
};

class SystemError_Def {  };
typedef ErrorGroupT<SystemErrorClass, SystemError_Def> SystemErrorGroup;

class SystemErrorDescriptionTable : public ErrorDescriptionTableT< SystemErrorGroup >
{
    public :
        SystemErrorDescriptionTable() {  }

        virtual eErrorLangPriority GetDescriptionString( String& str, LONG code, LCID language ) const;
};

class SysErr : public ErrorCode
{
    public :
        SysErr( PCSTR pstrFct )
            : ErrorCode( &SystemErrorGroup::s_Access, pstrFct )
        {  }

        ErrorCode& operator()( LONG code )
        { m_Code = code; return *this; }
};

#define syserr          ArgonLibrary::SysErr( __FUNCTION__ )




// ____ COM (HRESULT) errors (E_* etc.) ____________________

#define GROUPID_HRESULTERROR        { 0xd32cfb91, 0xa7bc, 0x4a3b, { 0x98, 0x3c, 0xa4, 0xb0, 0xa5, 0x5c, 0x9c, 0x8f } }  // {D32CFB91-A7BC-4A3B-983C-A4B0A55C9C8F}
#define GROUPID_HRESULTERROR_STR    "{D32CFB91-A7BC-4A3B-983C-A4B0A55C9C8F}"
extern const GUID g_groupidHresultError;

class HresultErrorClass : public ErrorClassT< HresultErrorClass >
{
    public :
        HresultErrorClass()
            : ErrorClassT< HresultErrorClass > ( ErrLangTableStrStruct().Init(errltsid_COMError) )
        {  }

        static bool IsSevere( LONG code )
        { return FAILED( code ); }

        static bool IsPureSuccess( LONG code )
        { return (HRESULT) code == S_OK; }
};

class HresultError_Def {  };
typedef ErrorGroupT<HresultErrorClass, HresultError_Def> HresultErrorGroup;

class HresultErrorDescriptionTable : public ErrorDescriptionTableT< HresultErrorGroup >
{
    public :
        HresultErrorDescriptionTable() {  }

        virtual eErrorLangPriority GetDescriptionString( String& str, LONG code, LCID language ) const;
};

class HresErr : public ErrorCode
{
    public :
        HresErr( PCSTR pstrFct )
            : ErrorCode( &HresultErrorGroup::s_Access, pstrFct )
        {  }

        ErrorCode& operator()( HRESULT code )
        { m_Code = (LONG) (ULONG) code; return *this; }
};

#define hreserr         ArgonLibrary::HresErr( __FUNCTION__ )




}

#endif




#ifndef _argonErrors_no_Part2   // this is a switch used only by argonMemory.h
#ifndef __argonErrors_h_Part2
#define __argonErrors_h_Part2

#ifndef _argon_no_includes
    #include "argonString.h"
    #include "argonLanguageTable.h"
    #include "argonMoreMessageBox.h"
#endif

namespace ArgonLibrary
{




class ErrorSpecializedExtras
{
    public :
    // The more specialized (chain of) ErrorCode(s)
        ErrorCode   m_SpecializedError; // if #m_bSpecializedErrorSet# is true, points to a more detailed error code (e.g. useful if an own error value may reference a bunch of HRESULT error values). Can be cascaded.
                                        // Note: When an ErrorCode is copied/assigned to anotherone, a copy of the specialized object will be created.
        bool        m_bSpecializedErrorSet;

    // Additional Text for the current ErrorCode
        String      m_strAdditionalText; // additional text appended to the normal error description text, but only for the current ErrorCode object.

    public :
        ErrorSpecializedExtras() { m_bSpecializedErrorSet = false; }
        ErrorSpecializedExtras( const ErrorCode& ec ) : m_SpecializedError(ec) { m_bSpecializedErrorSet = true; }
        //  Note: The following copy operation we do not define by hand,
        //  because we want automatic memberwise copying.
        //ErrorSpecializedExtras( const ErrorSpecializedExtras& src )
        //    : m_SpecializedError(src.m_SpecializedError), m_strAdditionalText(src.m_strAdditionalText)
        //{ m_bSpecializedErrorSet = src.m_bSpecializedErrorSet; }

        //  Note: The following copy operation we do not define by hand,
        //  because we want automatic memberwise copying.
        //ErrorSpecializedExtras& operator =( const ErrorSpecializedExtras& src )
        //{ m_SpecializedError = src.m_SpecializedError; m_bSpecializedErrorSet = src.m_bSpecializedErrorSet; m_strAdditionalText = src.m_strAdditionalText; }

        void SetSpecializedError( const ErrorCode& ec )
            { m_SpecializedError = ec; m_bSpecializedErrorSet = true; }
};




class ErrorsLanguageTable_Def {  };
typedef LanguageTableT<ErrorsLanguageTable_Def> ErrorsLanguageTable;




// ____ Implementations _______________________________

//      ErrLangTableStrStruct

inline ErrLangTableStrStruct& ErrLangTableStrStruct::Init( LONG_PTR id )
{
    m_pAccess = &ErrorsLanguageTable::s_Access;
    m_ID      = id;
    return *this;
}

inline ErrLangTableStrStruct ErrLangTableStrStruct::fromLTS( const LanguageTableString& lts )
{
    ErrLangTableStrStruct ret;

    ret.m_pAccess = lts.getAccessPtr();
    ret.m_ID = lts.getID();

    return ret;
}


//      ErrorDescriptionTableT<>

template< class ErrorGroup_T >
void ErrorDescriptionTableT<ErrorGroup_T>::AddToList()
{
    if( m_prev = ErrorGroup_T::s_pDescsTail ) m_prev->m_next = this;
    m_next = NULL;
    if( !ErrorGroup_T::s_pDescsHead ) ErrorGroup_T::s_pDescsHead = this;
    ErrorGroup_T::s_pDescsTail = this;
}

template< class ErrorGroup_T >
void ErrorDescriptionTableT<ErrorGroup_T>::RemoveFromList()
{
    if( m_prev ) m_prev->m_next = m_next;
    else         ErrorGroup_T::s_pDescsHead = m_next;
    if( m_next ) m_next->m_prev = m_prev;
    else         ErrorGroup_T::s_pDescsTail = m_prev;
}

//      ErrorClassT<>

template< class Derived > ErrLangTableStrStruct ErrorClassT<Derived>::s_ltsClassName = ERRLANGTABLESTRSTRUCT_NULL;
template< class Derived > ErrorClassT<Derived> *ErrorClassT<Derived>::s_pointer      = NULL;

template< class Derived >
ErrorClassT<Derived>::ErrorClassT( const LanguageTableString& ltsClassName )
{
    LockModuleWide lmw;

    if( !s_pointer )
    {
        s_ltsClassName = ErrLangTableStrStruct::fromLTS( ltsClassName );
        s_pointer = this;
    }
}

template< class Derived >
ErrorClassT<Derived>::ErrorClassT( const ErrLangTableStrStruct& ltsClassName )
{
    LockModuleWide lmw;

    if( !s_pointer )
    {
        s_ltsClassName = ltsClassName;
        s_pointer = this;
    }
}

template< class Derived >
ErrorClassT<Derived>::~ErrorClassT()
{
    LockModuleWide lmw;

    // The contents of #s_Access# we do not deinit because
    // first of all there is no need to deinit, plus we still
    // want its content to be valid even after #this# object's
    // and/or the #s_Access# object's dtor have been called.
    // Since s_Access is static, it lies in the global memory
    // space and its memory location will still be valid then,
    // EXCEPT if the global space was in an unloaded Dll.

    if( s_pointer == this )
        s_pointer = NULL;
}

template< class Derived >
bool ErrorClassT<Derived>::GetClassName( String& str )
{
    LanguageTableString lts( s_ltsClassName.m_pAccess, s_ltsClassName.m_ID );
    return lts.getStringThreadLocale( str );
}

template< class Derived >
bool ErrorClassT<Derived>::GetClassName( String& str, LCID language )
{
    LanguageTableString lts( s_ltsClassName.m_pAccess, s_ltsClassName.m_ID );
    return lts.getString( str, language );
}

//      ErrorGroupT<>

template< class ErrorClass_T, class NewGroup > ErrorGroup_Access                                              ErrorGroupT<ErrorClass_T, NewGroup>::s_Access     = ERRORGROUP_ACCESS_INITDATA;
template< class ErrorClass_T, class NewGroup > ErrorDescriptionTableT< ErrorGroupT<ErrorClass_T, NewGroup> > *ErrorGroupT<ErrorClass_T, NewGroup>::s_pDescsHead = NULL;
template< class ErrorClass_T, class NewGroup > ErrorDescriptionTableT< ErrorGroupT<ErrorClass_T, NewGroup> > *ErrorGroupT<ErrorClass_T, NewGroup>::s_pDescsTail = NULL;
template< class ErrorClass_T, class NewGroup > ErrorGroupT<ErrorClass_T, NewGroup>                           *ErrorGroupT<ErrorClass_T, NewGroup>::s_pointer    = NULL;

template< class ErrorClass_T, class NewGroup > 
ErrorGroupT<ErrorClass_T, NewGroup>::ErrorGroupT( const GUID& groupId )
{
    LockModuleWide lmw;

    if( !s_pointer )
    {
        ErrLangTableStrStruct lts = ERRLANGTABLESTRSTRUCT_NULL;

        s_Access.Init( ErrorClass_T::s_ltsClassName,
                       lts,
                       groupId,
                       ErrorClass_T::IsSevere,
                       ErrorClass_T::IsPureSuccess,
                       GetDescriptionString,
                       GetDisplayMessage );

        s_pointer = this;
    }
}

template< class ErrorClass_T, class NewGroup > 
ErrorGroupT<ErrorClass_T, NewGroup>::ErrorGroupT( const GUID& groupId, const LanguageTableString& ltsGroupName )
{
    LockModuleWide lmw;

    if( !s_pointer )
    {
        s_Access.Init( ErrorClass_T::s_ltsClassName,
                       ErrLangTableStrStruct::fromLTS( ltsGroupName ),
                       groupId,
                       ErrorClass_T::IsSevere,
                       ErrorClass_T::IsPureSuccess,
                       GetDescriptionString,
                       GetDisplayMessage );

        s_pointer = this;
    }
}

template< class ErrorClass_T, class NewGroup > 
ErrorGroupT<ErrorClass_T, NewGroup>::ErrorGroupT( const GUID& groupId, const ErrLangTableStrStruct& ltsGroupName )
{
    LockModuleWide lmw;

    if( !s_pointer )
    {
        s_Access.Init( ErrorClass_T::s_ltsClassName,
                       ltsGroupName,
                       groupId,
                       ErrorClass_T::IsSevere,
                       ErrorClass_T::IsPureSuccess,
                       GetDescriptionString,
                       GetDisplayMessage );

        s_pointer = this;
    }
}

template< class ErrorClass_T, class NewGroup > 
ErrorGroupT<ErrorClass_T, NewGroup>::~ErrorGroupT()
{
    LockModuleWide lmw;

    // The contents of #s_Access# we do not deinit because
    // first of all there is no need to deinit, plus we still
    // want its content to be valid even after #this# object's
    // and/or the #s_Access# object's dtor have been called.
    // Since s_Access is static, it lies in the global memory
    // space and its memory location will still be valid then,
    // EXCEPT if the global space was in an unloaded Dll.

    if( s_pointer == this )
        s_pointer = NULL;

    // Remove all description tables.
    while( s_pDescsHead )
        s_pDescsHead->RemoveFromList();
}

template< class ErrorClass_T, class NewGroup > 
bool ErrorGroupT<ErrorClass_T, NewGroup>::GetDescriptionString( String& str, LONG code, LCID language )
{
    // We scan all description tables in reverse order
    //  (because of priority handling) till one returns a
    //  description string successfully matching as close
    //  to the specified language as possible.

    LockModuleWide lmw;

    ErrorDescriptionTableT<ErrorGroupT> *table;
    eErrorLangPriority                   ret = errlangprior_none;

    for( table = s_pDescsTail ; table ; table = table->m_prev )
    {
        String s;
        eErrorLangPriority pri = table->GetDescriptionString( s, code, language );

        if( pri > ret )
            { ret = pri; str = s; }
    }

    return ret > errlangprior_none;
}

template< class ErrorClass_T, class NewGroup > 
void ErrorGroupT<ErrorClass_T, NewGroup>::GetDisplayMessage( String& str, PCTSTR pstrDescriptionString, LONG code, PCSTR pstrFct, LCID language )
{
    LanguageTableString ltsCN( s_Access.m_ltsClassName.m_pAccess, s_Access.m_ltsClassName.m_ID );
    LanguageTableString ltsGN( s_Access.m_ltsGroupName.m_pAccess, s_Access.m_ltsGroupName.m_ID );

    GetDefaultDisplayMessage( str, pstrDescriptionString, code,
                              ltsCN.getString( language ), ltsGN.getString( language ), pstrFct,
                              language );
}

template< class ErrorClass_T, class NewGroup >
bool ErrorGroupT<ErrorClass_T, NewGroup>::GetGroupName( String& str )
{
    LanguageTableString lts( s_Access.m_ltsGroupName.m_pAccess, s_Access.m_ltsGroupName.m_ID );
    return lts.getStringThreadLocale( str );
}

template< class ErrorClass_T, class NewGroup >
bool ErrorGroupT<ErrorClass_T, NewGroup>::GetGroupName( String& str, LCID language )
{
    LanguageTableString lts( s_Access.m_ltsGroupName.m_pAccess, s_Access.m_ltsGroupName.m_ID );
    return lts.getString( str, language );
}

template< class ErrorClass_T, class NewGroup >
bool ErrorGroupT<ErrorClass_T, NewGroup>::GetClassName( String& str )
{
    LanguageTableString lts( s_Access.m_ltsClassName.m_pAccess, s_Access.m_ltsClassName.m_ID );
    return lts.getStringThreadLocale( str );
}

template< class ErrorClass_T, class NewGroup >
bool ErrorGroupT<ErrorClass_T, NewGroup>::GetClassName( String& str, LCID language )
{
    LanguageTableString lts( s_Access.m_ltsClassName.m_pAccess, s_Access.m_ltsClassName.m_ID );
    return lts.getString( str, language );
}

template< class ErrorClass_T, class NewGroup >
void ErrorGroupT<ErrorClass_T, NewGroup>::GetDefaultDisplayMessage( String& str, PCTSTR pstrDescriptionString, LONG code,
                                                                    PCTSTR pstrClassName, PCTSTR pstrGroupName, PCSTR pstrFct,
                                                                    LCID language )
{
    str.Empty();

    if( pstrFct )
    {
        str += ErrorsLanguageTable::getString( errltsid_InFunction, language );
        str += _T(' ');
        str += pstrFct;
        str += _T(":\n");
    }

    String strErrorName = pstrGroupName;

    if( strErrorName.IsEmpty() )
        strErrorName = pstrClassName;

    if( strErrorName.IsEmpty() )
        strErrorName = LanguageTableString( &ErrorsLanguageTable::s_Access, errltsid_UnidentifiedError ).getString( language );

    str += strErrorName;
    str += String().Sprintf( _T(",\n%s 0x%lx (%ld)"),
                             (PCTSTR) LanguageTableString( &ErrorsLanguageTable::s_Access, errltsid_ErrorCode ).getString( language ),
                             code, code );

    if( !_isstrempty_( pstrDescriptionString ) )
    {
        str += _T("\n\n");
        str += pstrDescriptionString;
    }
}

//      ErrorCode

///<summary> Create the object similar to calling <b>reterr_t( interr (errorcode) )</b>
///  but without the debugging infos (current function). </summary>
inline void ErrorCode::Ctor_as_Internal( internal_error_t errorcode )
{
    m_pGroup = &InternalErrorGroup::s_Access;
    m_Code = (LONG) (ULONG) errorcode;
    m_pstrFct = NULL;
    m_pSpecialized = NULL;
}

inline void ErrorCode::Set_as_Internal( internal_error_t errorcode )
    { if( m_pSpecialized ) delete m_pSpecialized; Ctor_as_Internal( errorcode ); }

inline ErrorCode::ErrorCode( const ErrorCode& err )
{
    m_pGroup = err.m_pGroup;
    m_Code = err.m_Code;
    m_pstrFct = err.m_pstrFct;
    m_pSpecialized = err.m_pSpecialized ? new ErrorSpecializedExtras( *err.m_pSpecialized ) : NULL;
}

inline ErrorCode& ErrorCode::operator =( const ErrorCode& err )
{
    m_pGroup = err.m_pGroup;
    m_Code = err.m_Code;
    m_pstrFct = err.m_pstrFct;
    if( m_pSpecialized ) delete m_pSpecialized;
    m_pSpecialized = err.m_pSpecialized ? new ErrorSpecializedExtras( *err.m_pSpecialized ) : NULL;
    return *this;
}

inline ErrorCode::~ErrorCode()
    { if( m_pSpecialized ) delete m_pSpecialized; }

inline ErrorCode& ErrorCode::operator() ( const reterr_t& retSpecialized )
    { AppendSpecializedError( retSpecialized.m_code ); return *this; }

///<remarks>
///  Since this method is used only by <b>ErrorCode::IsEqual()</b>, we implement it
///  exactly matching for those needs. And it needs true to be returned if one of
///  the specialized extras does not have set a specialized error.
///</remarks>
inline bool ErrorCode::AreSpecializedErrorsEqual( const ErrorSpecializedExtras& spec1, const ErrorSpecializedExtras& spec2 )
{
    return spec1.m_bSpecializedErrorSet && spec2.m_bSpecializedErrorSet ?
        spec1.m_SpecializedError == spec2.m_SpecializedError : true;    // the returned true is correct here.
}

inline bool operator ==( const ErrorCode& err1, const ErrorCode& err2 ) { return  err1.IsEqual( err2 ); }
inline bool operator !=( const ErrorCode& err1, const ErrorCode& err2 ) { return !err1.IsEqual( err2 ); }

//      reterr_t

inline bool operator ==( const reterr_t&  err1, const reterr_t&  err2 ) { return err1.m_code == err2.m_code; }
inline bool operator ==( const reterr_t&  err1, const ErrorCode& err2 ) { return err1.m_code == err2; }
inline bool operator ==( const ErrorCode& err1, const reterr_t&  err2 ) { return err1 == err2.m_code; }
inline bool operator !=( const reterr_t&  err1, const reterr_t&  err2 ) { return err1.m_code != err2.m_code; }
inline bool operator !=( const reterr_t&  err1, const ErrorCode& err2 ) { return err1.m_code != err2; }
inline bool operator !=( const ErrorCode& err1, const reterr_t&  err2 ) { return err1 != err2.m_code; }

//      GenericErrorDescriptionTableT<>

template< class ErrorGroup_T >
eErrorLangPriority GenericErrorDescriptionTableT<ErrorGroup_T>::GetDescriptionString( String& str, LONG code, LCID language ) const
{
    // #language# contains PRIMARYLANGID, SUBLANGID and SORTID.
    LCID   want_lcid         = language;
    LANGID want_lang_id      = LANGIDFROMLCID( want_lcid ); // the combined form with PRIMARYLANDID and SUBLANGID
    LANGID want_prim_lang_id = PRIMARYLANGID( want_lang_id );
    //LANGID want_sub_lang_id  = SUBLANGID( want_lang_id );
    //WORD   want_sort_id      = SORTIDFROMLCID( want_lcid );

    if( !m_pDescs ) return errlangprior_none;
    eErrorLangPriority priret = errlangprior_none;

    for( int i = 0 ;; ++i )
    {
        if( m_pDescs[i].m_MappingType == 0 )
            break;

        if( m_pDescs[i].m_Code == code )
        {
            LCID   lcid            = m_pDescs[i].m_Locale;
            LANGID lang_id         = LANGIDFROMLCID( lcid );
            LANGID prim_lang_id    = PRIMARYLANGID( lang_id );
            eErrorLangPriority pri = errlangprior_none;

            // If the entry has LANG_NEUTRAL exactly, then this indicates
            // that any language may be used for this entry.
            // But *only* when we did not yet find description string for
            // the desired error code.
            if( lcid == LANG_NEUTRAL && priret == errlangprior_none )
                pri = errlangprior_Lcid;

            // First we try to find a match for PRIMARYLANGID, SUBLANGID
            //  and SORTID alltogether.
            if( pri < errlangprior_Lcid && lcid == want_lcid )
                pri = errlangprior_Lcid;

            // Now we try to find a match for PRIMARYLANGID and SUBLANGID.
            if( pri < errlangprior_LangId && lang_id == want_lang_id )
                pri = errlangprior_LangId;

            // Now we try to find a match for PRIMARYLANGID only.
            if( pri < errlangprior_PrimaryId && prim_lang_id == want_prim_lang_id )
                pri = errlangprior_PrimaryId;

            // Now we try to find a match for LANG_NEUTRAL as
            //  primary language id.
            if( pri < errlangprior_Neutral && prim_lang_id == LANG_NEUTRAL )
                pri = errlangprior_Neutral;

            // Now we try to find a match for LANG_ENGLISH as
            //  primary language id.
            if( pri < errlangprior_English && prim_lang_id == LANG_ENGLISH )
                pri = errlangprior_English;

            // Now we try to get a description string regardless
            //  of its language.
            if( pri < errlangprior_any )
                pri = errlangprior_any;

            // Now we get the description string
            // according to the mapping type.
            if( pri > priret )
                switch( m_pDescs[i].m_MappingType )
                {
                    case gemapt_Direct :
                    {
                        if( (PCTSTR) m_pDescs[i].m_DescInfo && IS_INTRESOURCE(m_pDescs[i].m_DescInfo) && want_lang_id != LANG_NEUTRAL )
                        {
                            String s;
                            if( !s.Loadstring( (UINT) m_pDescs[i].m_DescInfo, want_lang_id ).IsEmpty() )
                                { str = s; priret = pri; break; }
                        }

                        if( (PCTSTR) m_pDescs[i].m_DescInfo )
                            { str = (PCTSTR) m_pDescs[i].m_DescInfo; priret = pri; }

                        break;
                    }

                    case gemapt_SysErr :
                    {
                        String s;
                        reterr_t err = syserr ((LONG) m_pDescs[i].m_DescInfo);

                        if( err.GetDescriptionString( s, language ) )
                            { str = s; priret = pri; }

                        break;
                    }

                    case gemapt_HresErr :
                    {
                        String s;
                        reterr_t err = hreserr ((HRESULT) m_pDescs[i].m_DescInfo);

                        if( err.GetDescriptionString( s, language ) )
                            { str = s; priret = pri; }

                        break;
                    }
                }
        }
    }

    return priret;
}




}

#endif
#endif
