/*
    File:               argonExpressionEvaluator.cpp
    Description:        Class to evaluate mathematical expressions fast.

    Author:             Roman Komary
    created on:         21.09.2002
    last updated:       11.11.2006 (Roman Komary)

    Copyright (c) 2002-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonExpressionEvaluator_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonExpressionEvaluator.h"

#ifndef _argon_no_warning_disables
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)
    #pragma warning( disable: 4996 )    // warning C4996: 'sscanf' was declared deprecated
    // Note: Usage of #sscanf()# and #swscanf()# are no security problem in here because
    // we do not use any "%s" sequences in the format strings.
    // So in our case here, there is no security issue.
    #endif

    #pragma warning( disable: 4100 )    // unreferenced parameter

    #pragma warning( disable: 4706 )    // warning C4706: assignment within conditional expression
#endif

namespace ArgonLibrary
{




const GenErrDesc g_EeErrorDescs[] = {
    #ifndef _argon_no_LangEnglish
    // English
    { errEeEmptyExpression,             LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Empty arithmetic expression.") },
    { errEeSyntax,                      LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Syntax error.") },
    { errEeNoArgs,                      LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("No arguments.") },
    { errEeWrongNumArgs,                LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Wrong number of arguments.") },
    { errEeUnbalancedBrackets,          LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Unbalanced brackets.") },
    { errEeUnknownFunction,             LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Unknown function.") },
    { errEeMissingExpressionEndChar,    LANG_ENGLISH,       gemapt_Direct,      (LONG_PTR) _T("Missing ; character as end-of-expression.") },
    #endif

    #ifndef _argon_no_LangGerman
    // German
    { errEeEmptyExpression,             LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Leerer arithmetischer Ausdruck.") },
    { errEeSyntax,                      LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Syntaxfeher.") },
    { errEeNoArgs,                      LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Keine Argumente.") },
    { errEeWrongNumArgs,                LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Falsche Anzahl an Argumenten.") },
    { errEeUnbalancedBrackets,          LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Unausgewogene Klammerebenen.") },
    { errEeUnknownFunction,             LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Unbekannte Funktion.") },
    { errEeMissingExpressionEndChar,    LANG_GERMAN,        gemapt_Direct,      (LONG_PTR) _T("Ein ; Zeichen fehlt als Ausdrucksende.") },
    #endif

    /*
    #ifndef _argon_no_LangItalian
    // Italian
    //****
    #endif
    */

    GENERRDESC_ENDMARK
};

GenericErrorDescriptionTableT<InternalErrorGroup> g_edtEe( g_EeErrorDescs );  // error description tables for ExpressionEvaluator errors




// Stores the variables's values into 'this' object.
void VariableSaver::Store()
{
    EESavedVar *p = m_aVars.Ptr();
    _idx_ i = (_idx_) m_aVars.GetSize();
    for( ; i-- ; p++ ) p->m_savedValue = *p->m_pVariable;
}

// Restores the variables from 'this' object's saved values.
void VariableSaver::Restore()
{
    EESavedVar *p = m_aVars.Ptr();
    _idx_ i = (_idx_) m_aVars.GetSize();
    for( ; i-- ; p++ ) *p->m_pVariable = p->m_savedValue;
}

// Adds the variables from another variable saver.
// Note: This method will automatically handle duplicate
// variables and store them only once each.
void VariableSaver::AddVariables( const VariableSaver& vs )
{
    const EESavedVar *p = vs.m_aVars.Ptr();
    _size_t_ c = vs.m_aVars.GetSize();
    for( _idx_ i = 0 ; i < (_idx_) c ; ++i )
        AddVar( (p++)->m_pVariable );
}

// Note: This method will automatically handle duplicate
// variables and store them only once each.
void VariableSaver::AddVar( teevar *pVar )
{
    if( !pVar ) return;
    for( _idx_ i = 0 ; i < (_idx_) m_aVars.GetSize() ; ++i )
        if( m_aVars[i].m_pVariable == pVar ) return;
    EESavedVar v = { pVar, 0 };
    m_aVars.Add( v );
}



teevar eesin    ( teevar *x )   { return sin(*x); }
teevar eecos    ( teevar *x )   { return cos(*x); }
teevar eetan    ( teevar *x )   { return tan(*x); }
teevar eecot    ( teevar *x )   { return cot(*x); }
teevar eeasin   ( teevar *x )   { return asin(*x); }
teevar eeacos   ( teevar *x )   { return acos(*x); }
teevar eeatan   ( teevar *x )   { return atan(*x); }
teevar eeatan2  ( teevar *x )   { return atan2(x[0], x[1]); }
teevar eeacot   ( teevar *x )   { return acot(*x); }
teevar eesinh   ( teevar *x )   { return sinh(*x); }
teevar eecosh   ( teevar *x )   { return cosh(*x); }
teevar eetanh   ( teevar *x )   { return tanh(*x); }
teevar eecoth   ( teevar *x )   { return coth(*x); }
teevar eeasinh  ( teevar *x )   { return asinh(*x); }
teevar eeacosh  ( teevar *x )   { return acosh(*x); }
teevar eeatanh  ( teevar *x )   { return atanh(*x); }
teevar eeacoth  ( teevar *x )   { return acoth(*x); }
teevar eesinc   ( teevar *x )   { return ( *x!=0 )? sin(*x)/(*x) : 1; }
teevar eesec    ( teevar *x )   { return fdiv(1.0, cos(*x)); }
teevar eecosec  ( teevar *x )   { return fdiv(1.0, sin(*x)); }

teevar eehypot  ( teevar *x )   { return _hypot(x[0], x[1]); }
teevar eeexp    ( teevar *x )   { return exp(*x); }
teevar eelog    ( teevar *x )   { return log(*x); }
teevar eelog10  ( teevar *x )   { return log10(*x); }
teevar eelog2   ( teevar *x )   { return log2(*x); }
teevar eesqr    ( teevar *x )   { return *x * *x; }
teevar eesqrt   ( teevar *x )   { return sqrt(fabs(*x)); } // note: Milkdrop preset "Krash - Chronoshift" seems to make a 'fabs()' first. So we do it as well.
teevar eepow    ( teevar *x )   { return pow(x[0], x[1]); }

teevar eeabs    ( teevar *x )   { return fabs(*x); }
teevar eefix    ( teevar *x )   { return ( *x>0.0 )? floor(*x) : ceil(*x); }
teevar eefloor  ( teevar *x )   { return floor(*x); }
teevar eeceil   ( teevar *x )   { return ceil(*x); }
teevar eeround  ( teevar *x )   { return round(*x); }
teevar eemod    ( teevar *x )   { return ( x[1]!=0.0 )? fmod(x[0],x[1]) : 0.0; }
teevar eesign   ( teevar *x )   { return fsign(*x); }
teevar eemin    ( teevar *x )   { return tmin( x[0], x[1] ); }
teevar eemax    ( teevar *x )   { return tmax( x[0], x[1] ); }
teevar eebound  ( teevar *x )   { return tbound( x[1], *x, x[2] ); }
teevar eesigmoid( teevar *x )   { return sigmoid(x[0], x[1]); }

teevar eeif     ( teevar *x )   { return *x ? x[1] : x[2]; }
teevar eebor    ( teevar *x )   { return ( x[0] || x[1] ); }
teevar eeband   ( teevar *x )   { return ( x[0] && x[1] ); }
teevar eebnot   ( teevar *x )   { return !*x; }
teevar eeequal  ( teevar *x )   { return ( x[0] == x[1] ); }
teevar eenoteq  ( teevar *x )   { return ( x[0] != x[1] ); }
teevar eeabove  ( teevar *x )   { return ( x[0] > x[1] ); }
teevar eebelow  ( teevar *x )   { return ( x[0] < x[1] ); }
teevar eeaboveeq( teevar *x )   { return ( x[0] >= x[1] ); }
teevar eebeloweq( teevar *x )   { return ( x[0] <= x[1] ); }

teevar eerect   ( teevar *x )   { return ( *x>=x[1] && *x<x[2] )? 1.0 : 0.0; }

teevar eerand   ( teevar *x )   { return ( *x>=1.0 )? rand()%(INT)*x : 0.0; }
teevar eefrand  ( teevar *x )   { return frand05(); }
teevar eefrandx ( teevar *x )   { return frandx(*x); }
teevar eefrandxy( teevar *x )   { return frandxy(x[0], x[1]); }
teevar eeseed   ( teevar *x )   { fseed(*x); return 0.0; }

teevar eedeg    ( teevar *x )   { return deg(*x); }
teevar eerad    ( teevar *x )   { return rad(*x); }

teevar eeqlen   ( teevar *x )   { return x[0]*x[0] + x[1]*x[1]; }
teevar eelen    ( teevar *x )   { return sqrt(eeqlen(x)); }
teevar eeangle  ( teevar *x )   { return atan2(x[1], x[0]); }



const EEConst ExpressionEvaluator::m_aEEConst[] = {
    { "pi",      PI },
    { "e",       E  },
    { NULL,      0  }
};

const EEFunction ExpressionEvaluator::m_aEEFunctions[] = {
    { "sin",     1,    eesin,      "Sinus" },
    { "cos",     1,    eecos,      "Cosinus" },
    { "tan",     1,    eetan,      "Tangens" },
    { "cot",     1,    eecot,      "Cotangens" },
    { "asin",    1,    eeasin,     "Arcus Sinus" },
    { "acos",    1,    eeacos,     "Arcus Cosinus" },
    { "atan",    1,    eeatan,     "Arcus Tangens" },
    { "atan2",   2,    eeatan2,    "Arcus Tangens of a division = atan( x / y )" },
    { "acot",    1,    eeacot,     "Arcus Cotangens" },
    { "sinh",    1,    eesinh,     "Sinus Hyperbolicus" },
    { "cosh",    1,    eecosh,     "Cosinus Hyperbolicus" },
    { "tanh",    1,    eetanh,     "Tangens Hyperbolicus" },
    { "coth",    1,    eecoth,     "Cotangens Hyperbolicus" },
    { "asinh",   1,    eeasinh,    "Arcus Sinus Hyperbolicus" },
    { "acosh",   1,    eeacosh,    "Arcus Cosinus Hyperbolicus" },
    { "atanh",   1,    eeatanh,    "Arcus Tangens Hyperbolicus" },
    { "acoth",   1,    eeacoth,    "Arcus Cotangens Hyperbolicus" },
    { "sinc",    1,    eesinc,     "sin(x) / x; for x==0 -> returns 1" },
    { "sec",     1,    eesec,      "Secans = 1 / cos(x)" },
    { "cosec",   1,    eecosec,    "Cosecans = 1 / sin(x)" },

    { "hypot",   2,    eehypot,    "Length of Hypotenuse = sqrt( x^2 + y^2 )" },
    { "exp",     1,    eeexp,      "Exponent = e ^ x" },
    { "log",     1,    eelog,      "Logarithmus Naturalis. Basis = e" },
    { "log10",   1,    eelog10,    "Logarithmus. Basis = 10" },
    { "log2",    1,    eelog2,     "Logarithmus Dualis. Basis = 2" },
    { "sqr",     1,    eesqr,      "Square = x ^ 2" },
    { "sqrt",    1,    eesqrt,     "Square Root" },
    { "pow",     2,    eepow,      "x ^ y" },

    { "int",     1,    eefloor,    "Convert to Integer" },
    { "abs",     1,    eeabs,      "Absolute Value" },
    { "fix",     1,    eefix,      "Fixed Value" },
    { "floor",   1,    eefloor,    "Round downwards" },
    { "ceil",    1,    eeceil,     "Round upwards" },
    { "round",   1,    eeround,    "Round" },
    { "mod",     2,    eemod,      "Modulo (floating point)" },
    { "sign",    1,    eesign,     "Signum = -1, 0, 1 depending on x" },
    { "min",     2,    eemin,      "Minimum of x and y" },
    { "max",     2,    eemax,      "Maximum of x and y" },
    { "bound",   3,    eebound,    "x will be bound inside the range y to z" },
    { "sigmoid", 2,    eesigmoid,  "Sigmoid function" },

    { "if",      3,    eeif,       "Returns y if x is true, otherwise returns z" },
    { "bor",     2,    eebor,      "Boolean Or. (converts to integer)" },
    { "band",    2,    eeband,     "Boolean And. (converts to integer)" },
    { "bnot",    1,    eebnot,     "Boolean Not. (converts to integer)" },
    { "equal",   2,    eeequal,    "Returns 1 if x is equal to y, otherwise 0" },
    { "noteq",   2,    eenoteq,    "Returns 1 if x is not equal to y, otherwise 0" },
    { "above",   2,    eeabove,    "Returns 1 if x is greater than y, otherwise 0" },
    { "below",   2,    eebelow,    "Returns 1 if x is less than y, otherwise 0" },
    { "aboveeq", 2,    eeaboveeq,  "Returns 1 if x is greater than or equal to y, otherwise 0" },
    { "beloweq", 2,    eebeloweq,  "Returns 1 if x is less than or equal to y, otherwise 0" },

    { "rect",    3,    eerect,     "Returns 1 if x is inside range y to z (including y and excluding z), otherwise 0" },

    { "rand",    1,    eerand,     "Integer random number in the range 0 to x. (x exclusive)" },
    { "frand",   0,    eefrand,    "Random floating point value from -0.5 to 0.5 (inclusive)" },
    { "frandx",  1,    eefrandx,   "Random floating point value from 0 to x (inclusive)" },
    { "frandxy", 2,    eefrandxy,  "Random floating point value from x to y (inclusive)" },
    { "seed",    1,    eeseed,     "Seeds the random generator with an integer number (and always returns 0)" },

    { "deg",     1,    eedeg,      "Converts radian angle to degree" },
    { "rad",     1,    eerad,      "Converts degree angle to radian" },

    { "qlen",    2,    eeqlen,     "Quadratic length of 2-point vector = x^2 + y^2" },
    { "len",     2,    eelen,      "Length of 2-point vector = sqrt( x^2 + y^2 )" },
    { "angle",   2,    eeangle,    "Angle of 2-point vector = atan( y / x )" },

    { NULL,      0,    NULL,       NULL }
};



ExpressionEvaluator::ExpressionEvaluator( bool bFctCaseSens, bool bVarCaseSens )
                    :m_htVariables(37),
                     m_htVarNameMapper(37)
{
    m_bFunctionNamesCaseSensitive = bFctCaseSens;
    m_bVariableNamesCaseSensitive = bVarCaseSens;
    m_dwMode = EEMODE_DEFAULT;
    m_nPushes = 0;
    m_aCompiledExpressions.SetBlockSize( 128 );
    m_dwTokenType = TOKENT_UNDEF;
    m_szToken[0] = '\0';
    m_cStack = 0;
    m_nResult = 0;
    m_pVarSvr = NULL;
}



// Note: This method should be called only during initialization.
// If you have already added variables or expressions, you should not
// change the case-sensitivity since internally the names could
// have been converted when they were created during case-insensitive mode.
void ExpressionEvaluator::SetCaseSensitive( bool bFctCaseSens, bool bVarCaseSens )
{
    m_bFunctionNamesCaseSensitive = bFctCaseSens;
    m_bVariableNamesCaseSensitive = bVarCaseSens;
}



bool ExpressionEvaluator::GetValue( PCSTR pszVarName, teevar& value ) const
{
    StringA vname = GetVariableName( pszVarName );

    EEVar *pVar = m_htVariables.GetPtrAt( vname );
    if( pVar )
        { value = pVar->m_value; return true; }

    const EEConst *pConst = FindConst( vname );
    if( pConst )
        { value = pConst->m_value; return true; }

    return false;
}

// IMPORTANT: The names of functions of constants will not be checked,
// so you can define a variable that has the name of a constant
// variable without error and that variable will have priority
// against the internal constant when querying its value later.
bool ExpressionEvaluator::SetValue( PCSTR pszVarName, teevar value )
{
    m_htVariables[GetVariableName(pszVarName)].m_value = value; // plus create variable if not present
    return true;
}

bool ExpressionEvaluator::GetNextVariable( Iterator& pos, StringA& varname, teevar& value ) const
{
    EEVar v;
    if( !m_htVariables.GetNext( pos, varname, v ) ) return false;
    value = v.m_value;
    return true;
}

// Adds variables to the internal list of variables.
//  : pVarItems = pointer to an array of 'EEVarItem's. The variables will be
//                created (if not present already) according to their names
//                and a pointer to each value will be stored in the array.
//                NOTE: The end of the list is indicated by a name-pointer
//                 being NULL.
//  : cVarItem = the size (in bytes) of one variable-item of the array
//               'pVarItems'. Normally this should be 'sizeof(EEVarItem)',
//               But if you pass an array that has items that are larger,
//               e.g. containing a 'EEVarItem' as well as some additional
//               data, then set the 'cVarItem' to that size in bytes
//               and set 'pVarItems' to the first 'EEVarItem' in the array.
bool ExpressionEvaluator::AddVariables( EEVarItem *pVarItems, _size_t_ cVarItem )
{
    if( !pVarItems || cVarItem < sizeof(EEVarItem) ) return false;
    for( ; pVarItems->m_pszName ; pVarItems = (EEVarItem*) ( ((BYTE*) pVarItems) + cVarItem ) )
        pVarItems->m_pValue = &m_htVariables[GetVariableName(pVarItems->m_pszName)].m_value; // plus create variable if not present
    return true;
}

// This pushes the current values of the variables onto extra storage.
// Notice that the extra storage is of size EEVARSTORAGESIZE so that
// you can call 'PopVariables()' only EEVARSTORAGESIZE times successfully
// as long as you do not make any further 'PushVariables()' in between.
void ExpressionEvaluator::PushVariables()
{
    if( m_nPushes >= EEVARSTORAGESIZE ) m_nPushes = EEVARSTORAGESIZE - 1;
    _size_t_ size = ( ++m_nPushes ) * sizeof(teevar);
    Iterator pos;
    const StringA *pStrKey;
    EEVar *pVar;
    while( m_htVariables.GetNextPtr(pos, pStrKey, pVar) )
        memmove( &pVar->m_storage[0], pVar, size );
}

// Pops the variables back from a previous call to 'PushVariables()'.
// The variables will contain their values as they had them before
// the last call to 'PushVariables()'.
// See also remarks to 'PushVariables()'.
bool ExpressionEvaluator::PopVariables()
{
    if( m_nPushes <= 0 ) return false;
    _size_t_ size = ( m_nPushes-- ) * sizeof(teevar);
    Iterator pos;
    const StringA *pStrKey;
    EEVar *pVar;
    while( m_htVariables.GetNextPtr(pos, pStrKey, pVar) )
        memmove( pVar, &pVar->m_storage[0], size );
    return true;
}

// This is similar to 'PopVariables()', but with the following
// differencies:
// - Any of the states stored to the stack can be grabbed
//   and stored as actual values.
// - Unlike 'PopVariables()', the stack will not be consumed
//   which means, that the chosen state will not be removed
//   from the stack.
//  : state = state of the stack to take.
//            0 = take the state from the last 'PushVariables()' call.
//            maximum state = EEVARSTORAGESIZE - 1
//            NOTE: You can specify a state to which nothing has
//             been pushed yet (because you can set it explicitely
//             by hand using 'StoreVariableState()').
//  > FALSE = state not found and nothing will have been done
// Note: To store only a necessary set of the variables,
//  use the 'VariableSaver' class instead of this method, because
//  this method here only can handle all variables at once.
bool ExpressionEvaluator::GrabVariableState( int state )
{
    if( state < 0 || state >= EEVARSTORAGESIZE ) return false;
    Iterator pos;
    const StringA *pStrKey;
    EEVar *pVar;
    while( m_htVariables.GetNextPtr(pos, pStrKey, pVar) )
        pVar->m_value = pVar->m_storage[state];
    return true;
}

// This is the reverse of 'GrabVariableState()'.
// See descriptions there.
bool ExpressionEvaluator::StoreVariableState( int state )
{
    if( state < 0 || state >= EEVARSTORAGESIZE ) return false;
    Iterator pos;
    const StringA *pStrKey;
    EEVar *pVar;
    while( m_htVariables.GetNextPtr(pos, pStrKey, pVar) )
        pVar->m_storage[state] = pVar->m_value;
    return true;
}

bool ExpressionEvaluator::GetNextConstant( Iterator& pos, StringA& constname, teevar& value ) const
{
    if( pos.Finished() ) return false;
    if( pos.IsBeforeStart() ) pos.m_pos = (Iterator::iteration_pos) 0;
    if( !m_aEEConst[(INT_PTR) pos.m_pos].m_pszName ) return false;
    constname = m_aEEConst[(INT_PTR) pos.m_pos].m_pszName;
    if( !m_bVariableNamesCaseSensitive ) constname.Lowercase();
    value = m_aEEConst[(INT_PTR) pos.m_pos].m_value;
    pos.m_pos = (Iterator::iteration_pos) (((INT_PTR) pos.m_pos) + 1);
    return true;
}

const EEConst *ExpressionEvaluator::FindConst( PCSTR pszName ) const
{
    if( m_bVariableNamesCaseSensitive ) {
        for( int i = 0 ; m_aEEConst[i].m_pszName ; i++ )
            if( _stris_( m_aEEConst[i].m_pszName, pszName ) )
                return &m_aEEConst[i];
    } else {
        for( int i = 0 ; m_aEEConst[i].m_pszName ; i++ )
            if( _striis_( m_aEEConst[i].m_pszName, pszName ) )
                return &m_aEEConst[i];
    }
    return NULL;
}



// call this (maybe more times) *before* adding any expressions
bool ExpressionEvaluator::MapVariableName( PCSTR pszNameToMap, PCSTR pszNameToMapTo )
{
    if( m_bVariableNamesCaseSensitive ) m_htVarNameMapper[pszNameToMap] = pszNameToMapTo;
    else m_htVarNameMapper[StringA(pszNameToMap).Lowercase()] = StringA(pszNameToMapTo).Lowercase();
    return true;
}

// Gets the name of a variable. Mostly it returns the name passed,
// but if the name is mapped, it returns the mapped name.
StringA ExpressionEvaluator::GetVariableName( PCSTR pszVarName ) const
{
    if( !pszVarName ) return "";
    if( !*pszVarName ) return "";
    StringA name = pszVarName, *pname;
    if( !m_bVariableNamesCaseSensitive ) name.Lowercase();
    while( pname = m_htVarNameMapper.GetPtrAt(name) ) name = *pname;
    return name;
}



bool ExpressionEvaluator::GetNextFunction( Iterator& pos, StringA& funcname, int& nArgs ) const
{
    if( pos.Finished() ) return false;
    if( pos.IsBeforeStart() ) pos.m_pos = (Iterator::iteration_pos) 0;
    if( !m_aEEFunctions[(INT_PTR) pos.m_pos].m_pszName ) return false;
    funcname = m_aEEFunctions[(INT_PTR) pos.m_pos].m_pszName;
    nArgs    = m_aEEFunctions[(INT_PTR) pos.m_pos].m_nArgs;
    pos.m_pos = (Iterator::iteration_pos) (((INT_PTR) pos.m_pos) + 1);
    if( !m_bFunctionNamesCaseSensitive ) funcname.Lowercase();
    return true;
}

bool ExpressionEvaluator::GetNextFunction( Iterator& pos, StringA& funcname, int& nArgs, StringA& funcdescription ) const
{
    if( pos.Finished() ) return false;
    if( pos.IsBeforeStart() ) pos.m_pos = (Iterator::iteration_pos) 0;
    if( !m_aEEFunctions[(INT_PTR) pos.m_pos].m_pszName ) return false;
    funcname = m_aEEFunctions[(INT_PTR) pos.m_pos].m_pszName;
    nArgs    = m_aEEFunctions[(INT_PTR) pos.m_pos].m_nArgs;
    funcdescription = m_aEEFunctions[(INT_PTR) pos.m_pos].m_pszDescription;
    pos.m_pos = (Iterator::iteration_pos) (((INT_PTR) pos.m_pos) + 1);
    if( !m_bFunctionNamesCaseSensitive ) funcname.Lowercase();
    return true;
}

const EEFunction *ExpressionEvaluator::FindFunc( PCSTR pszName ) const
{
    if( m_bFunctionNamesCaseSensitive ) {
        for( int i = 0 ; m_aEEFunctions[i].m_pszName ; i++ )
            if( _stris_( m_aEEFunctions[i].m_pszName, pszName ) )
                return &m_aEEFunctions[i];
    } else {
        for( int i = 0 ; m_aEEFunctions[i].m_pszName ; i++ )
            if( _striis_( m_aEEFunctions[i].m_pszName, pszName ) )
                return &m_aEEFunctions[i];
    }
    return NULL;
}



// This empties completely the whole ExpressionEvaluator.
// All expressions and all variables (if 'bEmptyVars' is TRUE)
// will be deleted.
// This method will *not* empty variable name mappings.
void ExpressionEvaluator::Empty( bool bEmptyVars )
{
    if( bEmptyVars )
        { m_htVariables.Empty(); m_nPushes = 0; }
    m_aCompiledExpressions.Empty();
    m_aStack.Empty();
    m_cStack = 0;
}

///<summary>
///  <b>AddExpressions()</b> adds expression(s) to the internal
///  list of compiled expressions. <br/>
///  Internally, the passed expressions will compiled (stored in an intermediate memory form). <br/>
///  After having added all the desired expressions, call <b>AddVariables()</b> to add
///  variable definitions if you want, call <b>SetValue()</b> for all the
///  variables you want to create and initialize with specific values, call <b>Evaluate()</b>
///  to perform the calculations, call <b>GetResult()</b> to get the resulting value,
///  and call <b>GetValue()</b> to get values of variables.
///</summary>
///<param name="pszExpr"> a string with one or more expressions to add </param>
///<param name="bMore"> if TRUE, more than one expression (seperated by ; chars each)
///  will be added. </param>
///<param name="pVariableSaver"> if not NULL, it will be filled with all the variables
///  that are being assigned in the expressions passed in here. </param>
///<returns> error information </returns>
reterr_t ExpressionEvaluator::AddExpressions( PCSTR pszExpr, bool bMore, VariableSaver *pVariableSaver )
{
    if( !pszExpr ) return syserr (ERROR_INVALID_PARAMETER);

    m_aStack.SetSize( EEINITIALSTACKSIZE ); // just to prepare for 'Evaluate()' (so that it is done earlier when not so time-ciritcal in an application)

    m_pszExpression = m_pszE = pszExpr;
    m_pVarSvr = pVariableSaver;

    do
    {
        m_dwTokenType = TOKENT_UNDEF;
        m_szToken[0] = '\0';
        reterr_t e = Parse();
        if( !e ) { m_pVarSvr = NULL; return e; }
        if( !*m_szToken ) {
            if( bMore ) break;
            else return interr (errEeEmptyExpression);
        }
        if( *m_szToken == ';' ) {
            if( bMore ) continue;
            else return interr (errEeEmptyExpression);
        }
        e = Level1();
        if( !e ) { m_pVarSvr = NULL; return e; }
        AddCmd( EECMD_ENDOFEXPR );
    } while( bMore && *m_szToken == ';' );

    m_pVarSvr = NULL;
    if( *m_szToken ) return interr (errEeMissingExpressionEndChar);

    return true;
}

///<summary>
///  Do the evaluations beginning at the specified marker
///  or directly from the first expression. <br/>
///  Note: Since the expression are stored in compiled form,
///  they are executed very fast (hopefully).
///</summary>
///<param name="nStartAtMarker"> the marker where to start the evaluation
///  in the compiled expressions. <br/>
///  0 = start with first expression of all expressions. </param>
///<param name="nEndAtMarker"> the marker of an expression which should *not*
///  be evaluated (and following ones should also not
///  be evaluated). <br/>
///  0 = do till last expression of all expressions. </param>
///<returns> True on success, false on error. </returns>
bool ExpressionEvaluator::Evaluate( _idx_ nStartAtMarker, _idx_ nEndAtMarker )
{
    _size_t_ cnt = m_aCompiledExpressions.GetSize();
    if( cnt == 0 ) return ( nStartAtMarker == 0 );

    if( nEndAtMarker <= 0 || nEndAtMarker > (_idx_) cnt ) nEndAtMarker = (_idx_) cnt;
    if( nStartAtMarker >= (_idx_) cnt || nStartAtMarker < 0 ) return false;

    if( m_aStack.GetSize() < EEINITIALSTACKSIZE ) m_aStack.SetSize( EEINITIALSTACKSIZE );
    teevar *pStack = m_aStack.Ptr();

    m_cStack = 0;
    m_nResult = 0;

    BYTE *ptr = (BYTE*) m_aCompiledExpressions.Ptr();
    if( !ptr ) return false;

    BYTE *end = ptr + nEndAtMarker;
    ptr += nStartAtMarker;

    while( ptr < end )
    {
        switch( *(WORD*) ptr )
        {
            case EECMD_PUSHPTR : {
                pStack[ m_cStack++ ] = *(*(teevar**) (ptr += sizeof(WORD)));
                ptr += sizeof(teevar*);
                break;
            }
            case EECMD_PUSHVALUE : {
                pStack[ m_cStack++ ] = *(teevar*) (ptr += sizeof(WORD));
                ptr += sizeof(teevar);
                break;
            }
            case EECMD_FUNCCALL : {
                TeeFunc fct = *(TeeFunc*) (ptr += sizeof(WORD));
                WORD cargs = *(WORD*) (ptr + sizeof(TeeFunc));
                ptr += sizeof(TeeFunc) + sizeof(WORD);
                teevar *args = &pStack[ m_cStack -= cargs ];
                pStack[ m_cStack++ ] = fct(args);
                break;
            }
            case EECMD_ENDOFEXPR : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ --m_cStack ];
                m_nResult = *arg;
                m_cStack = 0;
                break;
            }
            case EECMD_ASSIGN : {
                teevar *arg = &pStack[ --m_cStack ];
                pStack[ m_cStack++ ] = *(*(teevar**) (ptr += sizeof(WORD))) = *arg;
                ptr += sizeof(teevar *);
                break;
            }
            case EECMD_LE : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] <= arg[1] );
                break;
            }
            case EECMD_LT : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] < arg[1] );
                break;
            }
            case EECMD_GE : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] >= arg[1] );
                break;
            }
            case EECMD_GT : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] > arg[1] );
                break;
            }
            case EECMD_EQ : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] == arg[1] );
                break;
            }
            case EECMD_NOTEQ : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] != arg[1] );
                break;
            }
            case EECMD_ADD : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] + arg[1] );
                break;
            }
            case EECMD_SUB : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] - arg[1] );
                break;
            }
            case EECMD_LOGICOR : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] || arg[1] );
                break;
            }
            case EECMD_LOGICAND : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] && arg[1] );
                break;
            }
            case EECMD_BINOR : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( ((teeint) round(arg[0])) | ((teeint) round(arg[1])) );
                break;
            }
            case EECMD_BINAND : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( ((teeint) round(arg[0])) & ((teeint) round(arg[1])) );
                break;
            }
            case EECMD_MUL : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = ( arg[0] * arg[1] );
                break;
            }
            case EECMD_DIV : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = fdiv( arg[0], arg[1] );
                break;
            }
            case EECMD_MOD : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                bool b_neg = false;
                teeint a = (teeint) round(arg[0]), b = (teeint) round(arg[1]);  // the roundings must be done to be equal to milkdrop
                if( b < 0 ) { a = -a; b = -b; b_neg = true; }
                if( a < 0 && b ) { a = b - (a / b) * b + a; }
                a = b ? a % b : 0;  // the integer-modulo must be done to be equal to milkdrop
                if( b_neg ) a = -a;
                pStack[ m_cStack++ ] = (teevar) a;
                break;
            }
            case EECMD_POW : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = pow( arg[0], arg[1] );
                break;
            }
            case EECMD_NEG : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ --m_cStack ];
                pStack[ m_cStack++ ] = -*arg;
                break;
            }
            case EECMD_LOGICNOT : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ --m_cStack ];
                pStack[ m_cStack++ ] = !*arg;
                break;
            }
            case EECMD_BINNOT : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ --m_cStack ];
                pStack[ m_cStack++ ] = ~(teeint) round(*arg);
                break;
            }
            case EECMD_ROUND_2 : {
                ptr += sizeof(WORD);
                teevar *arg = &pStack[ m_cStack -= 2 ];
                pStack[ m_cStack++ ] = round(arg[1]);
                break;
            }
            default :
                return false;
        }
    }

    return true;
}

// If the 'ch' is an operator character.
// Note: This does *not* include any brackets.
bool ExpressionEvaluator::IsOperatorChar( char ch )
{
    static const char szOperators[] = "=+-*/%&|><!~^";
    return( strchr( szOperators, (int) (UINT) (UCHAR) ch ) != NULL );
}

// This function is used to grab the next token from
// the expression that is being evaluated.
// Note: In here, case-sensitivity will not be used. The
// token will only be read without changing case anyway.
reterr_t ExpressionEvaluator::Parse()
{
    PSTR t = m_szToken;
    m_dwTokenType = TOKENT_UNDEF;
    SkipSpaces();

    char c = *m_pszE;
    if( c == '0' ? (m_pszE[1] == 'x' || m_pszE[1] == 'X') : false )
    {
        m_dwTokenType = TOKENT_HEXNUM;
        *t++ = *m_pszE++;
        *t++ = *m_pszE++;
        if( !isxdigit(*m_pszE) ) return interr (errEeSyntax);
        while( isxdigit(*m_pszE) )
            if( (int) (t - m_szToken) < 2 + 8 ) *t++ = *m_pszE++;
            else return interr (errEeSyntax);
    }
    else
    if( isdigit(c) || c == '.' )
    {
        int ncomma = 0, nexp = 0;
        m_dwTokenType = TOKENT_NUMBER;
        for( ;; )
        {
            c = *m_pszE;
            if( c == '.' ) {
                if( ncomma ) return interr (errEeSyntax);
                else ncomma++;
            } else if( c == 'e' || c == 'E' ) {
                if( nexp ) return interr (errEeSyntax);
                else nexp = 2;
            } else if( (c == '-' || c == '+') && (nexp == 2) ) {
                nexp = 1;
            } else if( !isdigit(c) ) break;
            if( (int) (t - m_szToken) < EETOKENLEN ) *t++ = *m_pszE++;
            else if( ncomma && !nexp ) m_pszE++;
            else return interr (errEeSyntax);
        }
    }
    else
    if( isalpha(c) || c == '_' )
    {
        m_dwTokenType = TOKENT_NAME;
        while( isalnum(*m_pszE) || *m_pszE == '_' )
            if( (int) (t - m_szToken) < EETOKENLEN ) *t++ = *m_pszE++;
            else return interr (errEeSyntax);
    }
    else
    if( c && !isspace(c) )
    {
        m_dwTokenType = TOKENT_DELIMITER;
        c = *t++ = *m_pszE++;
        if( c != ';' && c != '(' && c != ')' ) {
            if( *m_pszE == '=' ) *t++ = *m_pszE++;  // for >=, <=, !=, == cases
            else if( *m_pszE == m_pszE[-1] ) *t++ = *m_pszE++;  // for ||, && cases
        }
    }

    *t = '\0';
    SkipSpaces();

    return true;
}

// Handles any variable assignment.
reterr_t ExpressionEvaluator::Level1()
{
    if( m_dwTokenType != TOKENT_NAME ) return Level2();
    if( m_pszE[0] != '=' ) return Level2();
    if( m_pszE[1] == '=' ) return Level2();
    reterr_t e;
    char varname[EETOKENLEN+1];
    _strcpy_( varname, m_szToken );
    if( ( e = Parse() ).Failed() ) return e;
    if( ( e = Parse() ).Failed() ) return e;
    if( !*m_szToken ) return interr (errEeEmptyExpression);
    if( ( e = Level2() ).Failed() ) return e;
    teevar *pv = (teevar*) &m_htVariables[GetVariableName(varname)];  // plus create variable if not present
    AddCmd( EECMD_ASSIGN );
    AddPtr( pv );
    // Now check, if we have to serve a 'VariableSaver'.
    // If so, we add the current variable to its array.
    if( m_pVarSvr ) m_pVarSvr->AddVar( pv );
    return true;
}

// handles any logical or
reterr_t ExpressionEvaluator::Level2()
{
    reterr_t e = Level3();
    if( !e ) return e;

    for( ;; )
    {
        if( m_szToken[0] == '|' && m_szToken[1] == '|' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level3() ).Failed() ) return e;
            AddCmd( EECMD_LOGICOR );
        } else break;
    }

    return true;
}

// handles any logical and
reterr_t ExpressionEvaluator::Level3()
{
    reterr_t e = Level4();
    if( !e ) return e;

    for( ;; )
    {
        if( m_szToken[0] == '&' && m_szToken[1] == '&' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level4() ).Failed() ) return e;
            AddCmd( EECMD_LOGICAND );
        } else break;
    }

    return true;
}

// handles any relational operations
reterr_t ExpressionEvaluator::Level4()
{
    reterr_t e = Level5();
    if( !e ) return e;

    for( ;; )
    {
        char c0 = m_szToken[0], c1 = m_szToken[1];
        if( c0 == '<' || c0 == '>' || (c1 == '=' && (c0 == '!' || c0 == '=')) ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level5() ).Failed() ) return e;
            if(      c0 == '<' && c1 == '='  ) AddCmd( EECMD_LE );
            else if( c0 == '<' && c1 == '\0' ) AddCmd( EECMD_LT );
            else if( c0 == '>' && c1 == '='  ) AddCmd( EECMD_GE );
            else if( c0 == '>' && c1 == '\0' ) AddCmd( EECMD_GT );
            else if( c0 == '=' && c1 == '='  ) AddCmd( EECMD_EQ );
            else if( c0 == '!' && c1 == '='  ) AddCmd( EECMD_NOTEQ );
            else return interr (errEeSyntax);
        } else break;
    }

    return true;
}

// handles any binary or
reterr_t ExpressionEvaluator::Level5()
{
    reterr_t e = Level6();
    if( !e ) return e;

    for( ;; )
    {
        if( m_szToken[0] == '|' && m_szToken[1] == '\0' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level6() ).Failed() ) return e;
            AddCmd( m_dwMode == EEMODE_MILKDROP ? EECMD_ROUND_2 : EECMD_BINOR );
        } else break;
    }

    return true;
}

// handles any binary and
reterr_t ExpressionEvaluator::Level6()
{
    reterr_t e = Level7();
    if( !e ) return e;

    for( ;; )
    {
        if( m_szToken[0] == '&' && m_szToken[1] == '\0' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level7() ).Failed() ) return e;
            AddCmd( m_dwMode == EEMODE_MILKDROP ? EECMD_ROUND_2 : EECMD_BINAND );
        } else break;
    }

    return true;
}

// handles any addition and subtraction operations
reterr_t ExpressionEvaluator::Level7()
{
    reterr_t e = Level8();
    if( !e ) return e;

    for( ;; )
    {
        char c0 = m_szToken[0], c1 = m_szToken[1];
        if( (c0 == '+' || c0 == '-') && c1 == '\0' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level8() ).Failed() ) return e;
            if(      c0 == '+' ) AddCmd( EECMD_ADD );
            else if( c0 == '-' ) AddCmd( EECMD_SUB );
            else return interr (errEeSyntax);
        } else break;
    }

    return true;
}

// handles any multiplication, division, or modulo
reterr_t ExpressionEvaluator::Level8()
{
    reterr_t e = Level9();
    if( !e ) return e;

    for( ;; )
    {
        char c0 = m_szToken[0], c1 = m_szToken[1];
        if( (c0 == '*' || c0 == '/' || c0 == '%') && c1=='\0' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level9() ).Failed() ) return e;
            if(      c0 == '*' ) AddCmd( EECMD_MUL );
            else if( c0 == '/' ) AddCmd( EECMD_DIV );
            else if( c0 == '%' ) AddCmd( EECMD_MOD );
            else return interr (errEeSyntax);
        } else break;
    }

    return true;
}

// handles any "to the power of" operations
reterr_t ExpressionEvaluator::Level9()
{
    reterr_t e = Level10();
    if( !e ) return e;

    for( ;; )
    {
        if( m_szToken[0] == '^' && m_szToken[1] == '\0' ) {
            if( ( e = Parse() ).Failed() ) return e;
            if( ( e = Level10() ).Failed() ) return e;
            AddCmd( EECMD_POW );
        } else break;
    }

    return true;
}

// handles any unary +, -, !, ~ signs
reterr_t ExpressionEvaluator::Level10()
{
    reterr_t e;
    char c0 = m_szToken[0], c1 = m_szToken[1];
    if( (c0 == '+' || c0 == '-' || c0 == '!' || c0 == '~') && c1 == '\0' )
        { if( ( e = Parse() ).Failed() ) return e; }
    else c0 = 0;
    if( ( e = Level11() ).Failed() ) return e;
    if(      c0 == '-' ) AddCmd( EECMD_NEG      );
    else if( c0 == '!' ) AddCmd( EECMD_LOGICNOT );
    else if( c0 == '~' ) AddCmd( EECMD_BINNOT   );
    return true;
}

// handles any literal numbers, variables, or functions
reterr_t ExpressionEvaluator::Level11()
{
    reterr_t e;

    if( *m_szToken == '(' )
    {
        if( ( e = Parse() ).Failed() ) return e;
        if( *m_szToken == ')' ) return interr (errEeNoArgs);
        if( ( e = Level1() ).Failed() ) return e;
        if( *m_szToken != ')' ) return interr (errEeUnbalancedBrackets);
        if( ( e = Parse() ).Failed() ) return e;
    }
    else
    if( m_dwTokenType == TOKENT_NUMBER )
    {
        AddCmd( EECMD_PUSHVALUE );
        AddValue( (teevar) atof( m_szToken ) );
        if( ( e = Parse() ).Failed() ) return e;
    }
    else
    if( m_dwTokenType == TOKENT_HEXNUM )
    {
        ULONG n;
        sscanf( &m_szToken[2], "%lx", &n );
        AddCmd( EECMD_PUSHVALUE );
        AddValue( (teevar) n );
        if( ( e = Parse() ).Failed() ) return e;
    }
    else
    if( m_dwTokenType == TOKENT_NAME )
    {
        if( *m_pszE == '(' ) {
            const EEFunction *pFunc = FindFunc( m_szToken );
            if( !pFunc ) return interr (errEeUnknownFunction);
            if( ( e = Parse() ).Failed() ) return e;
            int nargs = 0;
            if( *m_pszE != ')' )
                while( !(m_szToken[0] == ')' && m_szToken[1] == '\0') ) {
                    nargs++;
                    if( nargs > pFunc->m_nArgs ) return interr (errEeWrongNumArgs);
                    if( ( e = Parse() ).Failed() ) return e;
                    if( ( e = Level1() ).Failed() ) return e;
                    if( !((m_szToken[0] == ',' || m_szToken[0] == ')') && m_szToken[1] == '\0') ) return interr (errEeSyntax);
                }
            else if( ( e = Parse() ).Failed() ) return e;
            if( nargs != pFunc->m_nArgs ) return interr (errEeWrongNumArgs);
            AddCmd( EECMD_FUNCCALL );
            AddPtr( pFunc->m_pFunc );
            AddWord( (WORD) pFunc->m_nArgs );
        } else {
            StringA vname = GetVariableName( m_szToken );
            const EEConst *pConst = FindConst( vname );
            if( pConst ) {
                AddCmd( EECMD_PUSHVALUE );
                AddValue( pConst->m_value );
            } else {
                AddCmd( EECMD_PUSHPTR );
                AddPtr( &m_htVariables[vname] );  // plus create variable if not present
            }
        }
        if( ( e = Parse() ).Failed() ) return e;
    }
    else
        return interr (errEeSyntax);

    return true;
}

void ExpressionEvaluator::AddWord( WORD w )
{
    * (WORD*) m_aCompiledExpressions.AddNew( sizeof(WORD) ) = w;
}

void ExpressionEvaluator::AddPtr( void *ptr )
{
    * (void**) m_aCompiledExpressions.AddNew( sizeof(void*) ) = ptr;
}

void ExpressionEvaluator::AddValue( teevar v )
{
    * (teevar*) m_aCompiledExpressions.AddNew( sizeof(teevar) ) = v;
}




}
