/*
    File:               argonFile.cpp
    Description:        File I/O handling classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         20.11.2003
    last updated:       11.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonFile_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonFile.h"

#ifndef _argon_no_includes
    #include <io.h>
    #include <fcntl.h>
#endif

namespace ArgonLibrary
{




File::File()
{
    m_FileType              = FileT_File;
    m_hFile                 = INVALID_HANDLE_VALUE;
    m_bCloseHFileOnDelete   = false;
    m_bEof                  = false;
}

File::File( HANDLE hfile )
{
    m_FileType              = FileT_File;
    m_hFile                 = hfile;
    m_bCloseHFileOnDelete   = false;
    m_bEof                  = false;
}

File::File( PCTSTR pstrFile, DWORD openflags )
{
    m_FileType              = FileT_File;
    m_hFile                 = INVALID_HANDLE_VALUE;
    m_bCloseHFileOnDelete   = false;
    m_bEof                  = false;
    Open( pstrFile, openflags );
}

File::~File()
{
    if( m_hFile != INVALID_HANDLE_VALUE && m_bCloseHFileOnDelete )
        Close();
}



String File::getFullFile() const
{
    //FileStatus status;
    //if( GetStatus(status) ) return status.m_strFullFile;
    return m_strFullFile;
}



// IMPORTANT: 'typeText' is *not* supported in this class for 'openflags'.
//            'typeText' will be used in derived classes only.
reterr_t File::Open( PCTSTR pstrFile, DWORD openflags )
{
    if( isOpen() ) return interr (errFileAlreadyOpen);
    if( _isstrempty_(pstrFile) ) return syserr (ERROR_INVALID_PARAMETER);

    // Note: 'File' objects always are of type binary.
    // Any 'typeText' or 'typeBinary' flag will be ignored.

    m_strFullFile.GetFullpathname( pstrFile );
    m_hFile                 = INVALID_HANDLE_VALUE;
    m_bCloseHFileOnDelete   = false;
    m_bEof                  = false;

    DWORD dwAccess = 0;
    switch( openflags & 0xf ) {
        case accessRead      : dwAccess = GENERIC_READ;  break;
        case accessWrite     : dwAccess = GENERIC_WRITE; break;
        case accessReadWrite : dwAccess = GENERIC_READ|GENERIC_WRITE; break;
    }

    DWORD dwShareMode = 0;
    switch( openflags & 0xf0 ) {
        case shareExclusive : dwShareMode = 0; break;
        case shareDenyWrite : dwShareMode = FILE_SHARE_READ;  break;
        case shareDenyRead  : dwShareMode = FILE_SHARE_WRITE; break;
        case shareDenyNone  : dwShareMode = FILE_SHARE_WRITE|FILE_SHARE_READ; break;
    }

    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(sa);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = ( (openflags & handleNoInherit) == 0 );

    DWORD dwCreateFlag = 0;
    switch( openflags & 0xf00 ) {
        case modeOpen   : dwCreateFlag = OPEN_EXISTING; break;
        case modeCreate : dwCreateFlag = CREATE_ALWAYS; break;
        case modeAppend : dwCreateFlag = OPEN_ALWAYS;   break;
        default         : dwCreateFlag = OPEN_EXISTING; break;
    }

    m_hFile = ::CreateFile( m_strFullFile,
                            dwAccess,
                            dwShareMode,
                            &sa,
                            dwCreateFlag,
                            FILE_ATTRIBUTE_NORMAL,
                            NULL );
    if( m_hFile == INVALID_HANDLE_VALUE )
        { m_strFullFile.Empty(); return syserr (GetLastError()); }

    m_bCloseHFileOnDelete = true;

    return true;
}

void File::Close()
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return;
    ::CloseHandle( m_hFile ); // ignore error
    m_hFile = INVALID_HANDLE_VALUE;
    m_bCloseHFileOnDelete = false;
    m_bEof = false;
}

reterr_t File::Flush()
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return syserr (ERROR_INVALID_HANDLE);
  	if( !::FlushFileBuffers( m_hFile )  ) return syserr (GetLastError());
    return true;
}



///<param name="count"> number of bytes to read </param>
///<returns> returns the number of bytes read.
///  Use <b>.Succeeded()</b> to determine if reading failed. </returns>
_retread_t_ File::Read( void *pBuffer, _size_t_ count )
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return _retread_t_(syserr (ERROR_INVALID_HANDLE), 0);
    if( !pBuffer || count <= 0 ) return _retread_t_(syserr (ERROR_INVALID_PARAMETER), 0);

    DWORD dwRead = 0;
    if( !::ReadFile( m_hFile, pBuffer, count, &dwRead, NULL ) )
        return _retread_t_(syserr (GetLastError()), dwRead);

    if( dwRead == 0 && count > 0 ) m_bEof = true;
    return _retread_t_(true, dwRead);
}

reterr_t File::ReadExactCount( void *pBuffer, _size_t_ count )
{
    _retread_t_ read = Read( pBuffer, count );
    if( !read ) return read.getError();

    if( read != count )
        return interr (errFileReadFailed);

    return true;
}

///<summary> This is similar to <b>Read()</b> except the file pointer position will not be changed. </summary>
///<param name="count"> number of bytes to read </param>
///<returns> returns the number of bytes read.
///  Use <b>.Succeeded()</b> to determine if reading failed. </returns>
_retread_t_ File::Peek( void *pBuffer, _size_t_ count )
{
    RestoreSeek rs( *this );
    return Read( pBuffer, count );
}

///<summary> This is similar to <b>ReadExactCount()</b> except the file pointer position will not be changed. </summary>
reterr_t File::PeekExactCount( void *pBuffer, _size_t_ count )
{
    RestoreSeek rs( *this );
    return ReadExactCount( pBuffer, count );
}

///<param name="count"> number of bytes to write </param>
reterr_t File::Write( const void *pBuffer, _size_t_ count )
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return syserr (ERROR_INVALID_HANDLE);
    if( count == 0 ) return true;

    DWORD dwWritten;
    if( !::WriteFile( m_hFile, pBuffer, count, &dwWritten, NULL ) )
        return syserr (GetLastError());

    if( (_size_t_) dwWritten != count ) return interr (errFileWriteFailed);
    return true;
}

//  : zlenMax = must be the size of the #pstrBuf# buffer in characters
//  > NULL = Error or EOF (determine which by using #Eof()#)
// Note: This method is not recommended for a 'File' object,
//  because it cannot handle typeText mode.
//  But it is good for derived classes.
// Note: 'StdioFile' will use its own (overloaded) implementation
//  using C-runtime functions.
PSTR File::FgetsA( PSTR pstrBuf, _size_t_ zlenMax )
{
    if( !isOpen() ) return NULL;
    if( !pstrBuf || zlenMax <= 1 ) return NULL;

    _idx_ ix = 0;
    char  ch = 0;
    bool  ok = true;

    while( ix < (_idx_) zlenMax - 1 )
    {
        if( !ReadChar( &ch ) )
            { ok = false; break; }

        if( ch == 0 ) continue; // just skip null-bytes (I am not sure if this is conform to 'fgets()', but it is the best without leading into erroneous handling)

        pstrBuf[ ix++ ] = ch;
    }

    pstrBuf[ix] = 0;

    if( !ok ) return NULL;
    if( Eof() && ix == 0 ) return NULL;

    return pstrBuf;
}

//  : zlenMax = must be the size of the #pstrBuf# buffer in characters
//  > NULL = Error or EOF
// Note: This method is not recommended for a 'File' object,
//  because it cannot handle typeText mode.
//  But it is good for derived classes.
// Note: 'StdioFile' will use its own (overloaded) implementation
//  using C-runtime functions.
PWSTR File::FgetsW( PWSTR pstrBuf, _size_t_ zlenMax )
{
    if( !isOpen() ) return NULL;
    if( !pstrBuf || zlenMax <= 1 ) return NULL;

    _idx_ ix = 0;
    WCHAR ch = 0;
    bool  ok = true;

    while( ix < (_idx_) zlenMax - 1 )
    {
        if( !ReadWchar( &ch ) )
            { ok = false; break; }

        if( ch == 0 ) continue; // just skip null-bytes (I am not sure if this is conform to 'fgets()', but it is the best without leading into erroneous handling)

        pstrBuf[ ix++ ] = ch;
    }

    pstrBuf[ix] = 0;

    if( !ok ) return NULL;
    if( Eof() && ix == 0 ) return NULL;

    return pstrBuf;
}

//  : zlenMax = must be the size of the #pstrBuf# buffer in characters
//  > NULL = Error or EOF
// Note: This method is not recommended for a 'File' object,
//  because it cannot handle typeText mode.
//  But it is good for derived classes.
// Note: 'StdioFile' will use its own (overloaded) implementation
//  using C-runtime functions.
PTSTR File::Fgets( PTSTR pstrBuf, _size_t_ zlenMax )
{
    #ifdef UNICODE
    return FgetsW( pstrBuf, zlenMax );
    #else
    return FgetsA( pstrBuf, zlenMax );
    #endif
}

///<summary> Reads an ASCII line of text from the file. Each byte of the file
///  will be treated as an ASCII char. </summary>
///<param name="bRemoveLastNewline"> if true and if the last char read into <b>str</b>
///  is a newline character -> then it will be removed </param>
///<param name="bAppend"> If true, the read textline will be appended to <b>str</b>.
///  If false, <b>str</b> will be replaced by the read line of text. </param>
///<returns> success or failure. In case of EOF -> errEOF is returned. </returns>
///<remarks> This method is not recommended for a <b>File</b> object,
///  because it cannot handle typeText mode.
///  But it is good for derived classes. </remarks>
reterr_t File::FgetstrA( StringA& str, bool bRemoveLastNewline, bool bAppend )
{
    if( !isOpen() ) return interr (errFileNotOpen);

    if( !bAppend )
        str.Empty();

    char s[1024];
    _size_t_ maxlen = _countof(s) - 1, len;

    for( ;; )
    {
        if( !FgetsA( s, _countof(s) ) ) {
            if( !Eof() ) return interr (errFileReadFailed);
            else { len = _strlen_(s); break; }
        }
        len = _strlen_(s);
        if( len == 0 ) break;
        if( len < maxlen || s[len-1] == '\n' ) break;
        str += s;
    }

    // remove '\n' from end of string if there
    if( len > 0 )
        { if( s[len - 1] == '\n' && bRemoveLastNewline ) s[len - 1] = 0; }
    else if( str.GetAt(str.length() - 1) == '\n' && bRemoveLastNewline )
        str.Delete( str.length() - 1 );

    str += s;

    return interr (Eof() ? errEOF : errOk);
}

///<summary> Reads a Unicode line of text from the file. Since <b>File</b> (and derived
///  classes except for <b>StdioFile</b>) can only act in binary mode, each
///  2-bytes-pair of the file will be treated as a Unicode char.
///  Only <b>StdioFile</b> can act in textmode and therefore it can read
///  to conversions from multibyte chars to unicode chars. </summary>
///<param name="bRemoveLastNewline"> if true and if the last char read into <b>str</b>
///  is a newline character -> then it will be removed </param>
///<param name="bAppend"> If true, the read textline will be appended to <b>str</b>.
///  If false, <b>str</b> will be replaced by the read line of text. </param>
///<returns> success or failure. In case of EOF -> errEOF is returned. </returns>
///<remarks> This method is not recommended for a <b>File</b> object,
///  because it cannot handle typeText mode.
///  But it is good for derived classes. </remarks>
reterr_t File::FgetstrW( StringW& str, bool bRemoveLastNewline, bool bAppend )
{
    if( !isOpen() ) return interr (errFileNotOpen);

    if( !bAppend )
        str.Empty();

    WCHAR s[1024];
    _size_t_ maxlen = _countof(s) - 1, len;

    for( ;; )
    {
        if( !FgetsW( s, _countof(s) ) ) {
            if( !Eof() ) return interr (errFileReadFailed);
            else { len = _strlen_(s); break; }
        }
        len = _strlen_(s);
        if( len == 0 ) break;
        if( len < maxlen || s[len-1] == L'\n' ) break;
        str += s;
    }

    // remove '\n' from end of string if there
    if( len > 0 )
        { if( s[len - 1] == L'\n' && bRemoveLastNewline ) s[len - 1] = 0; }
    else if( str.GetAt(str.length() - 1) == L'\n' && bRemoveLastNewline )
        str.Delete( str.length() - 1 );

    str += s;

    return interr (Eof() ? errEOF : errOk);
}

///<summary> Reads an ASCII or Unicode line of text from the file. Since <b>File</b> (and derived
///  classes except for <b>StdioFile</b>) can only act in binary mode, in Unicode compilation each
///  2-bytes-pair of the file will be treated as a Unicode char.
///  Only <b>StdioFile</b> can act in textmode and therefore it can read
///  to conversions from multibyte chars to unicode chars.
///  In ASCII compilations, each byte of the file will be interpreted as an ASCII char. </summary>
///<param name="bRemoveLastNewline"> if true and if the last char read into <b>str</b>
///  is a newline character -> then it will be removed </param>
///<param name="bAppend"> If true, the read textline will be appended to <b>str</b>.
///  If false, <b>str</b> will be replaced by the read line of text. </param>
///<returns> success or failure. In case of EOF -> errEOF is returned. </returns>
///<remarks> This method is not recommended for a <b>File</b> object,
///  because it cannot handle typeText mode.
///  But it is good for derived classes. </remarks>
reterr_t File::Fgetstr( String& str, bool bRemoveLastNewline, bool bAppend )
{
    if( !isOpen() ) return interr (errFileNotOpen);

    if( !bAppend )
        str.Empty();

    TCHAR s[1024];
    _size_t_ maxlen = _countof(s) - 1, len;

    for( ;; )
    {
        if( !Fgets( s, _countof(s) ) ) {
            if( !Eof() ) return interr (errFileReadFailed);
            else { len = _strlen_(s); break; }
        }
        len = _strlen_(s);
        if( len == 0 ) break;
        if( len < maxlen || s[len-1] == _T('\n') ) break;
        str += s;
    }

    // remove '\n' from end of string if there
    if( len > 0 )
        { if( s[len - 1] == _T('\n') && bRemoveLastNewline ) s[len - 1] = 0; }
    else if( str.GetAt(str.length() - 1) == _T('\n') && bRemoveLastNewline )
        str.Delete( str.length() - 1 );

    str += s;

    return interr (Eof() ? errEOF : errOk);
}

// Note: If string in 'pstr' does *not* end with a newline char,
//  then this method will *not* append a newline char automatically.
// Note: This method is not recommended for a 'File' object,
//  because it cannot handle typeText mode.
//  But it is good for derived classes.
// Note: 'StdioFile' will use its own (overloaded) implementation
//  using C-runtime functions.
reterr_t File::FputsA( PCSTR pstr )
{
    if( !isOpen() ) return interr (errFileNotOpen);
    if( !pstr ) return syserr (ERROR_INVALID_PARAMETER);

    return Write( pstr, _strlen_(pstr) * sizeof(char) );
}

// Note: If string in 'pstr' does *not* end with a newline char,
//  then this method will *not* append a newline char automatically.
// Note: This method is not recommended for a 'File' object,
//  because it cannot handle typeText mode.
//  But it is good for derived classes.
// Note: 'StdioFile' will use its own (overloaded) implementation
//  using C-runtime functions.
reterr_t File::FputsW( PCWSTR pstr )
{
    if( !isOpen() ) return interr (errFileNotOpen);
    if( !pstr ) return syserr (ERROR_INVALID_PARAMETER);

    return Write( pstr, _strlen_(pstr) * sizeof(WCHAR) );
}

// Note: If string in 'pstr' does *not* end with a newline char,
//  then this method will *not* append a newline char automatically.
// Note: This method is not recommended for a 'File' object,
//  because it cannot handle typeText mode.
//  But it is good for derived classes.
// Note: 'StdioFile' will use its own (overloaded) implementation
//  using C-runtime functions.
reterr_t File::Fputs( PCTSTR pstr )
{
    #ifdef UNICODE
    return FputsW( pstr );
    #else
    return FputsA( pstr );
    #endif
}



reterr_t File::readText( String& text, _size_t_ maxTextlines, bool bAppend )
{
    if( !bAppend )
        text.Empty();

    bool bLimit = maxTextlines > 0;

    while( !Eof() )
    {
        if( bLimit )
            if( maxTextlines-- == 0 )
                return true;

        reterr_t err = Fgetstr( text, false, true );
        if( !err ) return err;
    }

    return true;
}

reterr_t File::readTextA( StringA& text, _size_t_ maxTextlines, bool bAppend )
{
    if( !bAppend )
        text.Empty();

    bool bLimit = maxTextlines > 0;

    while( !Eof() )
    {
        if( bLimit )
            if( maxTextlines-- == 0 )
                return true;

        reterr_t err = FgetstrA( text, false, true );
        if( !err ) return err;
    }

    return true;
}

reterr_t File::readTextW( StringW& text, _size_t_ maxTextlines, bool bAppend )
{
    if( !bAppend )
        text.Empty();

    bool bLimit = maxTextlines > 0;

    while( !Eof() )
    {
        if( bLimit )
            if( maxTextlines-- == 0 )
                return true;

        reterr_t err = FgetstrW( text, false, true );
        if( !err ) return err;
    }

    return true;
}

PBYTE File::readBinary( UINT64& bufferSize, UINT64 maxBytes )
{
    bufferSize = 0;
    if( !isOpen() ) return NULL;

    INT64 startpos = Tell();
    if( startpos < 0 ) return NULL;

    INT64 len = getLength();
    if( len < 0 || len <= startpos ) return NULL;

    UINT64 bytes_get = len - startpos;
    if( maxBytes && bytes_get > maxBytes ) bytes_get = maxBytes;

    // now we must check malloc limits.
    if( bytes_get != (UINT64) (size_t) bytes_get || (size_t) bytes_get <= (size_t) 0 )
        return NULL;

    PBYTE ptr = (PBYTE) malloc( (size_t) bytes_get );
    if( !ptr ) return NULL;
    bufferSize = bytes_get;

    // Now we must determine the maximum value for the read operation
    UINT64 max_one_read = (UINT64) type<_size_t_>::getmax();

    PBYTE p = ptr;
    while( !Eof() )
    {
        _size_t_ cnt = (_size_t_) tmin( bytes_get, max_one_read );
        if( cnt <= 0 ) break;

        _retread_t_ ret = Read( p, cnt );
        if( !ret )
            { free(ptr); bufferSize = 0; return NULL; }

        p         += ret.getValue();
        bytes_get -= ret.getValue();
    }

    if( bytes_get > 0 ) memset( ptr + (bufferSize - bytes_get), 0, (size_t) bytes_get );
    bufferSize -= bytes_get;

    return ptr;
}



///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the number of bytes copied. </returns>
_retread_t_ File::CopyFromFile( File *f, _size_t_ countToCopy, _size_t_ mincountToCopy )
{
    if( !f ) return _retread_t_(syserr (ERROR_INVALID_PARAMETER), 0);
    if( !isOpen() || !f->isOpen() ) return _retread_t_(interr (errFileNotOpen), 0);
    if( countToCopy <= 0 ) return countToCopy == 0 ? _retread_t_(true, 0) : _retread_t_(syserr (ERROR_INVALID_PARAMETER), 0);

    if( mincountToCopy > countToCopy ) mincountToCopy = countToCopy;

    BYTE *ptrDA = NULL; _size_t_ countDA = 0;
    if( f->directaccessReadMemory( (const BYTE*&) ptrDA, countDA ) )
    {
        if( countDA < mincountToCopy || !ptrDA ) return _retread_t_(false, 0);
        if( countDA < countToCopy ) countToCopy = countDA;

        if( countToCopy <= 0 ) return _retread_t_(countToCopy == 0, 0);

        reterr_t err = Write( ptrDA, countToCopy );
        if( !err ) return _retread_t_(err, 0);

        f->Seek( countToCopy, current );

        return _retread_t_( true, countToCopy );
    }

    INT64 oldpos_f = f->Tell();

    if( directaccessWriteMemory( ptrDA, countToCopy ) )
    {
        if( !ptrDA ) return _retread_t_(false, 0);

        _retread_t_ ret = f->Read( ptrDA, countToCopy );
        if( !ret.Succeeded() || ret.getValue() < mincountToCopy )
            { f->Seek( oldpos_f, begin ); return ret ? _retread_t_(interr (errFileReadNotEnoughData), 0) : _retread_t_(ret, 0); }

        return _retread_t_( true, ret.getValue() );
    }

    PBYTE ptr = (PBYTE) malloc( countToCopy );
    if( !ptr ) return _retread_t_(syserr (ERROR_OUTOFMEMORY), 0);

    _retread_t_ ret = f->Read( ptr, countToCopy );
    if( !ret.Succeeded() || ret.getValue() < mincountToCopy )
        { f->Seek( oldpos_f, begin ); free( ptr ); return ret ? _retread_t_(interr (errFileReadNotEnoughData), 0) : _retread_t_(ret, 0); }

    reterr_t err = Write( ptr, ret.getValue() );
    if( !err )
        { f->Seek( oldpos_f, begin ); free( ptr ); return _retread_t_(err, 0); }

    free( ptr );

    return _retread_t_( true, ret.getValue() );
}



///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the seek position (will also be -1 on error). </returns>
_retseek_t_ File::Seek( INT64 nOfs, SeekPosition nFrom )
{
    if( m_hFile == INVALID_HANDLE_VALUE )
        return _retseek_t_(syserr (ERROR_INVALID_HANDLE), -1);

    LARGE_INTEGER n;
    n.QuadPart = nOfs;

    n.LowPart = ::SetFilePointer( m_hFile, n.LowPart, &n.HighPart, (DWORD) nFrom );

    if( n.LowPart == INVALID_SET_FILE_POINTER && ::GetLastError() != NO_ERROR )
        return _retseek_t_(syserr (GetLastError()), -1);

    if( nFrom == begin || nFrom == end )
    {
        if( ::GetLastError() == ERROR_NEGATIVE_SEEK )
            { m_bEof = false; return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1); }
    }

    m_bEof = false;

    return _retseek_t_(true, n.QuadPart);
}

///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the current file position (will also be -1 on error). </returns>
_retseek_t_ File::Tell() const
{
    if( m_hFile == INVALID_HANDLE_VALUE )
        return _retseek_t_(syserr (ERROR_INVALID_HANDLE), -1);

    LARGE_INTEGER n;
    n.QuadPart = 0;

    n.LowPart = ::SetFilePointer( m_hFile, n.LowPart, &n.HighPart, FILE_CURRENT );

    if( n.LowPart == INVALID_SET_FILE_POINTER && ::GetLastError() != NO_ERROR )
        return _retseek_t_(syserr (GetLastError()), -1);

    return _retseek_t_(true, n.QuadPart);
}

reterr_t File::setLength( UINT64 length )
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return syserr (ERROR_INVALID_HANDLE);

    reterr_t err = Seek( length, begin );
    if( !err ) return err;

    if( !::SetEndOfFile( m_hFile ) )
        return syserr (GetLastError());

    m_bEof = false;

    return true;
}

///<returns> -1 = error </returns>
/// IMPORTANT: After a call to this method, the
///  file pointer will be still on the same location.
INT64 File::getLength() const
{
    // Seek is a non const operation
    File *f = (File*) this;

    INT64 curr = f->Tell();
    if( curr == -1 ) return -1;

    INT64 len = f->SeekToEnd();

    if( !f->Seek( curr, begin ) )
        return -1;

    return len;
}



reterr_t File::LockRange( UINT64 pos, UINT64 count )
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return syserr (ERROR_INVALID_HANDLE);

    ULARGE_INTEGER lpos, lcount;
    lpos.QuadPart   = pos;
    lcount.QuadPart = count;

    if( !::LockFile( m_hFile, lpos.LowPart, lpos.HighPart, lcount.LowPart, lcount.HighPart ) )
        return syserr (GetLastError());

    return true;
}

reterr_t File::UnlockRange( UINT64 pos, UINT64 count )
{
    if( m_hFile == INVALID_HANDLE_VALUE ) return syserr (ERROR_INVALID_HANDLE);

    ULARGE_INTEGER lpos, lcount;
    lpos.QuadPart   = pos;
    lcount.QuadPart = count;

    if( !::UnlockFile( m_hFile, lpos.LowPart, lpos.HighPart, lcount.LowPart, lcount.HighPart ) )
        return syserr (GetLastError());

    return true;
}



reterr_t File::getStatus( FileStatus& status ) const
{
    status.Empty();
    status.m_strFullFile = m_strFullFile;

    if( m_hFile == INVALID_HANDLE_VALUE ) return interr (suxNoOp); // we return ok in this case

    // get time current file size
    FILETIME ftCreate, ftAccess, ftModify;
    if( !::GetFileTime(m_hFile, &ftCreate, &ftAccess, &ftModify) )
        return syserr (GetLastError());

    ULARGE_INTEGER siz;
    siz.LowPart = ::GetFileSize(m_hFile, &siz.HighPart);
    if( siz.LowPart == INVALID_FILE_SIZE && ::GetLastError() != NO_ERROR )
        return syserr (GetLastError());
    status.m_size = siz.QuadPart;

    if( m_strFullFile.IsEmpty() ) {
        status.m_attribute = 0;
    } else {
        DWORD dwAttribute = ::GetFileAttributes( m_strFullFile );
        if( dwAttribute == 0xFFFFFFFF ) status.m_attribute = 0; // ignore error
        else status.m_attribute = dwAttribute & ~normal_; // strip attribute of NORMAL bit
    }

    status.m_created    = ftCreate;
    status.m_modified   = ftModify;
    status.m_lastAccess = ftAccess;
    if( status.m_created.dwLowDateTime     == 0 && status.m_created.dwLowDateTime     == 0 ) status.m_created    = status.m_modified;
    if( status.m_lastAccess.dwHighDateTime == 0 && status.m_lastAccess.dwHighDateTime == 0 ) status.m_lastAccess = status.m_modified;

    return true;
}

reterr_t File::getStatus( PCTSTR pstrFile, FileStatus& status )
{
    status.Empty();
    if( _isstrempty_(pstrFile) ) return syserr (ERROR_INVALID_PARAMETER);
    status.m_strFullFile.GetFullpathname( pstrFile );

    WIN32_FIND_DATA fd;
    HANDLE hfind = FindFirstFile( status.m_strFullFile, &fd );
    if( hfind == INVALID_HANDLE_VALUE )
        return syserr (GetLastError());
    FindClose( hfind );

    // strip attribute of NORMAL bit
    status.m_attribute = fd.dwFileAttributes & ~FILE_ATTRIBUTE_NORMAL;

    ULARGE_INTEGER siz;
    siz.LowPart  = fd.nFileSizeLow;
    siz.HighPart = fd.nFileSizeHigh;
    status.m_size = siz.QuadPart;

    status.m_created    = fd.ftCreationTime;
    status.m_modified   = fd.ftLastWriteTime;
    status.m_lastAccess = fd.ftLastAccessTime;
    if( status.m_created.dwLowDateTime     == 0 && status.m_created.dwLowDateTime     == 0 ) status.m_created    = status.m_modified;
    if( status.m_lastAccess.dwHighDateTime == 0 && status.m_lastAccess.dwHighDateTime == 0 ) status.m_lastAccess = status.m_modified;

    return true;
}



// Similar to 'CreateDirectory()' but does create all the sub-dir structure
// as well if it does not exist yet. I mean:
// If "C:\\test" does exist and you want to create "C:\\test\\one\\two", then
// 'CreateDirectory()' would fail since "C:\\test\\one" does not exist.
// But 'createDirectoryPath()' will do it.
//  > If no error -> dir has been created *or* was already present
reterr_t File::createDirectoryPath( PCTSTR pstrDir, LPSECURITY_ATTRIBUTES lpSecurityAttributes )
{
    if( _isstrempty_(pstrDir) ) return syserr (ERROR_INVALID_PARAMETER);
    String strDir;
    strDir.GetFullpathname( pstrDir );

    if( directoryExists( strDir, true ) )
        return true;

    reterr_t err = createDirectoryPath( String(strDir).ParentPath(), lpSecurityAttributes );
    if( !err ) return err;

    return ::CreateDirectory( strDir, lpSecurityAttributes ) != FALSE;
}

// Deletes a directory and all of its contents
//  : pstrDir = a relative or absolute path ( MAX_PATH chars long at maximum )
reterr_t File::deleteWholeDirectory( PCTSTR pstrDir )
{
    if( _isstrempty_(pstrDir) ) return syserr (ERROR_INVALID_PARAMETER);
    String strDir;
    strDir.GetFullpathname( pstrDir );

    if( !directoryExists( strDir, false ) )
        return true;

    for( ;; )
    {
        WIN32_FIND_DATA w32fd;
        String str = strDir | _T("*.*");

        HANDLE h = FindFirstFile( str, &w32fd );
        if( h == INVALID_HANDLE_VALUE )  // because a directory also contains "." and "..", an invalid handle really means an error
            return syserr (GetLastError());

        bool found = false;
        do
        {
            if( !( w32fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) || ( !_patheq_(w32fd.cFileName, _T(".")) && !_patheq_(w32fd.cFileName, _T("..")) ) )
                { found = true; break; }
        } while( FindNextFile( h, &w32fd ) );

        FindClose( h );

        if( !found )
            break;

        if( w32fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) {
            reterr_t err = deleteWholeDirectory( strDir | w32fd.cFileName );
            if( !err ) return err;
        } else {
            if( !::DeleteFile( strDir | w32fd.cFileName ) )
                return syserr (GetLastError());
        }
    }

    return ::RemoveDirectory( strDir ) != FALSE;
}

///<summary> Checks, if a specific directory does exist. </summary>
///<param name="bAllowDrivesOnly"> If TRUE and <b>pstrDir</b> is just a
///  drive (like "C:\" or "C:"), then TRUE will be returned in that
///  case, too, if the drive does exist. </param>
bool File::directoryExists( PCTSTR pstrDir, bool bAllowDrivesOnly )
{
    if( _isstrempty_(pstrDir) ) return false;

    if( _getabsolutepathlen_(pstrDir) == _strlen_(pstrDir) )
    {
        if( !bAllowDrivesOnly ) return false;

        String searchmask = pstrDir;
        searchmask |= _T("*.*");

        WIN32_FIND_DATA w32fd;
        HANDLE hfind = FindFirstFile( searchmask, &w32fd );
        if( hfind == INVALID_HANDLE_VALUE ) return false;

        FindClose( hfind );
        return true;
    }

    bool dir_exists = false;
    WIN32_FIND_DATA w32fd;

    HANDLE hfind = FindFirstFile( pstrDir, &w32fd );
    if( hfind == INVALID_HANDLE_VALUE ) return false;

    do
    {
        if( ( w32fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) != 0 )
            { dir_exists = true; break; }
    } while( FindNextFile( hfind, &w32fd ) );

    FindClose( hfind );

    return dir_exists;
}

bool File::fileExists( PCTSTR pstrFile )
{
    if( _isstrempty_(pstrFile) ) return false;

    bool file_exists = false;
    WIN32_FIND_DATA w32fd;
    HANDLE hfind = FindFirstFile( pstrFile, &w32fd );
    if( hfind == INVALID_HANDLE_VALUE ) return false;

    do
    {
        if( ( w32fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ) == 0 )
            { file_exists = true; break; }
    } while( FindNextFile( hfind, &w32fd ) );

    FindClose( hfind );

    return file_exists;
}



///<returns> Empty string means canceled by user or an error occured. </returns>
String File::browseForDirectory( HWND hwndParent, PCTSTR pstrTitle, UINT flags )
{
    TCHAR dir[_MAX_PATH_];
    if( !browseForDirectory( NULL, hwndParent, dir, _countof(dir), pstrTitle, flags, NULL ) )
        return String();
    return dir;
}

//  : pMalloc = a MALLOC object to be used or
//              NULL -> it will use its own MALLOC object.
//  : hwndParent = parent of the dialog
//  : pstrDir = buffer which will contain the directory
//  : zlenDir = size of buffer in characters
//  : pszTitle = title of the Dialog or NULL
//  : flags = BIF_* flags as used for the BROWSEINFO struct's .ulFlags member.
//  : pbInvaildDir = pointer to a bool which will indicate if an invalid dir has been selected which
//                   only can be the case when this function has returned an error.
//                 If NULL -> this value will be ignored.
//  > if false returned = this means either Canceled by user or an Error occured
//    In that case, 'pbInvalidDir' can become true.
reterr_t File::browseForDirectory( LPMALLOC pMalloc, HWND hwndParent, PTSTR pstrDir, _size_t_ zlenDir, PCTSTR pstrTitle, UINT flags, bool *pbInvalidDir )
{
    if( pbInvalidDir ) *pbInvalidDir = false;
    if( !hwndParent || !pstrDir || !zlenDir ) return syserr (ERROR_INVALID_PARAMETER);
    *pstrDir = 0;

    bool ok = true;
    HRESULT hr;

    LPMALLOC pm = NULL;
    if( !pMalloc )
    {
        hr = SHGetMalloc( &pm );
        if( SUCCEEDED( hr ) ) pMalloc = pm;
        else return hreserr (hr);
    }

    // Alloc a buffer to receive browse info.
    PTSTR pstrBuffer = (PTSTR) pMalloc->Alloc( MAX_PATH * sizeof(TCHAR) );
    if( !pstrBuffer ) {
        if( pm ) pm->Release();
        return syserr (ERROR_OUTOFMEMORY);
    }

    /*
    // Get the PIDL for the Drives folder
    LPITEMIDLIST pidlDrives = NULL;
    hr = SHGetSpecialFolderLocation( hwndParent, CSIDL_DRIVES, &pidlDrives );
    if( !SUCCEEDED( hr ) ) {
        if( pstrBuffer ) pMalloc->Free( pstrBuffer );
        return hreserr (hr);
    }
    */

    BROWSEINFO bi;
    memset( &bi, 0, sizeof(bi) );

    bi.hwndOwner = hwndParent;
    //bi.pidlRoot = pidlDrives;
    bi.pszDisplayName = pstrBuffer;
    bi.lpszTitle = pstrTitle;
    bi.ulFlags = flags;

    // Now browse for a folder.
    LPITEMIDLIST pidl = SHBrowseForFolder( &bi );
    if( !pidl ) {
        //if( pidlDrives ) pMalloc->Free( pidlDrives );
        if( pstrBuffer ) pMalloc->Free( pstrBuffer );
        if( pm ) pm->Release();
        return false;
    }

    // #pstrBuffer# now contains the display-name, but it is not used in
    // here because (virtual) folder names do not always represent a directory.

    // Now get the directory.
    if( SHGetPathFromIDList( pidl, pstrBuffer ) ) {
        _strcpy_( pstrDir, pstrBuffer, zlenDir );
    } else {
        ok = false;
        if( pbInvalidDir ) *pbInvalidDir = true;
    }

    if( pidl ) pMalloc->Free( pidl );
    //if( pidlDrives ) pMalloc->Free( pidlDrives );
    if( pstrBuffer ) pMalloc->Free( pstrBuffer );
    if( pm ) pm->Release();

    return ok;
}




reterr_t File::RestoreSeek::StorePos()
{
    if( !m_pFile ) return interr (errFileNoStream);

    _retseek_t_ ret = m_pFile->Tell();
    if( !ret ) return ret;

    m_nSeekPos = ret;
    return true;
}

reterr_t File::RestoreSeek::RestorePos()
{
    if( !m_pFile ) return interr (errFileNoStream);
    if( !hasPos() ) return interr (errNoDataAvailable);

    return m_pFile->Seek( m_nSeekPos, File::begin );
}




///<summary> Finds the first file </summary>
///<remarks> The first file will have been retrieved already after
///  this call. </remarks>
///<param name="pstrWild">
///  files (w/ or w/o path, w/ or w/o wildcards) to find <br/>
///  if NULL -> "*.*" will be used </param>
///<returns> Returns an error on error *or* if file cannot be found. </returns>
///<remarks> I had to name it FindFirstfile istead of FindFirstFile, because
///  is a macro defined by the Win32 SDK and maps to either the A or W version.
///  We do not want a macro being used for our methods, so we name it a bit differently. </remarks>
reterr_t FileFinder::FindFirstfile( PCTSTR pstrWild )
{
    if( isOpen() ) return interr (errFileAlreadyOpen);
    if( !pstrWild ) pstrWild = _T("*.*");

    String strWild;
    strWild.GetFullpathname( pstrWild );

    m_strFullPath = strWild.DirectoryOf();

    memset( &m_wfd, 0, sizeof(m_wfd) );

    m_hFF = ::FindFirstFile( strWild, &m_wfd );

    if( m_hFF == INVALID_HANDLE_VALUE )
        return syserr (GetLastError());

    return true;
}

///<summary> Finds the next file </summary>
///<returns> Returns an error on error *or* if no more files (errEOF). </returns>
///<remarks> I had to name it FindNextfile istead of FindNextFile, because
///  is a macro defined by the Win32 SDK and maps to either the A or W version.
///  We do not want a macro being used for our methods, so we name it a bit differently. </remarks>
reterr_t FileFinder::FindNextfile()
{
    if( !isOpen() ) return interr (errFileNotOpen);

    if( ::FindNextFile( m_hFF, &m_wfd ) )
        return true;

    if( ::GetLastError() == ERROR_NO_MORE_FILES )
        return interr (errEOF);
        // No more files is not an error. It's just the end of data.

    return syserr (GetLastError());
}

void FileFinder::Close()
{
    if( m_hFF != INVALID_HANDLE_VALUE )
    {
        ::FindClose( m_hFF );
        m_hFF = INVALID_HANDLE_VALUE;
    }
}



// If we found a "." or ".." path.
bool FileFinder::isDots() const
{
    if( m_wfd.cFileName[0] == _T('.') ) {
        if( m_wfd.cFileName[1] == 0 ) return true;
        if( m_wfd.cFileName[1] == _T('.') && m_wfd.cFileName[2] == 0 ) return true;
    }
    return false;
}




StdioFile::StdioFile()
          :File()
{
    m_FileType = FileT_StioFile;
    m_pStream  = NULL;
    m_bCloseStreamOnDelete = false;
}

StdioFile::StdioFile( FILE *stream )
          :File( INVALID_HANDLE_VALUE )
{
    m_FileType = FileT_StioFile;
    m_pStream  = stream;
    m_hFile    = (HANDLE) _get_osfhandle( _fileno(m_pStream) );
    m_bCloseHFileOnDelete  = false;
    m_bCloseStreamOnDelete = false;
}

StdioFile::StdioFile( PCTSTR pstrFile, DWORD openflags )
          :File( INVALID_HANDLE_VALUE )
{
    m_FileType = FileT_StioFile;
    Open( pstrFile, openflags );
}

StdioFile::~StdioFile()
{
    if( m_pStream && m_bCloseStreamOnDelete )
        Close();
}



// Note: If none of 'typeText' and 'typeBinary' is specified,
//       'typeText' will be imagined.
reterr_t StdioFile::Open( PCTSTR pstrFile, DWORD openflags )
{
    if( isOpen() ) return interr (errFileAlreadyOpen);
    //if( _isstrempty_(pszFile) ) return syserr (ERROR_INVALID_PARAMETER);

    m_hFile   = INVALID_HANDLE_VALUE;
    m_pStream = NULL;
    m_bCloseHFileOnDelete  = false;
    m_bCloseStreamOnDelete = false;

    reterr_t err = File::Open( pstrFile, openflags & ~(typeText | typeBinary) );
    if( !err ) return err;

    TCHAR mode[8];
    int   imode = 0;
    switch( openflags & 0xf00 ) {
        case modeOpen :
            if( ( openflags & 0xf ) == accessWrite ) mode[imode++] = _T('a');
            else mode[imode++] = _T('r');
            break;
        case modeCreate : mode[imode++] = _T('w'); break;
        case modeAppend : mode[imode++] = _T('a'); break;
    }

    if(  mode[imode-1] == _T('r') && ( ( openflags & 0xf ) == accessReadWrite ) ||
         mode[imode-1] != _T('r') && ( ( openflags & 0xf ) != accessWrite     )  )
        mode[imode++] = _T('+');

    int flags = _O_RDONLY | _O_TEXT;
    if( (openflags & 0xf) != accessRead )
        flags ^= _O_RDONLY;

    if( openflags & typeBinary ) {
        mode[imode++] = _T('b');
        flags ^= _O_TEXT;
    } else
        mode[imode++] = _T('t');
    mode[imode++] = 0;

    int handle = _open_osfhandle( (intptr_t) m_hFile, flags );  // C-runtime low-level file handle
    if( handle != -1 ) m_pStream = _tfdopen( handle, mode );    // open a C-runtime stream from that handle

    if( !m_pStream ) {
        File::Close();
        return syserr (ERROR_OPEN_FAILED);
    }
    m_bCloseStreamOnDelete = true;
    m_bCloseHFileOnDelete  = false; // 'fclose()' will close the handle in 'Close()'.

    return true;
}

void StdioFile::Close()
{
    if( !m_pStream ) return;
    fclose( m_pStream );
    m_hFile   = INVALID_HANDLE_VALUE;
    m_pStream = NULL;
    m_bCloseHFileOnDelete  = false;
    m_bCloseStreamOnDelete = false;
}

reterr_t StdioFile::Flush()
{
    if( !m_pStream ) return interr (errFileNoStream);
    return fflush( m_pStream ) == 0;
}



///<param name="count"> number of bytes to read </param>
///<returns> returns the number of bytes read.
///  Use <b>.Succeeded()</b> to determine if reading failed. </returns>
_retread_t_ StdioFile::Read( void *pBuffer, _size_t_ count )
{
    if( !m_pStream ) return _retread_t_(interr (errFileNoStream), 0);
    if( count == 0 ) return _retread_t_(true, 0);
    if( !pBuffer ) _retread_t_(syserr (ERROR_INVALID_PARAMETER), 0);

    _size_t_ read = (_size_t_) fread( pBuffer, sizeof(BYTE), count, m_pStream );

    if( ferror( m_pStream ) ) return _retread_t_(interr (errFileReadFailed), read);
    return _retread_t_(true, read);
}

///<param name="count"> number of bytes to write </param>
reterr_t StdioFile::Write( const void *pBuffer, _size_t_ count )
{
    if( !m_pStream ) return interr (errFileNoStream);
    if( count == 0 ) return true;
    if( !pBuffer ) return syserr (ERROR_INVALID_PARAMETER);

    if( (_size_t_) fwrite( pBuffer, sizeof(BYTE), count, m_pStream ) != count )
        return interr (errFileWriteFailed);

    if( ferror( m_pStream ) ) return interr (errFileWriteFailed);
    return true;
}

//  : zlenMax = must be the size of the 'pstrBuf' buffer
//  > NULL = error or EOF
PSTR StdioFile::FgetsA( PSTR pstrBuf, _size_t_ zlenMax )
{
    if( !m_pStream ) return NULL;
    if( !pstrBuf ) return NULL;

    pstrBuf[0] = 0;

    if( !fgets( pstrBuf, zlenMax, m_pStream ) ) {
        //if( ferror( m_pStream ) ) return NULL;
        return NULL;
    }

    return pstrBuf;
}

//  : zlenMax = must be the size of the 'pstrBuf' buffer
//  > NULL = error or EOF
PWSTR StdioFile::FgetsW( PWSTR pstrBuf, _size_t_ zlenMax )
{
    if( !m_pStream ) return NULL;
    if( !pstrBuf || zlenMax <= 0 ) return NULL;

    pstrBuf[0] = 0;

    if( !fgetws( pstrBuf, zlenMax, m_pStream ) ) {
        //if( ferror( m_pStream ) ) return NULL;
        return NULL;
    }

    return pstrBuf;
}

// Note: If string in 'pstr' does *not* end with a newline char,
//  then this method will *not* append a newline char automatically.
reterr_t StdioFile::FputsA( PCSTR pstr )
{
    if( !m_pStream ) return interr (errFileNoStream);
    if( !pstr ) return syserr (ERROR_INVALID_PARAMETER);
    //if( !*pstr ) return true;

    if( fputs( pstr, m_pStream ) == EOF )
        return interr (errFileWriteFailed);

    return true;
}

// Note: If string in 'pstr' does *not* end with a newline char,
//  then this method will *not* append a newline char automatically.
reterr_t StdioFile::FputsW( PCWSTR pstr )
{
    if( !m_pStream ) return interr (errFileNoStream);
    if( !pstr ) return syserr (ERROR_INVALID_PARAMETER);
    //if( !*pstr ) return true;

    if( fputws( pstr, m_pStream ) == EOF )
        return interr (errFileWriteFailed);

    return true;
}



///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the seek position (will also be -1 on error). </returns>
///<remarks> IMPORTANT: When the file has been opened in typeText
///  mode, then 'Seek()' and 'Tell()' have limited usage
///  (exactly like C-runtime's 'fseek()' and 'ftell()'):
///  'Seek()' can be used only with 'nOfs' set to zero,
///  or 'nOfs' set to a returned value from 'Tell()' when
///  also 'nFrom' is set to 'begin' (seeking from begin-of-file).
///  PLUS: 'GetLength()' will return the length in bytes
///  *without* converting CR+LF pairs into LF chars. </remarks>
_retseek_t_ StdioFile::Seek( INT64 nOfs, SeekPosition nFrom )
{
    if( !m_pStream ) return _retseek_t_(interr (errFileNoStream), -1);

    if( fseek( m_pStream, (long) nOfs, nFrom ) != 0 )
        return _retseek_t_(interr (errFileSeekFailed), -1);

    INT64 pos = ftell( m_pStream );
    if( pos == -1 ) return _retseek_t_(syserr (ERROR_SEEK_ON_DEVICE), -1);

    return _retseek_t_(true, pos);
}

///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the current file position (will also be -1 on error). </returns>
///<remarks> IMPORTANT: When the file has been opened in typeText
///  mode, then 'Seek()' and 'Tell()' have limited usage
///  (exactly like C-runtime's 'fseek()' and 'ftell()'):
///  'Seek()' can be used only with 'nOfs' set to zero,
///  or 'nOfs' set to a returned value from 'Tell()' when
///  also 'nFrom' is set to 'begin' (seeking from begin-of-file).
///  PLUS: 'GetLength()' will return the length in bytes
///  *without* converting CR+LF pairs into LF chars. </remarks>
_retseek_t_ StdioFile::Tell() const
{
    if( !m_pStream ) return _retseek_t_(interr (errFileNoStream), -1);

    INT64 pos = ftell( m_pStream );
    if( pos == -1 ) return _retseek_t_(interr (errFileTellFailed), -1);

    return _retseek_t_(true, pos);
}




ResourceFile::ResourceFile()
             :File()
{
    m_FileType  = FileT_ResourceFile;
    m_pResStart = NULL;
    m_pResPtr   = NULL;
    m_cResData  = 0;
}

ResourceFile::ResourceFile( UINT resid, PCTSTR pResType )
             :File( INVALID_HANDLE_VALUE )
{
    m_FileType  = FileT_ResourceFile;
    m_pResStart = NULL;
    m_pResPtr   = NULL;
    m_cResData  = 0;
    Open( resid, pResType );
}

ResourceFile::ResourceFile( HMODULE hModule, UINT resid, PCTSTR pResType )
             :File( INVALID_HANDLE_VALUE )
{
    m_FileType  = FileT_ResourceFile;
    m_pResStart = NULL;
    m_pResPtr   = NULL;
    m_cResData  = 0;
    Open( hModule, resid, pResType );
}

ResourceFile::ResourceFile( HMODULE hModule, PCTSTR pstrResName, PCTSTR pResType )
             :File( INVALID_HANDLE_VALUE )
{
    m_FileType  = FileT_ResourceFile;
    m_pResStart = NULL;
    m_pResPtr   = NULL;
    m_cResData  = 0;
    Open( hModule, pstrResName, pResType );
}



reterr_t ResourceFile::Open( UINT resid, PCTSTR pResType )
{
    return Open( (HMODULE) Get_Resource_Instance(), resid, pResType );
}

reterr_t ResourceFile::Open( HMODULE hModule, UINT resid, PCTSTR pResType )
{
    return Open( hModule, MAKEINTRESOURCE(resid), pResType );
}

reterr_t ResourceFile::Open( HMODULE hModule, PCTSTR pstrResName, PCTSTR pResType )
{
    if( isOpen() ) return interr (errFileAlreadyOpen);
    if( !pstrResName ) return syserr (ERROR_INVALID_PARAMETER);

    m_bEof  = false;

    HRSRC hrsrc = ::FindResource( hModule, pstrResName, pResType );
    if( !hrsrc ) return syserr (GetLastError());

    _size_t_ cRes = (_size_t_) ::SizeofResource( hModule, hrsrc );
    if( cRes == 0 ) return syserr (GetLastError());

    HGLOBAL hglobal = ::LoadResource( hModule, hrsrc );
    if( !hglobal ) return syserr (GetLastError());

    m_pResPtr  = m_pResStart = (PBYTE) ::LockResource( hglobal );
    m_cResData = cRes;

    return true;
}

void ResourceFile::Close()
{
    m_pResStart = NULL;
    m_pResPtr   = NULL;
    m_cResData  = 0;
    m_bEof      = false;
}

reterr_t ResourceFile::Flush()
{
    return isOpen();
}



///<param name="count"> number of bytes to read </param>
///<returns> returns the number of bytes read.
///  Use <b>.Succeeded()</b> to determine if reading failed. </returns>
_retread_t_ ResourceFile::Read( void *pBuffer, _size_t_ count )
{
    if( !isOpen() ) return _retread_t_(interr (errFileNotOpen), 0);
    if( !pBuffer ) return _retread_t_(syserr (ERROR_INVALID_PARAMETER), 0);

    if( m_pResPtr < m_pResStart ||
        m_pResPtr > m_pResStart + m_cResData ) return _retread_t_(interr (errOutOfRange), 0);

    _size_t_ available = m_cResData - (_size_t_) (m_pResPtr - m_pResStart);
    if( available <= 0 )
        { m_bEof = true; return _retread_t_(true, 0); }

    _size_t_ readbytes = available;
    if( readbytes > count ) readbytes = count;
    if( readbytes <= 0 )
        return _retread_t_(true, 0);

    memcpy( pBuffer, m_pResPtr, readbytes );
    m_pResPtr += readbytes;

    return _retread_t_(true, readbytes);
}



bool ResourceFile::directaccessReadMemory( const BYTE *& ptr, _size_t_& count )
{
    // In here, we always return success to indicate that this object
    // usually supports direct memory access.
    // Errors or EOF are returned through #ptr# set to NULL.

    ptr = NULL; count = 0;
    if( !isOpen() ) return true;

    if( m_pResPtr < m_pResStart ||
        m_pResPtr > m_pResStart + m_cResData ) return true;

    _size_t_ available = m_cResData - (_size_t_) (m_pResPtr - m_pResStart);
    if( available <= 0 )
        { m_bEof = true; return true; }

    ptr   = m_pResPtr;
    count = available;

    return true;
}



///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the seek position (will also be -1 on error). </returns>
_retseek_t_ ResourceFile::Seek( INT64 nOfs, SeekPosition nFrom )
{
    if( !isOpen() ) return _retseek_t_(interr (errFileNotOpen), -1);

    switch( nFrom )
    {
        case begin :
            m_pResPtr = m_pResStart + nOfs;
            //m_bEof = false;
            break;
        case end :
            m_pResPtr = m_pResStart + m_cResData - nOfs;
            //m_bEof = false;
            break;
        case current :
            if( m_pResPtr < m_pResStart || m_pResPtr > m_pResStart + m_cResData )
                return _retseek_t_(interr (errOutOfRange), -1);
            m_pResPtr += nOfs;
            break;
        default :
            return _retseek_t_(syserr (ERROR_INVALID_PARAMETER), -1);
    }

    //if( m_pResPtr < m_pResStart ||
    //    m_pResPtr > m_pResStart + m_cResData ) return -1;
    if( m_pResPtr < m_pResStart )
        return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1);  // filepos may also be beyond end-of-file, but not before start-of-file

    m_bEof = false; // always set to false, like C-runtime's #fseek()#.

    return _retseek_t_(true, m_pResPtr - m_pResStart);
}

///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the current file position (will also be -1 on error). </returns>
_retseek_t_ ResourceFile::Tell() const
{
    if( !isOpen() ) return _retseek_t_(interr (errFileNotOpen), -1);

    //if( m_pResPtr < m_pResStart ||
    //    m_pResPtr > m_pResStart + m_cResData ) return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1);
    if( m_pResPtr < m_pResStart )
        return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1);  // filepos can also be beyond end-of-file, but not before start-of-file

    return _retseek_t_(true, m_pResPtr - m_pResStart);
}

///<returns> NULL = error or EOF </returns>
const BYTE *ResourceFile::TellPtr() const
{
    if( !isOpen() ) return NULL;

    if( m_pResPtr < m_pResStart ||
        m_pResPtr >= m_pResStart + m_cResData ) return NULL;

    return m_pResPtr;
}




MemFile::MemFile()
{
    m_FileType   = FileT_MemFile;
    m_pMemStart  = NULL;
    m_pMemPtr    = NULL;
    m_cMemData   = 0;
    m_cAllocated = 0;
    m_bReference = false;
    m_cBlockSize = 1024;
}

///<param name="cBlockSize"> when new memory is (re)allocated,
///  it will be aligned to the next upper block boundary (in bytes). </param>
MemFile::MemFile( _size_t_ cBlockSize )
{
    m_FileType   = FileT_MemFile;
    m_pMemStart  = NULL;
    m_pMemPtr    = NULL;
    m_cMemData   = 0;
    m_cAllocated = 0;
    m_bReference = false;
    m_cBlockSize = cBlockSize > 0 ? cBlockSize : 16;
}

///<param name="memflags">
///  You can specify how to use the passed pointer
///  and its memory. You have the choice:
///  <ul>
///      <li> memf_CopyMemory: A copy of the memory will be created
///          internally. So the MemFile object will have its own
///          memory then. The memory passed in <i>pBuffer</i> can be
///          deleted after this ctor. This is the default </li>
///      <li> memf_TakeOverMemoryFromMalloc: Here, *no* copy of the
///          passed memory will be created. Instead, the MemFile object
///          will *take over* the memory. That means, it will free
///          the memory on its own (so that MemFile object occupies
///          the memory then). <br/>
///          IMPORTANT: For this to work correctly, the memory *must*
///          have been created using <b>malloc()</b> or <b>realloc()</b>.
///          You must not free this memory on your own any more. </li>
///      <li> memf_Reference: Here, also *no* copy of the memory will
///          be created. But the MemFile object will also *not* take
///          over the memory. So it will *not* occupy the memory which
///          means, it will not free the memory. Instead, it will
///          treat the memory as external memory. <br/>
///          IMPORTANT: The passed memory must stay valid as long
///          as the MemFile object exists or as long as this MemFile
///          object references the passed memory. </li>
///  </ul>
///</param>
///<param name="cBlockSize"> when new memory is (re)allocated,
///  it will be aligned to the next upper block boundary (in bytes). </param>
MemFile::MemFile( void *pBuffer, _size_t_ cBuffer, eMemFlags memflags, _size_t_ cBlockSize )
{
    m_FileType   = FileT_MemFile;

    if( !pBuffer || cBuffer <= 0 )
        { pBuffer = NULL; cBuffer = 0; memflags = memf_CopyMemory; }

    m_pMemStart  = NULL;
    m_pMemPtr    = NULL;
    m_cMemData   = 0;
    m_cAllocated = 0;
    m_bReference = false;
    m_cBlockSize = cBlockSize > 0 ? cBlockSize : 16;

    switch( memflags )
    {
        case memf_CopyMemory               : CopyMemoryFrom          ( pBuffer, cBuffer ); break;
        case memf_TakeOverMemoryFromMalloc : TakeOverMemoryFromMalloc( pBuffer, cBuffer ); break;
        case memf_Reference                : AttachReference         ( pBuffer, cBuffer ); break;
    }
}



///<remarks> See descriptions to MemFile's ctor and its parameter <i>memflags</i>. </remarks>
reterr_t MemFile::CopyMemoryFrom( void *pBuffer, _size_t_ cBuffer )
{
    if( !pBuffer || cBuffer <= 0 )
        return syserr (ERROR_INVALID_PARAMETER);

    Close();

    m_pMemStart  = NULL;
    m_pMemPtr    = NULL;
    m_cMemData   = 0;
    m_cAllocated = 0;
    m_bReference = false;

    reterr_t err = Realloc( cBuffer );
    if( !err ) return err;

    memcpy( m_pMemStart, pBuffer, cBuffer );

    return true;
}

///<remarks> See descriptions to MemFile's ctor and its parameter <i>memflags</i>. </remarks>
reterr_t MemFile::TakeOverMemoryFromMalloc( void *pBuffer, _size_t_ cBuffer )
{
    if( !pBuffer || cBuffer <= 0 )
        return syserr (ERROR_INVALID_PARAMETER);

    Close();

    m_pMemStart  = (PBYTE) pBuffer;
    m_pMemPtr    = (PBYTE) pBuffer;
    m_cMemData   = cBuffer;
    m_cAllocated = 0;
    m_bReference = false;

    return true;
}

///<remarks> See descriptions to MemFile's ctor and its parameter <i>memflags</i>. </remarks>
reterr_t MemFile::AttachReference( void *pBuffer, _size_t_ cBuffer )
{
    if( !pBuffer || cBuffer <= 0 )
        return syserr (ERROR_INVALID_PARAMETER);

    Close();

    m_pMemStart  = (PBYTE) pBuffer;
    m_pMemPtr    = (PBYTE) pBuffer;
    m_cMemData   = cBuffer;
    m_cAllocated = 0;
    m_bReference = true;

    return true;
}



void MemFile::Close()
{
    Realloc( 0 );
    m_bEof = false;
}

reterr_t MemFile::Flush()
{
    return isOpen();
}



///<param name="count"> number of bytes to read </param>
///<returns> returns the number of bytes read.
///  Use <b>.Succeeded()</b> to determine if reading failed. </returns>
_retread_t_ MemFile::Read( void *pBuffer, _size_t_ count )
{
    if( !isOpen() ) return _retread_t_(interr (errFileNotOpen), 0);
    if( !pBuffer ) return _retread_t_(syserr (ERROR_INVALID_PARAMETER), 0);

    if( m_pMemPtr < m_pMemStart ||
        m_pMemPtr > m_pMemStart + m_cMemData ) return _retread_t_(interr (errOutOfRange), 0);

    _size_t_ available = m_cMemData - (_size_t_) (m_pMemPtr - m_pMemStart);
    if( available <= 0 )
        { m_bEof = true; return _retread_t_(true, 0); }

    _size_t_ readbytes = available;
    if( readbytes > count ) readbytes = count;
    if( readbytes <= 0 )
        return _retread_t_(true, 0);

    memcpy( pBuffer, m_pMemPtr, readbytes );
    m_pMemPtr += readbytes;

    return _retread_t_(true, readbytes);
}

///<param name="count"> number of bytes to write </param>
///<remarks> You can write to memory even if it is only referenced.
///  In that case, failure will occur only when you try to write
///  outside of the memory boundaries. </remarks>
reterr_t MemFile::Write( const void *pBuffer, _size_t_ count )
{
    if( !isOpen() ) return interr (errFileNotOpen);
    if( count == 0 ) return true;
    if( !pBuffer ) return syserr (ERROR_INVALID_PARAMETER);

    if( m_pMemPtr < m_pMemStart ) return interr (errOutOfRange);

    // check if we need to grow the memory.
    _size_t_ newSize = (_size_t_) (m_pMemPtr - m_pMemStart) + count;
    if( newSize > m_cMemData )
    {
        reterr_t err = Realloc( newSize );
        if( !err ) return err;
        //m_bEof = (_size_t_) (m_pMemPtr - m_pMemStart) >= m_cMemData;
    }

    if( !m_pMemPtr ) return false;

    memcpy( m_pMemPtr, pBuffer, count );
    m_pMemPtr += count;

    return true;
}



bool MemFile::directaccessReadMemory( const BYTE *& ptr, _size_t_& count )
{
    // In here, we always return true to indicate that this object
    // usually supports direct memory access.
    // Errors or EOF are returned through #ptr# set to NULL.

    ptr = NULL; count = 0;
    if( !isOpen() ) return true;

    if( m_pMemPtr < m_pMemStart ||
        m_pMemPtr > m_pMemStart + m_cMemData ) return true;

    _size_t_ available = m_cMemData - (_size_t_) (m_pMemPtr - m_pMemStart);
    if( available <= 0 )
        { m_bEof = true; return true; }

    ptr   = m_pMemPtr;
    count = available;

    return true;
}

bool MemFile::directaccessWriteMemory( BYTE *& ptr, _size_t_ count )
{
    // In here, we always return true to indicate that this object
    // usually supports direct memory access.
    // Errors are returned through #ptr# set to NULL.

    ptr = NULL;
    if( !isOpen() ) return true;
    if( count == 0 ) return true;

    if( m_pMemPtr < m_pMemStart ) return true;

    // check if we need to grow the memory.
    _size_t_ newSize = (_size_t_) (m_pMemPtr - m_pMemStart) + count;
    if( newSize > m_cMemData )
    {
        if( !Realloc(newSize) ) return true;
        //m_bEof = (_size_t_) (m_pMemPtr - m_pMemStart) >= m_cMemData;
    }

    if( !m_pMemPtr ) return true;

    ptr = m_pMemPtr;
    m_pMemPtr += count;

    return true;
}



///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the seek position (will also be -1 on error). </returns>
_retseek_t_ MemFile::Seek( INT64 nOfs, SeekPosition nFrom )
{
    if( !isOpen() ) return _retseek_t_(interr (errFileNotOpen), -1);

    switch( nFrom )
    {
        case begin :
            m_pMemPtr = m_pMemStart + nOfs;
            //m_bEof = false;
            break;
        case end :
            m_pMemPtr = m_pMemStart + m_cMemData - nOfs;
            //m_bEof = false;
            break;
        case current :
            if( m_pMemPtr < m_pMemStart || m_pMemPtr > m_pMemStart + m_cMemData )
                return _retseek_t_(interr (errOutOfRange), -1);
            m_pMemPtr += nOfs;
            break;
        default :
            return _retseek_t_(syserr (ERROR_INVALID_PARAMETER), -1);
    }

    //if( m_pMemPtr < m_pMemStart ||
    //    m_pMemPtr >= m_pMemStart + m_cMemData ) return -1;
    if( m_pMemPtr < m_pMemStart )
        return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1);  // filepos can also be beyond end-of-file, but not before start-of-file

    m_bEof = false; // always set to false, like C-runtime's #fseek()#.

    return _retseek_t_(true, m_pMemPtr - m_pMemStart);
}

///<returns> use .Succeeded() to check if call suceeded,
///  and .getValue() to get the current file position (will also be -1 on error). </returns>
_retseek_t_ MemFile::Tell() const
{
    if( !isOpen() ) return _retseek_t_(interr (errFileNotOpen), -1);

    //if( m_pMemPtr < m_pMemStart ||
    //    m_pMemPtr >= m_pMemStart + m_cMemData ) return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1);
    if( m_pMemPtr < m_pMemStart )
        return _retseek_t_(syserr (ERROR_NEGATIVE_SEEK), -1);  // filepos can also be beyond end-of-file, but not before start-of-file

    return _retseek_t_(true, m_pMemPtr - m_pMemStart);
}

///<returns> NULL = error or EOF </returns>
BYTE *MemFile::TellPtr() const
{
    if( !isOpen() ) return NULL;

    if( m_pMemPtr < m_pMemStart ||
        m_pMemPtr >= m_pMemStart + m_cMemData ) return NULL;

    return m_pMemPtr;
}

reterr_t MemFile::setLength( UINT64 length )
{
    reterr_t err = Realloc( (_size_t_) length );
    if( !err ) return err;
    m_bEof = false;
    return true;
}



reterr_t MemFile::Realloc( _size_t_ newSize )
{
    if( newSize == m_cMemData ) return true;

    // Size 0 means to free the memory.
    if( newSize <= 0 )
    {
        if( newSize != 0 ) return syserr (ERROR_INVALID_PARAMETER);
        if( !m_bReference ) free( m_pMemStart );
        m_pMemStart  = NULL;
        m_pMemPtr    = NULL;
        m_cMemData   = 0;
        m_cAllocated = 0;
        m_bReference = false;
        return true;
    }

    if( m_bReference ) return interr (errMemFileReferenceNotReallocable);

    _size_t_ newAllocated = m_cBlockSize < 1 ?
        newSize : ((newSize + m_cBlockSize - 1) / m_cBlockSize) * m_cBlockSize;

    if( newAllocated == m_cAllocated )
        { m_cMemData = newSize; return true; }

    // reallocate new memory
    PBYTE pNewMem = (PBYTE) realloc( m_pMemStart, newAllocated );
    if( !pNewMem ) return syserr (ERROR_OUTOFMEMORY);

    // clear the new bytes
    if( newAllocated > m_cAllocated )
        memset( pNewMem + m_cAllocated, 0, newAllocated - m_cAllocated );

    m_pMemPtr    = pNewMem + (m_pMemPtr - m_pMemStart);
    m_pMemStart  = pNewMem;
    m_cMemData   = newSize;
    m_cAllocated = newAllocated;

    return true;
}




FilenameDialog::FilenameDialog( bool bSaveDialog,
                                PCTSTR pstrDefExt, PCTSTR pstrInitialFilename,
                                DWORD dwFlags, PCTSTR pstrFilters,
                                HWND hwndParent )
{
    m_bSaveDialog = bSaveDialog;
    m_hwndParent = hwndParent;

    memset( &m_ofn, 0, sizeof(m_ofn) );

    // Now we set an initial size of the OPENFILENAME struct.
    //  We always set it to the old size (Win v4.0), because with
    //  Win v5.0, the size increased and I do not want
    //  file dialog compilations to be limited to Win v5.0
    //  systems by default.
    // If you want the bigger size nevertheless, I recommend you
    //  set the #m_ofn.lStructSize# parameter yourself after
    //  having constructed a FilenameDialog object.
    #ifdef OPENFILENAME_SIZE_VERSION_400
    m_ofn.lStructSize = OPENFILENAME_SIZE_VERSION_400;
    #else
    m_ofn.lStructSize = sizeof(m_ofn);
    #endif

    m_ofn.hInstance = Get_Module_Instance();
    m_strDefaultExtension = pstrDefExt;
    m_strInitialFile = pstrInitialFilename;
    m_strFilters = pstrFilters;
    m_ofn.Flags = dwFlags;

    m_cResultFiles = 0;

    PrepareDialogKind( bSaveDialog, !dwFlags );
}



///<summary>
///  <b>PrepareDialogKind()</b> prepares the object for creating a
///  SaveFileName or OpenFileName dialog.
///  Additionally, appropriate OFN-Flags parameters can be chosen automatically.
///</summary>
///<param name="bSaveDialog"> if to create a Save- (true) or OpenFileName (false) dialog. </param>
///<param name="bOverwrite_OFN_Flags"> true = the <b>m_ofn.Flags</b> parameter will be
///  replaced with automatically chosen nice parameters suiting for the desired kind of dialog. </param>
void FilenameDialog::PrepareDialogKind( bool bSaveDialog, bool bOverwrite_OFN_Flags )
{
    m_bSaveDialog = bSaveDialog;

    if( bOverwrite_OFN_Flags )
    {
        m_ofn.Flags  = OFN_DONTADDTORECENT | OFN_ENABLESIZING | OFN_EXPLORER | OFN_LONGNAMES | OFN_HIDEREADONLY;
        m_ofn.Flags |= m_bSaveDialog ?
            OFN_OVERWRITEPROMPT | OFN_PATHMUSTEXIST : OFN_FILEMUSTEXIST;
        // Note: Use OFN_EXPLORER *always* because only then, the
        // resulting (multi-selected) files will be a stringlist
        // separated by '\0' chars.
        // Note: If you clear the OFN_EXPLORER flags yourself, then you will have
        // to handle the result filenames on your own if you chose to allow multiselection.
    }
}



///<summary>
///  <b>DoModal()</b> creates and opens the modal Filename dialog.
///</summary>
///<returns> True = OK button pressed,
///  false = Cancel button pressed, or an error occured. </returns>
///<remarks>
///  The results will be stored in <b>m_strResultDirectory</b>, <b>m_strResultFiles</b>, <b>m_cResultFiles</b>
///  and <b>m_strResultFirstFullfilename</b>. <b>m_strResultFiles</b> will be a pseudo-string-list
///  (using | char as separator and is terminated with 2 | chars) with the filenames but without path.
///  To get a particular filename out of <b>m_strResultFiles</b>, you can use <b>getResultFilenameAt()</b>
///  or <b>getResultFullFilenameAt()</b>, <br/>
///  or you can get it on your own using <b>StringT</b>'s <b>Findword()</b> method similar to following sample:
///  <pre>
///      FilenameDialog d;
///      d.AllowMultiselect();
///      d.DoModal();
///      String file = d.m_strResultFiles.Findword( 3, findword_SeparatorChars, _T("|") );
///  </pre>
///  In this sample, <b>file</b> will contain the 3rd filename, that is the one with index 2.
///  Why you have to pass 3 instead of 2 in that case, you can read in the descriptions to <b>StringT::Findword()</b>
///  and <b>StringT::Findnextword()</b>.
///</remarks>
bool FilenameDialog::DoModal()
{
    m_strResultDirectory.Empty();
    m_strResultFiles.Empty();
    m_cResultFiles = 0;
    m_strResultFirstFullfilename.Empty();

    m_ofn.hwndOwner = m_hwndParent;
    if( !m_ofn.hwndOwner ) m_ofn.hwndOwner = MoreMessageBox::FindNiceParent();

    PTSTR pstrlstFilters = NULL;
    if( !m_ofn.lpstrTitle && !m_strCaption.IsEmpty() ) m_ofn.lpstrTitle = m_strCaption;
    if( !m_ofn.lpstrDefExt && !m_strDefaultExtension.IsEmpty() ) m_ofn.lpstrDefExt = m_strDefaultExtension;
    if( m_strFilters.IsEmpty() ) AppendAllFilesFilter();
    if( !m_ofn.lpstrFilter && !m_strFilters.IsEmpty() ) m_ofn.lpstrFilter = pstrlstFilters = m_strFilters.toStrlst();
    if( !m_ofn.lpstrInitialDir && !m_strInitialDir.IsEmpty() && m_strInitialFile.DirectoryOf().StripPath().IsEmpty() )
        m_ofn.lpstrInitialDir = m_strInitialDir;

    bool bAllowMutisel = (m_ofn.Flags & OFN_ALLOWMULTISELECT) != 0;
    BOOL ret = FALSE;

    _size_t_ cchFileBuffer = bAllowMutisel ?
        2 * 1024 * 1024 : _MAX_PATH_;
    // In case of multiselection, we allocate a very large filebuffer
    // because the buffer will then contain a string list of all the selected files.
    PTSTR pstrFileBuffer = (PTSTR) malloc( sizeof(TCHAR) * cchFileBuffer );
    if( pstrFileBuffer )
    {
        _strcpy_( pstrFileBuffer, m_strInitialFile, cchFileBuffer );

        m_ofn.lpstrFile = pstrFileBuffer;
        m_ofn.nMaxFile  = cchFileBuffer;

        ret = m_bSaveDialog ?
            ::GetSaveFileName( &m_ofn ) : ::GetOpenFileName( &m_ofn );

        if( ret )
        {
            if( !(m_ofn.Flags & OFN_ALLOWMULTISELECT) )
            {
                // on single selection
                m_strResultFirstFullfilename = m_ofn.lpstrFile;
                m_cResultFiles = 1;
            }
            else
            if( m_ofn.Flags & OFN_EXPLORER )
            {
                // when we got a real string list
                m_cResultFiles = (int) _cstrlststrs_( m_ofn.lpstrFile );

                if( m_cResultFiles == 1 )
                {
                    m_strResultFirstFullfilename = m_ofn.lpstrFile;
                }
                else
                if( m_cResultFiles > 1 )
                {
                    m_strResultDirectory = m_ofn.lpstrFile;
                    --m_cResultFiles;

                    PCTSTR pstrNext = _fwd_to_next_strlst_str_( m_ofn.lpstrFile );
                    m_strResultFiles.fromStrlst( pstrNext );

                    m_strResultFirstFullfilename = m_strResultDirectory | pstrNext;
                }
            }
            else
            {
                // when we got a space-separated string list (a normal string where spaces are separators)
                PTSTR  pstrEnd = (PTSTR) _strspc_( m_ofn.lpstrFile );
                PCTSTR pstrNext = pstrEnd;

                if( pstrNext )
                    while( _isspace_(*pstrNext) ) ++pstrNext;

                if( _isstrempty_(pstrNext) )
                {
                    m_strResultFirstFullfilename = m_ofn.lpstrFile;
                    m_strResultFirstFullfilename.Trim();
                    m_cResultFiles = 1;
                }
                else
                {
                    if( pstrEnd ) *pstrEnd = (TCHAR) 0;

                    m_strResultDirectory = m_ofn.lpstrFile;

                    m_strResultFiles = pstrNext;
                    m_strResultFiles.Trim();
                    m_strResultFiles.Replace( _T(" \t\r\n"), _T('|') );

                    m_cResultFiles = _strcntchr_( m_strResultFiles, _T('|') ) + 1;
                    m_strResultFiles += _T("||");

                    pstrEnd = (PTSTR) _strspc_( pstrNext );
                    if( pstrEnd ) *pstrEnd = (TCHAR) 0;
                    m_strResultFirstFullfilename = m_strResultDirectory | pstrNext;
                }
            }

            if( m_cResultFiles == 1 && m_strResultDirectory.IsEmpty() )
            {
                m_strResultDirectory = m_strResultFirstFullfilename.DirectoryOf();
                m_strResultFiles = m_strResultFirstFullfilename.FilenameOf();
                m_strResultFiles += _T("||");
            }
        }

        free( pstrFileBuffer );
    }

    if( pstrlstFilters == m_ofn.lpstrFilter ) m_ofn.lpstrFilter = NULL;
    if( pstrlstFilters ) free( pstrlstFilters );

    return ret != FALSE;
}



String FilenameDialog::getResultFilenameAt( _idx_ index )
{
    if( index < 0 || index >= (_idx_) m_cResultFiles )
        return (PCTSTR) NULL;

    return m_strResultFiles.Findword( index + 1, findword_SeparatorChars, _T("|") );
}

String FilenameDialog::getResultFullFilenameAt( _idx_ index )
{
    String s = getResultFilenameAt( index );
    return !s.IsEmpty() ? m_strResultDirectory | s : s;
}




///<summary>
///  Reads text from the file stream. The returned string is an ASCII string. <br/>
///  The text in the file can be either ASCII or Unicode. <br/>
///  If the <b>m_EncodingType</b> member is <b>enctDetectByByteOrderMark</b>, the first
///  2 bytes in the stream read will be investigated and if identified as a unicode
///  start sequence, the <b>m_EncodingType</b> member will be set to unicode, otherwise
///  it will be set to ascii. In the case of unicode detection, also the byte order
///  will be set which depends on the byte order marker found.
///</summary>
///<param name="str"> returned ASCII string </param>
///<param name="countChars"> Maximum number of characters to read. This indicates the number
///  of chars as found in the file's text, independent from the <b>textmode_for_String</b>
///  param. So if <b>textmode_for_String</b> causes a transformation of the string,
///  the string can contain more or less characters as specified in the
///  <b>countChars</b> parameter. </param>
///<param name="textmode_for_String"> if to transform the string's content </param>
///<returns> the number of characters read from the text stream (not including
///  changes made by the <b>textmode_for_String</b> param). </returns>
_retread_t_ TextStream::ReadA( StringA& str, _size_t_ countChars, eTextMode textmode_for_String )
{
    str.Empty();
    if( !m_pFile ) return _retread_t_(interr (errFileNoStream), 0);
    if( countChars <= 0 ) return _retread_t_(true, 0);

    if( m_EncodingType == enctDetectByByteOrderMark )
    {
        File::RestoreSeek rs( *m_pFile );

        WCHAR ch;
        reterr_t err = m_pFile->ReadWchar( &ch );
        if( !err )
            m_EncodingType = enctAscii;
        else if( ch == (WCHAR) 0xfeff )
            { m_EncodingType = enctUnicode; m_ByteOrder = (eByteOrder) ByteOrder::isBigEndian(); rs.Dismiss(); }
        else if( ch == (WCHAR) 0xfffe )
            { m_EncodingType = enctUnicode; m_ByteOrder = (eByteOrder) ByteOrder::isLittleEndian(); rs.Dismiss(); }
        else
            m_EncodingType = enctAscii;
    }

    _retread_t_ ret( false, 0 );

    switch( m_EncodingType )
    {
        case enctAscii :
        {
            PSTR pdest = str.GetBuffer( countChars );
            if( !pdest ) return _retread_t_(syserr (ERROR_OUTOFMEMORY), 0);

            ret = m_pFile->Read( pdest, countChars * sizeof(char) );
            _size_t_ cnt = ret.getValue() / sizeof(char);

            pdest[cnt] = (char) 0;
            str.ReleaseBuffer();

            ret = _retread_t_(ret, cnt);

            if( !ret )
                { str.Empty(); return ret; }

            break;
        }

        case enctUnicode :
        {
            StringW s;
            PWSTR pdest = s.GetBuffer( countChars );
            if( !pdest ) return _retread_t_(syserr (ERROR_OUTOFMEMORY), 0);

            ret = m_pFile->Read( pdest, countChars * sizeof(WCHAR) );
            _size_t_ cnt = ret.getValue() / sizeof(WCHAR);

            if( (bool) ret && m_ByteOrder == (eByteOrder) ByteOrder::isLittleEndian() )
                for( _idx_ i = 0 ; i < (_idx_) cnt ; ++i )
                    pdest[i] = (WCHAR) swapBytesOfWord( (WORD) pdest[i] );

            pdest[cnt] = (WCHAR) 0;
            s.ReleaseBuffer();

            ret = _retread_t_(ret, cnt);

            if( !ret )
                { str.Empty(); return ret; }

            str = s;

            break;
        }

        default :
            return _retread_t_(interr (errFileReadFailed), 0);
    }

    switch( textmode_for_String )
    {
        case txtmdCRLF : str.MakeCRLF(); break;
        case txtmdLF   : str.MakeLF  (); break;
        case txtmdCR   : str.MakeCR  (); break;
    }

    return ret;
}

///<summary>
///  Reads text from the file stream. The returned string is an Unicode string. <br/>
///  The text in the file can be either ASCII or Unicode. <br/>
///  If the <b>m_EncodingType</b> member is <b>enctDetectByByteOrderMark</b>, the first
///  2 bytes in the stream read will be investigated and if identified as a unicode
///  start sequence, the <b>m_EncodingType</b> member will be set to unicode, otherwise
///  it will be set to ascii. In the case of unicode detection, also the byte order
///  will be set which depends on the byte order marker found.
///</summary>
///<param name="str"> returned Unicode string </param>
///<param name="countChars"> Maximum number of characters to read. This indicates the number
///  of chars as found in the file's text, independent from the <b>textmode_for_String</b>
///  param. So if <b>textmode_for_String</b> causes a transformation of the string,
///  the string can contain more or less characters as specified in the
///  <b>countChars</b> parameter. </param>
///<param name="textmode_for_String"> if to transform the string's content </param>
///<returns> the number of characters read from the text stream (not including
///  changes made by the <b>textmode_for_String</b> param). </returns>
_retread_t_ TextStream::ReadW( StringW& str, _size_t_ countChars, eTextMode textmode_for_String )
{
    str.Empty();
    if( !m_pFile ) return _retread_t_(interr (errFileNoStream), 0);
    if( countChars <= 0 ) return _retread_t_(true, 0);

    if( m_EncodingType == enctDetectByByteOrderMark )
    {
        File::RestoreSeek rs( *m_pFile );

        WCHAR ch;
        reterr_t err = m_pFile->ReadWchar( &ch );
        if( !err )
            m_EncodingType = enctAscii;
        else if( ch == (WCHAR) 0xfeff )
            { m_EncodingType = enctUnicode; m_ByteOrder = (eByteOrder) ByteOrder::isBigEndian(); rs.Dismiss(); }
        else if( ch == (WCHAR) 0xfffe )
            { m_EncodingType = enctUnicode; m_ByteOrder = (eByteOrder) ByteOrder::isLittleEndian(); rs.Dismiss(); }
        else
            m_EncodingType = enctAscii;
    }

    _retread_t_ ret( false, 0 );

    switch( m_EncodingType )
    {
        case enctAscii :
        {
            StringA s;
            PSTR pdest = s.GetBuffer( countChars );
            if( !pdest ) return _retread_t_(syserr (ERROR_OUTOFMEMORY), 0);

            ret = m_pFile->Read( pdest, countChars * sizeof(char) );
            _size_t_ cnt = ret.getValue() / sizeof(char);

            pdest[cnt] = (char) 0;
            s.ReleaseBuffer();

            ret = _retread_t_(ret, cnt);

            if( !ret )
                { str.Empty(); return ret; }

            str = s;

            break;
        }

        case enctUnicode :
        {
            PWSTR pdest = str.GetBuffer( countChars );
            if( !pdest ) return _retread_t_(syserr (ERROR_OUTOFMEMORY), 0);

            ret = m_pFile->Read( pdest, countChars * sizeof(WCHAR) );
            _size_t_ cnt = ret.getValue() / sizeof(WCHAR);

            if( (bool) ret && m_ByteOrder == (eByteOrder) ByteOrder::isLittleEndian() )
                for( _idx_ i = 0 ; i < (_idx_) cnt ; ++i )
                    pdest[i] = (WCHAR) swapBytesOfWord( (WORD) pdest[i] );

            pdest[cnt] = (WCHAR) 0;
            str.ReleaseBuffer();

            ret = _retread_t_(ret, cnt);

            if( !ret )
                { str.Empty(); return ret; }

            break;
        }

        default :
            return _retread_t_(interr (errFileReadFailed), 0);
    }

    switch( textmode_for_String )
    {
        case txtmdCRLF : str.MakeCRLF(); break;
        case txtmdLF   : str.MakeLF  (); break;
        case txtmdCR   : str.MakeCR  (); break;
    }

    return ret;
}

///<summary>
///  Reads text from the file stream. <br/>
///  The text in the file can be either ASCII or Unicode. <br/>
///  If the <b>m_EncodingType</b> member is <b>enctDetectByByteOrderMark</b>, the first
///  2 bytes in the stream read will be investigated and if identified as a unicode
///  start sequence, the <b>m_EncodingType</b> member will be set to unicode, otherwise
///  it will be set to ascii. In the case of unicode detection, also the byte order
///  will be set which depends on the byte order marker found.
///</summary>
///<param name="str"> returned string </param>
///<param name="countChars"> Maximum number of characters to read. This indicates the number
///  of chars as found in the file's text, independent from the <b>textmode_for_String</b>
///  param. So if <b>textmode_for_String</b> causes a transformation of the string,
///  the string can contain more or less characters as specified in the
///  <b>countChars</b> parameter. </param>
///<param name="textmode_for_String"> if to transform the string's content </param>
///<returns> the number of characters read from the text stream (not including
///  changes made by the <b>textmode_for_String</b> param). </returns>
_retread_t_ TextStream::Read( String& str, _size_t_ countChars, eTextMode textmode_for_String )
{
    #ifdef UNICODE
    return ReadW( reinterpret_cast<StringW&>(str), countChars, textmode_for_String );
    #else
    return ReadA( reinterpret_cast<StringA&>(str), countChars, textmode_for_String );
    #endif
}



_retread_t_ TextStream::ReadToEndA( StringA& str, eTextMode textmode_for_String, _size_t_ cIterationChars )
{
    str.Empty();
    if( !m_pFile ) return _retread_t_(interr (errFileNoStream), 0);

    if( cIterationChars <= 0 ) cIterationChars = 4096;
    _size_t_ cnt = 0;

    while( !m_pFile->Eof() )
    {
        StringA s;

        _retread_t_ ret = ReadA( s, cIterationChars, textmode_for_String );
        if( !ret )
            { str.Empty(); return ret; }

        str += s;
        cnt += ret.getValue();
    }

    return _retread_t_(true, cnt);
}

_retread_t_ TextStream::ReadToEndW( StringW& str, eTextMode textmode_for_String, _size_t_ cIterationChars )
{
    str.Empty();
    if( !m_pFile ) return _retread_t_(interr (errFileNoStream), 0);

    if( cIterationChars <= 0 ) cIterationChars = 4096;
    _size_t_ cnt = 0;

    while( !m_pFile->Eof() )
    {
        StringW s;

        _retread_t_ ret = ReadW( s, cIterationChars, textmode_for_String );
        if( !ret )
            { str.Empty(); return ret; }

        str += s;
        cnt += ret.getValue();
    }

    return _retread_t_(true, cnt);
}

_retread_t_ TextStream::ReadToEnd( String& str, eTextMode textmode_for_String, _size_t_ cIterationChars )
{
    #ifdef UNICODE
    return ReadToEndW( reinterpret_cast<StringW&>(str), textmode_for_String, cIterationChars );
    #else
    return ReadToEndA( reinterpret_cast<StringA&>(str), textmode_for_String, cIterationChars );
    #endif
}



reterr_t TextStream::WriteA( PCSTR pstr, _size_t_ countChars )
{
    if( !pstr || countChars <= 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( !m_pFile ) return interr (errFileNoStream);

    if( m_EncodingType == enctDetectByByteOrderMark )
        m_EncodingType = enctAscii;

    const _size_t_ cBlockChars = 4096;

    switch( m_EncodingType )
    {
        case enctAscii :
        {
            while( countChars > 0 )
            {
                _size_t_ cnt = tmin( countChars, cBlockChars );

                StringA s;
                s.Strcpy_unaligned( pstr, cnt );

                switch( m_TextMode )
                {
                    case txtmdCRLF : s.MakeCRLF(); break;
                    case txtmdLF   : s.MakeLF  (); break;
                    case txtmdCR   : s.MakeCR  (); break;
                    case txtmdDontChange : break;
                    default : return interr (errFileWriteFailed);
                }

                reterr_t err = m_pFile->Write( (PCSTR) s, s.length() * sizeof(char) );
                if( !err ) return err;

                pstr += cnt;
                countChars -= cnt;
            }

            break;
        }

        case enctUnicode :
        {
            bool bSwapBytes = m_ByteOrder == (eByteOrder) ByteOrder::isLittleEndian();

            if( m_bPrependByteOrderMarker )
            {
                WCHAR ch = bSwapBytes ?
                    (WCHAR) 0xfffe : (WCHAR) 0xfeff;

                reterr_t err = m_pFile->Write( &ch, sizeof(WCHAR) );
                if( !err ) return err;
            }

            while( countChars > 0 )
            {
                _size_t_ cnt = tmin( countChars, cBlockChars );

                StringA tmp;
                tmp.Strcpy_unaligned( pstr, cnt );

                StringW s = tmp;

                switch( m_TextMode )
                {
                    case txtmdCRLF : s.MakeCRLF(); break;
                    case txtmdLF   : s.MakeLF  (); break;
                    case txtmdCR   : s.MakeCR  (); break;
                    case txtmdDontChange : break;
                    default : return interr (errFileWriteFailed);
                }

                if( bSwapBytes )
                {
                    _size_t_ c = s.length();
                    PWSTR p = s.GetBuffer( c );
                    for( _idx_ i = 0 ; i < (_idx_) c ; ++i )
                        p[i] = (WCHAR) swapBytesOfWord( (WORD) p[i] );
                    s.ReleaseBuffer();
                }

                reterr_t err = m_pFile->Write( (PCWSTR) s, s.length() * sizeof(WCHAR) );
                if( !err ) return err;

                pstr += cnt;
                countChars -= cnt;
            }

            break;
        }

        default :
            return interr (errFileWriteFailed);
    }

    m_bPrependByteOrderMarker = false;

    return true;
}

reterr_t TextStream::WriteW( PCWSTR pstr, _size_t_ countChars )
{
    if( !pstr || countChars <= 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( !m_pFile ) return interr (errFileNoStream);

    if( m_EncodingType == enctDetectByByteOrderMark )
        m_EncodingType = enctUnicode;

    const _size_t_ cBlockChars = 4096;

    switch( m_EncodingType )
    {
        case enctAscii :
        {
            while( countChars > 0 )
            {
                _size_t_ cnt = tmin( countChars, cBlockChars );

                StringW tmp;
                tmp.Strcpy_unaligned( pstr, cnt );

                StringA s = tmp;

                switch( m_TextMode )
                {
                    case txtmdCRLF : s.MakeCRLF(); break;
                    case txtmdLF   : s.MakeLF  (); break;
                    case txtmdCR   : s.MakeCR  (); break;
                    case txtmdDontChange : break;
                    default : return interr (errFileWriteFailed);
                }

                reterr_t err = m_pFile->Write( (PCSTR) s, s.length() * sizeof(char) );
                if( !err ) return err;

                pstr += cnt;
                countChars -= cnt;
            }

            break;
        }

        case enctUnicode :
        {
            bool bSwapBytes = m_ByteOrder == (eByteOrder) ByteOrder::isLittleEndian();

            if( m_bPrependByteOrderMarker )
            {
                WCHAR ch = bSwapBytes ?
                    (WCHAR) 0xfffe : (WCHAR) 0xfeff;

                reterr_t err = m_pFile->Write( &ch, sizeof(WCHAR) );
                if( !err ) return err;
            }

            while( countChars > 0 )
            {
                _size_t_ cnt = tmin( countChars, cBlockChars );

                StringW s;
                s.Strcpy_unaligned( pstr, cnt );

                switch( m_TextMode )
                {
                    case txtmdCRLF : s.MakeCRLF(); break;
                    case txtmdLF   : s.MakeLF  (); break;
                    case txtmdCR   : s.MakeCR  (); break;
                    case txtmdDontChange : break;
                    default : return interr (errFileWriteFailed);
                }

                if( bSwapBytes )
                {
                    _size_t_ c = s.length();
                    PWSTR p = s.GetBuffer( c );
                    for( _idx_ i = 0 ; i < (_idx_) c ; ++i )
                        p[i] = (WCHAR) swapBytesOfWord( (WORD) p[i] );
                    s.ReleaseBuffer();
                }

                reterr_t err = m_pFile->Write( (PCWSTR) s, s.length() * sizeof(WCHAR) );
                if( !err ) return err;

                pstr += cnt;
                countChars -= cnt;
            }

            break;
        }

        default :
            return interr (errFileWriteFailed);
    }

    m_bPrependByteOrderMarker = false;

    return true;
}

reterr_t TextStream::Write( PCTSTR pstr, _size_t_ countChars )
{
    #ifdef UNICODE
    return WriteW( pstr, countChars );
    #else
    return WriteA( pstr, countChars );
    #endif
}




}
