/*
    File:               argonGdiObjects.cpp
    Description:        Gdi Object wrappers.

    Author:             Roman Komary
    created on:         15.08.2005
    last updated:       27.09.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary. All rights reserved.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonGdiObjects_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonGdiObjects.h"

namespace ArgonLibrary
{




bool GdiObj::DeleteObject()
{
    if( !m_hGdiObj ) return false;
    return ::DeleteObject( Detach() ) != FALSE;
}

bool GdiObj::Attach( HGDIOBJ hobj )
{
    if( m_hGdiObj || !hobj ) return false;

    m_hGdiObj = hobj;

    return true;
}

HGDIOBJ GdiObj::Detach()
{
    if( !HGdiObj() ) return NULL;

    HGDIOBJ hobj = m_hGdiObj;
    m_hGdiObj = NULL;

    return hobj;
}

///<summary> Like <b>Attach()</b>, but it will automatically delete the
///  <b>hobj</b> object, if the attaching failed. </summary>
///<returns> same as for <b>Attach()</b>. </returns>
bool GdiObj::AttachAutoDelete( HGDIOBJ hobj )
{
    bool ret = Attach( hobj );
    if( !ret && hobj ) ::DeleteObject( hobj );
    return ret;
}




bool Bitmap::GetBitmap( BITMAP *pbm ) const
{
    if( !m_hGdiObj || !pbm ) return false;
    return ::GetObject( m_hGdiObj, sizeof(BITMAP), pbm ) == sizeof(BITMAP);
}

SIZE Bitmap::GetSize() const
{
    SIZE siz = { 0, 0 };

    BITMAP bm;
    if( GetBitmap(&bm) )
    {
        siz.cx = bm.bmWidth;
        siz.cy = bm.bmHeight;
    }

    return siz;
}

bool Bitmap::Draw( HDC hdc, int x, int y, DWORD dwRop ) const
{
    if( !m_hGdiObj || !hdc ) return false;

    HDC hmemdc = CreateCompatibleDC( hdc );
    if( !hmemdc ) return false;
    HGDIOBJ oldobj = SelectObject( hmemdc, m_hGdiObj );

    SIZE siz = GetSize();
    bool ret = BitBlt( hdc, x, y, siz.cx, siz.cy,
                       hmemdc, 0, 0,
                       dwRop ) != FALSE;

    SelectObject( hmemdc, oldobj );
    DeleteDC( hmemdc );

    return ret;
}

bool Bitmap::DrawStretched( HDC hdc, int x, int y, int cx, int cy, DWORD dwRop ) const
{
    if( !m_hGdiObj || !hdc ) return false;

    HDC hmemdc = CreateCompatibleDC( hdc );
    if( !hmemdc ) return false;
    HGDIOBJ oldobj = SelectObject( hmemdc, m_hGdiObj );

    SIZE siz = GetSize();
    bool ret = StretchBlt( hdc, x, y, cx, cy,
                           hmemdc, 0, 0, siz.cx, siz.cy,
                           dwRop ) != FALSE;

    SelectObject( hmemdc, oldobj );
    DeleteDC( hmemdc );

    return ret;
}

///<summary>
///  <b>DrawPortion()</b> draws a portion of the bitmap into
///  the desired rectangle and performs stretching if needed.
///</summary>
///<param name="hdc"> DC </param>
///<param name="rectDest"> destination rect relative to the DC </param>
///<param name="rectSource"> source rect raltive to the bitmap </param>
///<param name="dwRop"> raster operation code </param>
///<returns> True on success, false on error. </returns>
bool Bitmap::DrawPortion( HDC hdc, RECT rectDest, RECT rectSource, DWORD dwRop ) const
{
    if( !m_hGdiObj || !hdc ) return false;

    HDC hmemdc = CreateCompatibleDC( hdc );
    if( !hmemdc ) return false;
    HGDIOBJ oldobj = SelectObject( hmemdc, m_hGdiObj );

    bool ret = StretchBlt( hdc, rectDest.left, rectDest.top, rectDest.right - rectDest.left, rectDest.bottom - rectDest.top,
                           hmemdc, rectSource.left, rectSource.top, rectSource.right - rectSource.left, rectSource.bottom - rectSource.top,
                           dwRop ) != FALSE;

    SelectObject( hmemdc, oldobj );
    DeleteDC( hmemdc );

    return ret;
}

///<summary>
///  <b>DrawTiled()</b> draws the Image multiple times to fill the destination rect.
///</summary>
///<param name="hdc"> DC </param>
///<param name="rectDest"> rect where to clip the drawing to; coords are relative to the DC </param>
///<param name="ptDCOffset"> coordinates of a (any) corner point of the bitmap relative to the DC ( may lie outside the <b>rectDest</b> rect ) </param>
///<param name="dwRop"> raster operation code </param>
///<returns> True on success, false on error. </returns>
bool Bitmap::DrawTiled( HDC hdc, RECT rectDest, POINT ptDCOffset, DWORD dwRop ) const
{
    if( !m_hGdiObj || !hdc || IsRectEmpty(&rectDest) ) return false;

    HDC hmemdc = CreateCompatibleDC( hdc );
    if( !hmemdc ) return false;
    HGDIOBJ oldobj = SelectObject( hmemdc, m_hGdiObj );

    SaveDC( hdc );
    IntersectClipRect( hdc, rectDest.left, rectDest.top, rectDest.right, rectDest.bottom );

    SIZE siz = GetSize();

    long x0 = ptDCOffset.x - rectDest.left;
    long y0 = ptDCOffset.y - rectDest.top;
    x0 = ( x0 + siz.cx * (1 + labs(x0) / siz.cx) ) % siz.cx;
    y0 = ( y0 + siz.cy * (1 + labs(y0) / siz.cy) ) % siz.cy;
    if( x0 > 0 ) x0 -= siz.cx;
    if( y0 > 0 ) y0 -= siz.cy;
    x0 += rectDest.left;
    y0 += rectDest.top;

    bool ret = true;
    for( long x = x0 ; x < rectDest.right && ret ; x += siz.cx )
        for( long y = y0 ; y < rectDest.bottom && ret ; y += siz.cy )
            if( !BitBlt( hdc, x, y, siz.cx, siz.cy,
                         hmemdc, 0, 0,
                         dwRop ) )
                ret = false;

    RestoreDC( hdc, -1 );

    SelectObject( hmemdc, oldobj );
    DeleteDC( hmemdc );

    return ret;
}

///<summary>
///  <b>DrawTiled()</b> draws the a portion of the Image multiple times to fill the destination rect.
///</summary>
///<param name="hdc"> DC </param>
///<param name="rectDest"> rect where to clip the drawing to; coords are relative to the DC </param>
///<param name="rectSource"> source rect raltive to the bitmap </param>
///<param name="ptDCOffset"> coordinates of a (any) corner point of the bitmap's portion relative to the DC ( may lie outside the <b>rectDest</b> rect ) </param>
///<param name="dwRop"> raster operation code </param>
///<returns> True on success, false on error. </returns>
bool Bitmap::DrawTiledPortion( HDC hdc, RECT rectDest, RECT rectSource, POINT ptDCOffset, DWORD dwRop ) const
{
    if( !m_hGdiObj || !hdc || IsRectEmpty(&rectDest) ) return false;

    HDC hmemdc = CreateCompatibleDC( hdc );
    if( !hmemdc ) return false;
    HGDIOBJ oldobj = SelectObject( hmemdc, m_hGdiObj );

    SaveDC( hdc );
    IntersectClipRect( hdc, rectDest.left, rectDest.top, rectDest.right, rectDest.bottom );

    SIZE siz = { rectSource.right - rectSource.left, rectSource.bottom - rectSource.top };

    long x0 = ptDCOffset.x - rectDest.left;
    long y0 = ptDCOffset.y - rectDest.top;
    x0 = ( x0 + siz.cx * (1 + labs(x0) / siz.cx) ) % siz.cx;
    y0 = ( y0 + siz.cy * (1 + labs(y0) / siz.cy) ) % siz.cy;
    if( x0 > 0 ) x0 -= siz.cx;
    if( y0 > 0 ) y0 -= siz.cy;
    x0 += rectDest.left;
    y0 += rectDest.top;

    bool ret = true;
    for( long x = x0 ; x < rectDest.right && ret ; x += siz.cx )
        for( long y = y0 ; y < rectDest.bottom && ret ; y += siz.cy )
            if( !StretchBlt( hdc, x, y, siz.cx, siz.cy,
                             hmemdc, rectSource.left, rectSource.top, siz.cx, siz.cy,
                             dwRop ) )
                ret = false;
    // Note: Idea for speedup:
    // Instead of calling StretchBlt all the time, we could create a 2nd memory DC
    // and StretchBlt the bitmap into this one and then use that memory DC to
    // draw it into the final DC in our double loop. Then the double loop would
    // need to call only BitBlt multiple times which would result in higher speed.

    RestoreDC( hdc, -1 );

    SelectObject( hmemdc, oldobj );
    DeleteDC( hmemdc );

    return ret;
}




bool Pen::GetLogPen( LOGPEN *plogpen ) const
{
    if( !m_hGdiObj || !plogpen ) return false;
    return ::GetObject( m_hGdiObj, sizeof(LOGPEN), plogpen ) == sizeof(LOGPEN);
}

bool Pen::GetExtLogPen( EXTLOGPEN *plogpen ) const
{
    if( !m_hGdiObj || !plogpen ) return false;
    return ::GetObject( m_hGdiObj, sizeof(EXTLOGPEN), plogpen ) == sizeof(EXTLOGPEN);
}




bool Brush::GetLogBrush( LOGBRUSH *plogbrush ) const
{
    if( !m_hGdiObj || !plogbrush ) return false;
    return ::GetObject( m_hGdiObj, sizeof(LOGBRUSH), plogbrush ) == sizeof(LOGBRUSH);
}




bool Font::CreateFont( int nHeight, int nWidth,
                       int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                       PCTSTR pszFacename )
{
    return CreateFont( nHeight, nWidth, 0, 0,
                       nWeight, bItalic, bUnderline, bStrikeOut,
                       pszFacename );
}

bool Font::CreateFont( int nHeight, int nWidth, int nEscapement, int nOrientation,
                       int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                       PCTSTR pszFacename )
{
    return CreateFont( nHeight, nWidth, nEscapement, nOrientation,
                       nWeight, bItalic, bUnderline, bStrikeOut,
                       ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
                       DEFAULT_QUALITY, DEFAULT_PITCH|FF_DONTCARE,
                       pszFacename );
}

bool Font::CreateFont( int nHeight, int nWidth, int nEscapement, int nOrientation,
                       int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                       DWORD dwCharSet, DWORD dwOutPrecision, DWORD dwClipPrecision,
                       DWORD dwQuality, DWORD dwPitchAndFamily,
                       PCTSTR pszFacename )
{
    HFONT hfont = ::CreateFont( nHeight, nWidth, nEscapement, nOrientation,
                                nWeight, bItalic, bUnderline, bStrikeOut,
                                dwCharSet, dwOutPrecision, dwClipPrecision,
                                dwQuality, dwPitchAndFamily,
                                pszFacename );

    return AttachAutoDelete( hfont );
}

///<param name="nPointSize10"> actually scaled 10x a font's point size.
///  It will be converted into logical height. </param>
///<param name="pszFacename"> If NULL, the GDI uses a device-independent typeface. </param>
///<param name="hdc"> device context to be used to convert
///  the height in <b>nPointSize</b> to logical units.
///  If NULL, a screen device context is used for the conversion. </param>
bool Font::CreatePointFont( int nPointSize10, PCTSTR pszFaceName, HDC hdc )
{
    return CreatePointFont( nPointSize10,
                            FW_NORMAL, false, false, false,
                            pszFaceName, hdc );
}

///<param name="nPointSize10"> actually scaled 10x a font's point size.
///  It will be converted into logical height. </param>
///<param name="pszFacename"> If NULL, the GDI uses a device-independent typeface. </param>
///<param name="hdc"> device context to be used to convert
///  the height in <b>nPointSize</b> to logical units.
///  If NULL, a screen device context is used for the conversion. </param>
bool Font::CreatePointFont( int nPointSize10,
                            int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                            PCTSTR pszFaceName, HDC hdc )
{
    LOGFONT lf;
    memset( &lf, 0, sizeof(lf) );

    lf.lfHeight = nPointSize10;
    lf.lfWeight = nWeight;
    lf.lfItalic = bItalic;
    lf.lfUnderline = bUnderline;
    lf.lfStrikeOut = bStrikeOut;
    lf.lfCharSet = DEFAULT_CHARSET;
    // Note: I *must* specify DEFAULT_CHARSET in here and not ANSI_CHARSET, or
    // else the "Terminal" font (and maybe also others) could not be chosen correctly.

    if( pszFaceName )
        _strcpy_( lf.lfFaceName, pszFaceName, _countof(lf.lfFaceName) );

    return CreatePointFontIndirect( &lf, hdc );
}

///<param name="nPointSize10"> actually scaled 10x a font's point size.
///  It will be converted into logical height. </param>
///<param name="pszFacename"> If NULL, the GDI uses a device-independent typeface. </param>
///<param name="hdc"> device context to be used to convert
///  the height in <b>nPointSize</b> to logical units.
///  If NULL, a screen device context is used for the conversion. </param>
///<remarks>
///  Most of the params are important only when the name of
///  the specified font cannot be found (if there is a name).
///  If there is no facename (NULL or empty string), then the
///  other params directly describe, how the font shall look like.
///</remarks>
bool Font::CreatePointFont( int nPointSize10, int nWidth, int nEscapement, int nOrientation,
                            int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                            DWORD dwCharSet, DWORD dwOutPrecision, DWORD dwClipPrecision,
                            DWORD dwQuality, DWORD dwPitchAndFamily,
                            PCTSTR pszFaceName, HDC hdc )
{
    LOGFONT lf;
    memset( &lf, 0, sizeof(lf) );

    lf.lfHeight = nPointSize10;
    lf.lfWidth = nWidth;
    lf.lfEscapement = nEscapement;
    lf.lfOrientation = nOrientation;
    lf.lfWeight = nWeight;
    lf.lfItalic = bItalic;
    lf.lfUnderline = bUnderline;
    lf.lfStrikeOut = bStrikeOut;
    lf.lfCharSet = (BYTE) dwCharSet;
    lf.lfOutPrecision = (BYTE) dwOutPrecision;
    lf.lfClipPrecision = (BYTE) dwClipPrecision;
    lf.lfQuality = (BYTE) dwQuality;
    lf.lfPitchAndFamily = (BYTE) dwPitchAndFamily;

    if( pszFaceName )
        _strcpy_( lf.lfFaceName, pszFaceName, _countof(lf.lfFaceName) );

    return CreatePointFontIndirect( &lf, hdc );
}

///<param name="plogfont"> Note: <b>plogfont->lfHeight</b> is interpreted as pointsize * 10 </param>
///<param name="hdc"> device context to be used to convert
///  the pointsize in <b>plogfont->lfHeight</b> to logical units.
///  If NULL, a screen device context is used for the conversion. </param>
///<remarks>
///  <b>plogfont->lfHeight</b> is interpreted as pointsize * 10
///</remarks>
bool Font::CreatePointFontIndirect( const LOGFONT *plogfont, HDC hdc )
{
    if( !plogfont ) return false;

    LOGFONT lf = *plogfont;
    if( lf.lfHeight )
        if( ( lf.lfHeight = EvalLogHeightFromPointSize( lf.lfHeight, hdc ) ) == 0 )
            return false;

    return CreateFontIndirect( &lf );
}

///<summary>
///  <b>CreateAdoptedFont()</b> creates a font similar to the
///  passed font, except for some parameters will differ.
///</summary>
///<param name="hfontFrom"> source font. If NULL, an error will be returned. </param>
///<param name="nWeight"> font weight (an FW_* value). Mostly will want to use FW_NORMAL or FW_BOLD. </param>
///<param name="bItalic"> italic font </param>
///<param name="bUnderline"> underline font </param>
///<param name="bStrikeOut"> strikeout font </param>
///<returns> True on success, false on error. </returns>
bool Font::CreateAdoptedFont( HFONT hfontFrom, int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut )
{
    if( !hfontFrom ) return false;

    LOGFONT lf;
    memset( &lf, 0, sizeof(lf) );
    if( ::GetObject( hfontFrom, sizeof(lf), &lf ) == 0 )
        return false;

    lf.lfWeight    = nWeight;
    lf.lfItalic    = bItalic;
    lf.lfUnderline = bUnderline;
    lf.lfStrikeOut = bStrikeOut;

    return CreateFontIndirect( &lf );
}

bool Font::GetLogFont( LOGFONT *plogfont ) const
{
    if( !m_hGdiObj || !plogfont ) return false;
    return ::GetObject( m_hGdiObj, sizeof(LOGFONT), plogfont ) == sizeof(LOGFONT);
}

bool Font::GetTextMetrics( HDC hdc, TEXTMETRIC *ptm ) const
{
    if( !ptm ) return false;

    HDC htmpdc = hdc ? NULL : ( hdc = ::GetDC(NULL) );
    HGDIOBJ oldfont = ::SelectObject( hdc, HFont() );

    bool ret = ::GetTextMetrics( hdc, ptm ) != FALSE;

    ::SelectObject( hdc, oldfont );
    if( !htmpdc ) ::ReleaseDC( NULL, htmpdc );

    return ret;
}

long Font::GetFontHeight( HDC hdc ) const
{
    TEXTMETRIC tm;
    memset( &tm, 0, sizeof(tm) );

    GetTextMetrics( hdc, &tm );

    return tm.tmHeight;
}

///<summary>
///  <b>EvalLogHeightFromPointSize()</b> evaluates the logical height
///  of a font (this will *always* be a negative value as needed
///  by the LOGFONT struct) from a font's pointsize (*10).
///</summary>
///<param name="nPointSize10"> actually scaled 10x the pointsize of a font </param>
///<param name="hdc"> device context to be used to convert
///  the height in <b>nPointSize</b> to logical units.
///  If NULL, a screen device context is used for the conversion. </param>
///<returns> logical font height. <br/>
///  This will *always* be a negative value as needed
///  by the LOGFONT struct. </returns>
long Font::EvalLogHeightFromPointSize( int nPointSize10, HDC hdc )
{
    HDC htmpdc = hdc ? NULL : ( hdc = ::GetDC(NULL) );

    POINT pt;
    pt.x = 0;
    pt.y = -MulDiv( nPointSize10, ::GetDeviceCaps(hdc, LOGPIXELSY), 720 );  // 72 points/inch, 10 decipoints/point
    if( !::DPtoLP( hdc, &pt, 1 ) ) return 0;

    POINT ptOrg = { 0, 0 };
    if( !::DPtoLP( hdc, &ptOrg, 1 ) ) return 0;

    long logheight = -labs( pt.y - ptOrg.y );

    if( !htmpdc ) ::ReleaseDC( NULL, htmpdc );
    return logheight;
}




bool Rgn::CreateMaskRgn( HDC hdc, HBITMAP hbmMask, COLORREF rgbRef, COLORREF rgbMasking, int scale )
{
    if( !hdc || !hbmMask ) return false;
    BITMAP bmp;
    if( ::GetObject(hbmMask, sizeof(bmp), &bmp) == 0 ) return false;

    if( bmp.bmWidth <= 0 || bmp.bmHeight <= 0 ) return false;

    RECT rc =  { 0, 0, bmp.bmWidth, bmp.bmHeight };
    return CreateMaskRgn( hdc, hbmMask, rc, -rc.left, -rc.top, rgbRef, rgbMasking, scale );
}

bool Rgn::CreateMaskRgn( HDC hdc, HBITMAP hbmMask, RECT rect, int xofs, int yofs, COLORREF rgbRef, COLORREF rgbMasking, int scale )
{
    if( ::IsRectEmpty(&rect) ) return false;
    if( !hdc || !hbmMask ) return false;

    BITMAP bmp;
    if( ::GetObject(hbmMask, sizeof(bmp), &bmp) == 0 ) return false;

    if( bmp.bmWidth <= 0 || bmp.bmHeight <= 0 ) return false;

    RECT rc = { 0, 0, bmp.bmWidth, bmp.bmHeight };
    RECT rTmp = rc;
    if( !IntersectRect( &rc, &rTmp, &rect ) )
        return false;

    HRGN hrgn = CreateMaskRgn( hdc, hbmMask,
                               NULL, NULL, NULL,
                               rc, xofs, yofs, rgbRef, rgbRef, rgbMasking, scale );

    return AttachAutoDelete( hrgn );
}

bool Rgn::CreateMaskRgnRange( HDC hdc, HBITMAP hbmMask, RECT rect, int xofs, int yofs, COLORREF rgbRefLo, COLORREF rgbRefHi, COLORREF rgbMasking, int scale )
{
    if( ::IsRectEmpty(&rect) ) return false;
    if( !hdc || !hbmMask ) return false;

    BITMAP bmp;
    if( ::GetObject(hbmMask, sizeof(bmp), &bmp) == 0 ) return false;

    if( bmp.bmWidth <= 0 || bmp.bmHeight <= 0 ) return false;

    RECT rc = { 0, 0, bmp.bmWidth, bmp.bmHeight };
    RECT rTmp = rc;
    if( !IntersectRect( &rc, &rTmp, &rect ) )
        return false;

    HRGN hrgn = CreateMaskRgn( hdc, hbmMask,
                               NULL, NULL, NULL,
                               rc, xofs, yofs, rgbRefLo, rgbRefHi, rgbMasking, scale );

    return AttachAutoDelete( hrgn );
}

///<summary>
///  <b>CreateMaskRgn()</b> creates a Region from a monochrome, color or
///  other kind of Bitmap. <br/>
///  One color is to be transparent (or it is also possible that no color is
///  transparent which will lead into a rectangular region). <br/>
///  The input can be a Bitmap (or DIBitmap) represented by a HBITMAP or a DIBitmap
///  represented by <b>pbmihMask</b>, <b>pMaskColTab</b> and <b>pMask</b>. <br/>
///  Note: The <b>hbmMask</b> can either be a normal bitmap (DDB) or a DIB bitmap.
///</summary>
///<param name="hdc"> DC to be used for creating a compatible DC which will be used
///  to select the Mask Bitmap into it. <br/>
///  NULL = <b>pbmihMask</b>, <b>pMaskColTab</b> and <b>pMask</b> will be used instead. </param>
///<param name="hbmMask"> Mask Bitmap <br/>
///  NULL = <b>pbmihMask</b>, <b>pMaskColTab</b> and <b>pMask</b> will be used instead. </param>
///<param name="pbmihMask"> BITMAPINFOHEADER struct containing the header
///  of the DIBitmap. <br/>
///  <b>pbmihMask</b> will be ignored if <b>hdc</b> and <b>hbmMask</b> are not NULL. <br/>
///  Supported Color Depths: <br/>
///      1, 4, 8, 16, 24, 32 <br/>
///  NOTE: Compressed bitmaps are *not* supported.
///  Also color-tables for color depths higher than 8 are *not* supported. </param>
///<param name="pMaskColTab"> pointer to colortable for the <b>pbmihMask</b> and <b>pMask</b>.
///  If NULL -> the colortable (if any used) will be imagined
///  directly behind the <b>pbmihMask</b> structure (depending on
///  the size <b>pbmihMask->biSize</b>). </param>
///<param name="pMask"> points to the DIBitmap bits containing the image data.
///  The format depends on the values of the <b>pbmihMask</b> BITMAPINFOHEADER struct.
///  <b>pMask</b> will be ignored if <b>hdc</b> and <b>hbmMask</b> are not NULL </param>
///<param name="rect"> rectangle inside the mask image to be the bounding of the final region <br/>
///  NOTE: These rect *must not* exceed the Mask Bitmap dimensions. </param>
///<param name="xoff"> x-offset: Origin offset of the new Region </param>
///<param name="yoff"> y-offset: Origin offset of the new Region </param>
///<param name="rgbRefLo"> low Reference Color </param>
///<param name="rgbRefHi"> high Reference Color <br/>
///  Pixels in the range from <b>rgbRefLo</b> to <b>rgbRefHi</b> (inclusive)
///  will *not* be included in the mask, so then will be transparent.
///  Set them both to one color if you want only that color to be transparent. </param>
///<param name="rgbMasking"> contains the bits of the r,g,b components of each pixel color
///  to use in the algorithm. Only those bits are relevant to detect
///  which pixel's color is to be interpreted as transparent.
///  Mostly you will set it to RGB(255, 255, 255) (which is 0x00ffffff)
///  which means that all bits are important (so that only pixels
///  which exactly match the rgbRefLo/Hi-range are made transparent). </param>
///<param name="scale"> will scale all the coordinates of the rects which form the region </param>
///<returns> the region handle or NULL on error </returns>
///<remarks>
///  The <b>hbmMask</b> can either be a normal bitmap (DDB) or a DIB bitmap.
///</remarks>
HRGN Rgn::CreateMaskRgn( HDC hdc, HBITMAP hbmMask, const BITMAPINFOHEADER *pbmihMask, const RGBQUAD *pMaskColTab, const BYTE *pMask,
                         RECT rect, int xoff, int yoff, COLORREF rgbRefLo, COLORREF rgbRefHi, COLORREF rgbMasking,
                         int scale )
{
    if( scale <= 0 ) return NULL;

    HDC hdc2 = ( hdc && hbmMask ) ? CreateCompatibleDC( hdc ) : NULL;
    if( !hdc2 && (!pbmihMask || !pMask) ) return NULL;

    if( pbmihMask && !hdc2 )
        if( pbmihMask->biCompression != BI_RGB ) {
            if(hdc2) DeleteDC( hdc2 );
            return NULL;  // no compressed and no bitfield bitmaps supported
        }

    HGDIOBJ hGdiObj_Orig_Bitmap = hdc2 ? SelectObject( hdc2, hbmMask ) : NULL;
    const BITMAPINFOHEADER *pbmih = pbmihMask;
    const RGBQUAD *pcoltab = pMaskColTab ? pMaskColTab : ( pbmih ? (RGBQUAD*) (((PBYTE) pbmih) + pbmih->biSize) : NULL );
    const BYTE *ppic = pMask;

    rgbRefLo &= rgbMasking;
    rgbRefHi &= rgbMasking;

    int maxrects = 1024; // should not be larger than 4000 (or so) because else under Win95/98, 'ExtCreateRegion()' could fail.
    int crects = 0;

    PBYTE pData = (PBYTE) malloc( sizeof(RGNDATAHEADER) + sizeof(RECT) * maxrects );
    if( !pData )
    {
        if( hdc2 ) {
            SelectObject( hdc2, hGdiObj_Orig_Bitmap );
            DeleteDC( hdc2 );
        }
        return NULL;
    }

    PRECT pRects = (PRECT) ((PRGNDATA) pData)->Buffer;

    RECT rc = { 0, 0, 0, 0 };   // current Rect
    bool inrect = false;
    // 'cxalign', 'bTopDown' are necessary *only* for 'pbmihMask' and 'pMask' but *not* for 'hbmMask'
    int cxalign = 0;  // contains size of one pixel line in bytes
    if( pbmih )
        cxalign = ((( (((int) pbmih->biWidth) * (int) pbmih->biBitCount + 7) / 8 ) + 3) / 4) * 4;
    bool bTopDown = pbmih ? (pbmih->biHeight < 0) : false;
    RECT rcBound;
    SetRectEmpty( &rcBound );

    HRGN hRgn=NULL;

    for( long y = rect.top ; y < rect.bottom ; ++y )
    {
        long yy = pbmih ? ( bTopDown ? y : ((long) pbmih->biHeight) - y - 1 ) : y;
        for( long x = rect.left ; x < rect.right ; ++x )
        {
            bool newrect = false;   // if the current Rect contains a new Rect
            COLORREF rgb;

            if( hdc2 )
            {
                rgb = GetPixel( hdc2, x, y );
            }
            else
            {
                switch( pbmih->biBitCount )
                {
                    case 1 :
                    {
                        BYTE ix8 = *((PBYTE) &ppic[(x >> 3) + cxalign * yy]);
                        int ix = ( ix8 >> ((~x) & 7) ) & 1;    // '(~x)&7' is the same like '7-(x&7)'
                        const RGBQUAD *pix = &pcoltab[ix];
                        rgb = RGB( pix->rgbRed, pix->rgbGreen, pix->rgbBlue );
                        break;
                    }
                    case 4 :
                    {
                        BYTE ix2 = *((PBYTE) &ppic[(x >> 1) + cxalign * yy]);
                        int ix=( (x & 1) ? ix2 : (ix2 >> 4) ) & 0xf;
                        const RGBQUAD *pix = &pcoltab[ix];
                        rgb = RGB( pix->rgbRed, pix->rgbGreen, pix->rgbBlue );
                        break;
                    }
                    case 8 :
                    {
                        const RGBQUAD *pix = &pcoltab[*((PBYTE) &ppic[x + cxalign * yy])];
                        rgb = RGB( pix->rgbRed, pix->rgbGreen, pix->rgbBlue );
                        break;
                    }
                    case 16 :
                    {
                        WORD pix = *((WORD*) &ppic[sizeof(WORD) * x + cxalign * yy]);
                        rgb = RGB( (pix >> 10) & 0x1f,
                                   (pix >>  5) & 0x1f,
                                    pix        & 0x1f );
                        break;
                    }
                    case 24 :
                    {
                        const RGBTRIPLE *pix = (RGBTRIPLE*) &ppic[sizeof(RGBTRIPLE) * x + cxalign * yy];
                        rgb = RGB( pix->rgbtRed, pix->rgbtGreen, pix->rgbtBlue );
                        break;
                    }
                    case 32 :
                    {
                        const RGBQUAD *pix = (RGBQUAD*) &ppic[sizeof(RGBQUAD) * x + cxalign * yy];
                        rgb = RGB( pix->rgbRed, pix->rgbGreen, pix->rgbBlue );
                        break;
                    }
                }
            }

            rgb &= rgbMasking;

            bool bMakeTransparentPixel;
            if( rgbRefLo == rgbRefHi ) {
                bMakeTransparentPixel = ( rgb == rgbRefLo );
            } else {
                BYTE r = GetRValue(rgb), g = GetGValue(rgb), b = GetBValue(rgb);
                bMakeTransparentPixel = (( GetRValue(rgbRefLo) <= r && r <= GetRValue(rgbRefHi) ) &&
                                         ( GetGValue(rgbRefLo) <= g && g <= GetGValue(rgbRefHi) ) &&
                                         ( GetBValue(rgbRefLo) <= b && b <= GetBValue(rgbRefHi) ));
            }

            if( inrect ) {
                if( bMakeTransparentPixel || x == rect.right - 1 ) {
                    inrect = false;
                    rc.right = x + xoff + (bMakeTransparentPixel ? 0 : 1);
                    rc.bottom = y + yoff + 1;
                    newrect = true;
                }
            } else {
                if( !bMakeTransparentPixel ) {
                    inrect = true;
                    rc.left = x + xoff;
                    rc.top = y + yoff;
                    if( x == rect.right - 1 ) {
                        inrect = false;
                        rc.right = x + xoff + 1;
                        rc.bottom = y + yoff + 1;
                        newrect = true;
                    }
                }
            }

            if( newrect )
            {
                // New Rect will be added now to the array.
                if( crects >= maxrects )
                {
                    for( int i = 0 ; i < crects ; ++i )
                        { pRects[i].left *= scale; pRects[i].top *= scale; pRects[i].right *= scale; pRects[i].bottom *= scale; }
                    rcBound.left *= scale; rcBound.top *= scale; rcBound.right *= scale; rcBound.bottom *= scale;

                    PRGNDATA pRgnData = (PRGNDATA) pData;
                    pRgnData->rdh.dwSize = sizeof(RGNDATAHEADER);
                    pRgnData->rdh.iType = RDH_RECTANGLES;
                    pRgnData->rdh.nCount = crects;
                    pRgnData->rdh.nRgnSize = sizeof(RECT) * crects;
                    pRgnData->rdh.rcBound = rcBound;

                    HRGN hrgn = ::ExtCreateRegion( NULL, sizeof(RGNDATAHEADER) + sizeof(RECT) * crects, pRgnData );

                    if( hrgn )
                    {
                        if( hRgn ) {
                            ::CombineRgn( hRgn, hRgn, hrgn, RGN_OR );
                            ::DeleteObject( hrgn );
                        } else
                            hRgn = hrgn;
                    }

                    crects = 0;
                    SetRectEmpty( &rcBound );
                }

                pRects[ crects++ ] = rc;
                RECT rTmp = rcBound;
                UnionRect( &rcBound, &rTmp, &rc );  // using 'rTmp' just because I am not sure if we could pass the output rect as input rect
            }
        }
    }

    if( crects > 0 )
    {
        for( int i = 0 ; i < crects ; ++i )
            { pRects[i].left *= scale; pRects[i].top *= scale; pRects[i].right *= scale; pRects[i].bottom *= scale; }
        rcBound.left *= scale; rcBound.top *= scale; rcBound.right *= scale; rcBound.bottom *= scale;

        PRGNDATA pRgnData = (PRGNDATA) pData;
        pRgnData->rdh.dwSize = sizeof(RGNDATAHEADER);
        pRgnData->rdh.iType = RDH_RECTANGLES;
        pRgnData->rdh.nCount = crects;
        pRgnData->rdh.nRgnSize = sizeof(RECT) * crects;
        pRgnData->rdh.rcBound = rcBound;

        HRGN hrgn = ::ExtCreateRegion( NULL, sizeof(RGNDATAHEADER) + sizeof(RECT) * crects, pRgnData );

        if( hrgn )
        {
            if( hRgn ) {
                ::CombineRgn( hRgn, hRgn, hrgn, RGN_OR );
                ::DeleteObject( hrgn );
            } else
                hRgn = hrgn;
        }

        crects = 0;
        SetRectEmpty( &rcBound );
    }

    free( pData );

    if( hdc2 )
    {
        SelectObject( hdc2, hGdiObj_Orig_Bitmap );
        DeleteDC( hdc2 );
    }

    return hRgn;
}

bool Rgn::MoveRgnTo( int x, int y )
{
    if( !HRgn() ) return false;

    RECT rc = { 0, 0, 0, 0 };
    GetRgnBox( &rc );

    return OffsetRgn( x - rc.left, y - rc.top ) != ERROR;
}

///<summary> unlike <b>CopyRgn()</b>, this method will first create a HRGN to copy <b>pRgnSrc</b> into. </summary>
bool Rgn::QCopyRgn( Rgn *pRgnSrc )
{
    if( !pRgnSrc ) return false;

    if( !HRgn() )
        if( !CreateRectRgn( 0, 0, 1, 1 ) ) return false;

    return CopyRgn( pRgnSrc ) != ERROR;
}

///<summary> unlike <b>GetWindowRgn()</b>, this method will first create a HRGN to copy the window's region into. </summary>
bool Rgn::QGetWindowRgn( HWND hwnd )
{
    if( !hwnd ) return false;

    if( !HRgn() )
        if( !CreateRectRgn( 0, 0, 1, 1 ) ) return false;

    return GetWindowRgn( hwnd ) != ERROR;
}




}
