/*
    File:               argonGdiObjects.h
    Description:        Gdi Object wrappers.

    Author:             Roman Komary
    created on:         15.08.2005
    last updated:       27.09.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary. All rights reserved.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonGdiObjects_h
#define __argonGdiObjects_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonInstance.h"
    #include "argonString.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




#ifndef _countof
#define _countof(array) (sizeof(array) / sizeof((array)[0]))
#endif



///<summary>
///  Class <b>GdiObj</b> holds a Gdi Object. It is used as base class for
///  the different specializations of Gdi Objects. <br/>
///  This class will delete the Gdi Object it holds on its destruction.
///  If you want to prevent that, call <b>Detach()</b> first.
///</summary>
class GdiObj
{
    protected :
        HGDIOBJ m_hGdiObj;

    public :
        GdiObj() { m_hGdiObj = NULL; }
        virtual ~GdiObj() { DeleteObject(); }

        operator HGDIOBJ() const { return this ? m_hGdiObj : NULL; }
        HGDIOBJ  HGdiObj() const { return this ? m_hGdiObj : NULL; }

    // creation, attaching
        bool    CreateStockObject( int nIndex ) { m_hGdiObj = ::GetStockObject( nIndex ); return m_hGdiObj != NULL; }
        bool    DeleteObject();
        bool    Attach( HGDIOBJ hobj );
        HGDIOBJ Detach();

    // attributes, resetting
        int     GetObject( int nCount, void *pObject ) const { return ::GetObject( m_hGdiObj, nCount, pObject ); }
        DWORD   GetObjectType() const { return ::GetObjectType( m_hGdiObj ); }
        bool    UnrealizeObject() { return ::UnrealizeObject( m_hGdiObj ) != FALSE; }

    // compare
        bool operator ==( const GdiObj& obj ) const { return obj.HGdiObj() == m_hGdiObj; }
        bool operator !=( const GdiObj& obj ) const { return obj.HGdiObj() != m_hGdiObj; }

    protected :
        bool    AttachAutoDelete( HGDIOBJ hobj );
};



class Bitmap : public GdiObj
{
    public :
        Bitmap() {  }
        Bitmap( HBITMAP hbm ) { AttachAutoDelete( hbm ); }

        operator HBITMAP() const { return (HBITMAP) HGdiObj(); }
        HBITMAP  HBitmap() const { return (HBITMAP) HGdiObj(); }

    // creation
        bool LoadBitmap( UINT uIDResource )                        { return LoadBitmap( MAKEINTRESOURCE(uIDResource) ); }
        bool LoadBitmap( PCTSTR pszResourceName )                  { return LoadBitmap( Get_Resource_Instance(), pszResourceName ); }
        bool LoadBitmap( HMODULE hModule, PCTSTR pszResourceName ) { return AttachAutoDelete( ::LoadBitmap( hModule, pszResourceName ) ); }
        bool LoadOEMBitmap( UINT uIDBitmap )                       { return AttachAutoDelete( ::LoadBitmap( NULL, MAKEINTRESOURCE(uIDBitmap) ) ); } // for OBM_/OCR_/OIC_

    // attributes
        bool GetBitmap( BITMAP *pbm ) const;
        SIZE GetSize() const;

    // drawing
        bool Draw( HDC hdc, DWORD dwRop = SRCCOPY ) const           { return Draw( hdc, 0, 0, dwRop ); }
        bool Draw( HDC hdc, POINT pt, DWORD dwRop = SRCCOPY ) const { return Draw( hdc, pt.x, pt.y, dwRop ); }
        bool Draw( HDC hdc, int x, int y, DWORD dwRop = SRCCOPY ) const;
        bool DrawStretched( HDC hdc, int x, int y, int cx, int cy, DWORD dwRop = SRCCOPY ) const;
        bool DrawStretched( HDC hdc, RECT rect, DWORD dwRop = SRCCOPY ) const { return DrawStretched( hdc, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top, dwRop ); }
        bool DrawPortion( HDC hdc, RECT rectDest, RECT rectSource, DWORD dwRop = SRCCOPY ) const;
        bool DrawTiled( HDC hdc, RECT rectDest, POINT ptDCOffset, DWORD dwRop = SRCCOPY ) const;
        bool DrawTiledPortion( HDC hdc, RECT rectDest, RECT rectSource, POINT ptDCOffset, DWORD dwRop = SRCCOPY ) const;
};



class Pen : public GdiObj
{
    public :
        Pen() {  }
        Pen( HPEN hpen ) { AttachAutoDelete( hpen ); }
        Pen( COLORREF color ) { CreatePen( color ); }
        Pen( int nPenStyle, int nWidth, COLORREF color ) { CreatePen( nPenStyle, nWidth, color ); }
        Pen( int nPenStyle, int nWidth, const LOGBRUSH *pLogBrush,
             int nStyleCount = 0, const DWORD *pStyle = NULL )
            { CreatePen( nPenStyle, nWidth, pLogBrush, nStyleCount, pStyle ); }

        operator HPEN() const { return (HPEN) HGdiObj(); }
        HPEN     HPen() const { return (HPEN) HGdiObj(); }

    // creation:
        bool CreatePen( COLORREF color ) { return CreatePen( PS_SOLID, 0, color ); }
        bool CreatePen( int nPenStyle, int nWidth, COLORREF color ) { return AttachAutoDelete( ::CreatePen( nPenStyle, nWidth, color ) ); }
        bool CreatePen( int nPenStyle, int nWidth, const LOGBRUSH *pLogBrush,
                        int nStyleCount = 0, const DWORD *pStyle = NULL )
            { return AttachAutoDelete( ::ExtCreatePen( nPenStyle, nWidth, pLogBrush, nStyleCount, pStyle ) ); }
        bool CreatePenIndirect( const LOGPEN *pLogPen ) { return AttachAutoDelete( ::CreatePenIndirect( pLogPen ) ); }

    // attributes
        bool GetLogPen( LOGPEN *plogpen ) const;
        bool GetExtLogPen( EXTLOGPEN *plogpen ) const;
};



class Brush : public GdiObj
{
    public :
        Brush() {  }
        Brush( HBRUSH hbr ) { AttachAutoDelete( hbr ); }
        Brush( COLORREF color ) { CreateSolidBrush( color ); }
        Brush( int nHatchIndex, COLORREF color ) { CreateHatchBrush( nHatchIndex, color ); }
        Brush( const Bitmap *pPatternBmp ) { CreatePatternBrush( pPatternBmp ); }

        operator HBRUSH() const { return (HBRUSH) HGdiObj(); }
        HBRUSH   HBrush() const { return (HBRUSH) HGdiObj(); }

    // creation:
        bool CreateSolidBrush( COLORREF color ) { return AttachAutoDelete( ::CreateSolidBrush( color ) ); }
        bool CreateHatchBrush( int nHatchIndex, COLORREF color ) { return AttachAutoDelete( ::CreateHatchBrush( nHatchIndex, color ) ); }
        bool CreateBrushIndirect( const LOGBRUSH *pLogBrush ) { return AttachAutoDelete( ::CreateBrushIndirect( pLogBrush ) ); }
        bool CreatePatternBrush( HBITMAP hPatternBmp ) { return AttachAutoDelete( ::CreatePatternBrush( hPatternBmp ) ); }
        bool CreatePatternBrush( const Bitmap *pPatternBmp ) { return AttachAutoDelete( ::CreatePatternBrush( *pPatternBmp ) ); }
        bool CreateDIBPatternBrush( HGLOBAL hPackedDIB, UINT uUsage ) { return AttachAutoDelete( ::CreateDIBPatternBrush( hPackedDIB, uUsage ) ); }
        bool CreateDIBPatternBrush( const void *pPackedDIB, UINT uUsage ) { return AttachAutoDelete( ::CreateDIBPatternBrushPt( pPackedDIB, uUsage ) ); }
        bool CreateSysColorBrush( int nIndex ) { return AttachAutoDelete( ::GetSysColorBrush( nIndex ) ); }

    // attributes
        bool GetLogBrush( LOGBRUSH *plogbrush ) const;
};



class Font : public GdiObj
{
    public :
        Font() {  }
        Font( HFONT hfont ) { AttachAutoDelete( hfont ); }

        operator HFONT() const { return (HFONT) HGdiObj(); }
        HFONT    HFont() const { return (HFONT) HGdiObj(); }

    // creation
        bool CreateFont( int nHeight, int nWidth,
                         int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                         PCTSTR pszFacename );
        bool CreateFont( int nHeight, int nWidth, int nEscapement, int nOrientation,
                         int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                         PCTSTR pszFacename );
        bool CreateFont( int nHeight, int nWidth, int nEscapement, int nOrientation,
                         int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                         DWORD dwCharSet, DWORD dwOutPrecision, DWORD dwClipPrecision,
                         DWORD dwQuality, DWORD dwPitchAndFamily,
                         PCTSTR pszFacename );
        bool CreateFontIndirect( const LOGFONT *plogfont ) { return AttachAutoDelete( ::CreateFontIndirect( plogfont ) ); }
        bool CreatePointFont( int nPointSize10, PCTSTR pszFacename, HDC hdc = NULL );
        bool CreatePointFont( int nPointSize10,
                              int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                              PCTSTR pszFacename, HDC hdc = NULL );
        bool CreatePointFont( int nPointSize10, int nWidth, int nEscapement, int nOrientation,
                              int nWeight, bool bItalic, bool bUnderline, bool bStrikeOut,
                              DWORD dwCharSet, DWORD dwOutPrecision, DWORD dwClipPrecision,
                              DWORD dwQuality, DWORD dwPitchAndFamily,
                              PCTSTR pszFacename, HDC hdc = NULL );
        bool CreatePointFontIndirect( const LOGFONT *plogfont, HDC hdc = NULL );
        bool CreateAdoptedFont( HFONT hfontFrom,
                                int nWeight = FW_NORMAL, bool bItalic = false, bool bUnderline = false, bool bStrikeOut = false );

    // attributes
        bool GetLogFont( LOGFONT *plogfont ) const;
        bool GetTextMetrics( TEXTMETRIC *ptm ) const { return GetTextMetrics( NULL,ptm ); }
        bool GetTextMetrics( HDC hdc, TEXTMETRIC *ptm ) const;
        long GetFontHeight( HDC hdc = NULL ) const;
        static long EvalLogHeightFromPointSize( int nPointSize10, HDC hdc = NULL );
};



class Palette : public GdiObj
{
    public :
        Palette() {  }
        Palette( HPALETTE hpal ) { AttachAutoDelete( hpal ); }

        operator HPALETTE() const { return (HPALETTE) HGdiObj(); }
        HPALETTE HPalette() const { return (HPALETTE) HGdiObj(); }
};



class Rgn : public GdiObj
{
    public :
        Rgn() {  }
        Rgn( HRGN hrgn ) { AttachAutoDelete( hrgn ); }

        operator HRGN() const { return (HRGN) HGdiObj(); }
        HRGN     HRgn() const { return (HRGN) HGdiObj(); }

    // creation
        bool CreateRectRgn( int left, int top, int right, int bottom ) { return AttachAutoDelete( ::CreateRectRgn( left, top, right, bottom ) ); }
        bool CreateRectRgnIndirect( const RECT *prect ) { return AttachAutoDelete( ::CreateRectRgnIndirect( prect ) ); }
        bool CreateEllipticRgn( int left, int top, int right, int bottom ) { return AttachAutoDelete( ::CreateEllipticRgn( left, top, right, bottom ) ); }
        bool CreateEllipticRgnIndirect( const RECT *prect ) { return AttachAutoDelete( ::CreateEllipticRgnIndirect( prect ) ); }
        bool CreatePolygonRgn( const POINT *pPts, int count, int mode ) { return AttachAutoDelete( ::CreatePolygonRgn( pPts, count, mode ) ); }
        bool CreatePolyPolygonRgn( const POINT *pPts, const int *pPolyCounts, int count, int polyfillmode ) { return AttachAutoDelete( ::CreatePolyPolygonRgn( pPts, pPolyCounts, count, polyfillmode ) ); }
        bool CreateRoundRectRgn( int left, int top, int right, int bottom, int widthellipse, int heightellipse ) { return AttachAutoDelete( ::CreateRoundRectRgn( left, top, right, bottom, widthellipse, heightellipse ) ); }
        bool CreateRoundRectRgn( RECT rect, int widthellipse, int heightellipse ) { return CreateRoundRectRgn( rect.left, rect.top, rect.right, rect.bottom, widthellipse, heightellipse ); }
        bool CreateFromPath( HDC hdc ) { return AttachAutoDelete( ::PathToRegion( hdc ) ); }
        bool CreateFromData( const XFORM *pXForm, int cBytes, const RGNDATA *pRgnData ) { return AttachAutoDelete( ::ExtCreateRegion( pXForm, cBytes, pRgnData ) ); }
        bool CreateMaskRgn( HDC hdc, HBITMAP hbmMask, COLORREF rgbRef, COLORREF rgbMasking = 0x00ffffff, int scale = 1 );
        bool CreateMaskRgn( HDC hdc, HBITMAP hbmMask, RECT rect, int xofs, int yofs, COLORREF rgbRef, COLORREF rgbMasking = 0x00ffffff, int scale = 1 );
        bool CreateMaskRgnRange( HDC hdc, HBITMAP hbmMask, RECT rect, int xofs, int yofs, COLORREF rgbRefLo, COLORREF rgbRefHi, COLORREF rgbMasking = 0x00ffffff, int scale = 1 );
        static HRGN CreateMaskRgn( HDC hdc, HBITMAP hbmMask, const BITMAPINFOHEADER *pbmihMask, const RGBQUAD *pMaskColTab, const BYTE *pMask,
                                   RECT rect, int xoff, int yoff, COLORREF rgbRefLo, COLORREF rgbRefHi, COLORREF rgbMasking,
                                   int scale );

    // operations, attributes
        void  SetRectRgn( int left, int top, int right, int bottom ) { ::SetRectRgn( HRgn(), left, top, right, bottom ); }
        void  SetRectRgn( const RECT *prect ) { ::SetRectRgn( HRgn(), prect->left, prect->top, prect->right, prect->bottom ); }
        int   CombineRgn( HRGN hrgn1, HRGN hrgn2, int nCombineMode ) { return ::CombineRgn( HRgn(), hrgn1, hrgn2, nCombineMode ); }
            ///<remarks> see also <b>QCopyRgn()</b>. </remarks>
        int   CopyRgn( Rgn *pRgnSrc ) { return ::CombineRgn( HRgn(), pRgnSrc->HRgn(), NULL, RGN_COPY ); }
        bool  EqualRgn( Rgn *pRgn ) const { return ::EqualRgn( HRgn(), pRgn->HRgn() ) != FALSE; }
        int   OffsetRgn( int x, int y ) { return ::OffsetRgn( HRgn(), x, y ); }
        int   OffsetRgn( POINT pt ) { return ::OffsetRgn( HRgn(), pt.x, pt.y ); }
        bool  MoveRgnTo( int x, int y );
        bool  MoveRgnTo( POINT pt ) { return MoveRgnTo( pt.x, pt.y ); }
        int   GetRgnBox( PRECT prect ) const { return ::GetRgnBox( HRgn(), prect ); }
        bool  PtInRegion( int x, int y ) const { return ::PtInRegion( HRgn(), x, y ) != FALSE; }
        bool  PtInRegion( POINT pt ) const { return ::PtInRegion( HRgn(), pt.x, pt.y ) != FALSE; }
        bool  RectInRegion( const RECT *prect ) const { return ::RectInRegion( HRgn(), prect ) != FALSE; }
        DWORD GetRegionData( RGNDATA *pRgnData, DWORD cBytes ) const { return ::GetRegionData( HRgn(), cBytes, pRgnData ); }
            ///<summary> Gets the number of bytes for buffer needed by <b>GetRegionData()</b> </summary>
        DWORD GetRegionDataBytes() const { return GetRegionData( NULL, 0 ); }
            ///<remarks> see also <b>QGetWindowRgn()</b>. </remarks>
        int   GetWindowRgn( HWND hwnd ) { return ::GetWindowRgn( hwnd, HRgn() ); }

    // Fully qualified methods (starting with a 'Q'):
    // They will create the HRGN automatically if they need them.
            ///<summary> unlike <b>CopyRgn()</b>, this method will first create a HRGN to copy <b>pRgnSrc</b> into. </summary>
        bool QCopyRgn( Rgn *pRgnSrc );
            ///<summary> unlike <b>GetWindowRgn()</b>, this method will first create a HRGN to copy the window's region into. </summary>
        bool QGetWindowRgn( HWND hwnd );
};




}

#endif
