/*
    File:               argonInstance.h
    Description:        Module and Process instance simplifications.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         16.04.2004
    last updated:       24.10.2006 (Roman Komary)

    Copyright (c) 2004-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonInstance_h
#define __argonInstance_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

namespace ArgonLibrary
{




extern "C" HMODULE GetCurrentModuleHandle();

#define PACKDLLVERSION(major, minor)        ((DWORD) MAKELONG(minor, major))
extern "C" DWORD   GetDllVersion( PCTSTR pstrDllName );
extern "C" BOOL    IsThemingInAppEnabled();
extern "C" HRESULT Enable_Theme_Dialog_Texture( HWND hwnd, DWORD dwFlags );


extern "C" HINSTANCE Get_Module_Instance();
extern "C" HINSTANCE Get_Resource_Instance();




/*
    #SingleInstanceDetection# enables to easily check if
    a process or module exists solely in the system.

    Usage sample: Declare as global object:
        SingleInstanceDetection g_SingleInst( _T("MyApplicationInstanceName") );
    The name provided has similar conditions as a filename (so that you
    must not use some specific characters like /, \, |, <, >, etc.).

    To query if we run a single instance of our application with the
    provided instance name, call the following:
        if( g_SingleInst.IsSingleInstance() )
            { ... }

    According to the usage of Terminal Services and kernel object names,
    you can prefix the passed instance name by "Local\" or "Global\"
    to recognize single instance detection within the local (session) space
    (e.g. only within the currently logged in user) or within the global (system) space
    (within all currently logged in users alltogether and the system
    like for service applications).
    When no prefix specified, the kernel object name space will be
    same like "Local\" for applications started in local or session space, or
    it will be like "Global\" for applications started in global space.
    See the Terminal Services documentations for more on this.
*/

class SingleInstanceDetection
{
    public :
        #pragma pack(8)
        struct SingleInstanceData
        {
            DWORD   m_dwStructSize;         // size of the data in this struct that has been filled by #getInstanceData()#.
            DWORD   m_dwPid1stInstance;     // PID of the first (and single) instance
            HWND    m_hwndContactWindow;    // HWND window that can be set to use to send messages to the window of the first application/module instance

            void Empty() { m_dwStructSize = 0; m_dwPid1stInstance = 0; m_hwndContactWindow = NULL; }
        };
        #pragma pack()

    private :
        HANDLE  m_hMemory;
        bool    m_bMemoryCreated;   // if the memory has been created or if it has been opened (and therefore existed already)

        #pragma pack(8)
        struct MemHdr
        {
            WORD  m_wHdrId;             // header id. this must be "MD", so it must be 0x4D and 0x44 in that byte order
            WORD  m_wStructVersion;     // version of this struct
            DWORD m_cHeader;            // size of the memory header in bytes
            DWORD m_cInstanceData;      // size of the instance data stored in the memory in bytes
            DWORD m_ofsInstanceData;    // offset of the instance data in the memory in bytes
        };

        struct MemData
        {
            MemHdr              m_hdr;
            SingleInstanceData  m_dat;
        };
        #pragma pack()

    public :
        SingleInstanceDetection( PCTSTR pstrInstanceName );
        ~SingleInstanceDetection();

        bool  IsSingleInstance() const { return m_bMemoryCreated; }

        bool  getInstanceData  ( SingleInstanceData& dat ) const;
        DWORD getPid1stInstance() const { SingleInstanceData dat; getInstanceData( dat ); return dat.m_dwPid1stInstance; }
            ///<summary> <b>getContactWindow()</b> gets the contact window. It can be used to send messages to the first application/module instance. </summary>
        HWND  getContactWindow () const { SingleInstanceData dat; getInstanceData( dat ); return dat.m_hwndContactWindow; }
        bool  setContactWindow ( HWND hwnd );
};




}

#endif
