/*
    File:               argonLanguageString.cpp
    Description:        A class that holds different languages of
                        the "same" string and selects the language
                        by a locale.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         20.09.2005
    last updated:       19.09.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonLanguageString_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonLanguageString.h"

namespace ArgonLibrary
{




LinkedList< LanguageString::TmpItem > *LanguageString::s_pllTmpItems = NULL;
LanguageString::Creator                LanguageString::s_Creator;

LanguageString::LanguageString()
               :m_llItems( offsetof(Item, m_node) )
{
    LockModuleWide lmw;
    m_dwCheckmark = LANGSTR_VALIDITY_CHECKMARK;
    MoveTmpItemsToItemList();
}

///<param name="pInitData"> The end of the list must be indicated by
///  a LANGSTR_END_MARKER item data. </param>
LanguageString::LanguageString( LANGSTR_STRUCT *pInitData )
               :m_llItems( offsetof(Item, m_node) )
{
    LockModuleWide lmw;
    m_dwCheckmark = LANGSTR_VALIDITY_CHECKMARK;
    MoveTmpItemsToItemList();
    Append( pInitData );
}

LanguageString::LanguageString( PCTSTR pstr_or_residString, LCID lcid )
               :m_llItems( offsetof(Item, m_node) )
{
    LockModuleWide lmw;
    m_dwCheckmark = LANGSTR_VALIDITY_CHECKMARK;
    MoveTmpItemsToItemList();
    AddString( pstr_or_residString, lcid );
}

LanguageString::~LanguageString()
{
    LockModuleWide lmw;

    m_dwCheckmark = 0;
    // We reset the checkmark again, first to know that the object cannot
    // be used now any more, but also second:
    // Because when the process ends and then the program is started again,
    // it could be that the system allocates exactly the same memory
    // range of global data for the new process and if then the system does
    // not clear the global data first, then it could be that the new
    // global data has exactly the same data as it was when the previous
    // process has ended.
    // If in the previous process the checkmark would not have been
    // reset, then the new process would have the checkmark with already
    // the validity-value which could lead into a crash if the global
    // object is accessed (at global-object-creation-time) and the
    // checkmark says the object has been initialzed already although
    // this is not true.
    // Of course this would be the case only if the creation of
    // a LanguageString object would take place at the same memory
    // location like in the previous process instance of the application.

    while( m_llItems.m_pL )
        delete m_llItems.m_pL;
}



///<param name="pData"> The end of the list must be indicated by
///  a LANGSTR_END_MARKER item data. </param>
void LanguageString::Append( LANGSTR_STRUCT *pData )
{
    if( !pData ) return;

    LockModuleWide lmw;

    for( ; pData->m_pstr_or_residString || pData->m_lcidLanguage ; ++pData )
        AddString( pData->m_pstr_or_residString, pData->m_lcidLanguage );
}

///<remarks> If a string for the specified lcid does exist already,
///  it will be overwritten now. </remarks>
void LanguageString::AddString( PCTSTR pstr_or_residString, LCID lcid )
{
    LockModuleWide lmw;

    if( !isValidObj() )
    {
        // We add the string to the tmp items for now.

        if( !s_pllTmpItems )
            if( !( s_pllTmpItems = new LinkedList< TmpItem >( offsetof(TmpItem, m_node) ) ) )
                return;

        TmpItem *ti = new TmpItem;
        if( !ti ) return;

        s_pllTmpItems->Add( ti );

        ti->m_pOwner       = this;
        ti->m_lcidLanguage = lcid;
        ti->m_strString    = pstr_or_residString;

        return;
    }

    Item *itm = FindItemExact( lcid );

    if( !itm )
        m_llItems.Add( itm = new Item );

    if( !itm ) return;

    String s;
    if( pstr_or_residString && IS_INTRESOURCE(pstr_or_residString) && LANGIDFROMLCID(lcid) != LANG_NEUTRAL )
        s.Loadstring( (UINT) (UINT_PTR) pstr_or_residString, LANGIDFROMLCID(lcid) );

    if( s.IsEmpty() )
        s = pstr_or_residString;

    itm->m_lcidLanguage = lcid;
    itm->m_strString    = s;
}



///<summary> Tries to find a string as close to the desired
///  language as possible. </summary>
String LanguageString::getString( LCID lcid ) const
{
    // We scan the item list in reverse order
    //  (because of priority handling) till one returns a
    //  string successfully matching as close
    //  to the specified language as possible.

    LockModuleWide lmw;

    if( !isValidObj() ) return (PCTSTR) NULL;

    // #lcid# contains PRIMARYLANGID, SUBLANGID and SORTID.
    LANGID want_lang_id      = LANGIDFROMLCID( lcid ); // the combined form with PRIMARYLANDID and SUBLANGID
    LANGID want_prim_lang_id = PRIMARYLANGID( want_lang_id );
    //LANGID want_sub_lang_id  = SUBLANGID( want_lang_id );
    //WORD   want_sort_id      = SORTIDFROMLCID( want_lcid );

    Item *itm;

    // If the entry has LANG_NEUTRAL exactly, then this indicates
    // that any language may be used for this entry.

    for( itm = m_llItems.m_pL ; itm ; itm = m_llItems.prev(itm) )
    {
        if( itm->m_lcidLanguage == LANG_NEUTRAL )
            return itm->m_strString;
    }

    // First we try to find a match for PRIMARYLANGID, SUBLANGID
    //  and SORTID alltogether.

    for( itm = m_llItems.m_pL ; itm ; itm = m_llItems.prev(itm) )
    {
        if( itm->m_lcidLanguage == lcid )
            return itm->m_strString;
    }

    // Now we try to find a match for PRIMARYLANGID and SUBLANGID.

    for( itm = m_llItems.m_pL ; itm ; itm = m_llItems.prev(itm) )
    {
        LANGID lang_id = LANGIDFROMLCID( itm->m_lcidLanguage );

        if( lang_id == want_lang_id )
            return itm->m_strString;
    }

    // Now we try to find a match for PRIMARYLANGID only.

    for( itm = m_llItems.m_pL ; itm ; itm = m_llItems.prev(itm) )
    {
        LANGID lang_id      = LANGIDFROMLCID( itm->m_lcidLanguage );
        LANGID prim_lang_id = PRIMARYLANGID( lang_id );

        if( prim_lang_id == want_prim_lang_id )
            return itm->m_strString;
    }

    // Now we try to find a match for LANG_NEUTRAL as
    //  primary language id.

    for( itm = m_llItems.m_pL ; itm ; itm = m_llItems.prev(itm) )
    {
        LANGID lang_id      = LANGIDFROMLCID( itm->m_lcidLanguage );
        LANGID prim_lang_id = PRIMARYLANGID( lang_id );

        if( prim_lang_id == LANG_NEUTRAL )
            return itm->m_strString;
    }

    // Now we try to find a match for LANG_ENGLISH as
    //  primary language id.

    for( itm = m_llItems.m_pL ; itm ; itm = m_llItems.prev(itm) )
    {
        LANGID lang_id      = LANGIDFROMLCID( itm->m_lcidLanguage );
        LANGID prim_lang_id = PRIMARYLANGID( lang_id );

        if( prim_lang_id == LANG_ENGLISH )
            return itm->m_strString;
    }

    // Now we try to get a description string regardless
    //  of its language,
    //  but this time in non-reverse order.

    itm = m_llItems.m_p1;
    if( itm )
        return itm->m_strString;

    // If we still don't have a description string.
    return (PCTSTR) NULL;
}



LanguageString::Item *LanguageString::FindItemExact( LCID lcid ) const
{
    for( Item *itm = m_llItems.m_p1 ; itm ; itm = m_llItems.next(itm) )
        if( itm->m_lcidLanguage == lcid )
            return itm;
    return NULL;
}



void LanguageString::MoveTmpItemsToItemList()
{
    if( !s_pllTmpItems || !isValidObj() ) return;

    for( TmpItem *ti = s_pllTmpItems->m_p1 ; ti ; )
    {
        if( ti->m_pOwner == this )
        {
            AddString( ti->m_strString, ti->m_lcidLanguage );
            TmpItem *tmp = ti;
            ti = s_pllTmpItems->next( ti );
            delete tmp;
        }
        else
            ti = s_pllTmpItems->next( ti );
    }
}

void LanguageString::DeleteTmpItemsList()
{
    if( !s_pllTmpItems ) return;

    LockModuleWide lmw;

    if( s_pllTmpItems )
    {
        while( s_pllTmpItems->m_pL )
            delete s_pllTmpItems->m_pL;

        LinkedList< TmpItem > *tmp = s_pllTmpItems;
        s_pllTmpItems = NULL;
        delete tmp;
    }
}




}
