/*
    File:               argonLanguageString.h
    Description:        A class that holds different languages of
                        the "same" string and selects the language
                        by a locale.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         20.09.2005
    last updated:       17.07.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonLanguageString_h
#define __argonLanguageString_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4201 )    // nonstandard extension used : nameless struct/union
    // Will be necessary e.g. when using   #include <mmsystem.h>
#endif

#ifndef _argon_no_includes
    #include "argonBase.h"          // for makeFourcc etc.
    #include "argonLinkedList.h"
    #include "argonString.h"
    #include "argonLocking.h"
#endif

namespace ArgonLibrary
{




struct LANGSTR_STRUCT
{
    LCID    m_lcidLanguage;         // the (exact) language/locale for which the string is intended. Or LANG_NEUTRAL if it shall be used for all languages.
    PCTSTR  m_pstr_or_residString;  // string or resource-string-id (taken from module #Get_Resource_Instance()#)
};

#define LANGSTR_END_MARKER      { 0, NULL }     // end marker for a list of LANGSTR_STRUCT items

///<summary>
///  The <b>LanguageString</b> class is intended to map a locale to one langauge dependent string.
///  This class does not support object-copying. It is mainly intended to be used as
///  global objects where everyone can access the object without the need of copying.
///   <br/><br/>
///  Internally, copies of the passed strings will be created so you do not need
///  to hold valid any string memory passed to the ctors, <b>Append()</b> or <b>AddString()</b>.
///</summary>
class LanguageString
{
    protected :
        struct Item
        {
            Node< Item >    m_node;
            LCID            m_lcidLanguage; // the (exact) language/locale for which the string is intended. Or LANG_NEUTRAL if it shall be used for all languages.
            String          m_strString;
        };

        LinkedList< Item > m_llItems;

    // Now we want to check if the object has been initialized already.
        // It is necessary to know already at global-object-init or -deinit
        // time, which object may safely be accessed and which not.
        // We do this by checking some bytes for a specific content and hope, that this
        // content will never accidentially match already non-initialized member data.

        #define LANGSTR_VALIDITY_CHECKMARK  makeFourcc('L','S','t','r')
        DWORD m_dwCheckmark;

    // If #this# LanuageString object has not yet been initialized
        // but .AddString() or .Append() are called already (which can
        // be the case during global object init time if another global
        // object is initialized first and calls add methods of a global
        // LanuageString object that will be initialized later),
        // then the following temporary array will be used:

        struct TmpItem
        {
            Node< TmpItem > m_node;
            LanguageString *m_pOwner;
            LCID            m_lcidLanguage; // the (exact) language/locale for which the string is intended. Or LANG_NEUTRAL if it shall be used for all languages.
            String          m_strString;
        };

        static LinkedList< TmpItem > *s_pllTmpItems;

        class Creator
        {
            public :
                Creator() {  }
                ~Creator() { LanguageString::DeleteTmpItemsList(); }
        };

        static Creator s_Creator;

    public :
        LanguageString();
        LanguageString( LANGSTR_STRUCT *pInitData );
        LanguageString( PCTSTR pstr_or_residString, LCID lcid );
        ~LanguageString();

        void Append( LANGSTR_STRUCT *pData );
        void AddString( PCTSTR pstr_or_residString, LCID lcid );

        String getString( LCID lcid ) const;
        String getStringThreadLocale() const { return getString( GetThreadLocale() ); }

        operator String() const { return getStringThreadLocale(); }

            ///<summary> Call this to detect if the LanguageString object has
            ///  been initialized/constructed already. This makes sense during
            ///  initialization of global objects. </summary>
        bool isValidObj() const { return this ? (m_dwCheckmark == LANGSTR_VALIDITY_CHECKMARK) : false; }

        ///<summary>
        ///  Class <b>Adder</b> can be used to add a new language or
        ///  a whole list of languages to a specific LanguageString.
        ///  Using this object, the ctor will do the work so that you
        ///  can simply create several global Adder objects.
        ///</summary>
        class Adder
        {
            public :
                Adder( LanguageString *pLS, LANGSTR_STRUCT *pData ) { if( pLS ) pLS->Append( pData ); }
                Adder( LanguageString *pLS, PCTSTR pstr_or_residString, LCID lcid ) { if( pLS ) pLS->AddString( pstr_or_residString, lcid ); }
        };

    protected :
        LanguageString( const LanguageString& lstr );             // no implementation to prevent copying
        LanguageString& operator =( const LanguageString& lstr ); // no implementation to prevent copying

        Item *FindItemExact( LCID lcid ) const;

        void MoveTmpItemsToItemList();
        static void DeleteTmpItemsList();
};




}

#endif
