/*
    File:               argonLanguageTable.h
    Description:        A class that maps several strings to different languages.
                        There are 2 big differences to argonLanguageString:
                        - The argonLanguageTable contains a whole bunch of different strings at once.
                        - The strings are not copied internally, so they are static.
                          For faster initialization execution speed and memory usage reduction.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         17.10.2005
    last updated:       12.09.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    Switches:

    _argon_dont_use_SEH     Define _argon_dont_use_SEH if try/catch can also
                            catch asynchronous (hardware) exceptions. This is the
                            case for compiler option /EHac instead of /EHsc.
                            /EHsc usually is the default option.
                            If you don't define _argon_dont_use_SEH, then hardware
                            exceptions will be caught using __try/__except instead
                            of try/catch.
                            Hardware exceptions can be illegal pointer access, etc.
*/

#ifndef __argonLanguageTable_h
#define __argonLanguageTable_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

/*
    The classes have been implemented to ease handling and
    creation of language strings held in static tables.

    In contrast to LanguageString, the language string handling
    in here works with description arrays that are passed to
    the classes and objects and need to stay valid as long as
    the objects exist. Neither the arrays nor its contained strings
    will be copied.
    The advantage in contrast to LanguageString is easier and
    faster language string definition and creation as well as
    faster execution speed.


    Creating new/own language strings:
    =====================================

    You need following types
        template class LanguageTableT
        template class LanguageDescriptionTableT

    A LanguageTableT table contains several strings identified by
    IDs. It will map each ID to several strings in different languages.
    Such a set of strings for one ID we call a language string.

    The definitions (maps) of ID values to string values are held in
    several LanguageDescriptionTableT objects that will be long to
    a specifc LanguageTableT table.

    Sample:
    Write in your header file:

        class My1stLanguageTable_Def {  };
        typedef LanguageTableT<My1stLanguageTable_Def> My1stLanguageTable;

        enum my1stIDs
        {
            my1stid_MyFirstString
            ....
        };

    Write in your cpp file:

        const LangTableDesc g_My1stLangDescs[] = {
            { my1stid_MyFirstString,    LANG_ENGLISH,   ltmapt_Direct,  (LONG_PTR) _T("My First String.") },
            ....
            LANGTABLEDESC_ENDMARK
        };

        My1stLanguageTable                            g_My1stLanguageTable;
        LanguageDescriptionTableT<My1stLanguageTable> g_My1stLanguageDescriptionTable( g_My1stLangDescs );


    Adding more languages/strings to an existing language table:
    ===============================================================

    When we take the sample above,
    let's add a german string for my1stid_MyFirstString:

    Just write in your cpp file (this can be any cpp file, does not need
    to be the one where you did the definitions of the sample above):

        const LangTableDesc g_My1stLangDescs_German[] = {
            { my1stid_MyFirstString,    LANG_GERMAN,    ltmapt_Direct,  (LONG_PTR) _T("Mein erster String.") },
            LANGTABLEDESC_ENDMARK
        };

        LanguageDescriptionTableT<My1stLanguageTable> g_My1stLanguageDescriptionTable_German( g_My1stLangDescs_German );

    With that, the new description table will be added to My1stLanguageTable.

    Getting a german string for id my1stid_MyFirstString from My1stLanguageTable
    (using the ways described in the following chapter) will now always return the
    german string defined in the new description table, no matter from which
    sourcecode file you do the call to get the string.

    The advantage of this is that e.g. for a library that you are using, you
    can add new languages easily to language tables defined and created in
    that library.


    Usage of a language string:
    ==============================

    There are 2 ways:

    1/
        My1stLanguageTable::getStringThreadLocale( my1stid_MyFirstString );

    2/
    You need
        class LanguageTableString

    Use LanguageTableString to reference a specific ID within a table
    to allow a language dependent string to be extracted from it.

        String MyGetLangString()
        {
            LanguageTableString lts( &My1stLanguageTable::s_Access, my1stid_MyFirstString );
            return lts.getStringThreadLocale();
        }

    Note:
    When getting a string, the implementation will search for the desired language.
    But if this cannot be found, the implementation will look for any other available
    language and will then return one of those (preferring english if available).
*/

#ifndef _argon_no_includes
    #include "argonLocking.h"
    #include "argonString.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




// Now we define the tryHW and catchHW exception handler macros.
// We use them to catch asynchronous (that is hardware) exceptions
// like illegal pointer access.
// Define _argon_dont_use_SEH if try/catch can also catch asynchronous
// (hardware) exceptions. This is the case for compiler option /EHac instead
// of /EHsc.
// If you don't define _argon_dont_use_SEH, then hardware
// exceptions will be caught using __try/__except instead
// of try/catch.
// Usage:
//      tryHW { .... }
//      catchHW { .... }
#ifndef tryHW
#ifdef _argon_dont_use_SEH
    #define tryHW       try
    #define catchHW     catch(...)
#else
    #define tryHW       __try
    #define catchHW     __except( EXCEPTION_EXECUTE_HANDLER )
    // The disadvantages of using SEH (structured exception handling):
    // - It is not fully designed for C++.
    // - It is not in the C++ standard (so maybe not all compilers support it).
    // - In a function with __try/__except, no local object destructors
    //   can be executed.
    // The advantage:
    // - It can catch hardware exceptions (like illegal pointer access)
    //   even if the compiler option /EHsc is specified (this is the default,
    //   it means that try/catch can catch only synchronous exceptions).
#endif
#endif
// Note: By default (so by not defining _argon_dont_use_SEH) tryHW/catchHW
// map to __try/__except because the compiler does not use SEH for
// try/catch by default (/EHsc).
// IMPORTANT: Since tryHW/catchHW can map to __try/__except, beware of the
// restrictions of __try/__except usage: No local C++ object unwinding (dtor)
// is allowed in the function where the __try/__except statements are.
// In such a case, make sub-functions where in there, the object unwindings
// are done.

template<class> class LanguageTableT;

enum eLangTableMappingType {
    // 0 is used as end-of-array indicator (see LANGTABLEDESC_ENDMARK).
    ltmapt_Direct = 1,      // m_DescInfo is a PCTSTR or a String-ID
};

struct LangTableDesc
{
    LONG_PTR m_ID;          // the ID value that identifies a specific string (in multiple languages). Any value (even 0) is allowed as ID.
    LCID     m_Locale;      // specifying LANG_NETURAL will indicate to use this entry for *any* current language
    LONG     m_MappingType; // a ltmapt_* (eLangTableMappingType) value
    LONG_PTR m_DescInfo;
    // Note: if m_MappingType is zero -> end of array
};

#define LANGTABLEDESC_ENDMARK      { 0, 0, 0, 0 }

enum eLangPriority     // note: Higher values have higher priority
{
    langprior_none          = 0,
    langprior_any,          // any language will be used, because the desired language has not been found (so language is unimportant and will be ignored)
    langprior_English,      // english will be searched
    langprior_Neutral,      // neutral language will be searched
    langprior_PrimaryId,    // only primary lang Id will be searched
    langprior_LangId,       // primary lang Id and sublang Id will be searched
    langprior_Lcid          // primary lang Id, sublang Id and Sort Id will be searched
};

///<summary>
///  Use this template class for making a global variable
///  that will allow a <b>LanguageTableT</b> (derived) object
///  to retrieve the description string for a specific error
///  code. <br/>
///  You will have to specify as template parameter (<b>LanguageTable_T</b>)
///  that language table for which you want the description table to be
///  used.
///  So you can define multiple description tables for one language
///  table which will be identified by the language code.
///  <br/><br/>
///  During runtime, when a string for an ID is requested,
///  the language table will try to find the best description table to
///  return the string.
///  So it is not necessary for a description table in a 2nd or 3rd etc.
///  language to implement all ID values. If for one, no
///  string is available, the language table will try
///  to get it from the other description tables it has.
///</summary>
template< class LanguageTable_T >
class LanguageDescriptionTableT
{
    template<class> friend class LanguageTableT;

    private :
        LanguageDescriptionTableT *m_prev;
        LanguageDescriptionTableT *m_next;

    protected :
        const LangTableDesc *m_pDescs;

    public :
        LanguageDescriptionTableT( const LangTableDesc *pDescs )
            { m_pDescs = pDescs; LockModuleWide lmw; AddToList(); }
        virtual ~LanguageDescriptionTableT() { LockModuleWide lmw; RemoveFromList(); }

        virtual eLangPriority GetString( String& str, LONG_PTR id, LCID language ) const;

            ///<summary> Makes this description table as the one with the
            ///  highest priority for its language. </summary>
            ///<remarks> This can make sense if you have more than one
            ///  description table for the <u>same</b> Group-Id (depending
            ///  on the return value from <b>GetGroupId()</b>). </remarks>
        void MakeHighestPriority()
            { LockModuleWide lmw; RemoveFromList(); AddToList(); }

    private :
        void AddToList();
        void RemoveFromList();
};

// The following struct is intended to be stored only within global
// memory (so as static member of an object).
struct LanguageTable_Access     // for internal use only
{
    typedef bool (*tfnGetString)( String& str, LONG_PTR id, LCID language );

    bool m_bInitialized;

    tfnGetString m_pfnGetString;

    void Init( tfnGetString pfnGetString )
        { LockModuleWide lmw; if( !m_bInitialized ) { m_pfnGetString = pfnGetString; m_bInitialized = true; } }

    // Note: Since this struct may be used only as static data,
    // the following methods may be called even *after* #this#
    // struct object has been destructed or *before* Init() has been called,
    // because the memory then will still exist and be valid,
    // EXCEPT: if this object was in a Dll's global space and the Dll
    // has been unloaded.

    bool GetString( String& str, LONG_PTR id, LCID language ) const
        { return m_pfnGetString ? m_pfnGetString( str, id, language ) : false; }
};

#define LANGUAGETABLE_ACCESS_INITDATA   { false, NULL }

///<summary>
///  Use this template for making a global variable
///  that will represent a specific language table. <br/>
///  For each language table, you can use a LONG_PTR's range of
///  ID values independently to other language tables and classes,
///  preventing ID value conflicts but enabling
///  easy ID value definition and usage.
///  <br/><br/>
///  For your own language table definition, pass a NewTable class parameter
///  to differentiate between different LanguageTableT classes.
///  <br/><br/>
///  In addition, you will have to create one or more
///  <b>LanguageDescriptionTableT</b> (derived) objects for this class
///  (pass the <b>LanguageTableT</b> class to the <b>LanguageDescriptionTableT</b>'s
///  template parameter). From each description table template class,
///  you will then have to create one or more objects.
///</summary>
template< class NewTable >      // pass an arbitrary class as NewTable which will cause the LanguageTableT to differ from other LanguageTableTs. That's the only reason and sense of NewTable.
class LanguageTableT
{
    template<class> friend class LanguageDescriptionTableT;

    public :
        static LanguageTable_Access s_Access;
        operator const LanguageTable_Access*() { return &s_Access; }

    private :
        static LanguageDescriptionTableT<LanguageTableT> *s_pDescsHead;
        static LanguageDescriptionTableT<LanguageTableT> *s_pDescsTail;   // at the tail is the one with the highest priority

        static LanguageTableT *s_pointer;   // if we are using the correct object

    public :
        LanguageTableT();
        ~LanguageTableT();
    protected :
        LanguageTableT( const LanguageTableT& src );              // copying shall not be allowed -> no implementation
        LanguageTableT& operator =( const LanguageTableT& src );  // copying shall not be allowed -> no implementation

    public :
        static bool GetString( String& str, LONG_PTR id, LCID language );

        static String getString( LONG_PTR id, LCID language );
        static String getStringThreadLocale( LONG_PTR id ) { return getString( id, GetThreadLocale() ); }
};

///<summary>
///  <b>LanguageTableString</b> can be used to get one string
///  in multiple languages from a language table.
///</summary>
class LanguageTableString
{
    // The members are public mainly because we want to be able to set m_ID.
    public :
        const LanguageTable_Access *m_pAccess;
        LONG_PTR                    m_ID;

    public :
        LanguageTableString() { m_pAccess = NULL; m_ID = 0; }
        LanguageTableString( const LanguageTable_Access *pAccess, LONG_PTR id )
            { m_pAccess = pAccess; m_ID = id; }

        bool   getString( String& str, LCID lcid )  const { tryHW { return m_pAccess ? m_pAccess->GetString( str, m_ID, lcid ) : false; } catchHW { return false; } }
        String getString( LCID lcid )               const { String s; getString( s, lcid ); return s; }
        bool   getStringThreadLocale( String& str ) const { return getString( str, GetThreadLocale() ); }
        String getStringThreadLocale()              const { String s; getString( s, GetThreadLocale() ); return s; }

        operator String() const { return getStringThreadLocale(); }

        const LanguageTable_Access *getAccessPtr() const { return m_pAccess; }
        LONG_PTR                    getID       () const { return m_ID; }
};




// ____ Implementations _______________________________

//      LanguageDescriptionTableT<>

template< class LanguageTable_T >
eLangPriority LanguageDescriptionTableT<LanguageTable_T>::GetString( String& str, LONG_PTR id, LCID language ) const
{
    // #language# contains PRIMARYLANGID, SUBLANGID and SORTID.
    LCID   want_lcid         = language;
    LANGID want_lang_id      = LANGIDFROMLCID( want_lcid ); // the combined form with PRIMARYLANDID and SUBLANGID
    LANGID want_prim_lang_id = PRIMARYLANGID( want_lang_id );
    //LANGID want_sub_lang_id  = SUBLANGID( want_lang_id );
    //WORD   want_sort_id      = SORTIDFROMLCID( want_lcid );

    if( !m_pDescs ) return langprior_none;
    eLangPriority priret = langprior_none;

    for( int i = 0 ;; ++i )
    {
        if( m_pDescs[i].m_MappingType == 0 )
            break;

        if( m_pDescs[i].m_ID == id )
        {
            LCID   lcid         = m_pDescs[i].m_Locale;
            LANGID lang_id      = LANGIDFROMLCID( lcid );
            LANGID prim_lang_id = PRIMARYLANGID( lang_id );
            eLangPriority pri   = langprior_none;

            // If the entry has LANG_NEUTRAL exactly, then this indicates
            // that any language may be used for this entry.
            // But *only* when we did not yet find string for
            // the desired ID value.
            if( lcid == LANG_NEUTRAL && priret == langprior_none )
                pri = langprior_Lcid;

            // First we try to find a match for PRIMARYLANGID, SUBLANGID
            //  and SORTID alltogether.
            if( pri < langprior_Lcid && lcid == want_lcid )
                pri = langprior_Lcid;

            // Now we try to find a match for PRIMARYLANGID and SUBLANGID.
            if( pri < langprior_LangId && lang_id == want_lang_id )
                pri = langprior_LangId;

            // Now we try to find a match for PRIMARYLANGID only.
            if( pri < langprior_PrimaryId && prim_lang_id == want_prim_lang_id )
                pri = langprior_PrimaryId;

            // Now we try to find a match for LANG_NEUTRAL as
            //  primary language id.
            if( pri < langprior_Neutral && prim_lang_id == LANG_NEUTRAL )
                pri = langprior_Neutral;

            // Now we try to find a match for LANG_ENGLISH as
            //  primary language id.
            if( pri < langprior_English && prim_lang_id == LANG_ENGLISH )
                pri = langprior_English;

            // Now we try to get a description string regardless
            //  of its language.
            if( pri < langprior_any )
                pri = langprior_any;

            // Now we get the description string
            // according to the mapping type.
            if( pri > priret )
                switch( m_pDescs[i].m_MappingType )
                {
                    case ltmapt_Direct :
                    {
                        if( (PCTSTR) m_pDescs[i].m_DescInfo && IS_INTRESOURCE(m_pDescs[i].m_DescInfo) && want_lang_id != LANG_NEUTRAL )
                        {
                            String s;
                            if( !s.Loadstring( (UINT) m_pDescs[i].m_DescInfo, want_lang_id ).IsEmpty() )
                                { str = s; priret = pri; break; }
                        }

                        if( (PCTSTR) m_pDescs[i].m_DescInfo )
                            { str = (PCTSTR) m_pDescs[i].m_DescInfo; priret = pri; }

                        break;
                    }
                }
        }
    }

    return priret;
}

template< class LanguageTable_T >
void LanguageDescriptionTableT<LanguageTable_T>::AddToList()
{
    if( m_prev = LanguageTable_T::s_pDescsTail ) m_prev->m_next = this;
    m_next = NULL;
    if( !LanguageTable_T::s_pDescsHead ) LanguageTable_T::s_pDescsHead = this;
    LanguageTable_T::s_pDescsTail = this;
}

template< class LanguageTable_T >
void LanguageDescriptionTableT<LanguageTable_T>::RemoveFromList()
{
    if( m_prev ) m_prev->m_next = m_next;
    else         LanguageTable_T::s_pDescsHead = m_next;
    if( m_next ) m_next->m_prev = m_prev;
    else         LanguageTable_T::s_pDescsTail = m_prev;
}

//      LanguageTableT<>

template< class NewTable > LanguageTable_Access                                   LanguageTableT<NewTable>::s_Access     = LANGUAGETABLE_ACCESS_INITDATA;
template< class NewTable > LanguageDescriptionTableT< LanguageTableT<NewTable> > *LanguageTableT<NewTable>::s_pDescsHead = NULL;
template< class NewTable > LanguageDescriptionTableT< LanguageTableT<NewTable> > *LanguageTableT<NewTable>::s_pDescsTail = NULL;
template< class NewTable > LanguageTableT<NewTable>                              *LanguageTableT<NewTable>::s_pointer    = NULL;

template< class NewTable > 
LanguageTableT<NewTable>::LanguageTableT()
{
    LockModuleWide lmw;

    if( !s_pointer )
    {
        s_Access.Init( GetString );

        s_pointer = this;
    }
}

template< class NewTable > 
LanguageTableT<NewTable>::~LanguageTableT()
{
    LockModuleWide lmw;

    // The contents of #s_Access# we do not deinit because
    // first of all there is no need to deinit, plus we still
    // want its content to be valid even after #this# object's
    // and/or the #s_Access# object's dtor have been called.
    // Since s_Access is static, it lies in the global memory
    // space and its memory location will still be valid then,
    // EXCEPT if the global space was in an unloaded Dll.

    if( s_pointer == this )
        s_pointer = NULL;

    // Remove all description tables.
    while( s_pDescsHead )
        s_pDescsHead->RemoveFromList();
}

///<returns> True if we found a string, false if not found. </returns>
template< class NewTable > 
bool LanguageTableT<NewTable>::GetString( String& str, LONG_PTR id, LCID language )
{
    // We scan all description tables in reverse order
    //  (because of priority handling) till one returns a
    //  string successfully matching as close
    //  to the specified language as possible.

    LockModuleWide lmw;

    LanguageDescriptionTableT<LanguageTableT> *table;
    eLangPriority                              ret = langprior_none;

    for( table = s_pDescsTail ; table ; table = table->m_prev )
    {
        String s;
        eLangPriority pri = table->GetString( s, id, language );

        if( pri > ret )
            { ret = pri; str = s; }
    }

    return ret > langprior_none;
}

template< class NewTable > 
String LanguageTableT<NewTable>::getString( LONG_PTR id, LCID language )
{
    String s;
    GetString( s, id, language );
    return s;
}




}

#endif
