/*
    File:               argonLocking.h
    Description:        Useful extensions to locks and synchronization.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         16.04.2004
    last updated:       04.10.2006 (Roman Komary)

    Copyright (c) 2004-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    Switches:

    _argonLocking_use_moduleMutex       If you #define this, then LockModuleWide
                                        will use a mutex rather than a critical section.
*/

#ifndef __argonLocking_h
#define __argonLocking_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

namespace ArgonLibrary
{




///<summary>
///  You can use this class to make a global, module-space
///  wide mutex or critical section lock at any time (even
///  in the constructor or destructor of a global object). <br/><br/>
///
///  2 forms of usage: <br/>
///  1st: Create a local object and let the ctor and dtor handle
///  the locking/unlocking, or call <b>Lock()</b> and <b>Unlock()</b> explictely. <br/>
///
///  Or 2nd: Call <b>LockModuleWide::Enter()</b> and <b>LockModuleWide::Leave()</b>
///  without needing to create a <b>LockModuleWide</b> object, (because they are static).
///  This is useful if you need to enter in one function and leave in
///  some other function.
///</summary>
class LockModuleWide
{
    private :
        struct Creator
        {
            Creator();
            ~Creator();
        };

        static Creator s_Creator;
        static bool    s_bDeletingGlobalObjects;

        #ifdef _argonLocking_use_moduleMutex
        static HANDLE s_hLockMutex;
        #else
        static CRITICAL_SECTION s_crit_sec;
        static bool             s_crit_sec_initialized;
        #endif

        bool m_bLocked;

    public :
        LockModuleWide( bool bInitialLock = true )
            { Prepare(); m_bLocked = false; if( bInitialLock ) Lock(); }
        ~LockModuleWide() { Unlock(); }

        void Lock();
        void Unlock();

        static void Enter();
        static void Leave();

    private :
        static void Prepare();
        static void OnExit();
};




}

#endif
