/*
    File:               argonMath.h
    Description:        Mathematical values and functions.

    Author:             Roman Komary
    created on:         11.09.2002
    last updated:       06.11.2006 (Roman Komary)

    Copyright (c) 2002-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonMath_h
#define __argonMath_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include <math.h>
#endif

namespace ArgonLibrary
{




#ifdef PI
#undef PI
#endif

#define PI              3.14159265358979323846      // Pi
#define PIf             3.14159265358979323846f
#define PIPI            6.28318530717958623200      // 2 * Pi
#define PIPIf           6.28318530717958623200f
#define PI2             1.57079632679489655800      // Pi / 2
#define PI2f            1.57079632679489655800f
#define PI4             0.78539816339744827900      // Pi / 4
#define PI4f            0.78539816339744827900f
#define INV_PI          0.31830988618379069122      // 1 / Pi
#define INV_PIf         0.31830988618379069122f
#define DEGTORAD        0.01745329251994329547      // Degrees to Radians
#define DEGTORADf       0.01745329251994329547f
#define RADTODEG       57.29577951308232286465      // Radians to Degrees
#define RADTODEGf      57.29577951308232286465f
#define E               2.71828182845904523536
#define Ef              2.71828182845904523536f
#define SQRT_2          1.41421356237309504880
#define SQRT_2f         1.41421356237309504880f
#ifndef FLT__HUGE
#define FLT__HUGE       1.0e+38f                    // Huge number for FLOAT
#endif
#ifndef FLT__EPSILON
#define FLT__EPSILON    1.0e-5f                     // Tolerance for FLOATs
#endif
#define FLT__MAX        3.4e+38f
#define FLT__ZERO       3.4e-38f
#define DBL__MAX        1.7e+308
#define DBL__ZERO       1.7e-308


///<summary> division by zero is *not* endless, instead it is *undefined*. To fit the implementation of Milkdrop, we set it to zero (well what else?) </summary>
inline double fdiv   ( double y, double x ) { return( x!=0.0 ? y/x : 0.0 ); }
///<summary> radians to degree </summary>
inline double deg    ( double x ) { return( x * RADTODEG ); }
///<summary> degree to radians </summary>
inline double rad    ( double x ) { return( x * DEGTORAD ); }
inline double cot    ( double x ) { return fdiv(1.0, tan(x)); }
inline double acot   ( double x ) { return atan( fdiv(1.0, x) ); }
inline double coth   ( double x ) { return fdiv( 1.0, tanh(x) ); }
inline double asinh  ( double x ) { return log( x + sqrt(x*x+1.0) ); }
inline double acosh  ( double x ) { return log( x + sqrt(x*x-1.0) ); }
inline double atanh  ( double x ) { return log( fdiv((1.0+x), (1.0-x)) ) * 0.5; }
inline double acoth  ( double x ) { return log( fdiv((x+1.0), (x-1.0)) ) * 0.5; }
inline double log2   ( double x ) { return log(x) * 1.44269504088896; }
inline double round  ( double x ) { return (x > 0.0) ? floor(x+0.5) : ceil(x-0.5); }
inline double fsign  ( double x ) { return (x > 0.0) ? 1.0 : ( (x < 0.0) ? -1.0 : 0.0 ); }
///<summary> returns random number in the range from 0 to 1. </summary>
inline double frand1 () { return rand() / (double) RAND_MAX; }
///<summary> returns random number in the range from -0.5 to 0.5. </summary>
inline double frand05() { return (rand() / (double) RAND_MAX) - 0.5; }
///<summary> returns random number in the range from 0 to x. </summary>
inline double frandx ( double x ) { return x * rand() / (double) RAND_MAX; }
///<summary> returns random number in the range from x to y. </summary>
inline double frandxy( double x, double y ) { return x + (y - x) * rand() / (double) RAND_MAX; }
inline void   fseed  ( double x ) { srand( (UINT) x ); }
///<param name="a"> constraint </param>
inline double sigmoid( double x, double a ) { return 1.0 / ( 1.0 + exp(-a*(x-1.0)) ); }




}

#endif
