/*
    File:               argonMemory.cpp
    Description:        Memory handling classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         2.10.2003
    last updated:       14.07.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonMemory_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonMemory.h"

namespace ArgonLibrary
{




///<summary> Is similar to <b>memmove()</b>, but in here, memory
///  is described in size of bits and can be copied
///  from one bit position to another bit position. </summary>
///<param name="pdest"> pointer to first destination byte </param>
///<param name="destbit"> first destination bit within the
///  first destination byte <br/>
///  e.g. 0 means the bit0 of the first destination byte
///  as shown in the graph below. </param>
///<param name="psrce"> pointer to first source byte </param>
///<param name="srcebit"> first source bit within the first source byte <br/>
///  e.g. 0 means the bit0 of the first source byte
///  as shown in the graph below. </param>
///<param name="bitcount"> number of bits to copy </param>
///<returns> returns <b>pdest</b> </returns>
///<remarks> Note: The memory (viewed in bits) is interpreted as follows: <pre>
///  Addr  0       1       2       3       4       5       6        ...
///  Bit   76543210765432107654321076543210765432107654321076543210 ...
///  </pre> </remarks>
void *_memmovebits_( void *pdest, int destbit, const void *psrce, int srcebit, INT64 bitcount )
{
    static BYTE almask[8] = { 0xfe, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80, 0x00 };  // left  mask (which bits should not be changed by the copying)
    static BYTE armask[8] = { 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f };  // right mask (which bits should not be changed by the copying)

    BYTE *srce = (BYTE*) psrce;
    BYTE *dest = (BYTE*) pdest;

    if( !dest || !srce || bitcount < 1 || destbit < 0 || destbit >= 8 || srcebit < 0 || srcebit >= 8 )
        return pdest;

    // First check, if a normal byte-movement is suitable
    if( destbit == 7 && srcebit == 7 && (bitcount & 7) == 0 )
    {
        INT64 bytecount = (bitcount >> 3);
        if( dest < srce )
            for( INT64 i = 0 ; i < bytecount ; i++ )
                dest[i] = srce[i];
        else
            for( INT64 i = bytecount - 1 ; i >= 0 ; i-- )
                dest[i] = srce[i];
        return pdest;
    }

    bool dest_before_srce = false;  // if destination position has a lower address (regarding bits) as the source position
    if( dest == srce )
    {
        if( destbit == srcebit )
            return pdest;
        dest_before_srce = ( destbit > srcebit );
    }
    else
        dest_before_srce = ( dest < srce );

    INT64 cbytes_srce = ((INT64) (7 - srcebit) + (INT64) bitcount + 7) / 8;  // how many bytes in #srce# are occupied by the area to be copied
    INT64 cbytes_dest = ((INT64) (7 - destbit) + (INT64) bitcount + 7) / 8;  // how many bytes in #dest# are occupied by the area to be copied

    BYTE lmask_dest = almask[ destbit ];
    BYTE rmask_dest = armask[ ((UINT) ((INT64) destbit - (INT64) (bitcount - 1))) & 7 ];

    if( dest_before_srce )
    {
        short lshift = (short) (destbit - srcebit);  // how many bits to shift left
        for( INT64 ibyte_dest = 0 ; ibyte_dest < cbytes_dest ; ibyte_dest++, srce++, dest++ )
        {
            BYTE mask_dest = 0x00;          // mask (which bits should not be changed by the copying in the current #dest# byte)
            if( ibyte_dest == 0 )               mask_dest |= lmask_dest;
            if( ibyte_dest == cbytes_dest - 1 ) mask_dest |= rmask_dest;

            BYTE by = ( ibyte_dest < cbytes_srce )? *srce : (BYTE) 0;
            if( lshift > 0 )
            {
                by <<= lshift;
                if( ibyte_dest < cbytes_srce - 1 )
                    by |= srce[1] >> (8 - lshift);
            }
            else if( lshift < 0 )
            {
                by >>= -lshift;
                if( ibyte_dest > 0 )
                    by |= srce[-1] << (8 + lshift);
            }

            *dest = (*dest & mask_dest) | (by & ~mask_dest);
        }
    }
    else
    {
        short rshift = (short) (srcebit - destbit);  // how many bits to shift right
        srce += cbytes_dest - 1;
        dest += cbytes_dest - 1;
        for( INT64 ibyte_dest = cbytes_dest - 1 ; ibyte_dest >= 0 ; ibyte_dest--, srce--, dest-- )
        {
            BYTE mask_dest = 0x00;          // mask (which bits should not be changed by the copying in the current #dest# byte)
            if( ibyte_dest == 0 )               mask_dest |= lmask_dest;
            if( ibyte_dest == cbytes_dest - 1 ) mask_dest |= rmask_dest;

            BYTE by = ( ibyte_dest < cbytes_srce )? *srce : (BYTE) 0;
            if( rshift > 0 )
            {
                by >>= rshift;
                if( ibyte_dest > 0 )
                    by |= srce[-1] << (8 - rshift);
            }
            else if( rshift < 0 )
            {
                by <<= -rshift;
                if( ibyte_dest < cbytes_srce - 1 )
                    by |= srce[1] >> (8 + rshift);
            }

            *dest = (*dest & mask_dest) | (by & ~mask_dest);
        }
    }

    return pdest;
}

///<summary> Similar to <b>memset()</b>, but in here, memory
///  is described in size of bits. </summary>
///<param name="pdest"> pointer to first destination byte </param>
///<param name="destbit"> first destination bit within the
///  first destination byte </param>
///<param name="bit"> bit value with which to fill the destination bit area <br/>
///  Note: Only Bit0 will be used from this value </param>
///<param name="bitcount"> number of bits to copy </param>
///<returns> returns <b>pdest</b> </returns>
///<remarks> Note: The memory (viewed in bits) is interpreted as follows: <pre>
///  Addr  0       1       2       3       4       5       6        ...
///  Bit   76543210765432107654321076543210765432107654321076543210 ...
///  </pre> </remarks>
void *_memsetbits_( void *pdest, int destbit, char bit, INT64 bitcount )
{
    static BYTE almask[8] = { 0xfe, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x80, 0x00 };  // left  mask (which bits should not be changed by the operation)
    static BYTE armask[8] = { 0x00, 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f };  // right mask (which bits should not be changed by the operation)

    BYTE *dest = (BYTE*) pdest;

    if( !dest || bitcount < 1 || destbit < 0 || destbit >= 8 )
        return pdest;

    // First check, if a normal byte-filling is suitable
    if( destbit == 7 && (bitcount & 7) == 0 )
    {
        INT64 bytecount = (bitcount >> 3);
        BYTE by = ( bit & 1 )? 0xff : 0x00;
        for( INT64 i = 0 ; i < bytecount ; i++ )
            dest[i] = by;
        return pdest;
    }

    INT64 cbytes_dest = ((INT64) (7 - destbit) + (INT64) bitcount + 7) / 8;  // how many bytes in #dest# are occupied by the area to be operation

    BYTE lmask_dest = almask[ destbit ];
    BYTE rmask_dest = armask[ ((UINT) (destbit - (INT) (bitcount - 1))) & 7 ];

    BYTE by = ( bit & 1 )? 0xff : 0x00;
    for( INT64 ibyte_dest = 0 ; ibyte_dest < cbytes_dest ; ibyte_dest++, dest++ )
    {
        BYTE mask_dest = 0x00;          // mask (which bits should not be changed by the operation in the current #dest# byte)
        if( ibyte_dest == 0 )               mask_dest |= lmask_dest;
        if( ibyte_dest == cbytes_dest - 1 ) mask_dest |= rmask_dest;

        *dest = (*dest & mask_dest) | (by & ~mask_dest);
    }

    return pdest;
}

///<summary> Is similar to <b>memmove()</b>, but in here, memory
///  is described in size of bits and can be copied
///  from one bit position to another bit position. </summary>
///<param name="dest"> bit-pointer to first destination bit </param>
///<param name="srce"> bit-pointer to first source bit </param>
///<param name="size"> number of bits to copy </param>
///<returns> returns <b>dest</b> </returns>
///<remarks> Note: The memory (viewed in bits) is interpreted as follows: <pre>
///  Addr  0       1       2       3       4       5       6        ...
///  Bit   76543210765432107654321076543210765432107654321076543210 ...
///  </pre> </remarks>
BitPtr _memmovebits_( BitPtr dest, BitCPtr srce, BitLen size )
{
    _memmovebits_( dest.ptr(), dest.bitnr(), srce.ptr(), srce.bitnr(), size.cbits() );
    return dest;
}

///<summary> Similar to <b>memset()</b>, but in here, memory
///  is described in size of bits. </summary>
///<param name="dest"> bit-pointer to first destination bit </param>
///<param name="bit"> bit value with which to fill the destination bit area <br/>
///  Note: Only Bit0 will be used from this value </param>
///<param name="size"> number of bits to copy </param>
///<returns> returns <b>dest</b> </returns>
///<remarks> Note: The memory (viewed in bits) is interpreted as follows: <pre>
///  Addr  0       1       2       3       4       5       6        ...
///  Bit   76543210765432107654321076543210765432107654321076543210 ...
///  </pre> </remarks>
BitPtr _memsetbits_( BitPtr dest, char bit, BitLen size )
{
    _memsetbits_( dest.ptr(), dest.bitnr(), bit, size.cbits() );
    return dest;
}




BitPtr& BitPtr::Set( void *pointer, INT64 bit_offset )
{
    if( bit_offset >= 0 ) m_ptr = ((PBYTE) pointer) + (  bit_offset      >> 3);
    else                  m_ptr = ((PBYTE) pointer) - ((-bit_offset + 7) >> 3);
    m_bitnr = 7L - (LONG32) (((ULONG32) bit_offset) & 7);
    return *this;
}

BitPtr& BitPtr::operator +=( INT64 bit_offset )
{
    INT64 bit_offset_from_bytebegin = bit_offset + (INT64) (7 - m_bitnr);
    Set( m_ptr, bit_offset_from_bytebegin );
    return *this;
}

BitPtr& BitPtr::operator +=( const BitLen& len )
{
    if( len.IsError() ) { SetNull(); return *this; }
    return operator +=( (int) len.cbits() );
}

BitPtr& BitPtr::operator -=( const BitLen& len )
{
    if( len.IsError() ) { SetNull(); return *this; }
    return operator +=( -(int) len.cbits() );
}

BitPtr BitPtr::operator +( INT64 bit_offset )
{
    BitPtr result( *this );
    result += bit_offset;
    return result;
}

BitPtr BitPtr::operator +( const BitLen& len )
{
    if( len.IsError() ) return NULL;
    return operator +( (INT64) len.cbits() );
}

BitPtr BitPtr::operator -( const BitLen& len )
{
    if( len.IsError() ) return NULL;
    return operator +( -(INT64) len.cbits() );
}

///<returns> returns a bit-offset </returns>
INT64 BitPtr::operator -( const BitCPtr& bitptr2 )
{
    INT64 byte_offset = (INT64) (UINT_PTR) ((PSTR) m_ptr) - (INT64) (UINT_PTR) ((PSTR) bitptr2.m_ptr);
    //return (INT64) (byte_offset << 3) + (INT64) (7 - m_bitnr) - (INT64) (7 - bitptr2.m_bitnr);
    // or shorter:
    return (INT64) (byte_offset << 3) - (INT64) m_bitnr + (INT64) bitptr2.m_bitnr;
}

///<returns> returns a bit-offset </returns>
INT64 BitPtr::operator -( const BitPtr& bitptr2 )
{
    INT64 byte_offset = (INT64) (UINT_PTR) ((PSTR) m_ptr) - (INT64) (UINT_PTR) ((PSTR) bitptr2.m_ptr);
    //return (INT64) (byte_offset << 3) + (INT64) (7 - m_bitnr) - (INT64) (7 - bitptr2.m_bitnr);
    // or shorter:
    return (INT64) (byte_offset << 3) - (INT64) m_bitnr + (INT64) bitptr2.m_bitnr;
}



BitCPtr& BitCPtr::Set( const void *pointer, INT64 bit_offset )
{
    if( bit_offset >= 0 ) m_ptr = ((PBYTE) pointer) + (  bit_offset      >> 3);
    else                  m_ptr = ((PBYTE) pointer) - ((-bit_offset + 7) >> 3);
    m_bitnr = 7L - (LONG32) (((ULONG32) bit_offset) & 7);
    return *this;
}

BitCPtr& BitCPtr::operator +=( INT64 bit_offset )
{
    INT64 bit_offset_from_bytebegin = bit_offset + (INT64) (7 - m_bitnr);
    Set( m_ptr, bit_offset_from_bytebegin );
    return *this;
}

BitCPtr& BitCPtr::operator +=( const BitLen& len )
{
    if( len.IsError() ) { SetNull(); return *this; }
    return operator +=( (int) len.cbits() );
}

BitCPtr& BitCPtr::operator -=( const BitLen& len )
{
    if( len.IsError() ) { SetNull(); return *this; }
    return operator +=( -(int) len.cbits() );
}

BitCPtr BitCPtr::operator +( INT64 bit_offset )
{
    BitCPtr result( *this );
    result += bit_offset;
    return result;
}

BitCPtr BitCPtr::operator +( const BitLen& len )
{
    if( len.IsError() ) return NULL;
    return operator +( (INT64) len.cbits() );
}

BitCPtr BitCPtr::operator -( const BitLen& len )
{
    if( len.IsError() ) return NULL;
    return operator +( -(INT64) len.cbits() );
}

///<returns> returns a bit-offset </returns>
INT64 BitCPtr::operator -( const BitCPtr& bitptr2 )
{
    INT64 byte_offset = (INT64) (UINT_PTR) ((PSTR) m_ptr) - (INT64) (UINT_PTR) ((PSTR) bitptr2.m_ptr);
    //return (INT64) (byte_offset << 3) + (INT64) (7 - m_bitnr) - (INT64) (7 - bitptr2.m_bitnr);
    // or shorter:
    return (INT64) (byte_offset << 3) - (INT64) m_bitnr + (INT64) bitptr2.m_bitnr;
}

///<returns> returns a bit-offset </returns>
INT64 BitCPtr::operator -( const BitPtr& bitptr2 )
{
    INT64 byte_offset = (INT64) (UINT_PTR) ((PSTR) m_ptr) - (INT64) (UINT_PTR) ((PSTR) bitptr2.m_ptr);
    //return (INT64) (byte_offset << 3) + (INT64) (7 - m_bitnr) - (INT64) (7 - bitptr2.m_bitnr);
    // or shorter:
    return (INT64) (byte_offset << 3) - (INT64) m_bitnr + (INT64) bitptr2.m_bitnr;
}




void DimSize::SetSize( _size_t_ cSize )
{
    _size_t_ cSize_old = m_cElements;
    SetSize0( cSize );
    dim_t *p0 = Ptr();
    for( _idx_ i = cSize_old ; i < (_idx_) cSize ; ++i )
        p0[i] = 1;
}

void DimSize::Clear()
{
    dim_t *p0 = Ptr();
    for( _idx_ i = 0 ; i < (_idx_) m_cElements ; ++i )
        p0[i] = 1;
}

///<remarks> returns TRUE if even one element in the elements-array is zero (or less) </remarks>
bool DimSize::IsEmpty() const
{
    if( m_cElements == 0 ) return true;
    const dim_t *p0 = Ptr();
    for( _idx_ i = 0 ; i < (_idx_) m_cElements ; ++i )
        if( p0[i] <= 0 ) return true;
    return false;
}

///<summary> If <b>this</b> DimSize is smaller (either in dimension
///  or else in each and every extent of the dimension extent array). </summary>
bool DimSize::IsLessThan( const DimSize& dimsize2 ) const
{
    _size_t_ count = m_cElements;

    if( count > dimsize2.GetSize() )
        return false;

    const dim_t *p1 = Ptr();
    const dim_t *p2 = dimsize2.Ptr();

    for( _idx_ i = 0 ; i < (_idx_) count ; ++i ) {
        if( p1[i] >= p2[i] )
            return false;
    }

    return true;
}

///<summary> If <b>this</b> DimSize is smaller or equal (in dimension
///  and in even each and every extent of the dimension extent array). </summary>
bool DimSize::IsLessEqualThan( const DimSize& dimsize2 ) const
{
    _size_t_ count = m_cElements;

    if( count > dimsize2.GetSize() )
        return false;

    const dim_t *p1 = Ptr();
    const dim_t *p2 = dimsize2.Ptr();

    for( _idx_ i = 0 ; i < (_idx_) count ; ++i ) {
        if( p1[i] > p2[i] )
            return false;
    }

    return true;
}

///<summary> Takes the minimum of each extent from
///  <b>dimsize1</b> and <b>dimsize2</b> and
///  returns the result in <b>this</b> dimension. </summary>
///<remarks> Note: If one dimension is smaller,
///  then it will be imagined as large
///  as the other using an extent-value of 1
///  for the missing dim-extents. </remarks>
///<returns> if an empty <b>DimSize</b>, then can be that
///  <b>dimsize1/2</b> are both of size zero </returns>
DimSize DimSize::EvalMinimum( const DimSize& dimsize1, const DimSize& dimsize2 )
{
    _size_t_ count = tmax( dimsize1.GetSize(), dimsize2.GetSize() );
    DimSize rv( count );
    const dim_t *p1 = dimsize1.Ptr();
    const dim_t *p2 = dimsize2.Ptr();
    for( _idx_ i = 0 ; i < (_idx_) count ; i++ ) {
        dim_t element1 = ( i >= (_idx_) dimsize1.m_cElements )? 1 : p1[i];
        dim_t element2 = ( i >= (_idx_) dimsize2.m_cElements )? 1 : p2[i];
        rv.SetAt( i, tmin( element1, element2 ) );
    }
    return rv;
}

///<summary> Takes the maximum of each extent from
///  <b>dimsize1</b> and <b>dimsize2</b> and
///  returns the result in <b>this</b> dimension. </summary>
///<remarks> Note: If one dimension is smaller,
///  then it will be imagined as large
///  as the other using an extent-value of 1
///  for the missing dim-extents. </remarks>
///<returns> if an empty <b>DimSize</b>, then can be that
///  <b>dimsize1/2</b> are both of size zero </returns>
DimSize DimSize::EvalMaximum( const DimSize& dimsize1, const DimSize& dimsize2 )
{
    _size_t_ count = tmax( dimsize1.GetSize(), dimsize2.GetSize() );
    DimSize rv( count );
    const dim_t *p1 = dimsize1.Ptr();
    const dim_t *p2 = dimsize2.Ptr();
    for( _idx_ i = 0 ; i < (_idx_) count ; i++ ) {
        dim_t element1 = ( i >= (_idx_) dimsize1.m_cElements )? 1 : p1[i];
        dim_t element2 = ( i >= (_idx_) dimsize2.m_cElements )? 1 : p2[i];
        rv.SetAt( i, tmax( element1, element2 ) );
    }
    return rv;
}




///<summary> Creates a position array of the same size
///  as <b>dimsize</b> and sets its elements to each value
///  of the <b>dimsize</b>'s elements - 1.
///  This will cause <b>this</b> <b>DimPt</b> to be a
///  position pointing to the last element of an array
///  of size <b>dimsize</b>. This is a good value to start
///  a loop with decrementing a <b>DimPt</b>. </summary>
void DimPt::SetToEndOfDim( const DimSize& dimsize )
{
    SetSize( dimsize.GetSize() );
    dim_t *p0 = Ptr();
    for( _idx_ i = 0 ; i < (_idx_) m_cElements ; ++i )
        p0[i] = dimsize[i] - 1;
}

///<summary> Increases <b>this</b> <b>DimPt</b> by <b>step</b>
///  inside the memory size <b>dimsize</b>. If the result exceeds
///  the limits specified by <b>dimsize</b>, then <b>this</b>
///  <b>DimPt</b> will index a position like it had been calculated
///  modulo and in that case the return value will be FALSE to
///  indicate that the end-of-array had been reached. <br/>
///  Also FALSE will be returned on error (e.g. when the sizes of all
///  3 involved <b>Array</b> derived objects are not equal). </summary>
///<param name="dimpos"> see <b>dimsize</b> </param>
///<param name="dimsize"> <b>dimpos</b>, <b>dimsize</b> define the
///  range where <b>this</b> <b>DimPt</b> may lie inside </param>
///<param name="pbDimOverflow"> if not nULL, then it will contain
///  TRUE if the incrementation causes one (or more) dimension-extents
///  to overflow over the according dim-extent-size value in <b>dimsize</b>. </param>
///<returns> FALSE = reached end-of-array or error <br/>
///  TRUE = end-of-array not reached yet </returns>
///<remarks> IMPORTANT: <b>step</b> *must not* contain negative
///  values. This will not be catched. The <b>step</b> values
///  may only be positive or zero. </remarks>
bool DimPt::Increment( const DimPt& step, const DimPt& dimpos, const DimSize& dimsize, bool *pbDimOverflow )
{
    if( pbDimOverflow ) *pbDimOverflow = false;
    if(  GetSize() != step.GetSize() ||
         GetSize() != dimpos.GetSize() ||
         GetSize() != dimsize.GetSize() ||
         dimsize.IsEmpty()  )
        return false;
    _size_t_ count = GetSize();
    dim_t *p0 = Ptr();
    const dim_t *p0_step    = step.Ptr();
    const dim_t *p0_dimpos  = dimpos.Ptr();
    const dim_t *p0_dimsize = dimsize.Ptr();
    for( _idx_ i = 0 ; i < (_idx_) count ; ++i )
    {
        p0[i] += p0_step[i];
        if( p0[i] >= p0_dimpos[i] + p0_dimsize[i] )
        {
            if( pbDimOverflow ) *pbDimOverflow = true;
            if( i < (_idx_) count - 1 ) {
                p0[i+1] += (p0[i] - p0_dimpos[i]) / p0_dimsize[i];
                p0[i] = ( (p0[i] - p0_dimpos[i]) % p0_dimsize[i] ) + p0_dimpos[i];
            } else {
                p0[i] = ( (p0[i] - p0_dimpos[i]) % p0_dimsize[i] ) + p0_dimpos[i];
                return false;
            }
        }
    }
    return true;
}

///<summary> Decreases <b>this</b> <b>DimPt</b> by <b>step</b> inside
///  the memory size <b>dimsize</b>. If the result exceeds the limits
///  specified by <b>dimsize</b> (or when a dim-extent's position
///  value gets negative), then <b>this</b> <b>DimPt</b> will index
///  a position like it had been calculated modulo and in that case
///  the return value will be FALSE to indicate that the start-of-array
///  had been reached. <br/>
///  Also FALSE will be returned on error (e.g. when the sizes of all
///  3 involved <b>Array</b> derived objects are not equal). </summary>
///<param name="dimpos"> see <b>dimsize</b> </param>
///<param name="dimsize"> <b>dimpos</b>, <b>dimsize</b> define the
///  range where <b>this</b> <b>DimPt</b> may lie inside </param>
///<param name="pbDimUnderflow"> if not NULL, then it will contain
///  TRUE if the decrementation causes one (or more) dimension-extents
///  to underflow to a negative value. </param>
///<returns> False = reached start-of-array or error. <br/>
///  True = start-of-array not reached yet. </returns>
///<remarks> IMPORTANT: <b>step</b> *must not* contain negative
///  values. This will not be catched. The <b>step</b> values
///  may only be positive or zero. </remarks>
bool DimPt::Decrement( const DimPt& step, const DimPt& dimpos, const DimSize& dimsize, bool *pbDimUnderflow )
{
    if( pbDimUnderflow ) *pbDimUnderflow = false;
    if(  GetSize() != step.GetSize() ||
         GetSize() != dimpos.GetSize() ||
         GetSize() != dimsize.GetSize() ||
         dimsize.IsEmpty()  )
        return false;
    _size_t_ count = GetSize();
    dim_t *p0 = Ptr();
    const dim_t *p0_step    = step.Ptr();
    const dim_t *p0_dimpos  = dimpos.Ptr();
    const dim_t *p0_dimsize = dimsize.Ptr();
    for( _idx_ i = 0 ; i < (_idx_) count ; ++i )
    {
        p0[i] -= p0_step[i];
        if( p0[i] < p0_dimpos[i] )
        {
            if( pbDimUnderflow ) *pbDimUnderflow = true;
            if( i < (_idx_) count - 1 ) {
                dim_t cunderflows = ( ((dim_t)( -(p0[i] - p0_dimpos[i]) - 1 )) / p0_dimsize[i] ) + 1;
                p0[i+1] -= cunderflows;
                p0[i] = ( ((dim_t)( cunderflows * p0_dimsize[i] + (p0[i] - p0_dimpos[i]) )) % p0_dimsize[i] ) + p0_dimpos[i];
            } else {
                dim_t cunderflows = ( ((dim_t)( -(p0[i] - p0_dimpos[i]) - 1 )) / p0_dimsize[i] ) + 1;
                p0[i] = ( ((dim_t)( cunderflows * p0_dimsize[i]  +(p0[i] - p0_dimpos[i]) )) % p0_dimsize[i] ) + p0_dimpos[i];
                return false;
            }
        }
    }
    return true;
}

///<summary> Takes the minimum of each extent from <b>dimpt1</b>
///  and <b>dimpt2</b> and returns the result in <b>this</b> dim-point. </summary>
///<remarks> Note: If one dimension is smaller, then it will be
///  imagined as large as the other using an extent-value of 0 for
///  the missing dim-extents. </remarks>
///<returns> if an empty <b>DimPt</b>, then it can be that
///  <b>dimpt1/2</b> are both of size zero </returns>
DimPt DimPt::EvalMinimum( const DimPt& dimpt1, const DimPt& dimpt2 )
{
    _size_t_ count = tmax( dimpt1.GetSize(), dimpt2.GetSize() );
    DimPt rv( count );
    const dim_t *p1 = dimpt1.Ptr();
    const dim_t *p2 = dimpt2.Ptr();
    for( _idx_ i = 0 ; i < (_idx_) count ; ++i ) {
        dim_t element1 = ( i >= (_idx_) dimpt1.m_cElements )? 0 : p1[i];
        dim_t element2 = ( i >= (_idx_) dimpt2.m_cElements )? 0 : p2[i];
        rv.SetAt( i, tmin( element1, element2 ) );
    }
    return rv;
}

///<summary> Takes the minimum of each extent from <b>dimpt1</b>
///  and <b>dimpt2</b> and returns the result in <b>this</b> dim-point. </summary>
///<remarks> Note: If one dimension is smaller, then it will be
///  imagined as large as the other using an extent-value of 0 for
///  the missing dim-extents. </remarks>
///<returns> if an empty <b>DimPt</b>, then it can be that
///  <b>dimpt1/2</b> are both of size zero </returns>
DimPt DimPt::EvalMaximum( const DimPt& dimpt1, const DimPt& dimpt2 )
{
    _size_t_ count = tmax( dimpt1.GetSize(), dimpt2.GetSize() );
    DimPt rv(count);
    const dim_t *p1 = dimpt1.Ptr();
    const dim_t *p2 = dimpt2.Ptr();
    for( _idx_ i = 0 ; i < (_idx_) count ; ++i ) {
        dim_t element1 = ( i >= (_idx_) dimpt1.m_cElements )? 0 : p1[i];
        dim_t element2 = ( i >= (_idx_) dimpt2.m_cElements )? 0 : p2[i];
        rv.SetAt( i, tmax( element1, element2 ) );
    }
    return rv;
}




DWORD BitCounter::s_aUnusedMask[] = {
    0x00000000, 0x80000000, 0xc0000000, 0xe0000000,
    0xf0000000, 0xf8000000, 0xfc000000, 0xfe000000,
    0xff000000, 0xff800000, 0xffc00000, 0xffe00000,
    0xfff00000, 0xfff80000, 0xfffc0000, 0xfffe0000,
    0xffff0000, 0xffff8000, 0xffffc000, 0xffffe000,
    0xfffff000, 0xfffff800, 0xfffffc00, 0xfffffe00,
    0xffffff00, 0xffffff80, 0xffffffc0, 0xffffffe0,
    0xfffffff0, 0xfffffff8, 0xfffffffc, 0xfffffffe
};

void BitCounter::Inc()
{
    if( m_cElements == 0 ) return;

    _size_t_ cUnusedBits = ( bitcounter_t_bits - (m_cBits % bitcounter_t_bits) ) % bitcounter_t_bits;
    bitcounter_t *p0 = Ptr();

    bitcounter_t top_val_prev = p0[m_cElements - 1];

    _idx_ i;
    for( i = 0 ; i < (_idx_) m_cElements ; ++i )
        if( ++p0[i] != 0 )
            break;

    bitcounter_t top_val =
        ( p0[m_cElements - 1] &= ~s_aUnusedMask[cUnusedBits] );

    m_bOverflowed = ( top_val_prev != 0 && top_val == 0 );
}

void BitCounter::Dec()
{
    if( m_cElements == 0 ) return;

    _size_t_ cUnusedBits = ( bitcounter_t_bits - (m_cBits % bitcounter_t_bits) ) % bitcounter_t_bits;
    bitcounter_t *p0 = Ptr();

    bitcounter_t top_val_prev = p0[m_cElements - 1];

    _idx_ i;
    for( i = 0 ; i < (_idx_) m_cElements ; ++i )
        if( p0[i]-- != 0 )
            break;

    bitcounter_t top_val =
        ( p0[m_cElements - 1] &= ~s_aUnusedMask[cUnusedBits] );

    m_bOverflowed = ( top_val_prev == 0 && top_val != 0 );
}

//  > 0 or 1 if bit is cleared or set,
//    -1 = wrong index
int BitCounter::GetBit( _idx_ ix ) const
{
    if( ix < 0 || (_size_t_) ix >= m_cBits ) return -1;
    _idx_ ielement = ix / bitcounter_t_bits;
    _idx_ ibit     = ix % bitcounter_t_bits;
    return ( Ptr()[ielement] & (1L << ibit) )? 1 : 0;
}




void MemDat_Basic::Init( mem_t memtype )
{
    memset( this, 0, sizeof(MemDat_Basic) );
    m_MemType           = memtype;
}

void MemDat_Basic::Init( mem_t memtype, WORD wDataLockCount, WORD wReallocLockCount )
{
    memset( this, 0, sizeof(MemDat_Basic) );
    m_MemType           = memtype;
    m_wDataLockCount    = wDataLockCount;
    m_wReallocLockCount = wReallocLockCount;
}

// Takes only the initialzation params from the other basic data.
// The rest will be initalized with zeros and nulls.
void MemDat_Basic::Init( const MemDat_Basic& other_basic )
{
    memset( this, 0, sizeof(MemDat_Basic) );
    m_MemType           = other_basic.m_MemType;
    m_wDataLockCount    = other_basic.m_wDataLockCount;
    m_wReallocLockCount = other_basic.m_wReallocLockCount;
}

///<summary> Call this to init the union depending on the type
///  of memory being used. </summary>
bool MemDat_UnionExtended::Init( mem_t memtype )
{
    switch( memtype )
    {
        case MemT_FileMapping : m_filemapping.Init(); break;
        default : return false;
    }

    return true;
}

bool MemDat_UnionExtended::CopyFrom( mem_t memtype, const MemDat_Extended *ptr )
{
    if( !ptr )
        return false;

    switch( memtype )
    {
        case MemT_FileMapping : m_filemapping = *(MemDat_FileMapping*) ptr; break;
        default : return false;
    }

    return true;
}

_size_t_ MemDat_UnionExtended::StructSize( mem_t memtype )
{
    switch( memtype )
    {
        case MemT_FileMapping : return (_size_t_) sizeof(MemDat_FileMapping);
        default : return 0;
    }
}

MemReallocator::Realloc_T MemReallocator::s_aReallocFcts[] = {
    MemReallocator::Realloc_Malloc,
    MemReallocator::Realloc_New,
    MemReallocator::Realloc_FileMapping,
    MemReallocator::Realloc_External
};

UINT MemReallocator::GetNumReallocFcts()
{
    return sizeof(s_aReallocFcts) / sizeof(*s_aReallocFcts);
}

///<summary>
///  This can be used to reallocate memory without changing its
///  memory type. Setting <b>newsize</b> to zero will free the old memory.
///</summary>
///<remarks>
///  If you make the memory smaller, and you have assigned a block
///   size, it can happen that nothing is to reallocate, when the old
///   and the new size aligned to multiples of the defined block size
///   are of same size.
///   In that case, the padding bytes will or will not be cleared.
///   You never can expect the padding to have specific values. <br/><br/>
///  This method will try not to change the current memory if
///   an error occured. In almost all cases this will work.
///</remarks>
///<param name="headersize">
///  This is a header size specific to the memory type. It should either
///  be permanently zero or permanently on the same size for the
///  same memory type used in the same memory object.
///</param>
///<param name="newsize">
///  The new size parameter does not include the header size.
///</param>
///<param name="maxsize">
///  The maximum size to allow to allocate the memory (even the
///  padding bytes will be truncated to that size).
///  If it is <b>(_size_t_) -1L</b> then there is no maximum size limit.
///</param>
///<param name="bForbidRealloc">
///  If this is TRUE, then the reallocation method will return an error
///  when it really needs to reallocate the memory.
///  But if it does not need to reallocate its memory (e.g. because
///  same size and all other necessary params are equal),
///  then this method will succeed nevertheless.
///</param>
reterr_t MemReallocator::ReallocSameMemType( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc )
{
    if( (UINT) basic.m_MemType >= GetNumReallocFcts() )
        return interr (errIllegalMemType);

    if( pstrMemName )
        if( !*pstrMemName ) pstrMemName = NULL;

    return s_aReallocFcts[ (int) basic.m_MemType ]( basic, settings, ext, refcnt, headersize, newsize, pstrMemName, maxsize, bForbidRealloc );
}

reterr_t MemReallocator::Realloc_Malloc( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc )
{
    if( newsize > maxsize && maxsize != (_size_t_) -1L ) return syserr (ERROR_NOT_ENOUGH_MEMORY);
    if( basic.m_cMemory == newsize ) return true;
    if( newsize < 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( bForbidRealloc ) return interr (errMemoryReallocLocked);

    _size_t_ block     = settings.GetBlockSize();
    _size_t_ allocsize = ((newsize + block - 1) / block) * block;

    if( allocsize > maxsize && maxsize != (_size_t_) -1L )
        allocsize = maxsize;

    if( basic.m_cAllocated == allocsize ||
        (newsize >= basic.m_cMemory && newsize <= basic.m_cAllocated) )
    {
        // If memory has been enlarged (also inside steps of #block#),
        // then we must clear the new bytes (without padding bytes
        // because of speed).
        if( newsize > basic.m_cMemory && basic.m_pMemory && settings.m_bClearNewMem )
            memset( basic.m_pMemory + basic.m_cMemory, 0, newsize - basic.m_cMemory );
        basic.m_cMemory = newsize;
        return true;
    }

    PBYTE ptr = basic.m_pMemory;
    if( ptr ) ptr -= headersize;

    _size_t_ total_size = allocsize > 0 ? headersize + allocsize : 0;

    // Reallocate
    PBYTE ptr2 = (PBYTE) realloc( (void*) ptr, total_size );
    if( !ptr2 && total_size > 0 )
        return syserr (ERROR_OUTOFMEMORY);

    if( ptr2 ) ptr2 += headersize;

    // If memory has been enlarged (also inside steps of #block#),
    // then we must clear the new bytes (without padding bytes
    // because of speed).
    if( newsize > basic.m_cMemory && ptr2 && settings.m_bClearNewMem )
        memset( ptr2 + basic.m_cMemory, 0, newsize - basic.m_cMemory );

    basic.m_pMemory    = ptr2;
    basic.m_cMemory    = newsize;
    basic.m_cAllocated = allocsize;

    return true;
}

reterr_t MemReallocator::Realloc_New( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc )
{
    if( newsize > maxsize && maxsize != (_size_t_) -1L ) return syserr (ERROR_NOT_ENOUGH_MEMORY);
    if( basic.m_cMemory == newsize ) return true;
    if( newsize < 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( bForbidRealloc ) return interr (errMemoryReallocLocked);

    _size_t_ block     = settings.GetBlockSize();
    _size_t_ allocsize = ((newsize + block - 1) / block) * block;

    if( allocsize > maxsize && maxsize != (_size_t_) -1L )
        allocsize = maxsize;

    if( basic.m_cAllocated == allocsize ||
        (newsize >= basic.m_cMemory && newsize <= basic.m_cAllocated) )
    {
        // If memory has been enlarged (also inside steps of #block#),
        // then we must clear the new bytes (without padding bytes
        // because of speed).
        if( newsize > basic.m_cMemory && basic.m_pMemory && settings.m_bClearNewMem )
            memset( basic.m_pMemory + basic.m_cMemory, 0, newsize - basic.m_cMemory );
        basic.m_cMemory = newsize;
        return true;
    }

    PBYTE ptr = basic.m_pMemory;
    if( ptr ) ptr -= headersize;

    // First allocate the new memory.
    PBYTE ptr2 = NULL;
    MemDat_Basic newbasic;
    newbasic.Init( basic );
    if( allocsize > 0 )
    {
        ptr2 = new BYTE [ headersize + allocsize ];

        if( !ptr2 )
            return syserr (ERROR_OUTOFMEMORY);

        newbasic.m_pMemory    = ptr2 + headersize;
        newbasic.m_cMemory    = newsize;
        newbasic.m_cAllocated = allocsize;
    }

    if( ptr2 )
    {
        // Now copy the old to the new memory.
        _size_t_ ccopy = headersize + tmin( newbasic.m_cMemory, basic.m_cMemory );
        if( ptr && ccopy > 0 )
            memcpy( ptr2, ptr, ccopy );

        // If memory has been enlarged (also inside steps of #block#),
        // then we must clear the new bytes (without padding bytes
        // because of speed).
        if( newbasic.m_cMemory > basic.m_cMemory && settings.m_bClearNewMem )
            memset( newbasic.m_pMemory + basic.m_cMemory, 0, newbasic.m_cMemory - basic.m_cMemory );
    }

    // And free the old memory.
    if( ptr )
        delete [] ptr;

    basic = newbasic;

    return true;
}

reterr_t MemReallocator::Realloc_FileMapping( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc )
{
    if( !ext ) return syserr (ERROR_INVALID_PARAMETER);
    if( newsize > maxsize && maxsize != (_size_t_) -1L ) return syserr (ERROR_NOT_ENOUGH_MEMORY);
    MemDat_FileMapping *ex = (MemDat_FileMapping*) ext;
    bool bNaming = pstrMemName != NULL;
    if( basic.m_cMemory == newsize && ((bool) basic.m_bNamed == bNaming || basic.m_bNamed) ) return true;
    if( newsize < 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( bForbidRealloc ) return interr (errMemoryReallocLocked);

    _size_t_ block     = settings.GetBlockSize();
    _size_t_ allocsize = ((newsize + block - 1) / block) * block;

    if( allocsize > maxsize && maxsize != (_size_t_) -1L )
        allocsize = maxsize;

    if( ( basic.m_cAllocated == allocsize ||
          (newsize >= basic.m_cMemory && newsize <= basic.m_cAllocated) ||
          (basic.m_bNamed && newsize <= basic.m_cAllocated) ) &&
        ( (bool) basic.m_bNamed == bNaming || basic.m_bNamed ) )
    {
        // If memory has been enlarged (also inside steps of #block#),
        // then we must clear the new bytes (without padding bytes
        // because of speed).
        if( newsize > basic.m_cMemory && basic.m_pMemory && settings.m_bClearNewMem )
            memset( basic.m_pMemory + basic.m_cMemory, 0, newsize - basic.m_cMemory );
        basic.m_cMemory = newsize;
        return true;
    }

    // if memory does already exist and it is named,
    // we cannot support resizing memory outside the
    // first time allocation size (which depended on the block size).
    if( ex->m_hFileMapping && basic.m_bNamed && newsize != 0 )
        return interr (errMemoryNotResizeable);

    PBYTE ptr = basic.m_pMemory;
    if( ptr ) ptr -= headersize;

    // First allocate the new memory.
    PBYTE ptr2 = NULL;
    MemDat_Basic newbasic;
    MemDat_FileMapping newex;
    newbasic.Init( basic );
    newex.Init();
    bool bNamedMappingExisted = false;
    if( allocsize > 0 )
    {
        newex.m_hFileMapping = ::CreateFileMappingW( INVALID_HANDLE_VALUE,
                                                     NULL,
                                                     PAGE_READWRITE,
                                                     (DWORD) ( ((UINT64)(headersize + allocsize)) >> 32 ),
                                                     (DWORD) (headersize + allocsize),
                                                     pstrMemName );
        // Remarks: the high dword of the size is done by rolling the
        // top 32 bits downwards. But if the rolled value is 32 bits in
        // size only, the rolling will have no effect. So we must
        // first cast it to a UINT64 type and then roll it.
        if( !newex.m_hFileMapping )
            return syserr (::GetLastError());

        bNamedMappingExisted = ( pstrMemName && ::GetLastError() == ERROR_ALREADY_EXISTS );

        ptr2 = (PBYTE) ::MapViewOfFile( newex.m_hFileMapping, FILE_MAP_WRITE, 0, 0, 0 );

        newbasic.m_pMemory = ptr2 + headersize;
        newbasic.m_cMemory = newsize;
        newbasic.m_cAllocated = allocsize;
        newbasic.m_bNamed = pstrMemName != NULL;

        if( !ptr2 ) {
            LONG err = ::GetLastError();
            ::CloseHandle( newex.m_hFileMapping );
            return syserr (err);
        }

        // At last, if the memory is named, we disable our reference
        // counter, because the reference handling will be performed
        // by the file mapping object itself (having a reference
        // to the handle).
        if( refcnt && newbasic.m_bNamed )
            refcnt->InitDisabled();
        // Note: For non-named memory, we can let be enabled the
        // reference counter. The result will be that only *once*
        // the file mapping object handle will be retrieved regardless
        // how many memory objects share the memory (in the same process).
        // The last one that releases its reference counter will
        // close the handle (handling the references in #ReallocMemoryShareable#()).
        // So this is correct then.
    }

    if( ptr2 )
    {
        // Now copy the old to the new memory.
        _size_t_ ccopy = headersize + tmin( newbasic.m_cMemory, basic.m_cMemory );
        if( ptr && ccopy > 0 )
            memcpy( ptr2, ptr, ccopy );

        // If memory has been enlarged (also inside steps of #block#),
        // then we must clear the new bytes (without padding bytes
        // because of speed).
        if( newbasic.m_cMemory > basic.m_cMemory && !( bNamedMappingExisted && basic.m_cAllocated == 0 ) && settings.m_bClearNewMem ) {
            // Note: We do *not* clear the memory if it is named and existed
            // already and previously no memory was allocated in #basic#.
            memset( newbasic.m_pMemory + basic.m_cMemory, 0, newbasic.m_cMemory - basic.m_cMemory );
        }
    }

    // And free the old memory.
    if( basic.m_pMemory ) ::UnmapViewOfFile( basic.m_pMemory );
    if( ex->m_hFileMapping ) ::CloseHandle( ex->m_hFileMapping );

    basic = newbasic;
    *ex   = newex;

    return true;
}

reterr_t MemReallocator::Realloc_External( MemDat_Basic& basic, const MemDat_Settings& settings, MemDat_Extended *ext, MemDat_RefCnt *refcnt, _size_t_ headersize, _size_t_ newsize, PCWSTR pstrMemName, _size_t_ maxsize, bool bForbidRealloc )
{
    return interr (errMemoryNotReallocatable);
}

///<remarks>
///  If you make the memory smaller, and you have assigned a block
///   size, it can happen that nothing is to reallocate, when the old
///   and the new size aligned to multiples of the defined block size
///   are of same size.
///   In that case, the padding bytes will or will not be cleared.
///   You never can expect the padding to have specific values. <br/><br/>
///  This method will try not to change the current memory if
///   an error occured. In almost all cases this will work.
///</remarks>
///<param name="maxsize">
///  The maximum size to allow to allocate the memory (even the
///  padding bytes will be truncated to that size).
///  If it is <b>(_size_t_)-1L</b> then there is no maximum size limit.
///</param>
///<param name="bOverrideLocks">
///  Normally this should be FALSE. But you can set it to TRUE to force
///  ignoring all kinds of locks of the memory.
///  This is especially necessary and useful in the destructor of a
///  memory object to force freeing the memory regardless if the
///  memory is marked locked or not.
///</param>
reterr_t MemReallocator::ReallocNonShareableMemory( MemDat_Basic& basic, MemDat_Settings& settings, MemDat_UnionExtended& ext, _size_t_ newsize, mem_t memtype, PCWSTR pstrMemName, _size_t_ maxsize, bool bOverrideLocks )
{
    if( newsize < 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( newsize != 0 && memtype == MemT_None ) return syserr (ERROR_INVALID_PARAMETER);
    if( basic.m_MemType == MemT_None && memtype == MemT_None ) return true;

    if( memtype == MemT_Dont_Change_Dynamic && basic.m_MemType == MemT_External )
        memtype = (mem_t) settings.m_memtCreate;

    if( (memtype & MemT_Dont_Change) && newsize == 0 )
        memtype = (mem_t) settings.m_memtCreate;

    // Check if we have to alloc just without changing memory type.
    reterr_t err;
    if( (mem_t)basic.m_MemType == memtype || (memtype & MemT_Dont_Change) )
    {
        bool bForbidRealloc = false;

        if( !bOverrideLocks ) {
            if( basic.m_wReallocLockCount > 0 ) bForbidRealloc = true;
            // If content data is locked, we forbid decreasing memory size
            // because this would mean to loose some of the data bytes.
            if( basic.m_wDataLockCount > 0 && newsize < basic.m_cMemory ) return interr (errMemoryDataLocked);
        }

        err = ReallocSameMemType( basic, settings, (MemDat_Extended*)&ext, NULL, 0, newsize, pstrMemName, maxsize, bForbidRealloc );

        if( err )
            settings.m_memtCreate = basic.m_MemType;

        return err;
    }

    if( !bOverrideLocks ) {
        if( basic.m_wReallocLockCount > 0 ) return interr (errMemoryReallocLocked);
        // If content data is locked, we forbid decreasing memory size
        // because this would mean to loose some of the data bytes.
        if( basic.m_wDataLockCount > 0 && newsize < basic.m_cMemory ) return interr (errMemoryDataLocked);
    }

    if( memtype & MemT_Dont_Change )
        memtype = (mem_t) basic.m_MemType;

    // First allocate the new memory.
    MemDat_Basic newbasic;
    MemDat_UnionExtended newext;
    newbasic.Init( memtype, basic.m_wDataLockCount, basic.m_wReallocLockCount );
    newext.Init( memtype );
    // Normally we would call #ReallocSameMemType()' if #newsize > 0#
    // because only then we would need to alloc some memory.
    // But since other params could be set/changed as well
    // we call it even when #newsize == 0#.
    if( !( err = ReallocSameMemType( newbasic, settings, (MemDat_Extended*) &newext, NULL, 0, newsize, pstrMemName, maxsize, false ) ) )
        return err;

    if( newbasic.m_pMemory )
    {
        // Now copy the old to the new memory.
        _size_t_ ccopy = tmin( newbasic.m_cMemory, basic.m_cMemory );
        if( basic.m_pMemory && ccopy > 0 )
            memcpy( newbasic.m_pMemory, basic.m_pMemory, ccopy );
        // Clearing any memory is not necessary since this
        // has been done already in the allocation above.
    }

    // And free the old memory.
    if( basic.m_pMemory && IsMemTypeAllocateable( (mem_t) basic.m_MemType ) )
        if( !( err = ReallocSameMemType( basic, settings, (MemDat_Extended*) &ext, NULL, 0, 0, NULL, (_size_t_) -1L, false ) ) ) {
            // Normally freeing memory should not lead into an error.
            // But what shall we do if it happens nevertheless?
            // In the case we only had to free memory (#newsize# == 0),
            //  we can return the error without getting into troubles.
            // But when we had to re-alloc memory, the #newbasic# already
            //  has allocated the new memory.
            //  I decided to simply ignore any memory-freeing-error in
            //  that case.
            if( newsize == 0 && basic.m_cMemory > 0 )
                return err;
        }

    basic = newbasic;
    ext   = newext;

    settings.m_memtCreate = newbasic.m_MemType;

    return true;
}

///<summary>
///  This reallocates the memory taking into account the reference counter. <br/>
///   And for newly allocated memory, it will initialize the header data (and ids)
///   and the reference counter.
///</summary>
///<remarks>
///  If you make the memory smaller, and you have assigned a block
///   size, it can happen that nothing is to reallocate, when the old
///   and the new size aligned to multiples of the defined block size
///   are of same size.
///   In that case, the padding bytes will or will not be cleared.
///   You never can expect the padding to have specific values. <br/><br/>
///  This method will try not to change the current memory if
///   an error occured. In almost all cases this will work.
///</remarks>
///<param name="maxsize">
///  The maximum size to allow to allocate the memory (even the
///  padding bytes will be truncated to that size).
///  If it is <b>(_size_t_)-1L</b> then there is no maximum size limit.
///</param>
///<param name="bOverrideLocks">
///  Normally this should be FALSE. But you can set it to TRUE to force
///  ignoring all kinds of locks of the memory.
///  This is especially necessary and useful in the destructor of a
///  memory object to force freeing the memory regardless if the
///  memory is marked locked or not.
///</param>
///<param name="bForceRealloc">
///  This has *nothing* to do with <b>bOverrideLocks</b>. Is is only an
///  indicator that if normally memory will not be reallocated (because
///  it is shared and neither size nor memtype are changed), reallocation
///  will be done nevertheless. <br/>
///  But if reallocation is locked (and <b>bOverrideLocks</b> is FALSE),
///  then reallocation returns an error nevertheless.
///</param>
reterr_t MemReallocator::ReallocShareableMemory( PBYTE& ptrMem, MemDat_Settings& settings, MemDat_ForeignMem& foreign, _size_t_ newsize, mem_t memtype, _size_t_ maxsize, bool bOverrideLocks, bool bForceRealloc )
{
    if( newsize < 0 ) return syserr (ERROR_INVALID_PARAMETER);
    if( newsize != 0 && memtype == MemT_None ) return syserr (ERROR_INVALID_PARAMETER);

    MemDat_ShareHdr *hdr;
    MemDat_Basic basic;
    MemDat_RefCnt refcnt;
    MemDat_UnionExtended ext;
    _size_t_ total_header_size;

    if( !ptrMem )
    {
        hdr = NULL;

        basic.Init( memtype & MemT_Dont_Change ? (mem_t) settings.m_memtCreate : memtype );
        refcnt.InitEnabled();
        ext.Init( (mem_t) basic.m_MemType );

        total_header_size = 0;
    }
    else if( foreign.m_bForeignMem )
    {
        hdr = NULL;
        // Foreign memory we never occupy so we never have a header
        // for it that we would be allowed to update.

        basic.Init( MemT_External );
        basic.m_pMemory = ptrMem;
        basic.m_cMemory = (_size_t_) (UINT) foreign.m_cForeignSize;
        refcnt.InitDisabled();

        total_header_size = 0;
    }
    else
    {
        //hdr = GetHeaderPtr_Safe( ptrMem, FALSE, FALSE );
        hdr = GetHeaderPtr( ptrMem );
        if( !hdr )
            return interr (errFailed); // should never occur

        if( !MemReallocator::GetBasic_Safe( basic, hdr ) )
            return interr (errCritical); // should never occur
        refcnt = hdr->m_RefCnt;
        MemDat_Extended *ex = GetExtendedPtr( hdr );
        ext.CopyFrom( (mem_t) basic.m_MemType, ex );

        total_header_size = GetTotalHeaderSize( hdr );
    }

    if( basic.m_MemType == MemT_None && memtype == MemT_None )
        return true;

    if( memtype == MemT_Dont_Change_Dynamic && basic.m_MemType == MemT_External )
        memtype = (mem_t) settings.m_memtCreate;

    if( (memtype & MemT_Dont_Change) && newsize == 0 )
        memtype = (mem_t) settings.m_memtCreate;

    // Check if we have to alloc just without changing memory type.
    // (But if we have more references, than we must make a copy).
    // (If no memory was allocated yet, we must go to the make-copy
    // handling, because in there, the newly allocated memory's
    // header will be initialized completely).
    reterr_t err;
    if( ((mem_t)basic.m_MemType == memtype || (memtype & MemT_Dont_Change)) && refcnt.m_Counter == 1 && basic.m_pMemory )
    {
        bool bForbidRealloc = false;

        if( !bOverrideLocks ) {
            if( basic.m_wReallocLockCount > 0 ) bForbidRealloc = true;
            // If content data is locked, we forbid decreasing memory size
            // because this would mean to loose some of the data bytes.
            if( basic.m_wDataLockCount > 0 && newsize < basic.m_cMemory ) return interr (errMemoryDataLocked);
        }

        err = ReallocSameMemType( basic, settings, (MemDat_Extended*) &ext, &refcnt, total_header_size, newsize, NULL, maxsize, bForbidRealloc );

        if( !err )
            return err;

        hdr = GetHeaderPtr( basic.m_pMemory );
        if( hdr )
        {
            hdr->m_Basic  = basic;
            hdr->m_RefCnt = refcnt;
            MemDat_UnionExtended *ex = (MemDat_UnionExtended*) GetExtendedPtr( hdr );
            ex->CopyFrom( (mem_t) basic.m_MemType, (MemDat_Extended*) &ext );
        }

        ptrMem = basic.m_pMemory;
        foreign.Empty();

        settings.m_memtCreate = basic.m_MemType;

        return err;
    }

    if( memtype & MemT_Dont_Change )
        memtype = (mem_t) basic.m_MemType;

    // If we have memory shared but we do neither change the memory type
    // nor the memory size, then we do not need to reallocate at all
    // (as long as we do not force to reallocate).
    // (For nonallocateable memories, or if no memory is allocated right now,
    // we pass it on below because we want to run through the realloc functions
    // which will produce an error accordingly if reallocation is not possible.)
    if( (mem_t)basic.m_MemType == memtype && newsize == basic.m_cMemory && basic.m_pMemory &&
        IsMemTypeAllocateable(memtype) && !bForceRealloc )
    {
        settings.m_memtCreate = basic.m_MemType;
        return true;
    }

    if( !bOverrideLocks ) {
        if( basic.m_wReallocLockCount > 0 ) return interr (errMemoryReallocLocked);
        // If content data is locked, we forbid decreasing memory size
        // because this would mean to loose some of the data bytes.
        if( basic.m_wDataLockCount > 0 && newsize < basic.m_cMemory ) return interr (errMemoryDataLocked);
    }

    // First allocate the new memory.
    MemDat_Basic newbasic;
    MemDat_RefCnt newrefcnt;
    MemDat_UnionExtended newext;
    newbasic.Init( memtype, basic.m_wDataLockCount, basic.m_wReallocLockCount );
    newrefcnt.InitEnabled();
    newext.Init( memtype );
    _size_t_ new_total_header_size = GetTotalHeaderSize( memtype );
    // Normally we would call #ReallocSameMemType()' if #newsize > 0#
    // because only then we would need to alloc some memory.
    // But since other params could be set/changed as well
    // we call it even when #newsize == 0#.
    if( !( err = ReallocSameMemType( newbasic, settings, (MemDat_Extended*) &newext, &newrefcnt, new_total_header_size, newsize, NULL, maxsize, false ) ) )
        return err;

    MemDat_ShareHdr *newhdr = GetHeaderPtr( newbasic.m_pMemory );
    if( newhdr )
    {
        newhdr->m_4cc    = MEMDAT_SHHDRID;
        newhdr->m_pid    = GetCurrentProcessId();
        newhdr->m_Basic  = newbasic;
        newhdr->m_RefCnt = newrefcnt;
        MemDat_UnionExtended *newex = (MemDat_UnionExtended*) GetExtendedPtr( newhdr );
        newex->CopyFrom( (mem_t) newbasic.m_MemType, (MemDat_Extended*) &newext );
    }

    if( newbasic.m_pMemory )
    {
        // Now copy the old to the new memory.
        _size_t_ ccopy = tmin( newbasic.m_cMemory, basic.m_cMemory );
        if( basic.m_pMemory && ccopy > 0 )
            memcpy( newbasic.m_pMemory, basic.m_pMemory, ccopy );
        // Clearing any memory is not necessary since this
        // has been done already in the allocation above.
    }

    if( refcnt.IsLastRef_Unsafe() )
    {
        // Free the old memory if we have the last reference (or reference
        // counting is disabled).
        if( basic.m_pMemory && IsMemTypeAllocateable( (mem_t) basic.m_MemType ) )
            if( !( err = ReallocSameMemType( basic, settings, (MemDat_Extended*) &ext, &refcnt, total_header_size, 0, NULL, (_size_t_) -1L, false ) ) ) {
                // Normally freeing memory should not lead into an error.
                // But what shall we do if it happens nevertheless?
                // In the case we only had to free memory (#newsize# == 0),
                //  we can return the error without getting into troubles.
                // But when we had to re-alloc memory, the #newbasic# already
                //  has allocated the new memory.
                //  I decided to simply ignore any memory-freeing-error in
                //  that case.
                if( newsize == 0 && basic.m_cMemory > 0 )
                    return err;
            }
    }
    else
    {
        // Otherwise we only decrease the reference counter.
        if( hdr )
        {
            hdr->m_RefCnt.Dec();
            // CAUTION: In here, we do *not* unlock any locks
            // that we did lock because we don't know, which of
            // the locks have been done by us and which by other
            // (e.g.) #MemoryShared# objects sharing the same
            // memory.
            // So the user must ensure for himself that all locks
            // he did make are first undone before he releases
            // any memory.
        }
    }

    ptrMem = newbasic.m_pMemory;
    foreign.Empty();

    settings.m_memtCreate = newbasic.m_MemType;

    return true;
}

///<remarks> If <b>bIncRefCnt</b> is TRUE, then the reference counter will be
///  increased only if a non-NULL pointer is returned from this method. </remarks>
///<param name="bForbidLocked"> if TRUE, NULL will be returned if the
///  pointed memory is somehow locked. </param>
MemDat_ShareHdr *MemReallocator::GetHeaderPtr_Safe( const BYTE *ptrMem, bool bIncRefCnt, bool bForbidLocked )
{
    if( !ptrMem )
        return NULL;

    if( IsBadWritePtr( (PBYTE) ptrMem - sizeof(MemDat_ShareHdr), sizeof(MemDat_ShareHdr) ) )
        return NULL;

    MemDat_ShareHdr *pHdr = NULL;

    tryHW
    {
        MemDat_ShareHdr *h = ((MemDat_ShareHdr*) ptrMem) - 1;

        if( h->m_4cc == MEMDAT_SHHDRID &&
            h->m_pid == GetCurrentProcessId() &&
            (h->m_Basic.m_pMemory == ptrMem || !h->m_RefCnt.m_bEnabled) &&
            IsMemTypeAllocateable( (mem_t) h->m_Basic.m_MemType ) &&
            (!bForbidLocked || !h->m_Basic.IsAnythingLocked()) )
        {
            pHdr = h;
            if( bIncRefCnt ) h->m_RefCnt.Inc();
        }
    }
    catchHW
    {
        pHdr = NULL;
    }

    return pHdr;
}

///<remarks> If <b>bIncRefCnt</b> is TRUE, then the reference counter will be
///  increased only if a non-NULL pointer is returned from this method. </remarks>
///<param name="bForbidLocked"> if TRUE, NULL will be returned if the
///  pointed memory is somehow locked. </param>
MemDat_ShareHdr *MemReallocator::GetHeaderPtr_Safe( MemDat_Extended *& pExt, const BYTE *ptrMem, bool bIncRefCnt, bool bForbidLocked )
{
    pExt = NULL;

    if( !ptrMem )
        return NULL;

    if( IsBadWritePtr( (PBYTE) ptrMem - sizeof(MemDat_ShareHdr), sizeof(MemDat_ShareHdr) ) )
        return NULL;

    MemDat_ShareHdr *pHdr = NULL;
    mem_t memtype = MemT_None;

    tryHW
    {
        MemDat_ShareHdr *h = ((MemDat_ShareHdr*) ptrMem) - 1;

        if( h->m_4cc == MEMDAT_SHHDRID &&
            h->m_pid == GetCurrentProcessId() &&
            (h->m_Basic.m_pMemory == ptrMem || !h->m_RefCnt.m_bEnabled) &&
            IsMemTypeAllocateable( (mem_t) h->m_Basic.m_MemType ) &&
            (!bForbidLocked || !h->m_Basic.IsAnythingLocked()) )
        {
            pHdr = h;
            memtype = (mem_t) h->m_Basic.m_MemType;
        }
    }
    catchHW
    {
        pHdr = NULL;
    }

    if( !pHdr )
        return NULL;

    _size_t_ extsize = MemDat_UnionExtended::StructSize( memtype );
    MemDat_Extended *ext = (MemDat_Extended*) (((PBYTE) pHdr) - extsize);

    if( extsize > 0 )
        if( IsBadWritePtr( ext, extsize ) )
            return NULL;

    if( bIncRefCnt )
    {
        tryHW { pHdr->m_RefCnt.Inc(); }
        catchHW { pHdr = NULL; ext = NULL; }
    }

    pExt = ext;
    return pHdr;
}

MemDat_ShareHdr *MemReallocator::GetHeaderPtr( MemDat_Extended *& pExt, const BYTE *ptrMem )
{
    pExt = NULL;

    if( !ptrMem )
        return NULL;

    MemDat_ShareHdr *pHdr = ((MemDat_ShareHdr*) ptrMem) - 1;

    _size_t_ extsize = MemDat_UnionExtended::StructSize( (mem_t) pHdr->m_Basic.m_MemType );
    pExt = (MemDat_Extended*) (((PBYTE) pHdr) - extsize);

    return pHdr;
}

_size_t_ MemReallocator::GetTotalHeaderSize( const MemDat_ShareHdr *pHdr )
{
    if( !pHdr )
        return 0;

    _size_t_ total_header_size, extsize;

    tryHW { extsize = MemDat_UnionExtended::StructSize( (mem_t) pHdr->m_Basic.m_MemType ); }
    catchHW { return 0; }

    total_header_size = extsize + (_size_t_) sizeof(MemDat_ShareHdr);
    total_header_size = ((total_header_size + MEMDAT_TOTAL_HEADER_BLOCKING - 1) / MEMDAT_TOTAL_HEADER_BLOCKING) * MEMDAT_TOTAL_HEADER_BLOCKING;

    return total_header_size;
}

_size_t_ MemReallocator::GetTotalHeaderSize( mem_t memtype )
{
    if( memtype == MemT_External )
        return 0;

    _size_t_ total_header_size, extsize;

    extsize = MemDat_UnionExtended::StructSize( memtype );

    total_header_size = extsize + (_size_t_) sizeof(MemDat_ShareHdr);
    total_header_size = ((total_header_size + MEMDAT_TOTAL_HEADER_BLOCKING - 1) / MEMDAT_TOTAL_HEADER_BLOCKING) * MEMDAT_TOTAL_HEADER_BLOCKING;

    return total_header_size;
}

bool MemReallocator::GetBasic_Safe( MemDat_Basic& basic, MemDat_ShareHdr *pHdr )
{
    if( !pHdr ) return false;

    tryHW { basic = pHdr->m_Basic; }
    catchHW { return false; }

    return true;
}

bool MemReallocator::DecRefCnt_Safe( MemDat_ShareHdr *pHdr )
{
    if( !pHdr ) return false;

    tryHW { pHdr->m_RefCnt.Dec(); }
    catchHW { false; }

    return true;
}




_retPByte_t_ MemorySpace::WritePtr()
{
    // we get a write-to-pointer only if content data is not locked (#IsDataLocked()#)
    if( IsDataLocked() )
        return _retPByte_t_(interr (errMemoryDataLocked), NULL);

    return _retPByte_t_(true, m_Basic.m_pMemory);
}



reterr_t MemorySpace::SetSizeAndType( _size_t_ count, mem_t memtype, PCSTR pstrMemName )
{
    PWSTR pstrMemNameW = NULL;

    _size_t_ len = pstrMemName && count > 0 ? (_size_t_) _strlen_( pstrMemName ) : (_size_t_) 0;
    if( len > 0 ) {
        pstrMemNameW = (PWSTR) malloc( (len + 1) * sizeof(WCHAR) );
        if( !pstrMemNameW ) return syserr (ERROR_OUTOFMEMORY);
        MultiByteToWideChar( CP_ACP, 0, pstrMemName, (int) len + 1, pstrMemNameW, (int) len + 1 );
    }

    reterr_t err = SetSizeAndType( count, memtype, pstrMemNameW );

    if( pstrMemNameW )
        free( pstrMemNameW );

    return err;
}

///<summary> Use with care. It will ignore if the memory is locked and
///  will enlarge memory size. But making it smaller is never allowed
///  in any case. </summary>
reterr_t MemorySpace::ForceEnlargeToSize( _size_t_ count )
{
    if( count < m_Basic.m_cMemory )
        return interr (errMemoryCannotDecrease);

    return MemReallocator::ReallocNonShareableMemory( m_Basic, m_Settings, m_Extended,
                                                      count, MemT_Dont_Change_Dynamic, NULL,
                                                      (_size_t_) -1L, true );
}

reterr_t MemorySpace::SetExternalMem( LPCVOID ptr, _size_t_ count )
{
    if( !ptr || count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    reterr_t ret = Free();
    if( !ret ) return ret;

    m_Basic.Init( MemT_External );
    m_Basic.m_pMemory = (PBYTE) ptr;
    m_Basic.m_cMemory = m_Basic.m_cAllocated = count;
    m_Extended.Init( (mem_t) m_Basic.m_MemType );

    return true;
}



reterr_t MemorySpace::SetData( LPCVOID ptr, _size_t_ count )
{
    if( !ptr || count < 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    reterr_t ret = SetSize( count );
    if( !ret ) return ret;

    if( count > 0 )
        memcpy( m_Basic.m_pMemory, ptr, count );

    return true;
}

reterr_t MemorySpace::Add( _size_t_ count )
{
    if( count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize + count;

    return SetSize( newsize );
}

reterr_t MemorySpace::Add( LPCVOID ptr, _size_t_ count )
{
    if( !ptr || count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    memcpy( m_Basic.m_pMemory + oldsize, ptr, count );

    return true;
}

///<summary>
///  Adds new bytes at the end of the current memory.
///  The new bytes will be cleared.
///</summary>
///<returns>
///  Returns a pointer to the first byte of the new memory.
///  It will be a write-to-pointer. <br/>
///  Or NULL if allocation failed.
///</returns>
_retPByte_t_ MemorySpace::Add_Ptr( _size_t_ count )
{
    if( count <= 0 ) return _retPByte_t_(syserr (ERROR_INVALID_PARAMETER), NULL);

    // Even if only data is locked we return an error
    // because we would have to return a write-to-pointer
    // which is not allowed when data is locked.
    if( IsDataLocked() )
        return _retPByte_t_(interr (errMemoryDataLocked), NULL);

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return _retPByte_t_(ret, NULL);

    return _retPByte_t_(true, m_Basic.m_pMemory + oldsize);
}

///<remarks> This method supports also <b>mem</b> being
///  <b>this</b> <b>MemorySpace</b> object. </remarks>
reterr_t MemorySpace::Append( const MemorySpace& mem )
{
    if( mem.IsEmpty() ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ srcsize = mem.m_Basic.m_cMemory;

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize + srcsize;

    // The first operation (*before* retrieving the
    // source data pointer) is to resize the memory
    // so that we can support #mem# being #this#.
    // But the source data size we must have queried
    // already before.

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    memcpy( m_Basic.m_pMemory + oldsize, mem.m_Basic.m_pMemory, srcsize );

    return true;
}

reterr_t MemorySpace::Insert( _idx_ idx, _size_t_ count, LPCVOID ptr )
{
    if( !ptr ) return syserr (ERROR_INVALID_PARAMETER);
    if( idx < 0 || (_size_t_) idx > m_Basic.m_cMemory ) return interr (errOutOfRange);
    if( count == 0 ) return true;
    if( count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    // Now move away some parts of the content data
    // to make room for the inserted bytes.
    if( (_size_t_) idx < oldsize )
    {
        _size_t_ cmove = oldsize - idx;
        memmove( m_Basic.m_pMemory + idx + count, m_Basic.m_pMemory + idx, cmove );
    }

    memcpy( m_Basic.m_pMemory + idx, ptr, count );

    return true;
}

///<summary>
///  Inserts new bytes somewhere inside the current memory.
///  The new bytes will be cleared only if the settings are
///  set that way that new memory shall be cleared automatically.
///</summary>
///<returns>
///  Returns a pointer to the first byte of the newly inserted memory range.
///  It will be a write-to-pointer. <br/>
///  Or NULL if allocation failed or on wrong params.
///</returns>
///<remarks>
///  If <b>idx</b> specifies a location completely outside
///   the current memory, NULL will be returned. <br/>
///  If <b>idx</b> points exactly at the end of the current
///   memory (so it has the same value as returned by <b>GetSize()</b>),
///   the behavior of this method is the same as <b>Add()</b>. <br/>
///</remarks>
_retPByte_t_ MemorySpace::Insert_Ptr( _idx_ idx, _size_t_ count )
{
    if( idx < 0 || (_size_t_) idx > m_Basic.m_cMemory ) return _retPByte_t_(interr (errOutOfRange), NULL);
    if( count == 0 ) return _retPByte_t_(true, NULL);
    if( count <= 0 ) return _retPByte_t_(syserr (ERROR_INVALID_PARAMETER), NULL);

    // Even if only data is locked we return an error
    // because we would have to return a write-to-pointer
    // which is not allowed when data is locked.
    if( IsDataLocked() )
        return _retPByte_t_(interr (errMemoryDataLocked), NULL);

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return _retPByte_t_(ret, NULL);

    // Now move away some parts of the content data
    // to make room for the inserted bytes.
    if( (_size_t_) idx < oldsize )
    {
        _size_t_ cmove = oldsize - idx;
        memmove( m_Basic.m_pMemory + idx + count, m_Basic.m_pMemory + idx, cmove );
        // And clear the inserted bytes where previously old data was.
        if( m_Settings.m_bClearNewMem )
            memset( m_Basic.m_pMemory + idx, 0, tmin( cmove, count ) );
    }

    return _retPByte_t_(true, m_Basic.m_pMemory + idx);
}

reterr_t MemorySpace::Remove( _idx_ idx, _size_t_ count )
{
    if( !CheckRangeLimits( idx, count ) )
        return interr (errOutOfRange);

    // Removing any memory would mean also remove content data
    // which is not allowed when data is locked.
    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( count <  0 ) return interr (errOutOfRange);
    if( count == 0 ) return true;

    if( IsReallocLocked() ) return interr (errMemoryReallocLocked);

    _size_t_ oldsize = m_Basic.m_cMemory;
    _size_t_ newsize = oldsize - count;

    // Now move together the left and right side
    // of the memory area.
    if( idx + count < oldsize ) {
        _size_t_ cmove = oldsize - (idx + count);
        memmove( m_Basic.m_pMemory + idx, m_Basic.m_pMemory + idx + count, cmove );
    }

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    return true;
}

reterr_t MemorySpace::Clear()
{
    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( m_Basic.m_pMemory && m_Basic.m_cMemory > 0 )
        memset( m_Basic.m_pMemory, 0, m_Basic.m_cMemory );

    return true;
}

reterr_t MemorySpace::Clear( _idx_ idx, _size_t_ count )
{
    if( !CheckRangeLimits( idx, count ) )
        return interr (errOutOfRange);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( count > 0 )
        memset( m_Basic.m_pMemory + idx, 0, count );

    return true;
}

///<summary>
///  Normally you should not care about the padding bytes.
///  But if you want to, you can use this method to clear them.
///</summary>
///<remarks>
///  What is the padding: <br/>
///   The padding bytes are allocated but unused bytes. They
///   come from specifying a block size (e.g. using <b>SetBlockSize()</b>).
///   <br/>
///   Using blocks larger than 1 byte may cause the allocated
///   memory to be larger than the requested number of bytes (e.g.
///   when calling <b>SetSize()</b>). Although the valid size the
///   user should use still is the size he requested, there may
///   have been allocated some additional, unused padding bytes. <br/>
///   Normally the user should not use them in any case.
///   <br/>
///   But when he is decreasing the memory buffer size (e.g. using
///   <b>SetSize()</b> with a smaller size value), the padding
///   bytes may not be cleared and could possibly contain old
///   data from previous execution steps. <br/>
///   If the user does not want this, he has the opportunity to
///   clear them using <b>ClearPadding()</b>.
///   <br/><br/>
///  If content data is locked (<b>IsDataLocked()</b>), then
///   this does not include the padding bytes. So they may be
///   cleared at any point of time.
///</remarks>
reterr_t MemorySpace::ClearPadding()
{
    if( m_Basic.m_cAllocated <= m_Basic.m_cMemory ) return true;

    if( m_Basic.m_pMemory )
        memset( m_Basic.m_pMemory + m_Basic.m_cMemory, 0, m_Basic.m_cAllocated - m_Basic.m_cMemory );

    return true;
}

///<summary> Clears the complete buffer including the padding bytes </summary>
reterr_t MemorySpace::ClearAllocatedBuffer()
{
    if( m_Basic.m_cAllocated <= 0 ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( m_Basic.m_pMemory )
        memset( m_Basic.m_pMemory, 0, m_Basic.m_cAllocated );

    return true;
}



///<summary>
///  Copies only the properties (memtype, blocksize etc.)
///  from another <b>MemorySpace</b> object.
///</summary>
///<remarks>
///  This method will work regardless, if there is
///   memory allocated already in <b>this</b> object or not.
///</remarks>
reterr_t MemorySpace::CopyPropertiesFrom( const MemorySpace& mem )
{
    if( this == &mem ) return true;

    //if( !IsEmpty() )
    //    return interr (errMemoryMustBeEmpty);

    m_Settings.CopyPropertiesFrom( mem.m_Settings );
    // If no memory is currently allocated, we take over the
    // memory type for the basic data as well because its
    // memory type will be used when creating new memory.
    if( !m_Basic.m_pMemory )
        m_Basic.m_MemType = m_Settings.m_memtCreate;

    return true;
}

///<summary>
///  Copies the properties and the memory from
///  another <b>MemorySpace</b> object.
///</summary>
///<remarks>
///  The memory name (if any) will *not* be copied.
///   Any old name will be deleted.
///</remarks>
reterr_t MemorySpace::CopyFrom( const MemorySpace& mem )
{
    if( this == &mem ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    reterr_t ret = CopyPropertiesFrom( mem );
    if( !ret ) return ret;

    _size_t_ count = mem.GetSize();

    if( mem.GetMemType() == MemT_External )
        return SetExternalMem( mem.m_Basic.m_pMemory, count );

    if( !( ret = SetSizeAndType( count, (mem_t) mem.m_Basic.m_MemType ) ) )
        return ret;

    if( m_Basic.m_pMemory && count > 0 )
        memcpy( m_Basic.m_pMemory, mem.m_Basic.m_pMemory, count );

    return true;
}

///<summary>
///  Exchanges everything between 2 <b>MemorySpace</b> objects
///  including the memory, all properties, memory-names etc. <br/>
///  But <u>except</u> the SingletonSettings. They are the settings
///  related only to the single memory object here.
///</summary>
reterr_t MemorySpace::Exchange( MemorySpace& mem )
{
    if( this == &mem ) return true;

    MemDat_Settings      tmp_settings;
    MemDat_Basic         tmp_basic;
    MemDat_UnionExtended tmp_extended;

    memcpy( &tmp_settings, &m_Settings, sizeof(MemDat_Settings     ) );
    memcpy( &tmp_basic,    &m_Basic,    sizeof(MemDat_Basic        ) );
    memcpy( &tmp_extended, &m_Extended, sizeof(MemDat_UnionExtended) );

    memcpy( &m_Settings, &mem.m_Settings, sizeof(MemDat_Settings     ) );
    memcpy( &m_Basic,    &mem.m_Basic,    sizeof(MemDat_Basic        ) );
    memcpy( &m_Extended, &mem.m_Extended, sizeof(MemDat_UnionExtended) );

    memcpy( &mem.m_Settings, &tmp_settings, sizeof(MemDat_Settings     ) );
    memcpy( &mem.m_Basic,    &tmp_basic,    sizeof(MemDat_Basic        ) );
    memcpy( &mem.m_Extended, &tmp_extended, sizeof(MemDat_UnionExtended) );

    return true;
}



///<summary>
///  Checks if the specified index and count-of-bytes fit inside
///  the current memory array, and truncates or adopts their
///  values accordingly so that they do not exceed the current
///  memory's range.
///</summary>
///<returns>
///  Returns FALSE if either no memory is allocated at the moment
///   or if the specified range does not overlap the current memory
///   area even in one byte, except if the resuling <b>count</b> is
///   calculated to zero. <br/>
///  But if it does overlap in one or more bytes, the range specified
///   by <b>idx</b> and <b>count</b> will be truncated to fit
///   completely in the current memory's range, and TRUE is returned.
///</returns>
bool MemorySpace::CheckRangeLimits( _idx_& idx, _size_t_& count, _size_t_ memsize, _idx_ left_limit )
{
    if( memsize <= 0 )
        return false;

    if( idx < left_limit ) {
        count += idx - left_limit;
        idx = left_limit;
    }
    if( idx + count > memsize )
        count = memsize - idx;

    if( count < 0 ) return false;
    //if( count == 0 ) return true;

    return true;
}




bool MemoryShared::IsShared() const
{
    if( m_ForeignMem.m_bForeignMem ) return true;
    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );
    return hdr ? hdr->m_RefCnt.IsShared() : false;
}

///<summary>
///  This method will check if its memory is shared with other
///  <b>MemoryShared</b> objects. And if so, it will make a
///  copy for itself alone.
///  If the desired size and/or the desired memory type are different,
///  then a copy of the memory will be made regardless if the memory is shared.
///</summary>
///<remarks>
///  In the case of using External Memory, this method will always
///  make a local copy of the external memory when using either a
///  real memory type or <b>MemT_Dont_Change_Dynamic</b>.
///  But <b>MemT_Dont_Change</b> will not make a copy of the
///  external memory.
///</remarks>
reterr_t MemoryShared::IsolateToSizeAndType( _size_t_ count, mem_t memtype )
{
    return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem,
                                                   count, memtype,
                                                   (_size_t_) -1L, false, true );
}



_retPByte_t_ MemoryShared::WritePtr()
{
    // we get a write-to-pointer only if content data is not locked (#IsDataLocked()#)
    if( IsDataLocked() )
        return _retPByte_t_(interr (errMemoryDataLocked), NULL);

    reterr_t ret = Isolate();
    if( !ret ) return _retPByte_t_(ret, NULL);

    return _retPByte_t_(true, m_pMemory);
}

// Retain means that the contents must not get lost.
_retPByte_t_ MemoryShared::WritePtr_Retain()
{
    if( IsEmpty() ) return _retPByte_t_(interr (errNoDataAvailable), NULL);
    return WritePtr();
}

// Retain means that the contents must not get lost.
_retPByte_t_ MemoryShared::WritePtr_Retain( _size_t_ count )
{
    if( count <= 0 )
        { Free(); return _retPByte_t_(interr (errNoDataAvailable), NULL); }

    reterr_t ret = IsolateToSize( count );
    if( !ret ) return _retPByte_t_(ret, NULL);

    return WritePtr();
}

// Retain means that the contents must not get lost.
_retPByte_t_ MemoryShared::WritePtr_RetainMinimum( _size_t_ minsize )
{
    reterr_t ret;

    if( GetSize() < minsize ) {
        if( !( ret = ForceEnlargeToSize( minsize ) ) )
            return _retPByte_t_(ret, NULL);
    } else {
        if( !( ret = Isolate() ) )
            return _retPByte_t_(ret, NULL);
    }

    return WritePtr();
}

// Discard means that the contents may be lost or non-initialized.
_retPByte_t_ MemoryShared::WritePtr_Discard( _size_t_ count )
{
    if( count <= 0 )
        { Free(); return _retPByte_t_(interr (errNoDataAvailable), NULL); }

    reterr_t ret = IsolateToSize( count );
    if( !ret ) return _retPByte_t_(ret, NULL);

    return WritePtr();
}

// Discard means that the contents may be lost or non-initialized.
_retPByte_t_ MemoryShared::WritePtr_Discard( _size_t_ count, mem_t memtype )
{
    if( count <= 0 )
        { Free(); SetMemType( memtype ); return _retPByte_t_(interr (errNoDataAvailable), NULL); }

    if( GetMemType() != memtype && IsRealMemType( memtype ) )
        Free();

    reterr_t ret = IsolateToSizeAndType( count, memtype );
    if( !ret ) return _retPByte_t_(ret, NULL);

    return WritePtr();
}



bool MemoryShared::IsDataLocked() const
{
    MemDat_ShareHdr *hdr = m_ForeignMem.m_bForeignMem ? NULL : MemReallocator::GetHeaderPtr( m_pMemory );
    return hdr ? hdr->m_Basic.m_wDataLockCount > 0 : false;
}

bool MemoryShared::IsReallocLocked() const
{
    MemDat_ShareHdr *hdr = m_ForeignMem.m_bForeignMem ? NULL : MemReallocator::GetHeaderPtr( m_pMemory );
    return hdr ? hdr->m_Basic.m_wReallocLockCount > 0 : false;
}



_size_t_ MemoryShared::GetSize() const
{
    if( m_ForeignMem.m_bForeignMem )
        return (_size_t_) (UINT) m_ForeignMem.m_cForeignSize;

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );

    return hdr ? hdr->m_Basic.m_cMemory : 0;
}

_size_t_ MemoryShared::GetAllocatedSize() const
{
    if( m_ForeignMem.m_bForeignMem )
        return (_size_t_) (UINT) m_ForeignMem.m_cForeignSize;

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );

    return hdr ? hdr->m_Basic.m_cAllocated : 0;
}

///<summary> Use with care. It will ignore if the memory is locked and
///  will enlarge memory size. But making it smaller is never allowed
///  in any case. </summary>
reterr_t MemoryShared::ForceEnlargeToSize( _size_t_ count )
{
    if( count < GetSize() )
        return interr (errMemoryCannotDecrease);

    return MemReallocator::ReallocShareableMemory( m_pMemory, m_Settings, m_ForeignMem,
                                                   count, MemT_Dont_Change_Dynamic,
                                                   (_size_t_) -1L, true );
}

reterr_t MemoryShared::SetExternalMem( LPCVOID ptr, _size_t_ count )
{
    if( !ptr || count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    reterr_t ret = Free();
    if( !ret ) return ret;

    m_pMemory = (PBYTE) ptr;
    m_ForeignMem.m_cForeignSize = count;
    m_ForeignMem.m_bForeignMem = 1;

    return true;
}

///<summary>
///  This method will try to read the header before the specified
///   pointer and if it can retrieve it, it checks, if it is a
///   shareable memory (used by the current process). <br/>
///  If it succeeds, it takes over the pointer and creates
///   a reference to that memory, thus sharing it, too. <br/>
///  If the method fails, you could call <b>SetExternalMem()</b>
///   if you want to and if you have the size of the memory,
///   or you can make a normal copy of it (if you have its size). <br/>
///  Remember, if the passed pointer points to memory that is somehow
///   locked (see <b>IsDataLocked()</b> or <b>IsReallocLocked()</b>),
///   then attaching the pointer to <b>this</b> memory object will fail.
///</summary>
reterr_t MemoryShared::AttachPointer( LPCVOID ptr )
{
    if( !ptr ) return syserr (ERROR_INVALID_PARAMETER);

    MemDat_Extended *ext;
    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr_Safe( ext, (const BYTE*) ptr, true, true );
    // Note: The #ext# we pass only because we want the #GetHeaderPtr_Safe()#
    // force to check, if the data range of the extended data is accessible, too.

    if( !hdr )
        return interr (errMemoryCouldNotAttachPointer);

    reterr_t ret = Free();
    if( !ret )
        { MemReallocator::DecRefCnt_Safe( hdr ); return ret; }

    //mem_t memtype;
    //tryHW { memtype = (mem_t) hdr->m_Basic.m_MemType; }
    //catchHW { return interr (errCritical); }
    //m_Settings.m_memtCreate = memtype;

    //  Note: The sequence above is remarked out because m_memtCreate we
    //  do not want to change. The code is still there to remind us
    //  explicitely.
    //  But there is another fact not to forget if the code would
    //  be uncommented again: SEH (structured exception handling, so
    //  to catch asynchronous (hardware) exceptions).
    //  Since tryHW/catchHW can map to __try/__except, local object
    //  unwinding is not allowed in this function. But we do use
    //  reterr_t. So ahyway, we would have to export the tryHW/catchHW
    //  code into a separate function call.

    m_pMemory = (PBYTE) ptr;
    m_ForeignMem.m_bForeignMem = 0;

    return true;
}

///<summary>
///  This method will try to read the header before the specified
///   pointer and if it can retrieve it, it checks, if it is a
///   shareable memory (used by the current process).
///</summary>
///<param name="pBasic"> retrieves basic memory information.
///  Can be NULL if basic info is not desired. </param>
bool MemoryShared::IsAttachablePointer( LPCVOID ptr, MemDat_Basic *pBasic )
{
    if( !ptr ) return false;

    MemDat_Extended *ext;
    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr_Safe( ext, (const BYTE*) ptr, false, true );
    // Note: The #ext# we pass only because we want the #GetHeaderPtr_Safe()#
    // force to check, if the data range of the extended data is accessible, too.

    if( pBasic && hdr ) {
        if( !MemReallocator::GetBasic_Safe( *pBasic, hdr ) )
            hdr = NULL;
    }

    return hdr != NULL;
}



reterr_t MemoryShared::SetData( LPCVOID ptr, _size_t_ count )
{
    if( !ptr || count < 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    reterr_t ret = SetSize( count );
    if( !ret ) return ret;

    if( count == 0 ) return true;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    memcpy( pdest, ptr, count );

    return true;
}

reterr_t MemoryShared::Add( _size_t_ count )
{
    if( count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    _size_t_ oldsize = GetSize();
    _size_t_ newsize = oldsize + count;

    return SetSize( newsize );
}

reterr_t MemoryShared::Add( LPCVOID ptr, _size_t_ count )
{
    if( !ptr || count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ oldsize = GetSize();
    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    memcpy( pdest.getValue() + oldsize, ptr, count );

    return true;
}

///<summary>
///  Adds new bytes at the end of the current memory.
///  The new bytes will be cleared.
///</summary>
///<returns>
///  Returns a pointer to the first byte of the new memory.
///  It will be a write-to-pointer. <br/>
///  Or NULL if allocation failed.
///</returns>
_retPByte_t_ MemoryShared::Add_Ptr( _size_t_ count )
{
    if( count <= 0 ) return _retPByte_t_(syserr (ERROR_INVALID_PARAMETER), NULL);

    // Even if only data is locked we return an error
    // because we would have to return a write-to-pointer
    // which is not allowed when data is locked.
    if( IsDataLocked() )
        return _retPByte_t_(interr (errMemoryDataLocked), NULL);

    _size_t_ oldsize = GetSize();
    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return _retPByte_t_(ret, NULL);

    _retPByte_t_ ptr = WritePtr();
    if( !ptr ) return ptr;

    return _retPByte_t_(true, ptr.getValue() + oldsize);
}

///<remarks> This method supports also <b>mem</b> being
///  <b>this</b> <b>MemoryShared</b> object. </remarks>
reterr_t MemoryShared::Append( const MemoryShared& mem )
{
    if( mem.IsEmpty() ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ srcsize = mem.GetSize();

    _size_t_ oldsize = GetSize();
    _size_t_ newsize = oldsize + srcsize;

    // The first operation (*before* retrieving the
    // source data pointer) is to resize the memory
    // so that we can support #mem# being #this#.
    // But the source data size we must have queried
    // already before.

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    // Now (after retrieving the write-to-pointer)
    // we can query the source pointer.
    const BYTE *psrc = mem.ConstPtr();
    if( !psrc ) return interr (errFailed); // just for savety (should never occur)

    memcpy( pdest.getValue() + oldsize, psrc, srcsize );

    return true;
}

reterr_t MemoryShared::Insert( _idx_ idx, _size_t_ count, LPCVOID ptr )
{
    if( !ptr ) return syserr (ERROR_INVALID_PARAMETER);

    _size_t_ oldsize = GetSize();

    if( idx < 0 || (_size_t_) idx > oldsize ) return interr (errOutOfRange);
    if( count == 0 ) return true;
    if( count <= 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return ret;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    // Now move away some parts of the content data
    // to make room for the inserted bytes.
    if( (_size_t_) idx < oldsize )
    {
        _size_t_ cmove = oldsize - idx;
        memmove( pdest.getValue() + idx + count, pdest.getValue() + idx, cmove );
    }

    memcpy( pdest.getValue() + idx, ptr, count );

    return true;
}

///<summary>
///  Inserts new bytes somewhere inside the current memory.
///  The new bytes will be cleared only if the settings are
///  set that way that new memory shall be cleared automatically.
///</summary>
///<returns>
///  Returns a pointer to the first byte of the newly inserted memory range.
///  It will be a write-to-pointer. <br/>
///  Or NULL if allocation failed or on wrong params.
///</returns>
///<remarks>
///  If <b>idx</b> specifies a location completely outside
///   the current memory, NULL will be returned. <br/>
///  If <b>idx</b> points exactly at the end of the current
///   memory (so it has the same value as returned by <b>GetSize()</b>),
///   the behavior of this method is the same as <b>Add()</b>. <br/>
///</remarks>
_retPByte_t_ MemoryShared::Insert_Ptr( _idx_ idx, _size_t_ count )
{
    _size_t_ oldsize = GetSize();

    if( idx < 0 || (_size_t_) idx > oldsize ) return _retPByte_t_(interr (errOutOfRange), NULL);
    if( count == 0 ) return _retPByte_t_(true, NULL);
    if( count <= 0 ) return _retPByte_t_(syserr (ERROR_INVALID_PARAMETER), NULL);

    // Even if only data is locked we return an error
    // because we would have to return a write-to-pointer
    // which is not allowed when data is locked.
    if( IsDataLocked() )
        return _retPByte_t_(interr (errMemoryDataLocked), NULL);

    _size_t_ newsize = oldsize + count;

    reterr_t ret = SetSize( newsize );
    if( !ret ) return _retPByte_t_(ret, NULL);

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    // Now move away some parts of the content data
    // to make room for the inserted bytes.
    if( (_size_t_) idx < oldsize )
    {
        _size_t_ cmove = oldsize - idx;
        memmove( pdest.getValue() + idx + count, pdest.getValue() + idx, cmove );
        // And clear the inserted bytes where previously old data was.
        if( m_Settings.m_bClearNewMem )
            memset( pdest.getValue() + idx, 0, tmin( cmove, count ) );
    }

    return _retPByte_t_(true, pdest.getValue() + idx);
}

reterr_t MemoryShared::Remove( _idx_ idx, _size_t_ count )
{
    _size_t_ oldsize = GetSize();

    if( !CheckRangeLimits( idx, count, oldsize ) )
        return interr (errOutOfRange);

    // Removing any memory would mean also remove content data
    // which is not allowed when data is locked.
    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( count <  0 ) return interr (errOutOfRange);
    if( count == 0 ) return true;

    if( IsReallocLocked() ) return interr (errMemoryReallocLocked);

    _size_t_ newsize = oldsize - count;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    // Now move together the left and right side
    // of the memory area.
    if( idx + count < oldsize ) {
        _size_t_ cmove = oldsize - (idx + count);
        memmove( pdest.getValue() + idx, pdest.getValue() + idx + count, cmove );
    }

    return SetSize( newsize );
}

reterr_t MemoryShared::Clear()
{
    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ size = GetSize();

    if( size <= 0 )
        return true;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    memset( pdest, 0, size );

    return true;
}

reterr_t MemoryShared::Clear( _idx_ idx, _size_t_ count )
{
    _size_t_ size = GetSize();

    if( !CheckRangeLimits( idx, count, size ) )
        return interr (errOutOfRange);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( size <= 0 )
        return true;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    memset( pdest, 0, size );

    return true;
}

///<summary>
///  Normally you should not care about the padding bytes.
///  But if you want to, you can use this method to clear them.
///</summary>
///<remarks>
///  What is the padding: <br/>
///   The padding bytes are allocated but unused bytes. They
///   come from specifying a block size (e.g. using <b>SetBlockSize()</b>).
///   <br/>
///   Using blocks larger than 1 byte may cause the allocated
///   memory to be larger than the requested number of bytes (e.g.
///   when calling <b>SetSize()</b>). Although the valid size the
///   user should use still is the size he requested, there may
///   have been allocated some additional, unused padding bytes. <br/>
///   Normally the user should not use them in any case.
///   <br/>
///   But when he is decreasing the memory buffer size (e.g. using
///   <b>SetSize()</b> with a smaller size value), the padding
///   bytes may not be cleared and could possibly contain old
///   data from previous execution steps. <br/>
///   If the user does not want this, he has the opportunity to
///   clear them using <b>ClearPadding()</b>.
///   <br/><br/>
///  If content data is locked (<b>IsDataLocked()</b>), then
///   this does not include the padding bytes. So they may be
///   cleared at any point of time.
///</remarks>
reterr_t MemoryShared::ClearPadding()
{
    _size_t_ size       = GetSize();
    _size_t_ callocated = GetAllocatedSize();

    if( callocated <= size ) return true;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    memset( pdest.getValue() + size, 0, callocated - size );

    return true;
}

///<summary> Clears the complete buffer including the padding bytes </summary>
reterr_t MemoryShared::ClearAllocatedBuffer()
{
    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _size_t_ callocated = GetAllocatedSize();

    if( callocated <= 0 ) return true;

    _retPByte_t_ pdest = WritePtr();
    if( !pdest ) return pdest;

    memset( pdest, 0, callocated );

    return true;
}



mem_t MemoryShared::GetMemType() const
{
    if( m_ForeignMem.m_bForeignMem )
        return MemT_External;

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );

    return hdr ? (mem_t) hdr->m_Basic.m_MemType : MemT_None;
}



HANDLE MemoryShared::GetFileMappingHandle() const
{
    if( m_ForeignMem.m_bForeignMem )
        return NULL;

    MemDat_Extended *ext;
    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( ext, m_pMemory );

    if( !hdr || !ext )
        return NULL;

    return hdr->m_Basic.m_MemType == MemT_FileMapping
            ? ((MemDat_FileMapping*) ext)->m_hFileMapping
            : NULL;
}

///<remarks>
///  <b>MemoryShared</b> objects allocate memory including some header
///  information required for the sharing feature. <br/>
///  Therefore, an object allocates memory including the header information
///  and returns a pointer to read content data directly following the
///  header information. <br/>
///  You can use this method to get the complete size of the header (rounded
///  up to multiples of MEMDAT_TOTAL_HEADER_BLOCKING bytes).
///</remarks>
_size_t_ MemoryShared::GetInternalContentOffset() const
{
    if( m_ForeignMem.m_bForeignMem )
        return 0;

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );

    return MemReallocator::GetTotalHeaderSize( hdr );
}



///<summary>
///  Copies only the properties (memtype, blocksize etc.)
///  from another <b>MemoryShared</b> object.
///</summary>
///<remarks>
///  This method will work regardless, if there is
///   memory allocated already in <b>this</b> object or not.
///</remarks>
reterr_t MemoryShared::CopyPropertiesFrom( const MemoryShared& mem )
{
    if( this == &mem ) return true;

    //if( !IsEmpty() )
    //    return interr (errMemoryMustBeEmpty);

    m_Settings.CopyPropertiesFrom( mem.m_Settings );

    return true;
}

///<summary>
///  Copies the properties and the memory from
///  another <b>MemorySpace</b> object.
///</summary>
///<remarks>
///  The memory name (if any) will *not* be copied.
///   Any old name will be deleted. <br/>
///  IMPORTANT: If the source memory cannot be shared (e.g.
///   because reference counter is disabled so that sharing
///   is explicitely prohibited as it is for named
///   filemappingobject memories), a copy is made and in that
///   case, interr #suxMemoryCopyProduced# will be returned
///   rather than #retOk#.
///</remarks>
reterr_t MemoryShared::ShareFrom( const MemoryShared& mem )
{
    if( this == &mem ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    reterr_t ret = CopyPropertiesFrom( mem );
    if( !ret ) return ret;

    if( mem.IsEmpty() ) return true;

    _size_t_    count = mem.GetSize();
    const BYTE *psrc  = mem.ConstPtr();

    if( mem.GetMemType() == MemT_External )
        return SetExternalMem( psrc, count );

    MemDat_ShareHdr *srchdr = MemReallocator::GetHeaderPtr( psrc );

    if( !srchdr )
        return interr (errFailed); // just for savety. should never occur

    // If reference counter is disabled, we make a copy of
    //  rather than share the source memory.
    // Also if somehow the source memory is locked, we must
    //  make a copy because locked memory cannot be shared.
    if( !srchdr->m_RefCnt.m_bEnabled ||
        srchdr->m_Basic.IsAnythingLocked() )
    {
        if( !( ret = IsolateToSizeAndType( count, (mem_t) srchdr->m_Basic.m_MemType ) ) )
            return ret;

        _retPByte_t_ pdest = WritePtr();
        if( !pdest ) return pdest;

        memcpy( pdest, psrc, count );

        return interr (suxMemoryCopyProduced);
    }

    if( !( ret = Free() ) )
        return ret;

    srchdr->m_RefCnt.Inc();

    //m_Settings.m_memtCreate = (mem_t)srchdr->m_Basic.m_MemType;

    m_pMemory = (PBYTE) psrc;
    m_ForeignMem.m_bForeignMem = 0;

    return true;
}

///<summary>
///  Exchanges everything between 2 <b>MemorySpace</b> objects
///  including the memory, all properties, memory-names etc. <br/>
///  But <u>except</u> the SingletonSettings. They are the settings
///  related only to the single memory object here.
///</summary>
reterr_t MemoryShared::Exchange( MemoryShared& mem )
{
    if( this == &mem ) return true;

    MemDat_Settings   tmp_settings;
    PBYTE             tmp_ptr;
    MemDat_ForeignMem tmp_foreignmem;

    memcpy( &tmp_settings,   &m_Settings,   sizeof(MemDat_Settings  ) );
    tmp_ptr =                 m_pMemory;
    memcpy( &tmp_foreignmem, &m_ForeignMem, sizeof(MemDat_ForeignMem) );

    memcpy( &m_Settings,   &mem.m_Settings,   sizeof(MemDat_Settings  ) );
    m_pMemory =             mem.m_pMemory;
    memcpy( &m_ForeignMem, &mem.m_ForeignMem, sizeof(MemDat_ForeignMem) );

    memcpy( &mem.m_Settings,   &tmp_settings,   sizeof(MemDat_Settings  ) );
    mem.m_pMemory =             tmp_ptr;
    memcpy( &mem.m_ForeignMem, &tmp_foreignmem, sizeof(MemDat_ForeignMem) );

    return true;
}



///<remarks> CAUTION: Locking will cause the memory to isolated
///  and therefore could cause the memory to be reallocated.
///  So previous pointers to the memory content could be invalid. </remarks>
reterr_t MemoryShared::LockData()
{
    if( m_ForeignMem.m_bForeignMem ) return interr (suxNoOp);

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );
    if( !hdr ) return interr (errFailed);

    // Shared memories cannot be locked. So isolate first.
    if( hdr->m_RefCnt.IsShared() )
    {
        reterr_t ret = Isolate();
        if( !ret ) return ret;

        hdr = MemReallocator::GetHeaderPtr( m_pMemory );
        if( !hdr ) return interr (errFailed);
    }

    hdr->m_Basic.m_wDataLockCount++;

    return true;
}

reterr_t MemoryShared::UnlockData()
{
    if( m_ForeignMem.m_bForeignMem ) return interr (suxNoOp);

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );
    if( !hdr ) return interr (errFailed);

    hdr->m_Basic.m_wDataLockCount--;

    return true;
}

///<remarks> CAUTION: Locking will cause the memory to isolated
///  and therefore could cause the memory to be reallocated.
///  So previous pointers to the memory content could be invalid. </remarks>
reterr_t MemoryShared::LockRealloc()
{
    if( m_ForeignMem.m_bForeignMem ) return interr (suxNoOp);

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );
    if( !hdr ) return interr (errFailed);

    // Shared memories cannot be locked. So isolate first.
    if( hdr->m_RefCnt.IsShared() )
    {
        reterr_t ret = Isolate();
        if( !ret ) return ret;

        hdr = MemReallocator::GetHeaderPtr( m_pMemory );
        if( !hdr ) return interr (errFailed);
    }

    hdr->m_Basic.m_wReallocLockCount++;

    return true;
}

reterr_t MemoryShared::UnlockRealloc()
{
    if( m_ForeignMem.m_bForeignMem ) return interr (suxNoOp);

    MemDat_ShareHdr *hdr = MemReallocator::GetHeaderPtr( m_pMemory );
    if( !hdr ) return interr (errFailed);

    hdr->m_Basic.m_wReallocLockCount--;

    return true;
}




// Checks, if the complete data area can be handled by byte addressing,
// or otherwise it can be handled only by bit addressing.
// Note: In special cases, the data is byteable although its data
// consists of bits (e.g. when the mem-unit is 8 bits).
bool DynMem::IsByteable() const
{
    if( !m_cMemUnit.IsByteable() ) return false;
    // Now that we know that memory-units are byteable by themselves,
    // let's check the alignment and only if the alignment also is
    // byteable, then the whole data memory is byteable.
    // If alignment is disabled, or if it is only 1 bit, then the
    // data memory *is* byteable (if the other values are as well).
    if( m_cAlign.cbits() > 1 && !m_cAlign.IsByteable() ) return false;
    // Now the header must be byteable as well.
    if( !m_cHeader.IsByteable() ) return false;
    return true;
}



BitCPtr DynMem::ConstHeaderPtr() const
{
    if( m_cHeader.cbits() <= 0 ) return NULL;

    LPCVOID ptr = m_Mem.ConstPtr();
    if( !ptr ) return NULL;

    return ptr;
}

_retBitPtr_t_ DynMem::WriteHeaderPtr()
{
    if( m_cHeader.cbits() <= 0 ) return _retBitPtr_t_(interr (errDynMemNoHeader), NULL);

    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return _retBitPtr_t_(ptr.getError(), NULL);

    return _retBitPtr_t_(true, ptr.getValue());
}

BitCPtr DynMem::ConstPtr() const
{
    LPCVOID ptr = m_Mem.ConstPtr();
    if( !ptr ) return NULL;

    return BitCPtr(ptr) + m_cHeader;
}

_retBitPtr_t_ DynMem::WritePtr()
{
    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return _retBitPtr_t_(ptr.getError(), NULL);

    return _retBitPtr_t_(true, BitPtr(ptr) + m_cHeader);
}

BitCPtr DynMem::ConstPtrAt( _idx_ ix ) const
{
    dim_t apos[] = { (dim_t) ix };
    return ConstPtrAt( DimPt( apos, _countof(apos) ) );
}

_retBitPtr_t_ DynMem::WritePtrAt( _idx_ ix )
{
    dim_t apos[] = { (dim_t) ix };
    return WritePtrAt( DimPt( apos, _countof(apos) ) );
}

BitCPtr DynMem::ConstPtrAt( _idx_ ix1, _idx_ ix2 ) const
{
    dim_t apos[] = { (dim_t) ix1, (dim_t) ix2 };
    return ConstPtrAt( DimPt( apos, _countof(apos) ) );
}

_retBitPtr_t_ DynMem::WritePtrAt( _idx_ ix1, _idx_ ix2 )
{
    dim_t apos[] = { (dim_t) ix1, (dim_t) ix2 };
    return WritePtrAt( DimPt( apos, _countof(apos) ) );
}

//  > bitpointer of the item.
//    If error (e.g. position outside the array extents) ->
//     NULL-pointer ('BitPtr::IsNull()' will return TRUE).
// Note: If 'pos' has a smaller dimension than 'oDim', then the method
//  will internally increase 'pos's dimension by added new dim-extents and
//  settings them to a value of zero each.
// BUT: 'pos' *must not* have a greater dimenstion than 'oDim' or an error will occur.
BitCPtr DynMem::ConstPtrAt( const DimPt& pos ) const
{
    LPCVOID ptr = m_Mem.ConstPtr();
    if( !ptr ) return NULL;

    BitLen ofs = EvalOffset( pos );
    if( ofs.IsError() ) return NULL;

    return BitCPtr(ptr) + ofs;
}

//  > bitpointer of the item.
//    If error (e.g. position outside the array extents) ->
//     NULL-pointer ('BitPtr::IsNull()' will return TRUE).
// Note: If 'pos' has a smaller dimension than 'oDim', then the method
//  will internally increase 'pos's dimension by added new dim-extents and
//  settings them to a value of zero each.
// BUT: 'pos' *must not* have a greater dimenstion than 'oDim' or an error will occur.
_retBitPtr_t_ DynMem::WritePtrAt( const DimPt& pos )
{
    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return _retBitPtr_t_(ptr.getError(), NULL);

    BitLen ofs = EvalOffset( pos );
    if( ofs.IsError() ) return _retBitPtr_t_(interr (errOutOfRange), NULL);

    return _retBitPtr_t_(true, BitPtr(ptr) + ofs);
}

BitCPtr DynMem::ConstPtrOfArea( _idx_ ix, _size_t_ ci ) const
{
    dim_t apos [] = { (dim_t) ix };
    dim_t asize[] = { (dim_t) ci };
    return ConstPtrOfArea( DimPt  ( apos,  _countof(apos ) ),
                           DimSize( asize, _countof(asize) ) );
}

_retBitPtr_t_ DynMem::WritePtrOfArea( _idx_ ix, _size_t_ ci )
{
    dim_t apos [] = { (dim_t) ix };
    dim_t asize[] = { (dim_t) ci };
    return WritePtrOfArea( DimPt  ( apos,  _countof(apos ) ),
                           DimSize( asize, _countof(asize) ) );
}

BitCPtr DynMem::ConstPtrOfArea( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 ) const
{
    dim_t apos [] = { (dim_t) ix1, (dim_t) ix2 };
    dim_t asize[] = { (dim_t) ci1, (dim_t) ci2 };
    return ConstPtrOfArea( DimPt  ( apos,  _countof(apos ) ),
                           DimSize( asize, _countof(asize) ) );
}

_retBitPtr_t_ DynMem::WritePtrOfArea( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 )
{
    dim_t apos [] = { (dim_t) ix1, (dim_t) ix2 };
    dim_t asize[] = { (dim_t) ci1, (dim_t) ci2 };
    return WritePtrOfArea( DimPt  ( apos,  _countof(apos ) ),
                           DimSize( asize, _countof(asize) ) );
}

// Note: Different to 'ConstPtrOf()', in here the 'pos' and 'size' must have the
//  same dimension as 'oDim' or an error will occur (NULL-pointer returned).
BitCPtr DynMem::ConstPtrOfArea( const DimPt& pos, const DimSize& size ) const
{
    if( pos.GetSize() != m_oDim.GetSize() || size.GetSize() != m_oDim.GetSize() )
        return NULL;

    _size_t_ count = m_oDim.GetSize();

    for( _idx_ i = 0; i < (_idx_) count ; ++i )
        if( pos[i] < 0 || pos[i] + size[i] > m_oDim[i] )
            return NULL;

    return ConstPtrAt( pos );
}

// Note: Different to 'WritePtrOf()', in here the 'pos' and 'size' must have the
//  same dimension as 'oDim' or an error will occur (NULL-pointer returned).
_retBitPtr_t_ DynMem::WritePtrOfArea( const DimPt& pos, const DimSize& size )
{
    if( pos.GetSize() != m_oDim.GetSize() || size.GetSize() != m_oDim.GetSize() )
        return _retBitPtr_t_(interr (errDynMemWrongDimensions), NULL);

    _size_t_ count = m_oDim.GetSize();

    for( _idx_ i = 0; i < (_idx_) count ; ++i )
        if( pos[i] < 0 || pos[i] + size[i] > m_oDim[i] )
            return _retBitPtr_t_(interr (errOutOfRange), NULL);

    return WritePtrAt( pos );
}



///<summary> Gets the (real, internal) offset, where your
///  header information (if present) really starts beginning
///  from the start of the real allocated memory. </summary>
///<remarks> Normally, the <b>WriteHeaderPtr()</b> and <b>ConstHeaderPtr()</b>
///  return the correct pointer to your header data. <br/>
///  But if you want to use e.g. a file mapping object handle
///  returned from <b>GetFileMappingHandle()</b>, you will
///  want to know the correct internal offset (because <b>DynMem</b>
///  has an internal offset because it is using the
///  <b>MemoryShared</b> memory class). <br/>
///  This method here will return the correct offset in bytes. </remarks>
_size_t_ DynMem::GetInternalHeaderOffsetInBytes() const
{
    return m_Mem.GetInternalContentOffset();
}

///<summary> Gets the (real, internal) offset, where your array
///  content data information (if present) really starts beginning
///  from the start of the real allocated memory. </summary>
///<remarks> See remarks to <b>GetInternalHeaderOffsetInBytes()</b>. <br/>
///  The difference is that in here, you will not get the offset to
///  your header info but to your array data. </remarks>
BitLen DynMem::GetInternalContentDataOffset() const
{
    BitLen ofs( 8 * GetInternalHeaderOffsetInBytes() );

    ofs += m_cHeader;

    return ofs;
}



// Sets the size of one memory-unit in the dynamic array
//  : memunitsize = size of one memory-unit in bit/bytes
// NOTE: This method will work only when 'this' object currently contains
//  *no* array-memory (so it must be deallocated and have size 0) or
//  an error will occur. But if only header memory is allocated, this
//  is possible.
reterr_t DynMem::SetMemUnitSize( BitLen memunitsize )
{
    if( !m_oDim.IsEmpty() || memunitsize.IsEmpty() )
        return interr (errMemoryMustBeEmpty);

    m_cMemUnit = memunitsize;

    return true;
}



// Sets the alignment value in bit/bytes for each row
// (= just for the first dim-extent, for following
//  dim-extents you cannot specify additional alignments).
// The alignment effects the first dimension-extent. All
// following dim-extents (whose sizes (in bit/bytes) are multiples
// of the first dim-extent) do not have additional alignments.
// Alignment is useful e.g. when the DynMem is used to hold a
// Bitmap's pixels. Bitmaps need to align each pixel-row to a
// 32-bit boundary regardless of the real width of the bitmap.
// This can be done by setting an alignment of 32 bits (or 4 bytes).
//  : alignment = boundary-size for the alignment (in bit/bytes).
//                If empty (0 bits of size) -> alignment is switched off
//                (which is the same as an alignment of 1 bit of size).
// NOTE: This method will work only when 'this' object currently contains
//  *no* array-memory (so it must be deallocated and have size 0) or
//  an error will occur. But if only header memory is allocated, this
//  is possible.
reterr_t DynMem::SetAlignment( BitLen alignment )
{
    if( !m_oDim.IsEmpty() )
        return interr (errMemoryMustBeEmpty);

    m_cAlign = alignment;

    return true;
}



reterr_t DynMem::SetSize( _size_t_ size )
{
    dim_t asize[1] = { (dim_t) size };
    return SetSize( DimSize( asize, _countof(asize) ) );
}

reterr_t DynMem::SetSize( _size_t_ size1, _size_t_ size2)
{
    dim_t asize[2] = { (dim_t) size1, (dim_t) size2 };
    return SetSize( DimSize( asize, _countof(asize) ) );
}

reterr_t DynMem::SetSize( const DimSize& size, BitLen headersize, mem_t memtype )
{
    reterr_t ret = ReAlloc_Array( size, headersize, memtype );
    if( !ret ) return ret;

    m_cHeaderArea = headersize;

    return true;
}



// Note: This method works also when memory is allocated already.
reterr_t DynMem::SetHeaderSize( BitLen headersize )
{
    reterr_t ret = ReAlloc_Array( m_oDim, headersize, MemT_Dont_Change_Dynamic );
    if( !ret ) return ret;

    m_cHeaderArea = headersize;

    return true;
}



// Adds new elements to the array. If the desired dimension (1)
// is greater than the current array's dimension (of 'm_oDim'), then
// the dimension will be made as large as the desired one.
reterr_t DynMem::Add( _size_t_ add )
{
    if( m_oDim.GetSize() > 1 ) return interr (errDynMemWrongDimensions);

    DimSize dim = m_oDim;
    if( dim.GetSize() < 1 ) dim.SetSize0( 1 );
    dim[0] += (dim_t) add;

    return SetSize( dim );
}

// Adds new elements to the array. If the desired dimension (2)
// is greater than the current array's dimension (of 'm_oDim'), then
// the dimension will be made as large as the desired one.
reterr_t DynMem::Add( _size_t_ add1, _size_t_ add2 )
{
    if( m_oDim.GetSize() > 2 ) return interr (errDynMemWrongDimensions);

    DimSize dim = m_oDim;
    if( dim.GetSize() < 2 ) dim.SetSize0( 2 );
    dim[0] += (dim_t) add1;
    dim[1] += (dim_t) add2;

    return SetSize( dim );
}

// Adds new elements to the array. If the desired dimension (of 'add')
// is greater than the current array's dimension (of 'oDim'), then
// the dimension will be made as large as the desired one.
reterr_t DynMem::Add( const DimSize& add )
{
    if( m_oDim.GetSize() > add.GetSize() ) return interr (errDynMemWrongDimensions);

    DimSize dim = m_oDim;
    if( dim.GetSize() < add.GetSize() ) dim.SetSize0( add.GetSize() );

    for( _idx_ i = 0  ; i < (_idx_) dim.GetSize() ; ++i ) {
        if( add[i] < 0 ) return syserr (ERROR_INVALID_PARAMETER);
        dim[i] += add[i];
    }

    return SetSize( dim );
}

// Adds new elements to the array. If the desired dimension (1)
// is greater than the current array's dimension (of 'm_oDim'), then
// the dimension will be made as large as the desired one.
//  > NULL-ptr = failure.
_retBitPtr_t_ DynMem::Add_Ptr( _size_t_ add )
{
    if( m_oDim.GetSize() > 1 ) return _retBitPtr_t_(interr (errDynMemWrongDimensions), BitPtr());

    dim_t ixnew = m_oDim[0];
    DimSize dim = m_oDim;
    if( dim.GetSize() < 1 ) dim.SetSize0( 1 );
    dim[0] += (dim_t) add;

    reterr_t ret = SetSize( dim );
    if( !ret ) return _retBitPtr_t_(ret, NULL);

    _retBitPtr_t_ ptr = WritePtr();
    if( !ptr ) return ptr;

    return _retBitPtr_t_(true, ptr.getValue() + ixnew * m_cMemUnit.cbits());
}

//  > on failure = position and/or size are outside the valid memory array size or mem alloc failure.
reterr_t DynMem::Insert( _idx_ ix, _size_t_ ci )
{
    dim_t apos [] = { (dim_t) ix };
    dim_t asize[] = { (dim_t) ci };

    return Insert( DimPt  ( apos,  _countof(apos ) ),
                   DimSize( asize, _countof(asize) ) );
}

//  > on failure = position and/or size are outside the valid memory array size or mem alloc failure.
reterr_t DynMem::Insert( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 )
{
    dim_t apos [] = { (dim_t) ix1, (dim_t) ix2 };
    dim_t asize[] = { (dim_t) ci1, (dim_t) ci2 };

    return Insert( DimPt  ( apos,  _countof(apos ) ),
                   DimSize( asize, _countof(asize) ) );
}

// Note: Some (but not all) dim-extents of 'size' may be zero.
//  > on failure = position and/or size are outside the valid memory array size or mem alloc failure.
// Note: 'pos' may point directly *after* the end of one extent which will cause
//  new elements to be added at the end of the extent.
//  BUT: In such a case, 'pos' must point *directly* after the end of the extent
//  and not any further (not even one unit).
reterr_t DynMem::Insert( const DimPt& pos, const DimSize& size )
{
    if( pos.GetSize() != size.GetSize() ) return interr (errDynMemWrongDimensions);
    if( m_oDim.GetSize() > 0 &&
        m_oDim.GetSize() != pos.GetSize() ) return interr (errDynMemWrongDimensions);
    if( size.GetSize() == 0 && m_oDim.GetSize() == 0 ) return true;
    if( size.GetSize() == 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    _idx_ i;
    for( i = pos.GetSize() - 1 ; i >= 0 ; --i ) {
        if( pos[i] < 0 || pos[i] > m_oDim[i] || size[i] < 0 )
            return syserr (ERROR_INVALID_PARAMETER);
    }

    // To do the insertion, the following steps must be performed:
    // First the memory array must be resized to the desired size
    // (the additional 'size' will be added to the current size 'm_oDim').
    // This will cause the new memory space to be at the end of each
    // dimension-extent.
    // Now memory-areas must be moved so that the new memory space is
    // sitting at the desired position 'pos' for each dim-extent
    // (and when moved away, the new space must be cleared of course).

    // So first add the new space.

    reterr_t ret = Add( size );
    if( !ret ) return ret;

    if( m_oDim.IsEmpty() )
        return interr (errFailed);  // just for savety. should never occur

    // Now it is time to do the movements.
    // First of all, the number of movements depends on the dimension
    // of 'this' memory array.
    // For each dim-extent, two movements are necessary. A
    // movement may be skipped if the source and destination positions
    // are equal.

    _size_t_   count = m_oDim.GetSize();
    DimPt      msrce( count );
    DimPt      mdest( count );
    DimSize    msize( count );
    BitCounter counter( count );
    --counter;
    // The movement must begin at the highest address
    // (because the insertion only makes dim-extents larger),
    // so begin with last memory-area to be moved.
    do
    {
        for( i = 0 ; i < (_idx_) count ; ++i )
            if( counter.GetBit(i) )
                { msize[i] = m_oDim[i] - ( mdest[i] = ( msrce[i] = pos[i] ) + size[i] ); }
            else
                { mdest[i] = msrce[i] = 0;  msize[i] = pos[i]; }

        if( !( ret = Move( mdest, msrce, msize ) ) )
            return ret;

    } while( !(--counter).IsOverflowed() );

    // At last, clear the new memory space areas.
    // There are as many clears to do as the dimension of 'this'
    // memory array. For each dim-extent: The whole memory array
    // size must be cleared, except for the corresponding dim-extent
    // which should be set to the desired range (of the new space).

    mdest.Clear();
    msize = m_oDim;
    for( i = 0 ; i < (_idx_) count ; ++i )
    {
        mdest[i] = pos[i];
        msize[i] = size[i];
        if( i > 0 ) {
            mdest[i-1] = 0;
            msize[i-1] = m_oDim[i-1];
        }

        if( !( ret = Clear( mdest, msize ) ) )
            return ret;
    }

    return true;
}

//  > NULL-ptr = position and/or size are outside the valid memory array size or mem alloc failure.
_retBitPtr_t_ DynMem::Insert_Ptr( _idx_ ix, _size_t_ ci )
{
    dim_t apos [] = { (dim_t) ix };
    dim_t asize[] = { (dim_t) ci };

    reterr_t ret = Insert( DimPt  ( apos,  _countof(apos ) ),
                           DimSize( asize, _countof(asize) ) );
    if( !ret ) return _retBitPtr_t_(ret, NULL);

    _retBitPtr_t_ ptr = WritePtr();
    if( !ptr ) return ptr;

    return _retBitPtr_t_(true, ptr.getValue() + ix * m_cMemUnit.cbits());
}

//  > on failure = position and/or size are outside the valid memory array size or mem alloc failure.
reterr_t DynMem::Remove( _idx_ ix, _size_t_ ci )
{
    dim_t apos [] = { (dim_t) ix };
    dim_t asize[] = { (dim_t) ci };

    return Remove( DimPt  ( apos,  _countof(apos ) ),
                   DimSize( asize, _countof(asize) ) );
}

//  > on failure = position and/or size are outside the valid memory array size or mem alloc failure.
reterr_t DynMem::Remove( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2 )
{
    dim_t apos [] = { (dim_t) ix1, (dim_t) ix2 };
    dim_t asize[] = { (dim_t) ci1, (dim_t) ci2 };

    return Remove( DimPt  ( apos,  _countof(apos ) ),
                   DimSize( asize, _countof(asize) ) );
}

// Note: Some (but not all) dim-extents of 'size' may be zero.
//  > on failure = position and/or size are outside the valid memory array size or mem alloc failure.
reterr_t DynMem::Remove( const DimPt& pos, const DimSize& size )
{
    if( m_oDim.GetSize() == 0 ) return true;
    if( pos.GetSize() != size.GetSize() ) return interr (errDynMemWrongDimensions);
    if( m_oDim.GetSize() != pos.GetSize() ) return interr (errDynMemWrongDimensions);
    if( size.GetSize() == 0 ) return syserr (ERROR_INVALID_PARAMETER);

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    if( IsReallocLocked() ) return interr (errMemoryReallocLocked);

    _idx_ i;
    for( i = pos.GetSize() - 1 ; i >= 0 ; --i ) {
        if( pos[i] < 0 || pos[i] > m_oDim[i] - size[i] || size[i] < 0 )
            return syserr (ERROR_INVALID_PARAMETER);
    }

    reterr_t ret;

    // To do the removement, the following steps must be performed:
    // First memory-areas must be moved so that the memory space to remove
    // is not present any more at the desired position 'pos' for each dim-extent.
    // Then the memory array must be resized to the desired size.

    // Now it is time to do the movements.
    // First of all, the number of movements depends on the dimension
    // of 'this' memory array.
    // For each dim-extent, two movements are necessary. A
    // movement may be skipped if the source and destination positions
    // are equal.

    _size_t_   count = m_oDim.GetSize();
    DimPt      msrce( count );
    DimPt      mdest( count );
    DimSize    msize( count );
    BitCounter counter( count );
    // The movement must begin at the lowest address
    // (because the removement only makes dim-extents smaller),
    // so begin with first memory-area to be moved.
    do
    {
        for( i = 0 ; i < (_idx_) count ; ++i )
            if( counter.GetBit(i) )
                { msize[i] = m_oDim[i] - ( msrce[i] = ( mdest[i] = pos[i] ) + size[i] ); }
            else
                { mdest[i] = msrce[i] = 0;  msize[i] = pos[i]; }

        if( !( ret = Move( mdest, msrce, msize ) ) )
            return ret;

    } while( !(++counter).IsOverflowed() );

    // At last, resize the memory array.

    DimSize dim = m_oDim;
    if( dim.GetSize() < size.GetSize() )
        dim.SetSize0( size.GetSize() );
    for( i = dim.GetSize() - 1 ; i >= 0 ; --i )
        dim[i] -= size[i];

    return SetSize( dim );
}



reterr_t DynMem::Move( _idx_ idest, _idx_ isrce, _size_t_ size, bool bClip )
{
    dim_t adest[] = { (dim_t) idest };
    dim_t asrce[] = { (dim_t) isrce };
    dim_t asize[] = { (dim_t) size  };

    return Move( DimPt  ( adest, _countof(adest) ),
                 DimPt  ( asrce, _countof(asrce) ),
                 DimSize( asize, _countof(asize) ),
                 bClip );
}

reterr_t DynMem::Move( _idx_ idest1, _idx_ idest2, _idx_ isrce1, _idx_ isrce2, _size_t_ size1, _size_t_ size2, bool bClip )
{
    dim_t adest[] = { (dim_t) idest1, (dim_t) idest2 };
    dim_t asrce[] = { (dim_t) isrce1, (dim_t) isrce2 };
    dim_t asize[] = { (dim_t) size1,  (dim_t) size2  };

    return Move( DimPt  ( adest, _countof(adest) ),
                 DimPt  ( asrce, _countof(asrce) ),
                 DimSize( asize, _countof(asize) ),
                 bClip );
}

//  : size = size of the area to move.
//           This may be an empty area (or even negative) which will cause no
//           movement to be made, but also without returning an error.
//  : bClip = if TRUE -> the source- and destination-areas may lie outside the
//             memory array area. In that case, they will be clipped to fit inside
//             the memory area. If the source or destination is completely outside
//             the memory area, no movement will be done without producing an error.
//            if FALSE -> if the source- or destination-area is partially or completely
//             outside the memory array area, no movement will be done and an error
//             will be returned.
reterr_t DynMem::Move( DimPt dest, DimPt srce, DimSize size, bool bClip )
{
    if( dest.GetSize() != m_oDim.GetSize() ||
        srce.GetSize() != m_oDim.GetSize() ||
        size.GetSize() != m_oDim.GetSize() ) return interr (errDynMemWrongDimensions);
    if( m_oDim.GetSize() == 0 ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    // First check the borders and clip them if allowed.
    bool     dest_is_srce = true;
    dim_t   *pdest = dest.Ptr();
    dim_t   *psrce = srce.Ptr();
    dim_t   *psize = size.Ptr();
    dim_t   *pdim  = m_oDim.Ptr();
    _size_t_ count = m_oDim.GetSize();
    _idx_ i;

    for( i = 0 ; i < (_idx_) count ; ++i )
    {
        if( *pdest < 0 ) {
            if( !bClip ) return interr (errOutOfRange);
            *psize -= *pdest;
            *psrce -= *pdest;
            *pdest -= *pdest;
        }
        if( *pdest + *psize > *pdim ) {
            if( !bClip ) return interr (errOutOfRange);
            *psize = *pdim - *pdest;
        }
        if( *psrce < 0 ) {
            if( !bClip ) return interr (errOutOfRange);
            *psize -= *psrce;
            *pdest -= *psrce;
            *psrce -= *psrce;
        }
        if( *psrce + *psize > *pdim ) {
            if( !bClip ) return interr (errOutOfRange);
            *psize = *pdim - *psrce;
        }
        if( *psize <= 0 ) return true;
        if( *pdest != *psrce ) dest_is_srce = false;
        ++pdest;
        ++psrce;
        ++psize;
        ++pdim;
    }
    if( dest_is_srce )
        return true;
    // Now 'size' is *not* empty, and
    // 'dest' is *not* equal to 'srce'.

    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return ptr;

    // Now do the movement.
    // For this, move each row by row. But first the
    // direction of movement must be determined.
    DimPt step( count );
    if( count >= 2 ) step[1] = 1;
    else             step[0] = size[0];
    BitPtr bp_srce = BitPtr(ptr) + EvalOffset( srce );
    BitPtr bp_dest = BitPtr(ptr) + EvalOffset( dest );
    INT64  destofs = bp_dest - bp_srce;
    if( bp_dest < bp_srce )
    {
        // Destination is a lower address, so move the rows
        // starting with the first row.
        DimPt   isrce( srce );
        BitLen cblocktomove = m_cMemUnit.cbits() * size[0]; // not including alignment of course
        BitLen rowsize      = EvalArraySize( DimSize(m_oDim.Ptr(), 1) );
        bool    dim_overflow = false; // need not be TRUE, because 'srce==isrce' and so 'bp_srce' is equal
        do
        {
            if( dim_overflow )
                bp_srce = BitPtr(ptr) + EvalOffset( isrce );
            bp_dest = bp_srce + destofs;
            _memmovebits_( bp_dest, bp_srce, cblocktomove );
            bp_srce += rowsize;
        } while( isrce.Increment( step, srce, size, &dim_overflow ) );
    }
    else if( bp_dest > bp_srce )
    {
        // Destination is a greater address, so move the rows
        // starting with the last row.
        DimPt isrce( srce );
        for( _idx_ i = 1 ; i < (_idx_) count ; ++i )
            isrce[i] += size[i] - 1; // IMORTANT: 'i' must start at 1 (!)
        BitLen cblocktomove  = m_cMemUnit.cbits() * size[0];
        BitLen rowsize       = EvalArraySize( DimSize(m_oDim.Ptr(), 1) );
        bool   dim_underflow = true; // must be TRUE to cause 'bp_srce' to be set from 'isrce'
        do
        {
            if( dim_underflow )
                bp_srce = BitPtr(ptr) + EvalOffset( isrce );
            bp_dest = bp_srce + destofs;
            _memmovebits_( bp_dest, bp_srce, cblocktomove );
            bp_srce -= rowsize;
        } while( isrce.Decrement( step, srce, size, &dim_underflow ) );
    }

    return true;
}



reterr_t DynMem::ClearHeader()
{
    if( m_Mem.IsEmpty() ) return true;

    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return ptr;

    _memsetbits_( BitPtr(ptr.getValue()), 0, m_cHeader );

    return true;
}

reterr_t DynMem::Clear()
{
    if( m_Mem.IsEmpty() || m_oDim.IsEmpty() ) return true;

    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return ptr;

    _memsetbits_( BitPtr(ptr.getValue()) + EvalStartOfArray(), 0, m_cArraySize );

    return true;
}

reterr_t DynMem::Clear( _idx_ ix, _size_t_ ci, bool bClip )
{
    dim_t apos [] = { (dim_t) ix };
    dim_t asize[] = { (dim_t) ci };

    return Clear( DimPt  ( apos,  _countof(apos ) ),
                  DimSize( asize, _countof(asize) ),
                  bClip );
}

reterr_t DynMem::Clear( _idx_ ix1, _idx_ ix2, _size_t_ ci1, _size_t_ ci2, bool bClip )
{
    dim_t apos [] = { (dim_t) ix1, (dim_t) ix2 };
    dim_t asize[] = { (dim_t) ci1, (dim_t) ci2 };

    return Clear( DimPt  ( apos,  _countof(apos ) ),
                  DimSize( asize, _countof(asize) ),
                  bClip );
}

//  : size = size of the area to clear.
//           This may be an empty area (or even negative) which will cause no
//           clearing to be made, but also without returning an error.
//  : bClip = if TRUE -> the defined area may lie outside the
//             memory array area. In that case, it will be clipped to fit inside
//             the memory area. If the defined area is completely outside
//             the memory area, no clearing will be done without producing an error.
//            if FALSE -> if the defined area is partially or completely
//             outside the memory array area, no clearing will be done and an error
//             will be returned.
reterr_t DynMem::Clear( DimPt pos, DimSize size, bool bClip )
{
    if( pos.GetSize() != m_oDim.GetSize() || size.GetSize() != m_oDim.GetSize() )
        return interr (errDynMemWrongDimensions);
    if( m_oDim.GetSize() == 0 ) return true;

    if( IsDataLocked() )
        return interr (errMemoryDataLocked);

    // First check the borders and clip them if allowed.
    dim_t   *ppos  = pos.Ptr();
    dim_t   *psize = size.Ptr();
    dim_t   *pdim  = m_oDim.Ptr();
    _size_t_ count = m_oDim.GetSize();
    _idx_ i;
    for( i = 0 ; i < (_idx_) count ; ++i )
    {
        if( *ppos < 0 ) {
            if( !bClip ) return interr (errOutOfRange);
            *psize -= *ppos;
            *ppos  -= *ppos;
        }
        if( *ppos + *psize > *pdim ) {
            if( !bClip ) return interr (errOutOfRange);
            *psize = *pdim - *ppos;
        }
        if( *psize <= 0 ) return true;
        ++psize;
        ++ppos;
        ++pdim;
    }
    // Now 'size' is *not* empty.

    _retPByte_t_ ptr = m_Mem.WritePtr();
    if( !ptr ) return ptr;

    // Now do the clearing.
    // For this, clear each row by row.
    DimPt ipos( pos );
    DimPt step( count );
    if( count >= 2 ) step[1] = 1;
    else             step[0] = size[0];
    BitPtr bp_pos        = BitPtr(ptr.getValue()) + EvalOffset( ipos );
    BitLen cblocktoclear = m_cMemUnit.cbits() * size[0]; // not including alignment of course
    BitLen rowsize       = EvalArraySize( DimSize(m_oDim.Ptr(), 1) );
    bool   dim_overflow  = false;
    do
    {
        if( dim_overflow )
            bp_pos = BitPtr(ptr) + EvalOffset( ipos );
        _memsetbits_( bp_pos, 0, cblocktoclear );
        bp_pos += rowsize;
    } while( ipos.Increment( step, pos, size, &dim_overflow ) );

    return true;
}



///<summary>
///  Copies only the properties (memunit-size, mem-type, blocksize etc.)
///  from another <b>DynMem</b> object.
///</summary>
///<remarks>
///  This method will work regardless, if there is
///   memory allocated already in <b>this</b> object or not.
///</remarks>
reterr_t DynMem::CopyPropertiesFrom( const DynMem& dm )
{
    reterr_t ret = m_Mem.CopyPropertiesFrom( dm.m_Mem );
    if( !ret ) return ret;

    m_cMemUnit          = dm.m_cMemUnit;
    m_cAlign            = dm.m_cAlign;
    m_cHeaderArea       = dm.m_cHeaderArea;

    return true;
}

reterr_t DynMem::ShareFrom( const DynMem& dm )
{
    reterr_t ret = Free();
    if( !ret ) return ret;

    if( !( ret = m_Mem.ShareFrom( dm.m_Mem ) ) )
        return ret;

    m_cMemUnit          = dm.m_cMemUnit;
    m_cAlign            = dm.m_cAlign;
    m_cHeaderArea       = dm.m_cHeaderArea;

    m_cHeader           = dm.m_cHeader;
    m_oDim              = dm.m_oDim;
    m_cArraySize        = dm.m_cArraySize;

    return true;
}

// Exchanges everything of 'this' and a second
// 'DynMem' object (including the memory).
reterr_t DynMem::Exchange( DynMem& dm )
{
    reterr_t ret = m_Mem.Exchange( dm.m_Mem );
    if( !ret ) return ret;

    BitLen cMemUnit_t      = dm.m_cMemUnit;
    BitLen cAlign_t        = dm.m_cAlign;
    BitLen cHeaderArea_t   = dm.m_cHeaderArea;
    BitLen cHeader_t       = dm.m_cHeader;
    DimSize oDim_t          = dm.m_oDim;
    BitLen cArraySize_t    = dm.m_cArraySize;

    dm.m_cMemUnit           = m_cMemUnit;
    dm.m_cAlign             = m_cAlign;
    dm.m_cHeaderArea        = m_cHeaderArea;
    dm.m_cHeader            = m_cHeader;
    dm.m_oDim               = m_oDim;
    dm.m_cArraySize         = m_cArraySize;

    m_cMemUnit              = cMemUnit_t;
    m_cAlign                = cAlign_t;
    m_cHeaderArea           = cHeaderArea_t;
    m_cHeader               = cHeader_t;
    m_oDim                  = oDim_t;
    m_cArraySize            = cArraySize_t;

    return true;
}



void DynMem::Init()
{
    m_cMemUnit.SetCBytes( 1 );
    m_cAlign.SetCBytes( 0 );
    m_cHeaderArea.SetCBytes( 0 );

    m_cHeader.SetCBytes( 0 );
    m_oDim.Empty();
    m_cArraySize.SetCBits( 0 );

    m_Mem.EnableClearNewMem( TRUE );
    // #DynMem# objects always will clear new memory because
    // it is too uncomfy to handle otherwise even for the user.
    // #ReAlloc_Array()# will clear bits if necessary for a
    // move operation, but not in every case. So we must enable
    // the flag here in the internal memory object itself.
}

// Evaluates the size in bits/bytes of the specified dimension only
// (not including the header or any memory-block-size).
BitLen DynMem::EvalArraySize( const DimSize& oDim ) const
{
    if( oDim.GetSize() == 0 ) return 0;
    INT64 cbits = m_cMemUnit.cbits() * oDim[0];
    if( cbits <= 0 ) return 0;
    if( m_cAlign.cbits() > 1 )
        cbits = ( (cbits + (INT64) m_cAlign.cbits() - 1) / (INT64) m_cAlign.cbits() ) * (INT64) m_cAlign.cbits();
    _size_t_ count = oDim.GetSize();
    for( _idx_ i = 1 ; i < (_idx_) count ; ++i ) {
        dim_t cext = oDim[i];
        if( cext <= 0 ) return 0;
        cbits *= cext;
    }
    return cbits;
}

// Evaluates the size (in bit/bytes) of a memory for the whole
// array, but *without* a memory-block-size.
// So it will return the minimum necessary memory
// size to hold the whole array (but including alignment).
BitLen DynMem::EvalNettoMemSize( const DimSize& oDim, BitLen cHeader ) const
{
    return cHeader + EvalArraySize( oDim );
}

BitLen DynMem::EvalNettoMemSize( BitLen cArraySize, BitLen cHeader ) const
{
    return cHeader + cArraySize;
}

// Evaluates the offset (in bit/bytes) from the beginning
// of an array with the specified dimension 'oDim' (not
// including any other memory like a header etc.).
// The offset is described by position 'oPos' pointing
// into the array.
//  : bAllowOutsideArray = if TRUE -> 'oPos' may also point to a position *ouside*
//                         the 'oDim' limits without producing an error.
//                         If course, the returned offset value is not a really valid
//                         offset, but if 'oPos' is empty except for one dim-extent which
//                         is set to exactly the value of the corresponding dim-extent-value
//                         of 'oDim', then the position is outside, but the returned offset
//                         can be interpreted as a size value (: used by 'ReAlloc_Array()').
//  > the 'BitLen' value will return an error (through
//    'BitLen::IsError()') if the 'oPos' points
//    outside the dimension 'oDim'.
// Note: If 'oPos' has a smaller dimension than 'oDim', then the method
//  will internally increase 'oPos's dimension by added new dim-extents and
//  settings them to a value of zero each.
// BUT: 'oPos' *must not* have a greater dimenstion than 'oDim' or an error will occur.
BitLen DynMem::EvalArrayOffset( const DimPt& oPos, const DimSize& oDim, bool bAllowOutsideArray ) const
{
    //if( oPos.GetSize() != oDim.GetSize() ) return -1;
    if( oPos.GetSize() > oDim.GetSize() ) return -1;
    if( oDim.IsEmpty() ) return -1;
    INT64 bit_offset = 0;
    INT64 dim_bitsize = m_cMemUnit.cbits();
    _size_t_ count = oPos.GetSize();
    for( _idx_ i = 0 ; i < (_idx_) count ; ++i ) {
        if( !bAllowOutsideArray && (oPos[i] < 0 || oPos[i] >= oDim[i]) ) return -1;
        bit_offset += dim_bitsize * oPos[i];
        dim_bitsize *= oDim[i];
        if( i == 0 && m_cAlign.cbits() > 1 )
            dim_bitsize = ( (dim_bitsize + (INT64) m_cAlign.cbits() - 1) / (INT64) m_cAlign.cbits() ) * (INT64) m_cAlign.cbits();
    }
    return bit_offset;
}

// Evaluates the offset like 'EvalArrayOffset()' but from the
// beginning of the whole memory (so it is an offset
// including the header information).
//  : bAllowOutsideArray = if TRUE -> 'oPos' may also point to a position *ouside*
//                         the 'oDim' limits without producing an error.
//  > the 'BitLen' value will return an error (through
//    'BitLen::IsError()') if the 'oPos' points
//    outside the dimension 'oDim'.
BitLen DynMem::EvalOffset( const DimPt& oPos, const DimSize& oDim, BitLen cHeader, bool bAllowOutsideArray ) const
{
    return cHeader + EvalArrayOffset( oPos, oDim, bAllowOutsideArray );
}



// Reallocates whole data memory and sets 'm_oDim' accordingly.
// So the sizes of 'oDim_Desired' and 'm_oDim' may differ as
// well as each dim-extent of them.
// IMPORTANT: This method does *not* support changes in
//  'm_cMemUnit' and 'm_cAlign'.
// IMPORTANT:
//  When the memory reallocation failed, the old memory data
//  will or will not have been changed already.
reterr_t DynMem::ReAlloc_Array( const DimSize& oDim_Desired,
                                BitLen cHeader_Desired,
                                mem_t memtype_desired )
{
    // This implementation expects #m_Mem# to have the flag
    // for automatic new memory clearing enabled.

    if( m_Mem.IsReallocLocked() )
        return interr (errMemoryReallocLocked);

    //BOOL bBitField = !( IsByteable() && header_size.IsByteable() );

    // The '*_new' (and '*_Desired') values shall be the final values for the memory.
    // The '*_old' values are the values of the memory *before* anything has been done in here.
    // The '*_temp' values are the values which are used for the call to 'ReAllocMem()'.
    _size_t_ cMem_new = (_size_t_) EvalNettoMemSize( oDim_Desired, cHeader_Desired ).EvalByteSize();
    _size_t_ cMem_old = m_Mem.GetSize();
    if( cMem_new == 0 ) {
        return Free_Array( memtype_desired );
    }
    UINT64 cHeader_old  = m_cHeader.cbits();
    UINT64 cHeader_temp = cHeader_Desired.cbits() < cHeader_old ? cHeader_Desired.cbits() : cHeader_old;
    DimSize oDim_old = m_oDim;
    oDim_old.SetSize( tmax(m_oDim.GetSize(), oDim_Desired.GetSize()) );
    if( m_oDim.IsEmpty() ) oDim_old[0] = 0; // if 'm_oDim' was empty and size zero -> 'oDim_old' will be filled with ones, so set the first one to zero by hand
    DimSize oDim_temp = DimSize::EvalMinimum( oDim_old, oDim_Desired );
    // Note: 'oDim_old' and 'oDim_temp' must have the same size.

    // First search all dim-extents which shall be made *smaller* and for
    // those move the memory areas to their destinated positions and sizes.
    // In addition, if the new header size shall be smaller than the old one,
    // then move it as well.
    // All this ensures for the following processing, that only mem-movements
    // are done from higher addresses to lower ones.
    if( cMem_old > 0 &&
        !oDim_old.IsEmpty() &&
        !oDim_temp.IsEmpty() )
    {
        DimPt position( oDim_old.GetSize() );
        DimPt step( oDim_old.GetSize() );
        _idx_ ilowest_dimext_changed = -1; // which is the dim-extent with the lowest index which is to be changed
        _size_t_ cdimextents = oDim_old.GetSize();
        for( _idx_ idimext = 0 ; idimext < (_idx_) cdimextents - 1 ; ++idimext )
            if( oDim_temp[idimext] < oDim_old[idimext] ) {
                step[idimext + 1] = 1;
                ilowest_dimext_changed = idimext;
                break;
            }

        _retPByte_t_ ptr = m_Mem.WritePtr();
        if( !ptr ) return ptr;

        // Now: if 'ilowest_dimext_changed' <  0 -> no dim-extents are smaller, so move the whole array if temp header size is different.
        //      if 'ilowest_dimext_changed' == 0 -> when moving memory, do not forget to clear the blank bits to the alignment-boundary.
        //      if 'ilowest_dimext_changed' >  0 -> simply move the memory areas (those are automatically aligned to the alignment-boundary)
        if( ilowest_dimext_changed >= 0 )
        {
            // Now move all necessary memory block areas
            // beginning at the lowest addresses.
            BitLen cblocktomove   = EvalArrayOffset( step, oDim_temp, true ); // even is no error if the step is larger/equal the dim-size
            BitLen csrceblocksize = EvalArrayOffset( step, oDim_old,  true ); // even is no error if the step is larger/equal the dim-size
            BitPtr bp_srce = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_old,  cHeader_old  );
            BitPtr bp_dest = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_temp, cHeader_temp );
            bool dim_overflow = false;
            // And now calculate the value necessary to do the clearing of the bits
            // which are unused but necessary for the alignment. This is only necessary
            // for the first dim-extent.
            BitLen unusedofs, cunused;
            if( ilowest_dimext_changed == 0 )
                cunused = cblocktomove - ( unusedofs = m_cMemUnit.cbits() * oDim_temp[0] );
            do
            {
                if( dim_overflow )
                    bp_srce = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_old, cHeader_old );
                _memmovebits_( bp_dest, bp_srce, cblocktomove );
                if( cunused.cbits() > 0 ) _memsetbits_( bp_dest + unusedofs, 0, cunused ); // here do the clearing of the unused bits for the alignment
                bp_srce += csrceblocksize;
                bp_dest += cblocktomove;
            } while( position.Increment( step, oDim_temp, &dim_overflow ) );
        }
        else if( cHeader_temp < cHeader_old )
        {
            position.Clear();
            BitLen cblocktomove = EvalArraySize( oDim_temp ); // the block to move is the whole array
            BitPtr bp_srce = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_old,  cHeader_old  );
            BitPtr bp_dest = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_temp, cHeader_temp );
            _memmovebits_( bp_dest, bp_srce, cblocktomove );
        }
        // Now clear the bytes after the end of the (possibly smaller) temp-array.
        BitLen carray_temp = EvalNettoMemSize( oDim_temp, cHeader_temp );
        BitLen carray_old  = EvalNettoMemSize( oDim_old,  cHeader_old  );
        if( carray_temp.cbits() < carray_old.cbits() )
            _memsetbits_( BitPtr(ptr.getValue()) + carray_temp, 0, carray_old - carray_temp );
    }

    reterr_t ret = m_Mem.SetSizeAndType( cMem_new, memtype_desired );
    if( !ret ) return ret;

    oDim_temp.SetSize( oDim_Desired.GetSize() );
    // Note: 'oDim_temp' must have the same size like 'oDim_Desired'.
    // For this, it is imagined, that (if 'oDim_temp' is truncated by this
    // operation) the truncated dim-extent-values all were 1. This the
    // 'DimSize::EvalMinimum()' call somewhere above will have done automatically.

    // Now search all dim-extents which shall be made *larger* and for
    // those move the memory areas to their destinated positions and sizes.
    // In addition, if the new header size shall be larger than the old
    // then move it as well.
    // All this ensures for the following processing, that only mem-movements
    // are done from lower addresses to higher ones.
    bool is_dim_desired_empty = oDim_Desired.IsEmpty();
    if( cMem_new > 0 &&
        !oDim_temp.IsEmpty() &&
        !is_dim_desired_empty )
    {
        DimPt position;
        position.SetToEndOfDim( oDim_temp ); // start at the end of the (temp) array
        DimPt step( oDim_temp.GetSize() );
        _idx_ ilowest_dimext_changed = -1; // which is the dim-extent with the lowest index which is to be changed
        _size_t_ cdimextents = oDim_temp.GetSize();
        for( _idx_ idimext = 0 ; idimext < (_idx_) cdimextents - 1 ; ++idimext ) {
            position[idimext] = 0;
            if( oDim_temp[idimext] < oDim_Desired[idimext] ) {
                step[idimext + 1] = 1;
                ilowest_dimext_changed = idimext;
                break;
            }
        }

        _retPByte_t_ ptr = m_Mem.WritePtr();
        if( !ptr ) return ptr;

        // Now: if 'ilowest_dimext_changed' <  0 -> no dim-extents are larger, so move the whole array if temp header size is different.
        //      if 'ilowest_dimext_changed' == 0 -> when moving memory, do not forget to clear the blank bits to the alignment-boundary.
        //      if 'ilowest_dimext_changed' >  0 -> simply move the memory areas (those are automatically aligned to the alignment-boundary)
        if( ilowest_dimext_changed >= 0 )
        {
            // Now move all necessary memory block areas
            // beginning at the highest addresses.
            BitLen cblocktomove   = EvalArrayOffset( step, oDim_temp,    true ); // even is no error if the step is larger/equal the dim-size
            BitLen cdestblocksize = EvalArrayOffset( step, oDim_Desired, true ); // even is no error if the step is larger/equal the dim-size
            BitPtr bp_srce = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_temp,    cHeader_temp    );
            BitPtr bp_dest = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_Desired, cHeader_Desired );
            bool dim_underflow = false;
            // Note: Clearing of the unused bits for the alignment need not be done in here,
            // because it is only senseful when the dimension is made smaller and this has
            // been done *before* to call to 'ReAllocMem()'.
            // But in here, for *every* dimension-extent which is made larger, the newly
            // added bits must be cleared.
            // Note: If there are new bits after the last element of the whole array, then those
            // need not be cleared, because this will have been done already before the call
            // to 'ReAllocMem()' (and in addition, 'ReAllocMem()' will clear new memory, too).
            do
            {
                if( dim_underflow ) {
                    BitPtr bp_clearend = bp_dest + cdestblocksize;
                    bp_dest = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_Desired, cHeader_Desired );
                    BitPtr bp_clearstart = bp_dest + cblocktomove;
                    INT64 cclearbits = bp_clearend - bp_clearstart;
                    if( cclearbits > 0 ) _memsetbits_( bp_clearstart, 0, cclearbits );
                }
                _memmovebits_( bp_dest, bp_srce, cblocktomove );
                // Now clear the new bits
                if( cdestblocksize.cbits() > cblocktomove.cbits() )
                    _memsetbits_( bp_dest + cblocktomove, 0, cdestblocksize.cbits() - cblocktomove.cbits() );
                // And clear the bits from where the movement has been made, but
                // only those source-bits which do *not* overlap with the dest-bits.
                INT64 cclearbits = bp_dest - bp_srce;
                if( cclearbits > (INT64) cblocktomove.cbits() ) cclearbits = cblocktomove.cbits();
                if( cclearbits > 0 ) _memsetbits_( bp_srce, 0, cclearbits );
                bp_srce -= cblocktomove;
                bp_dest -= cdestblocksize;
            } while( position.Decrement( step, oDim_temp, &dim_underflow ) );
        }
        else if( cHeader_temp < cHeader_Desired.cbits() )
        {
            position.Clear();
            BitLen cblocktomove = EvalArraySize( oDim_temp ); // the block to move is the whole array
            BitPtr bp_srce = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_temp,    cHeader_temp    );
            BitPtr bp_dest = BitPtr(ptr.getValue()) + EvalOffset( position, oDim_Desired, cHeader_Desired );
            _memmovebits_( bp_dest, bp_srce, cblocktomove );
        }
        // Now also clear new bits added to the header (if done).
        if( cHeader_temp < cHeader_Desired.cbits() )
            _memsetbits_( BitPtr(ptr.getValue()) + cHeader_temp, 0, cHeader_Desired.cbits() - cHeader_temp );
    }

    if( is_dim_desired_empty ) m_oDim.Empty();
    else                       m_oDim = oDim_Desired;
    m_cHeader = cHeader_Desired;

    m_cArraySize = EvalArraySize();

    return true;
}

reterr_t DynMem::Free_Array( mem_t memtype )
{
    reterr_t ret = m_Mem.Free();
    if( !ret ) return ret;

    m_cHeader.Empty();
    m_oDim.Empty();
    m_cArraySize.Empty();

    return m_Mem.SetMemType( memtype );
}

reterr_t DynMem::ForceFree_Array( mem_t memtype )
{
    reterr_t ret = m_Mem.ForceFree();
    if( !ret ) return ret;

    if( !( ret = m_Mem.SetMemType( memtype ) ) )
        return ret;

    m_cHeader.Empty();
    m_oDim.Empty();
    m_cArraySize.Empty();

    return true;
}




MemLockData::MemLockData( MemorySpace *pMem, bool bInitialLock )
{
    m_pMemory = (PVOID) pMem;
    m_objectT = MemLockObjMemorySpace;
    m_bLocked = false;
    if( bInitialLock ) Lock();
}

MemLockData::MemLockData( MemoryShared *pMem, bool bInitialLock )
{
    m_pMemory = (PVOID) pMem;
    m_objectT = MemLockObjMemoryShared;
    m_bLocked = false;
    if( bInitialLock ) Lock();
}

MemLockData::MemLockData( DynMem *pMem, bool bInitialLock )
{
    m_pMemory = (PVOID) pMem;
    m_objectT = MemLockObjDynMem;
    m_bLocked = false;
    if( bInitialLock ) Lock();
}

bool MemLockData::Lock()
{
    if( m_bLocked || !m_pMemory )
        return false;

    tryHW { if( !_Lock() ) return false; }
    catchHW { m_bLocked = false; m_pMemory = NULL; return false; }

    return m_bLocked;
}

bool MemLockData::Unlock()
{
    if( !m_bLocked || !m_pMemory )
        return false;

    tryHW { if( !_Unlock() ) return false; }
    catchHW { m_bLocked = false; m_pMemory = NULL; return false; }

    return true;
}

bool MemLockData::_Lock()
{
    if( m_bLocked || !m_pMemory )
        return false;

    switch( m_objectT ) {
        case MemLockObjMemorySpace  : m_bLocked = (bool) ((MemorySpace *) m_pMemory)->LockData(); break;
        case MemLockObjMemoryShared : m_bLocked = (bool) ((MemoryShared*) m_pMemory)->LockData(); break;
        case MemLockObjDynMem       : m_bLocked = (bool) ((DynMem      *) m_pMemory)->LockData(); break;
        default : m_bLocked = false; return false;
    }

    return true;
}

bool MemLockData::_Unlock()
{
    if( !m_bLocked || !m_pMemory )
        return false;

    switch( m_objectT ) {
        case MemLockObjMemorySpace  : ((MemorySpace *) m_pMemory)->UnlockData(); break;
        case MemLockObjMemoryShared : ((MemoryShared*) m_pMemory)->UnlockData(); break;
        case MemLockObjDynMem       : ((DynMem      *) m_pMemory)->UnlockData(); break;
        default : m_bLocked = false; return false;
    }

    m_bLocked = false;

    return true;
}




MemLockRealloc::MemLockRealloc( MemorySpace *pMem, bool bInitialLock )
{
    m_pMemory = (PVOID) pMem;
    m_objectT = MemLockObjMemorySpace;
    m_bLocked = false;
    if( bInitialLock ) Lock();
}

MemLockRealloc::MemLockRealloc( MemoryShared *pMem, bool bInitialLock )
{
    m_pMemory = (PVOID) pMem;
    m_objectT = MemLockObjMemoryShared;
    m_bLocked = false;
    if( bInitialLock ) Lock();
}

MemLockRealloc::MemLockRealloc( DynMem *pMem, bool bInitialLock )
{
    m_pMemory = (PVOID) pMem;
    m_objectT = MemLockObjDynMem;
    m_bLocked = false;
    if( bInitialLock ) Lock();
}

bool MemLockRealloc::Lock()
{
    if( m_bLocked || !m_pMemory )
        return false;

    tryHW { if( !_Lock() ) return false; }
    catchHW { m_bLocked = false; m_pMemory = NULL; return false; }

    return m_bLocked;
}

bool MemLockRealloc::Unlock()
{
    if( !m_bLocked || !m_pMemory )
        return false;

    tryHW { if( !_Unlock() ) return false; }
    catchHW { m_bLocked = false; m_pMemory = NULL; return false; }

    return true;
}

bool MemLockRealloc::_Lock()
{
    if( m_bLocked || !m_pMemory )
        return false;

    switch( m_objectT ) {
        case MemLockObjMemorySpace  : m_bLocked = (bool) ((MemorySpace *) m_pMemory)->LockRealloc(); break;
        case MemLockObjMemoryShared : m_bLocked = (bool) ((MemoryShared*) m_pMemory)->LockRealloc(); break;
        case MemLockObjDynMem       : m_bLocked = (bool) ((DynMem      *) m_pMemory)->LockRealloc(); break;
        default : m_bLocked = false; return false;
    }

    return true;
}

bool MemLockRealloc::_Unlock()
{
    if( !m_bLocked || !m_pMemory )
        return false;

    switch( m_objectT ) {
        case MemLockObjMemorySpace  : ((MemorySpace *) m_pMemory)->UnlockRealloc(); break;
        case MemLockObjMemoryShared : ((MemoryShared*) m_pMemory)->UnlockRealloc(); break;
        case MemLockObjDynMem       : ((DynMem      *) m_pMemory)->UnlockRealloc(); break;
        default : m_bLocked = false; return false;
    }

    m_bLocked = false;

    return true;
}




}
