/*
    File:               argonMoreMessageBox.h
    Description:        Provides a class with an enhanced messagebox which
                        includes additional fields and icons and a more button.

    Author:             Roman Komary
    created on:         15.02.2006
    last updated:       06.10.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary. All rights reserved.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    Switches:

    _argon_no_Lang...       These defines allow excluding of specific languages predefined
                            in the ArgonLibrary.
                            By default, all languages available in the ArgonLibrary will
                            be included. If you define one or some of this set of defines,
                            the appropriate language strings will not be included in
                            the compilation.
                            The following languages and their excluding defines are available:
                                _argon_no_LangEnglish
                                _argon_no_LangGerman
                                _argon_no_LangItalian
*/

#ifndef __argonMoreMessageBox_h
#define __argonMoreMessageBox_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

/*
    MoreMessageBox:
    =================

    This is a class that will be used by reterr_t's DisplayMessage().
    But can be used in any cases you like.
    It acts similar to normal system messageboxes, but with an addition
    and some restrictions:
    +   A "More" button is shown beside the Ok button and when pressed,
        an additional (longer or more descriptive) text appears that
        can be scrolled.
    -   The width of the dialog will not be adjusted according to
        the normal text to be displayed.
        Only the height will be changed and possible lines in the normal
        message text will automatically be broken if they exceed the
        textfield's width.
    -   Only the following MB_* flags are supported:
        *   All MB_TYPEMASK contained flags (MB_OK, MB_YESNO, etc.).
            They define, which buttons to display.
        *   All MB_ICON* flags.
            Note: You will also hear the sound defined for an MB_ICON* flag
            but only if a sound is defined and only if System-Beeps are enabled
            in your system.
        *   MB_DEFBUTTON1, 2, 3 and 4.
        *   MB_APPLMODAL, MB_SYSTEMMODAL, MB_SETFOREGROUND, MB_TOPMOST.

    Beside that, the class offers new static methods (which you can use in
    any cases you like, even independent from message boxes):
        FindNiceParent
        GetOwnedPopups

    You can use FindNiceParent() in any case you like where you would like to search
    for a nice parent window within the current thread.

    GetOwnedPopups() you can use to retrieve all popup windows (selectable by flags)
    that a specific parent window owns. This parent window must be a valid parent
    window for popup windows (so a top-level-window, but not a child window).
    E.g. the MoreMessageBox will use that to retrieve all enabled popup windows of
    the parent window for the MoreMessageBox so that it can disable them and reenable
    them again after the MoreMessageBox has closed.
*/

#ifndef _argon_no_includes
    #include <shellapi.h>       // for ShellExecute()
    #include "argonArray.h"
    #include "argonInstance.h"
    #include "argonString.h"
    #include "argonLanguageTable.h"
    #include "argonResources.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




#ifndef MB_CANCELTRYCONTINUE
#define MB_CANCELTRYCONTINUE        0x00000006L
#endif

// Language Table strings for the MoreMessageBox (Buttons etc.).

class MMBLanguageTable_Def {  };
typedef LanguageTableT<MMBLanguageTable_Def> MMBLanguageTable;

// language table string IDs for the MoreMessageBox (Buttons etc.).
enum MmbIDs
{
    mmbid_Ok = 1,
    mmbid_Cancel,
    mmbid_Yes,
    mmbid_No,
    mmbid_Abort,
    mmbid_Retry,
    mmbid_Ignore,
    mmbid_TryAgain,
    mmbid_Continue,
    mmbid_Help,
    mmbid_More__,
    mmbid_AlwaysShowThisMessage,
    mmbid_Information
};

// IDs of the MoreMessageBox controls and the dialog
#define mmbIDD_MOREMESSAGEBOX           132
#define mmbIDC_MSGICON                  1005
#define mmbIDC_MSG                      1006
#define mmbIDC_LONGMSG                  1007
#define mmbIDC_BUTTON1                  1008
#define mmbIDC_BUTTON2                  1009
#define mmbIDC_BUTTON3                  1010
#define mmbIDC_MORE                     1011
#define mmbIDC_CHECK1                   1012
#define mmbIDC_CHECK2                   1013
#define mmbIDC_BUTTON4                  1014
#define mmbIDC_HELPBTN                  1015
#define mmbIDC_HEADER                   1016
#define mmbIDC_EXTRAICON                1017
#define mmbIDC_EXTRAMSG                 1018
#define mmbIDC_EXTRALINK                1019
#define mmbIDC_HR                       1020

// Messages sent by controls that the MoreMessageBox has subclassed to the MoreMessageBox.
#define WM_MMB_BASE             (WM_APP + 30)
// WM_MMB_DRAWSTATIC is sent by a subclassed static control to its
// parent window to cause the parent window to draw the control.
//  : wParam = HDC to draw into
//  : lParam = HWND of the control
#define WM_MMB_DRAWSTATIC       (WM_MMB_BASE + 0)
// WM_MMB_CALCSTATIC is similar to WM_MMB_DRAWSTATIC except that the
// rect of the text shall be calculated rather than drawn.
//  : wParam = PRECT (pointer to RECT) Rect of the text to be calculated.
//  : lParam = HWND of the control
#define WM_MMB_CALCSTATIC       (WM_MMB_BASE + 1)
// WM_MMB_URLCLICKED is sent by an Url static control when it has been clicked.
//  : lParam = HWND of the control
#define WM_MMB_URLCLICKED       (WM_MMB_BASE + 2)
// WM_MMB_DRAWBUTTON is sent by a subclassed button control to its
// parent window to cause the parent window to draw the control.
//  : wParam = HDC to draw into
//  : lParam = HWND of the control
#define WM_MMB_DRAWBUTTON       (WM_MMB_BASE + 3)

///<summary>
///  Class <b>MoreMessageBox</b> is similar to the system's MessageBox function,
///  but it contains an additional "More" button which, when pressed,
///  will display a long message text additional to the normal message
///  text. <br/>
///  <pre>
///  It acts similar to normal system messagebox, but with additions
///  and some restrictions:
///  +   A "More" button is shown beside the Ok button and when pressed,
///      an additional (longer or more descriptive) text appears that
///      can be scrolled.
///  -   The width of the dialog will not be adjusted according to
///      the normal text to be displayed.
///      Only the height will be changed and possible lines in the normal
///      message text will automatically be broken if they exceed the
///      textfield's width.
///  -   Only the following MB_* flags are supported:
///      *   All MB_TYPEMASK contained flags (MB_OK, MB_YESNO, etc.).
///          They define, which buttons to display.
///      *   All MB_ICON* flags.
///          Note: You will also hear the sound defined for an MB_ICON* flag
///          but only if a sound is defined and only if System-Beeps are enabled
///          in your system.
///      *   MB_DEFBUTTON1, 2, 3 and 4.
///      *   MB_APPLMODAL, MB_SYSTEMMODAL, MB_SETFOREGROUND, MB_TOPMOST.
///  </pre>
///</summary>
class MoreMessageBox
{
    public :
        #ifndef _WIN32_WCE
            ///<summary> A number of buttons (1 to 4) that indicates the minimum width
            ///  of the message box in any cases (regardless of the really displayed
            ///  number of buttons). 0 = choose automatically. </summary>
        int     m_minWidthAsNumButtons;
            ///<summary> How much to inflate the 4 borders of the dialog towards the controls. </summary>
        RECT    m_rectInflateBorders;
        // Following are strings that will be set on the according buttons
        // and checkboxes for those strings which are not empty.
        // If a string is empty, the normal text for that button or
        // checkbox will be used.
        String  m_strOwnBtn1, m_strOwnBtn2, m_strOwnBtn3, m_strOwnBtn4;
        String  m_strOwnHelpBtn, m_strOwnMoreBtn;
        String  m_strOwnCheck1;         // if not set, the checkbox will usually display text "Always show this message"
        String  m_strOwnCheck2;         // if not set, usually displays "Check2".
            ///<summary> On input, defines if the Check1 checkbox shall be checked initially.
            ///  On output, defines if the user has checked or unchecked the checkbox.
            ///  By default, it is checked. </summary>
        bool    m_bIs_Check1_Checked;
            ///<summary> On input, defines if the Check2 checkbox shall be checked initially.
            ///  On output, defines if the user has checked or unchecked the checkbox.
            ///  By default, it is unchecked. </summary>
        bool    m_bIs_Check2_Checked;
            ///<summary> Set this to non-zero as message ID to be sent to the <b>MoreMessageBox</b>'s
            ///  parent window when the user clicked the Url link.
            ///  Its wParam will be 0 and the lParam will be the HWND of MoreMessageBox.
            ///  If the ID is not zero, then only the message will be sent. ShellExecute()
            ///  will then not be called by the MoreMessageBox. </summary>
        int     m_idMessage_UrlClicked;

        HICON   m_hiconUpArrow;         // if you want to set another icon, first delete the old one in here, then put a new one. Don't forget, it will be destroyed in the dtor (unless you set it to NULL before the dtor is called).
        HICON   m_hiconDownArrow;       // if you want to set another icon, first delete the old one in here, then put a new one. Don't forget, it will be destroyed in the dtor (unless you set it to NULL before the dtor is called).
        SIZE    m_sizeUpDownArrow;      // if you have set another icon, you can set its size here
        long    m_nOffsetUpDownArrow;   // up or down arrow icon's x-offset in pixels from the right button border when drawn. (vertically, the icon will be centered in the button)
        #endif

        MoreMessageBox();
        virtual ~MoreMessageBox();

        // Flags for param dwMmbFlags of the MoreMessageBox class:
        #define MMB_min1Button              0x00000001  // the MoreMessageBox shall contain at minimum 1 button
        #define MMB_min2Buttons             0x00000002  // the MoreMessageBox shall contain at minimum 2 buttons
        #define MMB_min3Buttons             0x00000003  // the MoreMessageBox shall contain at minimum 3 buttons
        #define MMB_min4Buttons             0x00000004  // the MoreMessageBox shall contain at minimum 4 buttons
        #define MMB_Mask_MinNumButtons      0x00000007  // the bitmask to mask out the minimum number of buttons the MoreMessageBox shall display
        #define MMB_NoMoreButton            0x00000008  // will have no More button. (by default it does exist but only if the #pstrLongMsg# param is *not* NULL or an empty string)
        #define MMB_1Checkbox               0x00000010  // will include 1 checkbox (the first one is the one usually displaying "Always show this message") (by default no checkboxes do exist)
        #define MMB_2Checkboxes             0x00000020  // will include 2 checkboxes
        #define MMB_Mask_NumCheckboxes      0x00000030  // the bitmask to mask out the number of checkboxes the MoreMessageBox shall display
        #define MMB_AppendExeNameInCaption  0x00000040  // if set, the MoreMessageBox caption will include the executable name of the application that opened the MessageBox.
        #define MMB_DontDisablePopups       0x00000080  // by default the MoreMessageBox will disable all enabled popup windows the parent window owns and will reenable them after the messagebox closed. This flag prevents this behaviour.
        #define MMB_TextLikeMessageBox      0x00000100  // If set, normal message text will be displayed similar to the system's messagebox (box-width will be adjusted to the text width).
                                                        // If not set, text will be broken as needed.
        #define MMB_AppWindow               0x00000200  // the MoreMessageBox will include style WS_EX_APPWINDOW
        #define MMB_ToolWindow              0x00000400  // the MoreMessageBox will include style WS_EX_TOOLWINDOW
        #define MMB_AlignButtonsLeft        0x00000800
        #define MMB_AlignButtonsCentered    0x00000000
        #define MMB_AlignButtonsRight       0x00001000
        #define MMB_Mask_ButtonAlignX       0x00001800
        #define MMB_AlignMsgIconTop         0x00000000
        #define MMB_AlignMsgIconVCentered   0x00002000
        #define MMB_AlignMsgIconBottom      0x00004000
        #define MMB_Mask_MsgIconAlignY      0x00006000
        #define MMB_AlignExtraIconTop       0x00008000
        #define MMB_AlignExtraIconVCentered 0x00000000
        #define MMB_AlignExtraIconBottom    0x00010000
        #define MMB_Mask_ExtraIconAlignY    0x00018000
        #define MMB_HeaderMsgNonBold        0x00020000
        #define MMB_MsgBold                 0x00040000
        #define MMB_ExtraMsgBold            0x00080000
        #define MMB_MsgIcon_ExtraOffsetY    0x00100000  // the MsgIcon will be additionally offset down by approximately half of its height.
        #define MMB_CenterOverParent        0x00200000  // If set, the message box will be centered over its parent window. By default, it will centered in the screen.
        #define MMB_AlignMsgTextTop         0x00400000  // takes effect only if the message text height is smaller than the MsgIcon and only if an MsgIcon exists.
        #define MMB_AlignMsgTextVCentered   0x00000000  // takes effect only if the message text height is smaller than the MsgIcon and only if an MsgIcon exists.
        #define MMB_AlignMsgTextBottom      0x00800000  // takes effect only if the message text height is smaller than the MsgIcon and only if an MsgIcon exists.
        #define MMB_Mask_MsgTextAlignY      0x00c00000
        #define MMB_PushForeground          0x01000000  // this acts similar to MB_SETFOREGROUND, but instead of calling only #SetForegroundWindow()#, it uses #MoreMessageBox::PushForegroundWindow()# to bring the window really to the foreground without flashing the taskbar.
        #define MMB_ExpandableOnly          0x02000000  // if this is set, then the More button will be disabled once the MoreMessageBox has been expanded. This flag is ignored if there is no More button.

        INT_PTR MessageBox( PCTSTR pstrMsg, PCTSTR pstrLongMsg = NULL, PCTSTR pstrCaption = NULL,
                            UINT uType = MB_OK, DWORD dwMmbFlags = 0 )
            { return MessageBox( NULL, pstrMsg, pstrLongMsg, pstrCaption, uType, dwMmbFlags ); }
        INT_PTR MessageBox( HWND hwndParent,
                            PCTSTR pstrMsg, PCTSTR pstrLongMsg = NULL, PCTSTR pstrCaption = NULL,
                            UINT uType = MB_OK, DWORD dwMmbFlags = 0 )
            { return MessageBox( hwndParent, pstrMsg, pstrLongMsg, pstrCaption, uType, dwMmbFlags, GetThreadLocale() ); }
        INT_PTR MessageBox( HWND hwndParent,
                            PCTSTR pstrMsg, PCTSTR pstrLongMsg, PCTSTR pstrCaption,
                            UINT uType, DWORD dwMmbFlags, LCID language,
                            PCTSTR pstrHeader = NULL, PCTSTR pstrExtraMsg = NULL, PCTSTR pstrExtraLink_Text = NULL, PCTSTR pstrExtraLink_Url = NULL,
                            HMODULE hModuleToLoadIconsFrom = NULL, UINT_PTR nMsgIcon_IdOrHicon = 0, UINT_PTR nExtraIcon_IdOrHicon = 0 );

        #ifndef _WIN32_WCE
        static HWND FindNiceParent();
        static void PushForegroundWindow( HWND hwnd );
        #endif

        // flags for the GetOwnedPopups() function
        #define OwnedPopupsF_skipDisabledPopups     0x00000001
        #define OwnedPopupsF_skipInvisiblePopups    0x00000002

        template< class Tstruct >
        static void GetOwnedPopups( HWND hwndParent, Array<Tstruct>& aPopups, DWORD dwFlags = 0 );

        #ifndef _WIN32_WCE
        void Center( bool bCenterOnlyHorizontally = false, long nAdditionalHeight = 0 )
            { CenterWindowOverWindow( m_hWnd, m_dwMmbFlags & MMB_CenterOverParent ? GetParent(m_hWnd) : NULL, bCenterOnlyHorizontally, nAdditionalHeight ); }
        #endif
        static void CenterWindowOverWindow( HWND hwndToBeCentered, HWND hwndReferenceWindow, bool bCenterOnlyHorizontally = false, long nAdditionalHeight = 0 );
        static void CenterWindowOverItsParent( HWND hwndToBeCentered, bool bCenterOnlyHorizontally = false, long nAdditionalHeight = 0 );

        static HFONT CreateAdoptedFont( HFONT hfontFrom, int nWeight = FW_NORMAL, bool bItalic = false, bool bUnderline = false, bool bStrikeOut = false );

    protected :
        #ifndef _WIN32_WCE
        HWND        m_hWnd;         // HWND of the messagebox dialog

        struct Controls
        {
            DWORD   m_dwMmbFlags;
            int     m_minWidthAsNumButtons; // a number of buttons (1 to 4) that indicates the minimum width of the message box in any cases (regardless of the really displayed number of buttons)

            HWND    m_hWnd;             // HWND of the messagebox dialog
            HWND    m_hwndHeaderMsg;
            HWND    m_hwndMsgIcon, m_hwndMsg, m_hwndLongMsg;
            HWND    m_ahwndBtns[4];
            HWND    m_hwndHelpBtn, m_hwndMore;
            HWND    m_ahwndChecks[2];
            HWND    m_hwndHr;
            HWND    m_hwndExtraIcon, m_hwndExtraMsg, m_hwndExtraLink;

            RECT    m_rectWnd;          // rect of the messagebox dialog
            RECT    m_rectHeaderMsg;
            RECT    m_rectMsgIcon;
            RECT    m_rectMsg, m_rectLongMsg;
            RECT    m_arectBtns[4];
            RECT    m_rectHelpBtn, m_rectMore;
            RECT    m_arectChecks[2];
            RECT    m_rectHr;
            RECT    m_rectExtraIcon, m_rectExtraMsg, m_rectExtraLink;

            bool    m_bIsExpanded;
            long    m_dyExpand;         // y offset for collapsing/expanding the dialog
            long    m_dxButtons;        // x offset of one button to the next
            long    m_dyCheck1;         // y offset of Check1 checkbox to upper control.
            long    m_dyCheckboxes;     // y offset of one checkbox to the next
            RECT    m_rectControlsAreaBordersThicknesses; // each extent describes the thickness of the border of the control area to the dialog's client area
            POINT   m_ptControlsAreaTopLeft; // top left coords of the control area rect
            long    m_cxminControlsArea; // minimum width of the all-controls bounding rect (so dialog client area without a border)
            long    m_dyExtraLinkDist;  // y distance from Extra Msg's bottom to Extra Link's top coord.

            static const int MaxButtons    = 4;
            static const int MaxCheckboxes = 2;

            void Empty();
            void getFromDialog( HWND hwnd,
                                DWORD dwMmbFlags, int minWidthAsNumButtons );
            void deleteHeaderMsg();
            void deleteMsgIcon();
            void getMsgIconSize();
            void getExtraIconSize();
            void deleteExtraArea( bool bDelIcon, bool bDelMsg, bool bDelLink );
            void decreaseNumButtonsTo( int cButtons, bool bHelpBtn, bool bMoreBtn );
            RECT getAllButtonsBoundsRect();
            long getCxMinControlsArea();
            long correctControlsAreaNewRightCoord( long nControlsAreaNewRightCoord );
            void adjustWidth( long nControlsAreaNewRightCoord );
            void adjustHeightFromMsgs( long cyNewHeaderMsg, long cyNewMsg, long cyNewExtraMsg, long cyNewExtraLink );
            void offsetControls( long dx, long dy,
                                 bool bDialogSize, bool bHeader,
                                 bool bMsgIcon, bool bMsg, bool bLongMsg,
                                 bool bButtons, bool bCheckboxes,
                                 bool bHorzRuler, bool bExtraArea );
            void inflateBorders( RECT rectInflate );
            void decreaseNumCheckboxesTo( int cCheckboxes );
            void setRects( bool bKeepDialogVisible = true );
            static void SetControlRect( HWND hwndCtrl, RECT rect, bool bDisableIfEmptyRect = true );
            static void SetControlTopLeft( HWND hwndCtrl, POINT pt );
            static void SetControlTopLeft( HWND hwndCtrl, RECT rect ) { SetControlTopLeft( hwndCtrl, * (POINT*) &rect ); }
            void Collapse();
            void Expand();
        };

        LCID        m_Language;
        String      m_strMsg;
        String      m_strLongMsg;
        String      m_strCaption;
        UINT        m_uType;
        UINT_PTR    m_uMsgIcon;
        UINT        m_uMsgBeep;
        UINT        m_uMsgButtons;
        UINT        m_uMsgDefButton;
        DWORD       m_dwMmbFlags;
        Controls    m_Controls;
        HWND        m_hwndDefBtn;
        Array<HWND> m_aDisabledPopups;  // the popup windows that have been disabled and must be enabled again when the messagebox closed
        UINT        m_DT_MMB;           // combination of DT_* values (for DrawText()) to use for the normal message text
        UINT        m_DT_MMBHeader;     // combination of DT_* values (for DrawText()) to use for the Header message text
        UINT        m_DT_MMBExtra;      // combination of DT_* values (for DrawText()) to use for the Extra message and link text
        HFONT       m_hfontBold;        // the normal dialog font but in bold (will be deleted when dialog is destroyed)
        HFONT       m_hfontUnderlined;  // the normal dialog font but underlined (used for URLs) (will be deleted when dialog is destroyed)
        String      m_strHeaderMsg;
        String      m_strExtraMsg;
        String      m_strExtraLink_Text;
        String      m_strExtraLink_Url;
        HMODULE     m_hModuleToLoadIconsFrom;
        UINT_PTR    m_nMsgIcon_IdOrHicon, m_nExtraIcon_IdOrHicon;
        #endif

        template< class Tstruct >
        struct MMBGetOwnedPopupsStruct
        {
            HWND            m_hwndParent;
            DWORD           m_dwOwnedPopupsFlags;
            Array<Tstruct> *m_paPopups;
        };

        #ifndef _WIN32_WCE
        virtual BOOL onInitDialog();
        virtual void onDestroy();
        virtual void onMoreClicked();
        virtual void onUrlClicked();
        static INT_PTR CALLBACK WndProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam );

        static LRESULT CALLBACK StaticCtrlProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam );
        static LRESULT CALLBACK UrlStaticCtrlProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam );
        static LRESULT CALLBACK ButtonProc( HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam );

        void EnableDisabledPopups( bool bEnable );
        static BOOL CALLBACK FindWndProc( HWND hwnd, LPARAM lParam );
        #endif
        template< class Tstruct >
        static BOOL CALLBACK GetOwnedPopupsProc( HWND hwnd, LPARAM lParam );
};




// ____ Implementations _______________________________

//      MoreMessageBox

///<summary> Gets all popup windows owned by the specified parent. <br/>
///  <b>Tstruct</b> must be a class or struct or directly a HWND.
///  In case of class or struct, it must have a <b>Tstruct(HWND)</b> constructor and
///  a <b>operator =(HWND)</b> defined that allows direct assigning
///  of a popup-window HWND to the struct instance. </summary>
template< class Tstruct >
void MoreMessageBox::GetOwnedPopups( HWND hwndParent, Array<Tstruct>& aPopups, DWORD dwFlags )
{
    aPopups.Empty();
    if( !hwndParent ) return;

    MMBGetOwnedPopupsStruct<Tstruct> dat;
    dat.m_hwndParent         = hwndParent;
    dat.m_dwOwnedPopupsFlags = dwFlags;
    dat.m_paPopups           = &aPopups;

    EnumWindows( GetOwnedPopupsProc<Tstruct>, (LPARAM) &dat );
}

///<returns> To continue enumeration, return TRUE. To stop enumeration, return FALSE. </returns>
template< class Tstruct >
BOOL CALLBACK MoreMessageBox::GetOwnedPopupsProc( HWND hwnd, LPARAM lParam )
{
    MMBGetOwnedPopupsStruct<Tstruct>& dat = * (MMBGetOwnedPopupsStruct<Tstruct>*) lParam;

    if( GetParent(hwnd) == dat.m_hwndParent )
    {
        if( (dat.m_dwOwnedPopupsFlags & OwnedPopupsF_skipDisabledPopups) && !IsWindowEnabled(hwnd) )
            return TRUE;

        if( (dat.m_dwOwnedPopupsFlags & OwnedPopupsF_skipInvisiblePopups) && !IsWindowVisible(hwnd) )
            return TRUE;

        Tstruct *p = dat.m_paPopups->AddNew();
        if( !p ) return FALSE;

        *p = hwnd;
    }

    return TRUE;
}




}

#endif
