/*
    File:               argonParser.h
    Description:        A basic string parser for the most common
                        cases like C syntax or Cmdline parameters.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         18.09.2005
    last updated:       25.08.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonParser_h
#define __argonParser_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include "argonString.h"
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4634 )    // warning C4634: XML document comment: cannot be applied:  Discarding XML document comment for invalid target
#endif

namespace ArgonLibrary
{




enum syntaxparserLanguage_t {
    syntaxparserlanguage_Cmdline    = 1,    // prepare the parser for Cmdline parsing.
    syntaxparserlanguage_C,                 // prepare the parser for C/C++ syntax like parsing.
    syntaxparserlanguage_Basic              // prepare the parser for BASIC programming language syntax like parsing.
};

enum basicSPsyntax_t {
    bspsyxt_none                    = 0,
    bspsyxt_anything,
    bspsyxt_Word,
    bspsyxt_SpecialChar,
    bspsyxt_SpacingChar,
    bspsyxt_String,
    bspsyxt_Remarkline,
    bspsyxt_RemarkSection
};

enum basicSPfindflags_t {
    // the bspfindf_find* values indicate which syntax types may be found and
    // returned by #FindNextSyntaxelm()#.
    bspfindf_findAnything           = 0x00000001,
    bspfindf_findWords              = 0x00000002,
    bspfindf_findSpecialChars       = 0x00000004,
    bspfindf_findStrings            = 0x00000008,
    bspfindf_findRemarks            = 0x00000010,
    // the bspfindf_count* values indicate which syntax types shall be detected
    // at all by #FindNextSyntaxelm()#. If e.g. #bspfindf_countRemarks# is *not*
    // specified, then remark sections will be treated like spacing characters.
    // So they will not be counted when using the #nGoElementsForward# parameter to count
    // elements (also named sequences) when using #FindNextSyntaxelm()#.
    bspfindf_countAnything          = 0x00000100,
    bspfindf_countWords             = 0x00000200,
    bspfindf_countSpecialChars      = 0x00000400,
    bspfindf_countStrings           = 0x00000800,
    bspfindf_countRemarks           = 0x00001000,
    //
    bspfindf_Default                = bspfindf_findAnything  | bspfindf_findWords  | bspfindf_findSpecialChars  | bspfindf_findStrings |
                                      bspfindf_countAnything | bspfindf_countWords | bspfindf_countSpecialChars | bspfindf_countStrings
    // The default does not find and not count remark lines and sections.
};

enum basicSPextractflags_t {
    bspextractf_replaceInsideStrings    = 0x00000001,
    bspextractf_expand_C_escapes        = 0x00000002,
    bspextractf_Default                 = bspextractf_replaceInsideStrings | bspextractf_expand_C_escapes
};

template< class String_T >
struct basicSPresult
{
    typedef typename String_T::Tchar Tchar;
    typedef typename String_T::Tchar *Ptstr;
    typedef const typename String_T::Tchar *Pctstr;

    basicSPsyntax_t m_syntaxType;   // out: type of the syntax found

    _idx_   m_idxOuterStart;        // out: Will contain the start index of the syntax-element found.
                                    //  In case of a string or a remark sequence, this index will point at
                                    //  the start *including* the string's or remark's start character(s).
    _idx_   m_idxInnerStart;        // out: Will contain the start index of the syntax-element found.
                                    //  In case of a string or a remark sequence, this index will point at
                                    //  the start *excluding* the string's or remark's start character(s).
    _idx_   m_idxInnerEnd;          // out: Will contain the character index of the syntax-element's end found.
                                    //  To be precise: It will point *after* the last character of the
                                    //  syntax-element found.
                                    //  Except in case of a string or a remark sequence, this index will point
                                    //  at the *first* character of the string's or remark's end character(s).
                                    // Note: For a found remark line that ends with the end-of-line, this index
                                    //  point at the newline character (or at end-of-#.m_pstrParseString#).
                                    // Note: For a found special character, the inner end will point after
                                    //  the special character.
    _idx_   m_idxOuterEnd;          // out: Will contain the character index of the syntax-element's end found.
                                    //  To be precise: It will point *after* the last character of the
                                    //  syntax-element found.
                                    //  In case of a string or a remark sequence, this index will point
                                    //  *after* the last character of the string's or remark's end character(s).
                                    // Note: For a found remark line that ends with the end-of-line, this index
                                    //  point at the newline character (or at end-of-#.m_pstrParseString#).
                                    // Note: For a found special character, the inner end will point after
                                    //  the last special character found in a sequence of special chars beginning
                                    //  from #m_idxOuterStart#.

    _idx_   m_idxStrlstSubstring;   // out: If the found sytax type has been detected using a stringlist
                                    //  (like #.m_pstrlstStringStart# or #.m_pstrlstRemarkSectionStart# etc.),
                                    //  then in here is the index of the stringlist's substring found.

    Pctstr  m_pstrParseString;      // out: On return, this will point to the whole string that has been
                                    //  parsed to find the desired syntax-element.

    basicSPresult() { Reset(); }
    void Reset()
    { m_syntaxType = bspsyxt_none; m_idxOuterStart = m_idxInnerStart = m_idxInnerEnd = m_idxOuterEnd = m_idxStrlstSubstring = -1; m_pstrParseString = NULL; }
};

///<summary>
///  The <b>basicSyntaxParser</b> is just basic. That means, it can parse
///  the syntax basically, but not hierarchical bracket levels. <br/>
///  What can be distinguished is words (command words or numbers) or
///  special characters or strings within string-encapsulation chars,
///  skipping remark lines and remark sections. <br/><br/>
///
///  Note: <br/>
///  This parser is not capable of parsing backwards. The only thing possible is:
///  when the current parsing start location points somewhere into a word,
///  then the algorithm can find the start of that word by going backward some chars.
///  But that's it.
///  No backward remark-section or string sequence recognition. <br/><br/>
///
///  Note: <br/>
///  As template parameter String_T, you can pass argonString's String, StringA
///  or StringW (or _string_, etc.) or all other classes derived from StringT.
///</summary>
template< class String_T >
class basicSyntaxParser
{
    public :
        typedef typename String_T::Tchar Tchar;
        typedef typename String_T::Tchar *Ptstr;
        typedef const typename String_T::Tchar *Pctstr;


    private :
    // String to parse:
        Pctstr  m_pstrParseString;
        mutable _size_t_ m_lenParseString;


    public :
    // syntax engine settings:
    // change them as needed to specify the syntax to parse for
        bool    m_bCaseinsensitive;

        Pctstr  m_pstrWordChars;                // Characters a word (a command word or a numeric value) may consist of.
                                                //  Don't forget to include the number digits in here, if you want
                                                //  to detect number values, too.
                                                //  Set to NULL (or empty string) if there are no specific word chars.
        bool    m_bInclWordAlphaChars;          // If to include alphabetic characters in the word-chars automatically
        bool    m_bInclWordNumChars;            // If to include numberic characters in the word-chars automatically

        Pctstr  m_pstrSpecialChars;             // Special chars that can also be found as word but only one char each.
                                                //  Can be used for operators.
                                                //  Can be NULL (or empty string).

        Pctstr  m_pstrlstStringStart;           // A stringlist with all the start-string indicating the start of a string.
                                                //  For the n-th string found in the stringlist, the appropriate string in
                                                //  #m_pstrlstStringEnd#, #m_pstrlstStringInside# and
                                                //  #m_pstrlstStringReplace# will be used.
                                                //  Can be NULL (or empty stringlist) if no string-encapsulations to search for.
        Pctstr  m_pstrlstStringEnd;             // A stringlist with the according end-string encapsulations indicating the end of a string.
                                                //  Must have same number of strings like #m_pstrlstStringStart#.
        Pctstr  m_pstrlstStringInside;          // A stringlist with the according inside-strings. If inside a string, this string
                                                //  is detected (which normally starts like the string in #m_pstrlstStringStart#),
                                                //  it will be replaced by the according string in #m_pstrlstStringReplace#),
                                                //  Can be NULL or otherwise can have any number of strings.
        Pctstr  m_pstrlstStringReplace;         // A stringlist with the according inside-replace-strings. If inside a string, the
                                                //  according string in #m_pstrlstStringInside# has been found, then it will be
                                                //  replaced by the according string from #m_pstrlstStringReplace#.
                                                //  it will be replaced by the according string in #m_pstrlstStringReplace#),
                                                //  Must have same number of strings like #m_pstrlstStringReplace#.
        bool    m_bString_C_escapes;            // Set to true if C-escape sequences ("\...") inside an encapsulated string are allowed.

        Pctstr  m_pstrlstRemarkStartline;       // A stringlist with strings that indicate the start of a remark line
                                                //  that ends at the end-of-line.
                                                //  Can be NULL (or empty stringlist).
        Pctstr  m_pstrlstRemarkSectionStart;    // A stringlist with strings that indicate the start of a remark section
                                                //  that ends with the according string from #m_pstrlstRemarkSectionEnd#.
                                                //  Can be NULL (or empty stringlist).
        Pctstr  m_pstrlstRemarkSectionEnd;      // A stringlist with strings that indicate the end of a remark section.
                                                //  Must contain same number of strings like #m_pstrlstRemarkSectionStart#.

        Pctstr  m_pstrSpacingChars;             // Characters used as spacing (or separation) between words, special chars etc.
                                                //  Can be NULL (or empty string) if no spacing chars at all or only those when #m_bInclSpaceSpacingChars# is true.
        bool    m_bInclSpaceSpacingChars;       // if space characters should be included in the spacing-chars automatically (#_isspace_()#).


    // parser settings:
        DWORD   m_findFlags;                    // combination of #basicSPfindflags_t# values as flags for the find algorithm.
        DWORD   m_extractFlags;                 // combination of #basicSPextractflags_t# values as flags for the extraction algorithm.


    public :
        basicSyntaxParser( syntaxparserLanguage_t lang )
            { m_pstrParseString = NULL; m_lenParseString = 0; m_findFlags = bspfindf_Default; m_extractFlags = bspextractf_Default; Prepare( lang ); }
        basicSyntaxParser( syntaxparserLanguage_t lang, Pctstr pstrParseString )
            { m_pstrParseString = NULL; m_lenParseString = 0; m_findFlags = bspfindf_Default; m_extractFlags = bspextractf_Default; Prepare( lang ); setParseStringPtr( pstrParseString ); }
        ~basicSyntaxParser() {  }

        bool     Prepare( syntaxparserLanguage_t lang );

            ///<summary>
            ///  <b>setParseStringPtr()</b> sets the string to be parsed.
            ///  WARNING: The memory of the string buffer must be valid as long as
            ///  this parser object is using it. The parser object will store just
            ///  a pointer to the memory and will not copy the string.
            ///  This is done because a desired string to parse could be very large.
            ///</summary>
        void     setParseStringPtr( Pctstr pstrParseString ) { m_pstrParseString = pstrParseString; m_lenParseString = 0; }
        Pctstr   getParseStringPtr() { return m_pstrParseString; }
        _size_t_ getParseStringLength() const
            { if( m_lenParseString == 0 ) m_lenParseString = _strlen_(m_pstrParseString); return m_lenParseString; }

            ///<returns>
            ///  Returns an index where to continue parsing (so it returns the same value
            ///  as <b>result.m_idxOuterEnd</b> or <b>result.m_idxInnerEnd</b> (e.g. in case
            ///  of special chars, because then, the index must point after the special char). <br/><br/>
            ///
            /// -1 = element not found (reached end of string), or <b>idxStartParsing</b> is out of range <br/>
            /// -2 = error: end of string reached but an ending-sequence not found
            ///      (e.g. the closing of a string or of a remark section is missing)
            ///</returns>
        _idx_    FindNextSyntaxelm( basicSPresult<String_T>& result, _idx_ idxStartParsing, _idx_ nGoElementsForward = 1 ) const;

            ///<remarks> When extracting an inner string, the <b>.m_extractFlags</b>
            ///  will take effect. </remarks>
        bool     ExtractInnerSyntaxelm( String_T& strResult, const basicSPresult<String_T>& result ) const;
        String_T ExtractInnerSyntaxelm( basicSPresult<String_T>& result ) const { String_T s; ExtractInnerSyntaxelm( s, result ); return s; }

            ///<remarks> When extracting an outer string, the <b>.m_extractFlags</b>
            ///  *cannot* take effect. </remarks>
        bool     ExtractOuterSyntaxelm( String_T& strResult, const basicSPresult<String_T>& result ) const;
        String_T ExtractOuterSyntaxelm( basicSPresult<String_T>& result ) const { String_T s; ExtractOuterSyntaxelm( s, result ); return s; }

    protected :
        basicSPsyntax_t identifySyntaxType( Pctstr pstr, _idx_& isubstr ) const;
        bool            isStartOfSyntaxType( Pctstr pstr, basicSPsyntax_t syntax_type ) const;
        Pctstr          getStartOfSyntaxType( Pctstr pstr, basicSPsyntax_t syntax_type ) const;
        Pctstr          skipFwdNoncontentType( Pctstr pstr, basicSPsyntax_t& syntax_type, _idx_& isubstr ) const;

        static bool     isContentType( basicSPsyntax_t syntax_type )    // returns true if the type is with visible content (so no spaces)
            { return syntax_type != bspsyxt_none && syntax_type != bspsyxt_SpacingChar; }

        bool            isWordChar( Pctstr pstr ) const;
        bool            isSpecialChar( Pctstr pstr ) const;
        bool            isSpacingChar( Pctstr pstr ) const;
        _idx_           isRemarkline( Pctstr pstr ) const;
        _idx_           isRemarkSectionStart( Pctstr pstr ) const;
        bool            isRemarkSectionEnd( Pctstr pstr, _idx_ isubstr ) const;
        _idx_           isStringStart( Pctstr pstr ) const;
        bool            isStringEnd( Pctstr pstr, _idx_ isubstr ) const;
        _idx_           isStringInside( Pctstr pstr ) const;
};




///<summary>
///  A Command line parser. <br/><br/>
///
///  Note: <br/>
///  As template parameter String_T, you can pass argonString's String, StringA
///  or StringW (or _string_, etc.) or all other classes derived from StringT.
///</summary>
template< class String_T >
class CmdlineParser : public basicSyntaxParser<String_T>
{
    public :
        bool    m_bCmdlineIncludesExe;  // true = the command line contains the executable name at the start of the string.
                                        // This will be used only by #findNormal()# to see if #nthNonParam# can be 0.

    public :
        CmdlineParser( bool bCmdlineIncludesExe = true ) : basicSyntaxParser<String_T>( syntaxparserlanguage_Cmdline ) { m_bCmdlineIncludesExe = bCmdlineIncludesExe; }
        CmdlineParser( Pctstr pstrCmdline, bool bCmdlineIncludesExe = true ) : basicSyntaxParser<String_T>( syntaxparserlanguage_Cmdline )
            { m_bCmdlineIncludesExe = bCmdlineIncludesExe; setParseStringPtr( pstrCmdline ); }

        bool     findNonValuedParam( Pctstr pstrParam );
        bool     findValuedParam( Pctstr pstrParam, String_T& strValue );
        bool     findOptionalValuedParam( Pctstr pstrParam, bool& bHasValue, String_T& strValue );

        bool     findNormal( String_T& strFound, _idx_ nthNonParam = 1, Pctstr pstrNonStartChars = _Tstr2("/-", String_T) );
        String_T findNormal( _idx_ nthNonParam = 1, Pctstr pstrNonStartChars = _Tstr2("/-", String_T) )
            { String_T s; findNormal( s, nthNonParam, pstrNonStartChars ); return s; }
};




// ____ Implementations _______________________________

//      basicSyntaxParser

template< class String_T >
bool basicSyntaxParser<String_T>::Prepare( syntaxparserLanguage_t lang )
{
    switch( lang )
    {
        case syntaxparserlanguage_Cmdline :
        {
            m_bCaseinsensitive          = false;
            m_pstrWordChars             = NULL;
            m_bInclWordAlphaChars       = false;
            m_bInclWordNumChars         = false;
            m_pstrSpecialChars          = _Tstr2("=", String_T);
            m_pstrlstStringStart        = _Tstr2("\"\0", String_T);
            m_pstrlstStringEnd          = _Tstr2("\"\0", String_T);
            m_pstrlstStringInside       = _Tstr2("\"\"\0", String_T);
            m_pstrlstStringReplace      = _Tstr2("\"\0", String_T);
            m_bString_C_escapes         = false;
            m_pstrlstRemarkStartline    = NULL;
            m_pstrlstRemarkSectionStart = NULL;
            m_pstrlstRemarkSectionEnd   = NULL;
            m_pstrSpacingChars          = NULL;
            m_bInclSpaceSpacingChars    = true;
            return true;
        }

        case syntaxparserlanguage_C :
        {
            m_bCaseinsensitive          = false;
            m_pstrWordChars             = _Tstr2("_", String_T);
            m_bInclWordAlphaChars       = true;
            m_bInclWordNumChars         = true;
            m_pstrSpecialChars          = _Tstr2("^&!|*/%+-{}[]()<>:,.;=~\\#", String_T);
            m_pstrlstStringStart        = _Tstr2("\"\0\'\0", String_T);
            m_pstrlstStringEnd          = _Tstr2("\"\0\'\0", String_T);
            m_pstrlstStringInside       = NULL;
            m_pstrlstStringReplace      = NULL;
            m_bString_C_escapes         = true;
            m_pstrlstRemarkStartline    = _Tstr2("//\0", String_T);
            m_pstrlstRemarkSectionStart = _Tstr2("/*\0", String_T);
            m_pstrlstRemarkSectionEnd   = _Tstr2("*/\0", String_T);
            m_pstrSpacingChars          = NULL;
            m_bInclSpaceSpacingChars    = true;
            return true;
        }

        case syntaxparserlanguage_Basic :
        {
            m_bCaseinsensitive          = true;
            m_pstrWordChars             = _Tstr2("_", String_T);
            m_bInclWordAlphaChars       = true;
            m_bInclWordNumChars         = true;
            m_pstrSpecialChars          = _Tstr2("^&!|*/%+-{}[]()<>:,.;=~\\", String_T);
            m_pstrlstStringStart        = _Tstr2("\"\0\'\0", String_T);
            m_pstrlstStringEnd          = _Tstr2("\"\0\'\0", String_T);
            m_pstrlstStringInside       = _Tstr2("\"\"\0\'\'\0", String_T);
            m_pstrlstStringReplace      = _Tstr2("\"\0\'\0", String_T);
            m_bString_C_escapes         = false;
            m_pstrlstRemarkStartline    = _Tstr2("\'\0", String_T);
            m_pstrlstRemarkSectionStart = NULL;
            m_pstrlstRemarkSectionEnd   = NULL;
            m_pstrSpacingChars          = NULL;
            m_bInclSpaceSpacingChars    = true;
            return true;
        }

        default :
        {
            m_bCaseinsensitive          = false;
            m_pstrWordChars             = NULL;
            m_bInclWordAlphaChars       = false;
            m_bInclWordNumChars         = false;
            m_pstrSpecialChars          = NULL;
            m_pstrlstStringStart        = NULL;
            m_pstrlstStringEnd          = NULL;
            m_pstrlstStringInside       = NULL;
            m_pstrlstStringReplace      = NULL;
            m_bString_C_escapes         = false;
            m_pstrlstRemarkStartline    = NULL;
            m_pstrlstRemarkSectionStart = NULL;
            m_pstrlstRemarkSectionEnd   = NULL;
            m_pstrSpacingChars          = NULL;
            m_bInclSpaceSpacingChars    = true;
            return false;
        }
    }
}



template< class String_T >
_idx_ basicSyntaxParser<String_T>::FindNextSyntaxelm( basicSPresult<String_T>& result, _idx_ idxStartParsing, _idx_ nGoElementsForward ) const
{
    result.Reset();
    result.m_pstrParseString = m_pstrParseString;

    if( nGoElementsForward < 0 || _isstrempty_(m_pstrParseString) ) return -1;

    _idx_    idx = idxStartParsing;
    _size_t_ len = getParseStringLength();
    if( idx < 0 || (_size_t_) idx > len ) return -1;

    Pctstr ptr = m_pstrParseString + idx;
    if( !*ptr ) return -1;

    _idx_ isubstr;
    basicSPsyntax_t syxt = identifySyntaxType( ptr, isubstr );

    if( nGoElementsForward == 0 )
    {
        // scan backwards to the start of the syntax type.
        ptr = getStartOfSyntaxType( ptr, syxt );
        if( !ptr ) return -1;

        ptr = skipFwdNoncontentType( ptr, syxt, isubstr );
        if( !ptr ) return -1;
    }

    if( nGoElementsForward > 0 )
    {
        ptr = skipFwdNoncontentType( ptr, syxt, isubstr );
        if( !ptr ) return -1;

        if( isStartOfSyntaxType( ptr, syxt ) )
            --nGoElementsForward;
    }

    while( nGoElementsForward >= 0 )
    {
        if( !*ptr ) return -1;

        ptr = skipFwdNoncontentType( ptr, syxt, isubstr );
        if( !ptr ) return -1;

        Pctstr pstart       = ptr;          // outer start
        Pctstr pend         = ptr + 1;      // outer end
        _idx_  isubstrStart = isubstr;

        switch( syxt )
        {
            case bspsyxt_anything :
            {
                while( *pend && identifySyntaxType(pend, isubstr) == syxt )
                    ++pend;

                if( nGoElementsForward == 0 && (m_findFlags & bspfindf_findAnything) )
                {
                    result.m_idxOuterStart = result.m_idxInnerStart = pstart - m_pstrParseString;
                    result.m_idxOuterEnd = result.m_idxInnerEnd = pend - m_pstrParseString;
                    result.m_syntaxType = syxt;
                    result.m_idxStrlstSubstring = isubstrStart;

                    return pend - m_pstrParseString;
                }

                if( m_findFlags & bspfindf_countAnything )
                    --nGoElementsForward;

                break;
            }

            case bspsyxt_Word :
            {
                while( identifySyntaxType(pend, isubstr) == syxt )
                    ++pend;

                if( nGoElementsForward == 0 && (m_findFlags & bspfindf_findWords) )
                {
                    result.m_idxOuterStart = result.m_idxInnerStart = pstart - m_pstrParseString;
                    result.m_idxOuterEnd = result.m_idxInnerEnd = pend - m_pstrParseString;
                    result.m_syntaxType = syxt;
                    result.m_idxStrlstSubstring = isubstrStart;

                    return pend - m_pstrParseString;
                }

                if( m_findFlags & bspfindf_countWords )
                    --nGoElementsForward;

                break;
            }

            case bspsyxt_SpecialChar :
            {
                while( identifySyntaxType(pend, isubstr) == syxt )
                    ++pend;

                _idx_ cSpecialChars = (_idx_) (pend - pstart);

                if( nGoElementsForward < cSpecialChars && (m_findFlags & bspfindf_findSpecialChars) )
                {
                    pstart += nGoElementsForward;

                    result.m_idxOuterStart = result.m_idxInnerStart = pstart - m_pstrParseString;
                    result.m_idxOuterEnd = pend - m_pstrParseString;
                    result.m_idxInnerEnd = (pstart - m_pstrParseString) + 1;
                    result.m_syntaxType = syxt;
                    result.m_idxStrlstSubstring = isubstrStart;

                    return (pstart - m_pstrParseString) + 1;
                }

                if( m_findFlags & bspfindf_countSpecialChars )
                    nGoElementsForward -= cSpecialChars;

                break;
            }

            case bspsyxt_String :
            {
                Pctstr pinstart = ptr = pend = pstart + _strlen_( _get_strlst_str_ptr_(m_pstrlstStringStart, isubstrStart) );

                // now find endmarker of string-sequence
                for( ;; )
                {
                    if( !*ptr ) return -2;

                    // skip so-called inside-strings
                    _idx_ isubstrInside = isStringInside( ptr );
                    if( isubstrInside >= 0 )
                        { ptr += _strlen_( _get_strlst_str_ptr_(m_pstrlstStringInside, isubstrInside) ); continue; }

                    if( isStringEnd( ptr, isubstrStart ) )
                        break;

                    if( m_bString_C_escapes ) ptr = _escfwdchr_ptr_( ptr );
                    else                      ++ptr;
                }

                Pctstr pinend = ptr;
                pend = pinend + _strlen_( _get_strlst_str_ptr_(m_pstrlstStringEnd, isubstrStart) );

                if( nGoElementsForward == 0 && (m_findFlags & bspfindf_findStrings) )
                {
                    result.m_idxOuterStart = pstart - m_pstrParseString;
                    result.m_idxInnerStart = pinstart - m_pstrParseString;
                    result.m_idxInnerEnd = pinend - m_pstrParseString;
                    result.m_idxOuterEnd = pend - m_pstrParseString;
                    result.m_syntaxType = syxt;
                    result.m_idxStrlstSubstring = isubstrStart;

                    return pend - m_pstrParseString;
                }

                if( m_findFlags & bspfindf_countStrings )
                    --nGoElementsForward;

                break;
            }

            case bspsyxt_Remarkline :
            {
                Pctstr pinstart = ptr = pend = pstart + _strlen_( _get_strlst_str_ptr_(m_pstrlstRemarkStartline, isubstrStart) );

                // go to end-of-line or end-of-string.
                while( *pend && *pend != _Tchar2('\r', String_T) && *pend != _Tchar2('\n', String_T) )
                    ++pend;

                if( nGoElementsForward == 0 && (m_findFlags & bspfindf_findRemarks) )
                {
                    result.m_idxOuterStart = pstart - m_pstrParseString;
                    result.m_idxInnerStart = pinstart - m_pstrParseString;
                    result.m_idxOuterEnd = result.m_idxInnerEnd = pend - m_pstrParseString;
                    result.m_syntaxType = syxt;
                    result.m_idxStrlstSubstring = isubstrStart;

                    return pend - m_pstrParseString;
                }

                if( m_findFlags & bspfindf_countRemarks )
                    --nGoElementsForward;

                break;
            }

            case bspsyxt_RemarkSection :
            {
                Pctstr pinstart = ptr = pend = pstart + _strlen_( _get_strlst_str_ptr_(m_pstrlstRemarkSectionStart, isubstrStart) );

                // now find endmarker of remark section
                for( ;; )
                {
                    if( !*ptr ) return -2;

                    if( isRemarkSectionEnd( ptr, isubstrStart ) )
                        break;
                }

                Pctstr pinend = ptr;
                pend = pinend + _strlen_( _get_strlst_str_ptr_(m_pstrlstRemarkSectionEnd, isubstrStart) );

                if( nGoElementsForward == 0 && (m_findFlags & bspfindf_findRemarks) )
                {
                    result.m_idxOuterStart = pstart - m_pstrParseString;
                    result.m_idxInnerStart = pinstart - m_pstrParseString;
                    result.m_idxInnerEnd = pinend - m_pstrParseString;
                    result.m_idxOuterEnd = pend - m_pstrParseString;
                    result.m_syntaxType = syxt;
                    result.m_idxStrlstSubstring = isubstrStart;

                    return pend - m_pstrParseString;
                }

                if( m_findFlags & bspfindf_countRemarks )
                    --nGoElementsForward;

                break;
            }

            default :
                return -1;
        }

        ptr = pend;

        syxt = identifySyntaxType( ptr, isubstr );
    }

    return -1;
}



template< class String_T >
bool basicSyntaxParser<String_T>::ExtractInnerSyntaxelm( String_T& strResult, const basicSPresult<String_T>& result ) const
{
    strResult.Empty();

    if( result.m_idxInnerStart < 0 || result.m_idxInnerEnd < 0 ) return false;

    _size_t_ len  = getParseStringLength();
    if( result.m_idxInnerStart >  (_idx_) len || result.m_idxInnerEnd <  result.m_idxInnerStart ) return false;
    if( result.m_idxInnerStart >= (_idx_) len || result.m_idxInnerEnd <= result.m_idxInnerStart ) return true;

    _size_t_ retlen = result.m_idxInnerEnd - result.m_idxInnerStart;

    String_T s;
    Ptstr pstr = s.GetBuffer( retlen );
    if( !pstr ) return false;

    _strcpy_( pstr, m_pstrParseString + result.m_idxInnerStart, retlen + 1 );

    s.ReleaseBuffer();

    Pctstr pstrlstInside = m_pstrlstStringInside, pstrlstReplace = m_pstrlstStringReplace;
    if( result.m_syntaxType == bspsyxt_String && (m_extractFlags & bspextractf_replaceInsideStrings) )
        while( pstrlstInside )
        {
            if( m_bCaseinsensitive ) s.Replaceistring( pstrlstInside, pstrlstReplace ? pstrlstReplace : _Tstr2("", String_T) );
            else                     s.Replacestring ( pstrlstInside, pstrlstReplace ? pstrlstReplace : _Tstr2("", String_T) );
            pstrlstInside  = _fwd_to_next_strlst_str_( pstrlstInside  );
            pstrlstReplace = _fwd_to_next_strlst_str_( pstrlstReplace );
        }

    if( result.m_syntaxType == bspsyxt_String && m_bString_C_escapes && (m_extractFlags & bspextractf_expand_C_escapes) )
        strResult = s.EscToStr();
    else
        strResult = s;

    return true;
}



template< class String_T >
bool basicSyntaxParser<String_T>::ExtractOuterSyntaxelm( String_T& strResult, const basicSPresult<String_T>& result ) const
{
    strResult.Empty();

    if( result.m_idxOuterStart < 0 || result.m_idxOuterEnd < 0 ) return false;

    _size_t_ len  = getParseStringLength();
    if( result.m_idxOuterStart >  (_idx_) len || result.m_idxOuterEnd <  result.m_idxOuterStart ) return false;
    if( result.m_idxOuterStart >= (_idx_) len || result.m_idxOuterEnd <= result.m_idxOuterStart ) return true;

    _size_t_ retlen = result.m_idxOuterEnd - result.m_idxOuterStart;

    Ptstr pstr = strResult.GetBuffer( retlen );
    if( !pstr ) return false;

    _strcpy_( pstr, m_pstrParseString + result.m_idxOuterStart, retlen + 1 );

    strResult.ReleaseBuffer();
    return true;
}



template< class String_T >
basicSPsyntax_t basicSyntaxParser<String_T>::identifySyntaxType( Pctstr pstr, _idx_& isubstr ) const
{
    if( _isstrempty_(pstr) ) return bspsyxt_none;

    if( ( isubstr = isRemarkline( pstr ) ) >= 0 ) return bspsyxt_Remarkline;
    if( ( isubstr = isRemarkSectionStart( pstr ) ) >= 0 ) return bspsyxt_RemarkSection;
    if( ( isubstr = isStringStart( pstr ) ) >= 0 ) return bspsyxt_String;
    if( isSpecialChar( pstr ) ) return bspsyxt_SpecialChar;
    if( isWordChar( pstr ) ) return bspsyxt_Word;
    if( isSpacingChar( pstr ) ) return bspsyxt_SpacingChar;

    return bspsyxt_anything;
}

///<returns> Returns false only if it is necessary for the type to
///  scan backwards to its start in the string. </returns>
template< class String_T >
bool basicSyntaxParser<String_T>::isStartOfSyntaxType( Pctstr pstr, basicSPsyntax_t syntax_type ) const
{
    if( !pstr || pstr < m_pstrParseString ) return true;

    if( syntax_type == bspsyxt_Word )
    {
        if( pstr == m_pstrParseString ) return true;
        return !isWordChar( pstr - 1 );
    }

    return true;
}

template< class String_T >
typename basicSyntaxParser<String_T>::Pctstr basicSyntaxParser<String_T>::getStartOfSyntaxType( Pctstr pstr, basicSPsyntax_t syntax_type ) const
{
    if( !pstr || pstr < m_pstrParseString ) return NULL;

    while( !isStartOfSyntaxType( pstr, syntax_type ) )
        --pstr;

    return pstr;
}

///<param name="syntax_type">
///  in: must be the syntax type of the char at <b>pstr</b>.
///  If you don't have it yet, call <b>identifySyntaxType()</b> first. <br/>
///  out: will be the found content type.
///</param>
///<returns> NULL if end of string reached </returns>
template< class String_T >
typename basicSyntaxParser<String_T>::Pctstr basicSyntaxParser<String_T>::skipFwdNoncontentType( Pctstr pstr, basicSPsyntax_t& syntax_type, _idx_& isubstr ) const
{
    if( _isstrempty_(pstr) ) return NULL;

    while( !isContentType( syntax_type ) )
    {
        syntax_type = identifySyntaxType( ++pstr, isubstr );
        if( !*pstr ) return NULL;
    }

    return pstr;
}



template< class String_T >
bool basicSyntaxParser<String_T>::isWordChar( Pctstr pstr ) const
{
    if( !pstr ) return false;

    if( m_bInclWordAlphaChars && m_bInclWordNumChars )
        { if( _isalnum_(*pstr) ) return true; }
    else
    if( m_bInclWordAlphaChars )
        { if( _isalpha_(*pstr) ) return true; }
    else
    if( m_bInclWordNumChars )
        { if( _isdigit_(*pstr) ) return true; }

    return m_bCaseinsensitive ?
            _strichr_(m_pstrWordChars, *pstr) != NULL :
            _strchr_ (m_pstrWordChars, *pstr) != NULL;
}

template< class String_T >
bool basicSyntaxParser<String_T>::isSpecialChar( Pctstr pstr ) const
{
    if( !pstr ) return false;

    return m_bCaseinsensitive ?
            _strichr_(m_pstrSpecialChars, *pstr) != NULL :
            _strchr_ (m_pstrSpecialChars, *pstr) != NULL;
}

template< class String_T >
bool basicSyntaxParser<String_T>::isSpacingChar( Pctstr pstr ) const
{
    if( !pstr ) return false;

    if( m_bInclSpaceSpacingChars )
        if( _isspace_(*pstr) ) return true;

    return m_bCaseinsensitive ?
            _strichr_(m_pstrSpacingChars, *pstr) != NULL :
            _strchr_ (m_pstrSpacingChars, *pstr) != NULL;
}

template< class String_T >
_idx_ basicSyntaxParser<String_T>::isRemarkline( Pctstr pstr ) const
{
    return m_bCaseinsensitive ?
            _findistrlststrhere_( pstr, m_pstrlstRemarkStartline ) :
            _findstrlststrhere_ ( pstr, m_pstrlstRemarkStartline );
}

template< class String_T >
_idx_ basicSyntaxParser<String_T>::isRemarkSectionStart( Pctstr pstr ) const
{
    return m_bCaseinsensitive ?
            _findistrlststrhere_( pstr, m_pstrlstRemarkSectionStart ) :
            _findstrlststrhere_ ( pstr, m_pstrlstRemarkSectionStart );
}

template< class String_T >
bool basicSyntaxParser<String_T>::isRemarkSectionEnd( Pctstr pstr, _idx_ isubstr ) const
{
    return m_bCaseinsensitive ?
            _striishere_( pstr, _get_strlst_str_ptr_( m_pstrlstRemarkSectionEnd, isubstr ) ) :
            _strishere_ ( pstr, _get_strlst_str_ptr_( m_pstrlstRemarkSectionEnd, isubstr ) );
}

template< class String_T >
_idx_ basicSyntaxParser<String_T>::isStringStart( Pctstr pstr ) const
{
    return m_bCaseinsensitive ?
            _findistrlststrhere_( pstr, m_pstrlstStringStart ) :
            _findstrlststrhere_ ( pstr, m_pstrlstStringStart );
}

template< class String_T >
bool basicSyntaxParser<String_T>::isStringEnd( Pctstr pstr, _idx_ isubstr ) const
{
    return m_bCaseinsensitive ?
            _striishere_( pstr, _get_strlst_str_ptr_( m_pstrlstStringEnd, isubstr ) ) :
            _strishere_ ( pstr, _get_strlst_str_ptr_( m_pstrlstStringEnd, isubstr ) );
}

template< class String_T >
_idx_ basicSyntaxParser<String_T>::isStringInside( Pctstr pstr ) const
{
    return m_bCaseinsensitive ?
            _findistrlststrhere_( pstr, m_pstrlstStringInside ) :
            _findstrlststrhere_ ( pstr, m_pstrlstStringInside );
}




//      CmdlineParser

///<summary>
///  <b>findNonValuedParam()</b> will search for the specified param
///  in the command line. It will successfully find it only if the
///  param does *not* contain a value which means no = operator must
///  follow the param name in the command line (even with spaces
///  between param name and = operator).
///</summary>
///<param name="pstrParam"> name of the param. You must include
///  a preceding / or - char if you want such a param. </param>
///<returns> True if the param has been found, otherwise false. </returns>
template< class String_T >
bool CmdlineParser<String_T>::findNonValuedParam( Pctstr pstrParam )
{
    if( _isstrempty_(pstrParam) ) return false;

    _idx_ idx = 0;
    _idx_ nInc = 1;

    for( ;; )
    {
        basicSPresult<String_T> result;
        idx = FindNextSyntaxelm( result, idx, nInc );
        if( idx < 0 ) return false;

        nInc = 1;

        // Now skip = operator and its following value.
        if( result.m_syntaxType == bspsyxt_SpecialChar )
        {
            nInc = 2;
            continue;
        }

        if( result.m_syntaxType == bspsyxt_anything ||
            result.m_syntaxType == bspsyxt_Word ||
            result.m_syntaxType == bspsyxt_String )
        {
            String_T param;
            if( !ExtractOuterSyntaxelm( param, result ) )
                continue;

            bool found = m_bCaseinsensitive ? param.Striis(pstrParam) : param.Stris(pstrParam);
            if( !found ) continue;

            // Let's check if no value is following.
            if( FindNextSyntaxelm( result, idx ) >= 0 )
                if( result.m_syntaxType == bspsyxt_SpecialChar )
                    return false;

            return true;
        }
    }
}

///<summary>
///  <b>findValuedParam()</b> will search for the specified param
///  in the command line. It will successfully find it only if the
///  param *does* contain a value which means a = operator must follow
///  the param name in the command line (spaces between param name
///  and = operator are allowed). <br/>
///  After the = operator, the next word (or "..."-sequence) found
///  will be interpreted as the value. If there is no value (end of
///  string), then the value will be interpreted as being an empty
///  string, but the param will be found successfully nevertheless.
///  For values, spaces between the = operator and the value are
///  allowed and valid and will not be included in the value itself.
///</summary>
///<param name="pstrParam"> name of the param. You must include
///  a preceding / or - char if you want such a param. </param>
///<param name="strValue"> will contain the extracted value after
///  the = character in the command line. The extracted value
///  will be extracted using <b>ExtractInnerSyntaxelm()</b>. </param>
///<returns> True if the param has been found, otherwise false. </returns>
template< class String_T >
bool CmdlineParser<String_T>::findValuedParam( Pctstr pstrParam, String_T& strValue )
{
    if( _isstrempty_(pstrParam) ) return false;

    _idx_ idx = 0;
    _idx_ nInc = 1;

    for( ;; )
    {
        basicSPresult<String_T> result;
        idx = FindNextSyntaxelm( result, idx, nInc );
        if( idx < 0 ) return false;

        nInc = 1;

        // Now skip = operator and its following value.
        if( result.m_syntaxType == bspsyxt_SpecialChar )
        {
            nInc = 2;
            continue;
        }

        if( result.m_syntaxType == bspsyxt_anything ||
            result.m_syntaxType == bspsyxt_Word ||
            result.m_syntaxType == bspsyxt_String )
        {
            String_T param;
            if( !ExtractOuterSyntaxelm( param, result ) )
                continue;

            bool found = m_bCaseinsensitive ? param.Striis(pstrParam) : param.Stris(pstrParam);
            if( !found ) continue;

            // Let's check if a value is following.
            idx = FindNextSyntaxelm( result, idx );
            if( idx < 0 ) return false;

            if( result.m_syntaxType != bspsyxt_SpecialChar )
                return false;

            // Get the value.
            idx = FindNextSyntaxelm( result, idx );
            if( idx < 0 )
                { strValue.Empty(); return true; }

            return ExtractInnerSyntaxelm( strValue, result );
        }
    }
}

///<summary>
///  <b>findOptionalValuedParam()</b> will find a param that has
///  or has not a value. It has a value, when a = operator follows
///  the param name in the command line. It is allowed and valid
///  to include spaces between the param name and = operator in
///  the command line.
///</summary>
///<param name="pstrParam"> name of the param. You must include
///  a preceding / or - char if you want such a param. </param>
///<param name="bHasValue"> will be true if a = operator follows
///  the param name in the command line, otherwise false. </param>
///<param name="strValue"> will contain the extracted value after
///  the = character in the command line. The extracted value
///  will be extracted using <b>ExtractInnerSyntaxelm()</b>. </param>
///<returns> True if the param has been found, otherwise false. </returns>
template< class String_T >
bool CmdlineParser<String_T>::findOptionalValuedParam( Pctstr pstrParam, bool& bHasValue, String_T& strValue )
{
    if( _isstrempty_(pstrParam) ) return false;

    _idx_ idx = 0;
    _idx_ nInc = 1;

    for( ;; )
    {
        basicSPresult<String_T> result;
        idx = FindNextSyntaxelm( result, idx, nInc );
        if( idx < 0 ) return false;

        nInc = 1;

        // Now skip = operator and its following value.
        if( result.m_syntaxType == bspsyxt_SpecialChar )
        {
            nInc = 2;
            continue;
        }

        if( result.m_syntaxType == bspsyxt_anything ||
            result.m_syntaxType == bspsyxt_Word ||
            result.m_syntaxType == bspsyxt_String )
        {
            String_T param;
            if( !ExtractOuterSyntaxelm( param, result ) )
                continue;

            bool found = m_bCaseinsensitive ? param.Striis(pstrParam) : param.Stris(pstrParam);
            if( !found ) continue;

            // Let's check if a value is following.
            idx = FindNextSyntaxelm( result, idx );
            if( idx < 0 )
                { bHasValue = false; strValue.Empty(); return true; }

            if( result.m_syntaxType != bspsyxt_SpecialChar )
                return false;

            bHasValue = true;

            // Get the value.
            idx = FindNextSyntaxelm( result, idx );
            if( idx < 0 )
                { strValue.Empty(); return true; }

            return ExtractInnerSyntaxelm( strValue, result );
        }
    }
}



///<summary>
///  <b>findNormal()</b> finds the nth (1st, or 2nd, or 3rd, etc.)
///  word in the command line that is *not* identified as parameter.
///  A word in the command line will be identified as parameter if
///  one of the following cases is met: <br/>
///      *)  A = operator is following the param name so that
///          it has a value and therefore must be a parameter. <br/>
///      *)  The word begins with one of the characters from
///          <b>pstrNonStartChars</b>. <br/>
///  Note: This method can find also empty-stringed words: if in the
///  command line, only "" is specified. In that case, the returned
///  <b>strFound</b> will be empty, but the return value will be
///  true nevertheless.
///</summary>
///<param name="strFound"> the word found. It will be retrieved
///  using <b>ExtractInnerSyntaxelm()</b> so that it will be
///  extracted to its final version without surrounding " chars. </param>
///<param name="nthNonParam"> the Nth non-param-word to find.
///  1 = 1st word, 2 = 2nd, etc., <br/>
///  0 = get executable at start of command line if <b>.m_bCmdlineIncludesExe</b>
///      is true (indicating that the exe name is present).
///      But if <b>.m_bCmdlineIncludesExe</b> is false, then
///      this method will return false. </param>
///<param name="pstrNonStartChars"> a set of characters that, if one
///  of them is the start char of a word, identify the word
///  as a param and therefore will be skipped then by this method. </param>
///<returns> True if found (even for an empty word like ""), false otherwise. </returns>
template< class String_T >
bool CmdlineParser<String_T>::findNormal( String_T& strFound, _idx_ nthNonParam, Pctstr pstrNonStartChars )
{
    strFound.Empty();

    if( nthNonParam == 0 && m_bCmdlineIncludesExe )
    {
        basicSPresult<String_T> result;
        _idx_ idx = FindNextSyntaxelm( result, 0 );
        if( idx < 0 ) return false;

        return ExtractInnerSyntaxelm( strFound, result );
    }

    if( nthNonParam < 1 ) return false;

    _idx_ idx = 0;
    _idx_ nInc = 1;

    if( m_bCmdlineIncludesExe )
        ++nInc;

    for( ;; )
    {
        basicSPresult<String_T> result;
        idx = FindNextSyntaxelm( result, idx, nInc );
        if( idx < 0 ) return false;

        nInc = 1;

        // Now skip = operator and its following value.
        if( result.m_syntaxType == bspsyxt_SpecialChar )
        {
            nInc = 2;
            continue;
        }

        if( result.m_syntaxType == bspsyxt_anything ||
            result.m_syntaxType == bspsyxt_Word ||
            result.m_syntaxType == bspsyxt_String )
        {
            String_T param;
            if( !ExtractOuterSyntaxelm( param, result ) )
                continue;

            if( !_isstrempty_(pstrNonStartChars) )
                if( _strchr_(pstrNonStartChars, param[0]) )
                    continue;

            // Let's check if no value is following.
            basicSPresult<String_T> tmp;
            if( FindNextSyntaxelm( tmp, idx ) >= 0 )
                if( tmp.m_syntaxType == bspsyxt_SpecialChar )
                {
                    nInc = 2;
                    continue;
                }

            if( --nthNonParam > 0 )
                continue;

            return ExtractInnerSyntaxelm( strFound, result );
        }
    }
}




}

#endif
