/*
    File:               argonRandom.cpp
    Description:        provides Random number generators
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         05.11.2006
    last updated:       05.11.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonRandom_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonRandom.h"

namespace ArgonLibrary
{




bool MersenneTwister::setSeed( DWORD seed, eMersenneTwisterKind kind )
{
    // IMPORTANT:
    //  Do *never* change the values in here nor change the
    //  calculation in the other functions.
    //  If you want a different handling, specify a new kind
    //  with different values, or increase the handling in
    //  the other functions as long as the old kinds produce
    //  exactly the same results.
    //  This is necessary to keep applications working properly
    //  with newer versions of the Argon Library.
    //  This can be important e.g. when you write an application
    //  that encodes something using the Mersenne Twister and
    //  you want the decoding algorithm to work with newer versions
    //  of the Argon Library.

    switch( kind )
    {
        case mersenneKind_default :
        {
            setSeed( seed,
                     624, 397,
                     0x9908b0df, 0x80000000, 0x7fffffff,
                     0x9d2c5680, 0xefc60000,
                     true,
                     0x52ef, 0x348a9, 69069, 2111111 );
            return true;
        }

        case mersenneKind_common :  // the original mersenne twister params
        {
            setSeed( seed,
                     624, 397,
                     0x9908b0df, 0x80000000, 0x7fffffff,
                     0x9d2c5680, 0xefc60000,
                     false,
                     0, 1, 69069, 1 );
            return true;
        }
    }

    return false;
}

void MersenneTwister::setSeed( DWORD seed,
                               int n, int m,
                               DWORD matrix_a,
                               DWORD upper_mask, DWORD lower_mask,
                               DWORD tempering_mask_b, DWORD tempering_mask_c,
                               bool swap_mag01,
                               DWORD seed_1, DWORD seed_2, DWORD seed_3, DWORD seed_4 )
{
    m_n = n;
    m_m = m;
    m_matrix_a = matrix_a;
    m_upper_mask = upper_mask;
    m_lower_mask = lower_mask;
    m_tempering_mask_b = tempering_mask_b;
    m_tempering_mask_c = tempering_mask_c;
    m_swap_mag01 = swap_mag01;
    m_seed_1 = seed_1;
    m_seed_2 = seed_2;
    m_seed_3 = seed_3;
    m_seed_4 = seed_4;

    m_mt.SetSize( m_n );
    DWORD *ptr = m_mt.Ptr();

    ptr[0] = seed ^ m_seed_1 ^ ( seed % m_seed_2 ); // to prevent the first value being zero
    for( m_mti = 1 ; m_mti < m_n ; m_mti++ )
        ptr[m_mti] = ( m_seed_3 * ptr[m_mti-1] ) ^ ( ptr[m_mti-1] % m_seed_4 );
    if( m_swap_mag01 ) { m_mag01[0] = m_matrix_a; m_mag01[1] = 0;          }
    else               { m_mag01[0] = 0;          m_mag01[1] = m_matrix_a; }

    /*  original mersenne twister initialization:
    m_mt[0] = seed;
    for( m_mti = 1 ; m_mti < m_n ; m_mti++ )
        m_mt[m_mti] = ( 69069 * m_mt[m_mti-1] );
    m_mag01[0] = 0;
    m_mag01[1] = m_matrix_a;
    */
}

DWORD MersenneTwister::nextDWord()
{
    DWORD v;

    // First generate m_n buffer values at one time
    // if buffer is empty.
    if( m_mti >= m_n )
    {
        DWORD *ptr = m_mt.Ptr();

        int i;
        for( i = 0 ; i < m_n - m_m ; i++ ) {
            v = ( ptr[i] & m_upper_mask ) | ( ptr[i+1] & m_lower_mask );
            ptr[i] = ptr[i+m_m] ^ (v>>1) ^ m_mag01[ v & 1 ];
        }

        for( ; i < m_n - 1 ; i++ ) {
            v = ( ptr[i] & m_upper_mask ) | ( ptr[i+1] & m_lower_mask );
            ptr[i] = ptr[i+(m_m-m_n)] ^ (v>>1) ^ m_mag01[ v & 1 ];
        }

        v = ( ptr[m_n-1] & m_upper_mask ) | ( ptr[0] & m_lower_mask );
        ptr[m_n-1] = ptr[m_m-1] ^ (v>>1) ^ m_mag01[ v & 1 ];
        m_mti = 0;
    }

    // Now calc the next pseudo random value.
    v = m_mt[ m_mti++ ];
    v ^= ( v >> 11 );
    v ^= ( v <<  7 ) & m_tempering_mask_b;
    v ^= ( v << 15 ) & m_tempering_mask_c;
    v ^= ( v >> 18 );

    return v;
}

float MersenneTwister::nextFloat()
{
    static const float rshift24 = 1.0f / (float) 0x1000000;
    return (float) (nextDWord() >> 8) * rshift24;
}

double MersenneTwister::nextDouble()
{
    // derived from nextDouble documentation in jdk 1.2 docs
    static const double rshift53 = 1.0 / ((double) 0x8000000 * (double) 0x4000000);
    DWORD y = nextDWord(), z = nextDWord();
    return ( (double) (y >> 6) * (double) 0x8000000 + (double) (z >> 5) ) * rshift53;
}




}
