/*
    File:               argonRect.cpp
    Description:        Point, Size and Rectangle classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         23.11.2003
    last updated:       03.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonRect_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonRect.h"

namespace ArgonLibrary
{




// Use this method if you want 'this' (imginary visible) point
// stay visible inside the rect. The result is that the
// point cannot lie on the right or bottom border of the rect.
// This is different to 'StayInPseudoRect()'.
_Pt_& _Pt_::StayInRect( RECT rc )
{
    if(      x <  rc.left   ) x = rc.left;
    else if( x >= rc.right  ) x = rc.right-1;
    if(      y <  rc.top    ) y = rc.top;
    else if( y >= rc.bottom ) y = rc.bottom-1;
    return *this;
}

// Use this method if you want 'this' point to be handled
// as being the corner of another rect.
// The point will stay inside the rect. But the result is that the
// point may also lie on the right or bottom border of the rect.
// This is different to 'StayInRect()'.
_Pt_& _Pt_::StayInPseudoRect( RECT rc )
{
    if(      x < rc.left   ) x = rc.left;
    else if( x > rc.right  ) x = rc.right;
    if(      y < rc.top    ) y = rc.top;
    else if( y > rc.bottom ) y = rc.bottom;
    return *this;
}

_Size_& _Pt_::GetAsSize()
{
    return * (_Size_*) this;
}

const _Size_& _Pt_::GetAsSize() const
{
    return * (_Size_*) this;
}

_Rect_ _Pt_::GetEmptyRect() const
{
    return _Rect_( x, y, x, y );
}

_Rect_ _Pt_::GetOneDotRect() const
{
    return _Rect_( x, y, x + 1, y + 1 );
}



_Size_ _Pt_::operator -( POINT pt ) const
{
    return _Size_( x - pt.x, y - pt.y );
}



_Rect_ _Pt_::operator +( PCRECT prc ) const
{
    if( !prc ) return GetEmptyRect();
    return _Rect_( x + prc->left, y + prc->top, x + prc->right, y + prc->bottom );
}

_Rect_ _Pt_::operator -( PCRECT prc ) const
{
    if( !prc ) return GetEmptyRect();
    return _Rect_( prc ).MirrorRect( *this );
}



// Sets only a new position for a window (does not change its size)
void _Pt_::SetWindowPos( HWND hwnd ) const
{
    ::SetWindowPos( hwnd, HWND_TOP, x, y, 0, 0,
                    SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}



_Pt_& _Pt_::MapSystemDialogCoords( bool bBackward )
{
    long baseunits = GetDialogBaseUnits();
    int baseunitX = (int) LOWORD( baseunits );
    int baseunitY = (int) HIWORD( baseunits );

    if( bBackward )
    {
        x = ::MulDiv( (int) x, baseunitX, 4 );
        y = ::MulDiv( (int) y, baseunitY, 8 );
    }
    else
    {
        x = ::MulDiv( (int) x, 4, baseunitX ? baseunitX : 1 );
        y = ::MulDiv( (int) y, 8, baseunitY ? baseunitY : 1 );
    }

    return *this;
}




///<param name="sizOfAspectRatio"> the size from which to take the aspect ratio </param>
///<param name="option"> a SAR_* value </param>
_Size_& _Size_::SetAspectRatio( SIZE sizOfAspectRatio, DWORD option )
{
    if( sizOfAspectRatio.cx <= 0 || sizOfAspectRatio.cy <= 0 )
    {
        if( sizOfAspectRatio.cx <= 0 ) cx = 0;
        if( sizOfAspectRatio.cy <= 0 ) cy = 0;
        return *this;
    }

    if( cx <= 0 || cy <= 0 )
        return *this;

    if( option == SAR_KEEPWIDTH )
    {
        cy = (cx * sizOfAspectRatio.cy) / sizOfAspectRatio.cx;
        return *this;
    }

    if( option == SAR_KEEPHEIGHT )
    {
        cx = (cy * sizOfAspectRatio.cx) / sizOfAspectRatio.cy;
        return *this;
    }

    if( option == SAR_ENLARGETOSMALLEST )
    {
        LONG cy2 = (cx * sizOfAspectRatio.cy) / sizOfAspectRatio.cx;
        if( cy2 >= cy ) cy = cy2;
        else            cx = (cy * sizOfAspectRatio.cx) / sizOfAspectRatio.cy;
        return *this;
    }

    if( option == SAR_SHRINKTOBIGGEST )
    {
        LONG cy2 = (cx * sizOfAspectRatio.cy) / sizOfAspectRatio.cx;
        if( cy2 <= cy ) cy = cy2;
        else            cx = (cy * sizOfAspectRatio.cx) / sizOfAspectRatio.cy;
        return *this;
    }

    return *this;
}

///<param name="aspect_ratio"> the aspect ratio to set (width divided by height) </param>
///<param name="option"> a SAR_* value </param>
_Size_& _Size_::SetAspectRatio( double aspect_ratio, DWORD option )
{
    if( aspect_ratio <= 0.0 )
    {
        if( aspect_ratio <= 0.0 ) cx = 0;
        if( aspect_ratio <  0.0 ) cy = 0;
        return *this;
    }

    if( cx <= 0 || cy <= 0 )
        return *this;

    if( option == SAR_KEEPWIDTH )
    {
        cy = (LONG) (cx / aspect_ratio);
        return *this;
    }

    if( option == SAR_KEEPHEIGHT )
    {
        cx = (LONG) (cy * aspect_ratio);
        return *this;
    }

    if( option == SAR_ENLARGETOSMALLEST )
    {
        LONG cy2 = (LONG) (cx / aspect_ratio);
        if( cy2 >= cy ) cy = cy2;
        else            cx = (LONG) (cy * aspect_ratio);
        return *this;
    }

    if( option == SAR_SHRINKTOBIGGEST )
    {
        LONG cy2 = (LONG) (cx / aspect_ratio);
        if( cy2 <= cy ) cy = cy2;
        else            cx = (LONG) (cy * aspect_ratio);
        return *this;
    }

    return *this;
}



_Rect_ _Size_::GetOriginRect() const
{
    return _Rect_( 0, 0, cx, cy );
}



_Rect_ _Size_::operator +( PCRECT prc ) const
{
    if( !prc ) return GetOriginRect();
    return _Rect_( prc->left, prc->top, cx + prc->right, cy + prc->bottom );
}



// Sets only a new size for a window (does not change its position)
void _Size_::SetWindowSize( HWND hwnd ) const
{
    ::SetWindowPos( hwnd, HWND_TOP, 0, 0, cx, cy,
                    SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );
}



_Size_& _Size_::MapSystemDialogCoords( bool bBackward )
{
    long baseunits = GetDialogBaseUnits();
    int baseunitX = (int) LOWORD( baseunits );
    int baseunitY = (int) HIWORD( baseunits );

    if( bBackward )
    {
        cx = ::MulDiv( (int) cx, baseunitX, 4 );
        cy = ::MulDiv( (int) cy, baseunitY, 8 );
    }
    else
    {
        cx = ::MulDiv( (int) cx, 4, baseunitX ? baseunitX : 1 );
        cy = ::MulDiv( (int) cy, 8, baseunitY ? baseunitY : 1 );
    }

    return *this;
}



bool _Size_::GetBitmapSize( HBITMAP hbm, BITMAP *pbm )
{
    if( !hbm ) return false;

    BITMAP bm;
    if( !pbm ) pbm = &bm;

    if( GetObject( hbm, sizeof(BITMAP), pbm ) < sizeof(BITMAP) )
        return false;

    cx = pbm->bmWidth;
    cy = labs( pbm->bmHeight );

    return true;
}

///<summary>
///  <b>GetIconSize()</b> can be used to get the size of an
///  icon *or* a cursor. <br/>
///  CAUTION: When <b>pii</b> is not NULL, then the 2 bitmaps
///  (<code>hbmMask</code> and <code>hbmColor</code>) will be
///  created by the internally called <code>GetIconInfo()</code>
///  function. You will then have to delete these bitmaps on your own.
///  Otherwise, if <b>pii</b> is NULL, then this method will delete
///  them automatically for you.
///</summary>
bool _Size_::GetIconSize( HICON hicon, ICONINFO *pii )
{
    if( !hicon ) return false;

    bool bExternalStruct = pii != NULL;

    ICONINFO ii;
    if( !pii ) pii = &ii;

    if( !GetIconInfo( hicon, pii ) )
        return false;

    BITMAP bm;
    bool ret = pii->hbmColor ?
        GetBitmapSize( pii->hbmColor, &bm ) : GetBitmapSize( pii->hbmMask, &bm );

    if( ret && !pii->hbmColor && bm.bmBitsPixel * bm.bmPlanes == 1 )
        cy /= 2;

    if( !bExternalStruct )
    {
        if( pii->hbmColor ) DeleteObject( pii->hbmColor );
        if( pii->hbmMask  ) DeleteObject( pii->hbmMask  );
    }

    return ret;
}



bool _Size_::GetTextPixelSize( PCTSTR pstrText, HFONT hfont, HDC hdc, UINT uDrawtextFormat )
{
    if( !pstrText || !hfont )
        { SetEmpty(); return false; }

    HDC tmp = NULL;
    if( !hdc )
        hdc = tmp = GetDC( NULL );

    HGDIOBJ old_font = SelectObject( hdc, hfont );
    RECT rect = { 0, 0, 0, 0 };
    DrawText( hdc, pstrText, -1, &rect, DT_CALCRECT | uDrawtextFormat );
    SelectObject( hdc, old_font );

    Set( rect.right - rect.left, rect.bottom - rect.top );

    if( tmp )
        ReleaseDC( NULL, tmp );
    return true;
}

bool _Size_::GetWindowFontPixelSize( HWND hwnd )
{
    if( !hwnd )
        { SetEmpty(); return false; }

    HFONT hfont = (HFONT) SendMessage( hwnd, WM_GETFONT, 0, 0 );

    HDC hdc = GetDC( hwnd );
    bool ret = GetFontPixelSize( hfont, hdc );
    ReleaseDC( hwnd, hdc );

    return ret;
}




bool _Range_::IntersectRange( const _Range_& range2 )
{
    int _end = privateRect::tmin( GetEnd(), range2.GetEnd() );
    m_start  = privateRect::tmax( m_start,  range2.m_start  );
    m_length = _end - m_start;

    return m_length > 0;
}

bool _Range_::UnionRange( const _Range_& range2 )
{
    if( range2.IsEmpty() )
        return !IsEmpty();

    int _end = privateRect::tmax( GetEnd(), range2.GetEnd() );
    m_start  = privateRect::tmin( m_start,  range2.m_start  );
    m_length = _end - m_start;

    return m_length > 0;
}

///<summary> This is similar to <b>UnionRange()</b>, but
///  with the difference that empty ranges (as input and
///  as output) are handled as if for non-empty ranges,
///  and an empty result range's coords not be set to zero.
///  They will then be the really *calculated* coords. </summary>
bool _Range_::UnionRangeEmul( const _Range_& range2 )
{
    int _end = privateRect::tmax( GetEnd(), range2.GetEnd() );
    m_start  = privateRect::tmin( m_start,  range2.m_start  );
    m_length = _end - m_start;

    return m_length > 0;
}




_Rect_& _Rect_::NormalizeRect()
{
    if( left > right  ) privateRect::tswap( left, right  );
    if( top  > bottom ) privateRect::tswap( top,  bottom );
    return *this;
}



// Moves the rect to the frame of 'rcFrame' according to 'flags'
// but always that way, that the rect stays inside of 'rcFrame'
// (if 'rcFrame' is larger than 'this' rect of course).
// The size of 'this' rect will never be changed.
//  : flags = combination of MRTF_* values which indicate which direction to move where to.
//            E.g. if this is only MRTF_LEFT, the x-coord of the rect will be moved to
//             the left side of 'rcFrame' but the y-coord will not be changed.
//  > false = Error
_Rect_& _Rect_::MoveRectToFrame( _Rect_ rcFrame, DWORD flags )
{
    if(      ( flags & MRTF_HORZ_MASK ) == MRTF_LEFT   ) MoveRectTo( rcFrame.left,          top );
    else if( ( flags & MRTF_HORZ_MASK ) == MRTF_CENTER ) MoveRectTo( rcFrame.Xm() - Cx()/2, top );
    else if( ( flags & MRTF_HORZ_MASK ) == MRTF_RIGHT  ) MoveRectTo( rcFrame.right - Cx(),  top );

    if(      ( flags & MRTF_VERT_MASK ) == MRTF_TOP     ) MoveRectTo( left, rcFrame.top           );
    else if( ( flags & MRTF_VERT_MASK ) == MRTF_VCENTER ) MoveRectTo( left, rcFrame.Ym() - Cy()/2 );
    else if( ( flags & MRTF_VERT_MASK ) == MRTF_BOTTOM  ) MoveRectTo( left, rcFrame.bottom - Cy() );

    return *this;
}

// Moves the rect that way that it is *completely* inside of 'rcFrame'.
// Note: It will only be moved if it is outside.
// Note: Only that coordinates will be moved which have been defined in 'flags'
//  : flags = combination of RSIFF_* values
// Note: If the rect is larger than the frame-rect, then it will be position
//  onto the top and/or left coord of the frame-rect.
_Rect_& _Rect_::RectStayInFrame( RECT rcFrame, DWORD flags )
{
    if( flags & RSIFF_X ) {
        if( right > rcFrame.right ) MoveRectTo( rcFrame.right - Cx(), top );
        if( left  < rcFrame.left  ) MoveRectTo( rcFrame.left,         top );
    }

    if( flags & RSIFF_Y ) {
        if( bottom > rcFrame.bottom ) MoveRectTo( left, rcFrame.bottom - Cy() );
        if( top    < rcFrame.top    ) MoveRectTo( left, rcFrame.top           );
    }

    return *this;
}

// Multiplies each coord of the rect
_Rect_& _Rect_::ZoomRectF( double xscale, double yscale )
{
    left   = (LONG)privateRect::round( left   * xscale );
    top    = (LONG)privateRect::round( top    * yscale );
    right  = (LONG)privateRect::round( right  * xscale );
    bottom = (LONG)privateRect::round( bottom * yscale );
    return *this;
}

// Divides each coord of the rect
_Rect_& _Rect_::DivRectF( double xinvscale, double yinvscale )
{
    if( xinvscale != 0.0 && yinvscale != 0.0 ) {
        left   = (LONG)privateRect::round( left   / xinvscale );
        top    = (LONG)privateRect::round( top    / yinvscale );
        right  = (LONG)privateRect::round( right  / xinvscale );
        bottom = (LONG)privateRect::round( bottom / yinvscale );
    }
    return *this;
}

// Multiplies only the size (width and height) of the rect
_Rect_& _Rect_::ZoomSizeF( double xscale, double yscale )
{
    right  = left + (LONG)privateRect::round( (right - left) * xscale );
    bottom = top  + (LONG)privateRect::round( (bottom - top) * yscale );
    return *this;
}

// Divides only the size (width and height) of the rect
_Rect_& _Rect_::DivSizeF( double xinvscale, double yinvscale )
{
    if( xinvscale != 0.0 && yinvscale != 0.0 ) {
        right  = left + (LONG)privateRect::round( (right - left) / xinvscale );
        bottom = top  + (LONG)privateRect::round( (bottom - top) / yinvscale );
    }
    return *this;
}

///<param name="sizOfAspectRatio"> the size from which to take the aspect ratio </param>
///<param name="option"> a SAR_* value </param>
///<remarks> In any case, the calculated rect will keep its top-left coordinate unchanged. </remarks>
_Rect_& _Rect_::SetAspectRatio( SIZE sizOfAspectRatio, DWORD option )
{
    _Size_ siz = Size();
    siz.SetAspectRatio( sizOfAspectRatio, option );
    SetRect( GetTopLeft(), siz );
    return *this;
}

///<param name="aspect_ratio"> the aspect ratio to set (width divided by height) </param>
///<param name="option"> a SAR_* value </param>
///<remarks> In any case, the calculated rect will keep its top-left coordinate unchanged. </remarks>
_Rect_& _Rect_::SetAspectRatio( double aspect_ratio, DWORD option )
{
    _Size_ siz = Size();
    siz.SetAspectRatio( aspect_ratio, option );
    SetRect( GetTopLeft(), siz );
    return *this;
}

// Mirrors the rect on a point 'pt'.
_Rect_ _Rect_::MirrorRect( POINT pt ) const
{
    _Rect_ rc;
    pt.x += pt.x; pt.y += pt.y; // double the 'pt'
    rc.left   = pt.x - right;
    rc.top    = pt.y - bottom;
    rc.right  = pt.x - left;
    rc.bottom = pt.y - top;
    return rc;
}

// Mirrors the rect on a *horizontal* axes
// whose y-coord is 'y'.
_Rect_ _Rect_::MirrorOnHorizontalAxis( int y ) const
{
    _Rect_ rc;
    y += y; // double the coord
    rc.left   = left;
    rc.top    = y - bottom;
    rc.right  = right;
    rc.bottom = y - top;
    return rc;
}

// Mirrors the rect on a *vertical* axes
// whose x-coord is 'x'.
_Rect_ _Rect_::MirrorOnVerticalAxis( int x ) const
{
    _Rect_ rc;
    x += x; // double the coord
    rc.left   = x - right;
    rc.top    = top;
    rc.right  = x - left;
    rc.bottom = bottom;
    return rc;
}



// The rect will be devided into 4 rects and
// they will be equally sized if possilbe.
//
//      +-------+-------+
//      |       |       |
//      | prcTL | prcTR |
//      |       |       |
//      +-------+-------+
//      |       |       |
//      | prcBL | prcBR |
//      |       |       |
//      +-------+-------+
//
_Rect_& _Rect_::QuarterizeRect( PRECT prcTL, PRECT prcTR, PRECT prcBL, PRECT prcBR )
{
    prcTL->left  = prcBL->left  = left;
    prcTR->left  = prcBR->left  =
    prcTL->right = prcBL->right = Xm();
    prcTR->right = prcBR->right = right;

    prcTL->top    = prcTR->top    = top;
    prcBL->top    = prcBR->top    =
    prcTL->bottom = prcTR->bottom = Ym();
    prcBL->bottom = prcBR->bottom = bottom;

    return *this;
}



bool _Rect_::AreAttachedRects( PCRECT prc1, PCRECT prc2 )
{
    return AreHorizontallyAttachedRects(prc1, prc2) ? true : AreVerticallyAttachedRects(prc1, prc2);
}

bool _Rect_::AreHorizontallyAttachedRects( PCRECT prc1, PCRECT prc2 )
{
    if( prc1->top    != prc2->top    ||
        prc1->bottom != prc2->bottom ) return false;
    return prc1->left  == prc2->right ||
           prc1->right == prc2->left;
}

bool _Rect_::AreVerticallyAttachedRects( PCRECT prc1, PCRECT prc2 )
{
    if( prc1->left  != prc2->left  ||
        prc1->right != prc2->right ) return false;
    return prc1->top    == prc2->bottom ||
           prc1->bottom == prc2->top;
}



// Note: This method automatically makes internal copies
//  of the passed rect(s) if the passed rect(s) are
//  'this' themselves.
//  This I did, because I don't know, if '::IntersectRect()',
//  '::UnionRect()' etc. may have same destination- and
//  source-rect pointers. For savety, I prevent this.
bool _Rect_::IntersectRect( PCRECT prc )
{
    if( prc == this ) return IsRectEmpty();
    RECT rctmp = *this;
    return FALSE != ::IntersectRect( this, &rctmp, prc );
}

// Note: This method automatically makes internal copies
//  of the passed rect(s) if the passed rect(s) are
//  'this' themselves.
//  This I did, because I don't know, if '::IntersectRect()',
//  '::UnionRect()' etc. may have same destination- and
//  source-rect pointers. For savety, I prevent this.
bool _Rect_::IntersectRect( PCRECT prc1, PCRECT prc2 )
{
    RECT rctmp1, rctmp2;
    if( prc1 == this ) { rctmp1 = *prc1; prc1 = &rctmp1; }
    if( prc2 == this ) { rctmp2 = *prc2; prc2 = &rctmp2; }
    return FALSE != ::IntersectRect( this, prc1, prc2 );
}

///<summary> This is similar to <b>IntersectRect()</b>, but
///  with the difference that empty rectangles (as input and
///  as output) are handled as if for non-empty rectangles,
///  and an empty result rect's coords not be set to zero.
///  They will then be the really *calculated* coords. </summary>
_Rect_& _Rect_::IntersectRectEmulRet( PCRECT prc1, PCRECT prc2 )
{
    _Rect_ rc1 = prc1 ? *prc1 : _Rect_(0, 0, 0, 0);
    _Rect_ rc2 = prc2 ? *prc2 : _Rect_(0, 0, 0, 0);

    left   = privateRect::tmax( rc1.left,   rc2.left   );
    right  = privateRect::tmin( rc1.right,  rc2.right  );
    top    = privateRect::tmax( rc1.top,    rc2.top    );
    bottom = privateRect::tmin( rc1.bottom, rc2.bottom );

    return *this;
}

// Note: This method automatically makes internal copies
//  of the passed rect(s) if the passed rect(s) are
//  'this' themselves.
//  This I did, because I don't know, if '::IntersectRect()',
//  '::UnionRect()' etc. may have same destination- and
//  source-rect pointers. For savety, I prevent this.
bool _Rect_::UnionRect( PCRECT prc )
{
    if( prc == this ) return IsRectEmpty();
    RECT rctmp = *this;
    return FALSE != ::UnionRect( this, &rctmp, prc );
}

// Note: This method automatically makes internal copies
//  of the passed rect(s) if the passed rect(s) are
//  'this' themselves.
//  This I did, because I don't know, if '::IntersectRect()',
//  '::UnionRect()' etc. may have same destination- and
//  source-rect pointers. For savety, I prevent this.
bool _Rect_::UnionRect( PCRECT prc1, PCRECT prc2 )
{
    RECT rctmp1, rctmp2;
    if( prc1 == this ) { rctmp1 = *prc1; prc1 = &rctmp1; }
    if( prc2 == this ) { rctmp2 = *prc2; prc2 = &rctmp2; }
    return FALSE != ::UnionRect( this, prc1, prc2 );
}

///<summary> This is similar to <b>UnionRect()</b>, but
///  with the difference that empty rectangles (as input and
///  as output) are handled as if for non-empty rectangles,
///  and an empty result rect's coords not be set to zero.
///  They will then be the really *calculated* coords. </summary>
_Rect_& _Rect_::UnionRectEmulRet( PCRECT prc1, PCRECT prc2 )
{
    _Rect_ rc1 = prc1 ? *prc1 : _Rect_(0, 0, 0, 0);
    _Rect_ rc2 = prc2 ? *prc2 : _Rect_(0, 0, 0, 0);

    left   = privateRect::tmin( rc1.left,   rc2.left   );
    right  = privateRect::tmax( rc1.right,  rc2.right  );
    top    = privateRect::tmin( rc1.top,    rc2.top    );
    bottom = privateRect::tmax( rc1.bottom, rc2.bottom );

    return *this;
}

// Note: This method automatically makes internal copies
//  of the passed rect(s) if the passed rect(s) are
//  'this' themselves.
//  This I did, because I don't know, if '::IntersectRect()',
//  '::UnionRect()' etc. may have same destination- and
//  source-rect pointers. For savety, I prevent this.
bool _Rect_::SubtractRect( PCRECT prc1, PCRECT prc2 )
{
    RECT rctmp1, rctmp2;
    if( prc1 == this ) { rctmp1 = *prc1; prc1 = &rctmp1; }
    if( prc2 == this ) { rctmp2 = *prc2; prc2 = &rctmp2; }
    return FALSE != ::SubtractRect( this, prc1, prc2 );
}



_Rect_ _Rect_::MulDiv( int nMultiplier, int nDivisor ) const
{
    return _Rect_( ::MulDiv( left,   nMultiplier, nDivisor ),
                   ::MulDiv( top,    nMultiplier, nDivisor ),
                   ::MulDiv( right,  nMultiplier, nDivisor ),
                   ::MulDiv( bottom, nMultiplier, nDivisor ) );
}



// Get area not obscured by tray and desktop toolbars (on the *primary* monitor only)
_Rect_& _Rect_::GetWorkArea()
{
    if( !SystemParametersInfo( SPI_GETWORKAREA, 0, *this, 0 ) )
        SetRectEmpty();
    return *this;
}

_Rect_& _Rect_::GetMonitorWorkAreaFromMon( HMONITOR hmon )
{
    MONITORINFO mi;
    memset( &mi, 0, sizeof(mi) );
    mi.cbSize = sizeof(mi);

    if( !GetMonitorInfo( hmon, &mi ) )
        return GetWorkArea();

    *this = mi.rcWork;

    return *this;
}

// Get screen size of primary display monitor.
_Rect_& _Rect_::GetPrimaryScreen()
{
    left   = 0;
    top    = 0;
    right  = GetSystemMetrics( SM_CXSCREEN );
    bottom = GetSystemMetrics( SM_CYSCREEN );
    return *this;
}

// Get the bounding rect of all display monitors.
_Rect_& _Rect_::GetVirtualScreen()
{
    int cx = GetSystemMetrics( SM_CXVIRTUALSCREEN );
    if( cx == 0 ) return GetPrimaryScreen();
    left   = GetSystemMetrics( SM_XVIRTUALSCREEN );
    top    = GetSystemMetrics( SM_YVIRTUALSCREEN );
    right  = left + cx;
    bottom = top  + GetSystemMetrics( SM_CYVIRTUALSCREEN );
    return *this;
}

_Rect_& _Rect_::GetMonitorRectFromMon( HMONITOR hmon )
{
    MONITORINFO mi;
    memset( &mi, 0, sizeof(mi) );
    mi.cbSize = sizeof(mi);

    if( !GetMonitorInfo( hmon, &mi ) )
        return GetPrimaryScreen();

    *this = mi.rcMonitor;

    return *this;
}



bool _Rect_::SetWindowRect( HWND hwnd, HWND hwndInsertAfter, UINT flags ) const
{
    if( !hwnd ) return false;
    return ::SetWindowPos( hwnd, hwndInsertAfter, left, top, Cx(), Cy(),
                           flags ) != FALSE;
}

// Sets only a new position and size for a window
void _Rect_::SetWindowRect( HWND hwnd ) const
{
    if( hwnd )
        ::SetWindowPos( hwnd, HWND_TOP, left, top, Cx(), Cy(),
                        SWP_NOZORDER | SWP_NOACTIVATE );
}

// Sets only a new position for a window (does not change its size)
void _Rect_::SetWindowTopLeft( HWND hwnd ) const
{
    if( hwnd )
        ::SetWindowPos( hwnd, HWND_TOP, left, top, 0, 0,
                        SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}

// Sets only a new size for a window (does not change its position)
void _Rect_::SetWindowSize( HWND hwnd ) const
{
    if( hwnd )
        ::SetWindowPos( hwnd, HWND_TOP, 0, 0, Cx(), Cy(),
                        SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );
}

// Gets the rect of the 'hwnd' but relative to the client coordinates of
// the 'hwndOfClient' window.
_Rect_& _Rect_::GetWindowRectInClient( HWND hwnd, HWND hwndOfClient )
{
    if( !hwnd ) return *this;

    if( !hwndOfClient )
        return GetWindowRect( hwnd );

    if( !::GetWindowRect( hwnd, this ) )
        return SetRectEmpty();

    ::MapWindowPoints( NULL, hwndOfClient, (PPOINT)this, 2 );

    return *this;
}

bool _Rect_::AdjustWindowRectEx( HWND hwnd, bool bIncludeScrollbars )
{
    if( !hwnd ) return false;

    #ifndef _WIN32_WCE

    DWORD style   = (DWORD) GetWindowLongPtr( hwnd, GWL_STYLE   );
    DWORD exstyle = (DWORD) GetWindowLongPtr( hwnd, GWL_EXSTYLE );
    HMENU hmenu   = GetMenu( hwnd );

    // Note: For some reason, #GetMenu()# on a child window can return
    // something non-NULL even if the parent window (or dialog) does
    // not have any menu (and a child window of course does not).
    if( style & WS_CHILD ) hmenu = NULL;

    MENUBARINFO mbi;
    if( hmenu )
    {
        memset( &mbi, 0, sizeof(mbi) );
        mbi.cbSize = sizeof(mbi);
        if( !GetMenuBarInfo( hwnd, OBJID_MENU, 0, &mbi ) )
            memset( &mbi.rcBar, 0, sizeof(mbi.rcBar) );
    }

    return AdjustWindowRectEx( style, hmenu ? &mbi : (MENUBARINFO*) NULL, exstyle, bIncludeScrollbars );

    #else

    DWORD style   = (DWORD) GetWindowLong( hwnd, GWL_STYLE   );
    DWORD exstyle = (DWORD) GetWindowLong( hwnd, GWL_EXSTYLE );

    return AdjustWindowRectEx( style, false, exstyle, bIncludeScrollbars );

    #endif
}

bool _Rect_::AdjustWindowRectEx( DWORD dwStyle, bool bMenu, DWORD dwExStyle, bool bIncludeScrollbars )
{
    #ifndef WS_EX_LAYOUTRTL
    #define WS_EX_LAYOUTRTL         0x00400000L // Right to left mirroring
    #endif

    bool ret = FALSE != ::AdjustWindowRectEx( this, dwStyle, bMenu, dwExStyle );

    if( bIncludeScrollbars && ret )
    {
        if( dwStyle & WS_VSCROLL ) {
            bool bLeftScrollbar = (dwExStyle & WS_EX_LAYOUTRTL) && (dwExStyle & WS_EX_LEFTSCROLLBAR);
            if( bLeftScrollbar ) left  -= GetSystemMetrics( SM_CXVSCROLL );
            else                 right += GetSystemMetrics( SM_CXVSCROLL );
        }

        if( dwStyle & WS_HSCROLL )
            bottom += GetSystemMetrics( SM_CYHSCROLL );
    }

    return ret;
}

#ifndef _WIN32_WCE
///<param name="pmbiMenubar"> if NULL, no menu will be included in the calculations. <br/>
///  If not NULL, then the <b>pmbiMenubar->rcBar</b>'s height only will be used. It will
///  then be used to retrieve the height of the menubar that is involved. </param>
///<remarks>
///  This call is different to the one where the parameter <b>bool bMenu</b>
///  does exist instead of the <b>pmbiMenubar</b> parameter here. The difference is,
///  that in here, any menu height can be used while the one with <b>bMenu</b>
///  takes a menu height of one row. The arbitrary menu height is necessary for
///  windows whose width is so small that their menus wrap into more than only
///  one row of height.
///</remarks>
bool _Rect_::AdjustWindowRectEx( DWORD dwStyle, const MENUBARINFO *pmbiMenubar, DWORD dwExStyle, bool bIncludeScrollbars )
{
    bool ret = FALSE != ::AdjustWindowRectEx( this, dwStyle, FALSE, dwExStyle );

    if( pmbiMenubar && ret )
    {
        top -= (pmbiMenubar->rcBar.bottom - pmbiMenubar->rcBar.top) + 1;    // +1 because the system adds a pixelline as menu separator
    }

    if( bIncludeScrollbars && ret )
    {
        if( dwStyle & WS_VSCROLL ) {
            bool bLeftScrollbar = (dwExStyle & WS_EX_LAYOUTRTL) && (dwExStyle & WS_EX_LEFTSCROLLBAR);
            if( bLeftScrollbar ) left  -= GetSystemMetrics( SM_CXVSCROLL );
            else                 right += GetSystemMetrics( SM_CXVSCROLL );
        }

        if( dwStyle & WS_HSCROLL )
            bottom += GetSystemMetrics( SM_CYHSCROLL );
    }

    return ret;
}
#endif

bool _Rect_::ReverseAdjustWindowRectEx( HWND hwnd, bool bIncludeScrollbars )
{
    _Rect_ rect(0, 0, 0, 0);
    if( !rect.AdjustWindowRectEx( hwnd, bIncludeScrollbars ) )
        return false;

    left   -= rect.left;
    top    -= rect.top;
    right  -= rect.right;
    bottom -= rect.bottom;

    return true;
}

bool _Rect_::ReverseAdjustWindowRectEx( DWORD dwStyle, bool bMenu, DWORD dwExStyle, bool bIncludeScrollbars )
{
    _Rect_ rect(0, 0, 0, 0);
    if( !rect.AdjustWindowRectEx( dwStyle, bMenu, dwExStyle, bIncludeScrollbars ) )
        return false;

    left   -= rect.left;
    top    -= rect.top;
    right  -= rect.right;
    bottom -= rect.bottom;

    return true;
}

#ifndef _WIN32_WCE
///<param name="pmbiMenubar"> if NULL, no menu will be included in the calculations. <br/>
///  If not NULL, then the <b>pmbiMenubar->rcBar</b>'s height only will be used. It will
///  then be used to retrieve the height of the menubar that is involved. </param>
bool _Rect_::ReverseAdjustWindowRectEx( DWORD dwStyle, const MENUBARINFO *pmbiMenubar, DWORD dwExStyle, bool bIncludeScrollbars )
{
    _Rect_ rect(0, 0, 0, 0);
    if( !rect.AdjustWindowRectEx( dwStyle, pmbiMenubar, dwExStyle, bIncludeScrollbars ) )
        return false;

    left   -= rect.left;
    top    -= rect.top;
    right  -= rect.right;
    bottom -= rect.bottom;

    return true;
}
#endif

HDWP _Rect_::DeferWindowRect( HDWP hdefer, HWND hwnd, HWND hwndInsertAfter, UINT flags ) const
{
    if( !hdefer || !hwnd ) return NULL;
    return ::DeferWindowPos( hdefer, hwnd, hwndInsertAfter, left, top, Cx(), Cy(),
                             flags );
}

// Defers only a new position and size for a window (using #DeferWindowPos()#)
HDWP _Rect_::DeferWindowRect( HDWP hdefer, HWND hwnd ) const
{
    if( !hdefer || !hwnd ) return NULL;
    return ::DeferWindowPos( hdefer, hwnd, HWND_TOP, left, top, Cx(), Cy(),
                             SWP_NOZORDER | SWP_NOACTIVATE );
}

// Defers only a new position for a window (does not change its size) (using #DeferWindowPos()#)
HDWP _Rect_::DeferWindowTopLeft( HDWP hdefer, HWND hwnd ) const
{
    if( !hdefer || !hwnd ) return NULL;
    return ::DeferWindowPos( hdefer, hwnd, HWND_TOP, left, top, 0, 0,
                             SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}

// Defers only a new size for a window (does not change its position) (using #DeferWindowPos()#)
HDWP _Rect_::DeferWindowSize( HDWP hdefer, HWND hwnd ) const
{
    if( !hdefer || !hwnd ) return NULL;
    return ::DeferWindowPos( hdefer, hwnd, HWND_TOP, 0, 0, Cx(), Cy(),
                             SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );
}

void _Rect_::CenterWindowOverWindow( HWND hwndToBeCentered, HWND hwndReferenceWindow )
{
    if( !hwndToBeCentered ) return;

    _Rect_ rectRef;
    if( hwndReferenceWindow ) rectRef.GetWindowRect( hwndReferenceWindow );
    else rectRef.GetWorkArea();

    _Rect_ rect;
    rect.GetWindowRect( hwndToBeCentered );
    rect.MoveRectCCTo( rectRef.GetCenter() );
    rect.SetWindowTopLeft( hwndToBeCentered );
}

void _Rect_::CenterWindowOverItsParent( HWND hwndToBeCentered )
{
    CenterWindowOverWindow( hwndToBeCentered, GetParent(hwndToBeCentered) );
}



_Rect_& _Rect_::MapSystemDialogCoords( bool bBackward )
{
    long baseunits = GetDialogBaseUnits();
    int baseunitX = (int) LOWORD( baseunits );
    int baseunitY = (int) HIWORD( baseunits );

    if( bBackward )
    {
        left   = ::MulDiv( (int) left,   4, baseunitX ? baseunitX : 1 );
        top    = ::MulDiv( (int) top,    8, baseunitY ? baseunitY : 1 );
        right  = ::MulDiv( (int) right,  4, baseunitX ? baseunitX : 1 );
        bottom = ::MulDiv( (int) bottom, 8, baseunitY ? baseunitY : 1 );
    }
    else
    {
        left   = ::MulDiv( (int) left,   baseunitX, 4 );
        top    = ::MulDiv( (int) top,    baseunitY, 8 );
        right  = ::MulDiv( (int) right,  baseunitX, 4 );
        bottom = ::MulDiv( (int) bottom, baseunitY, 8 );
    }

    return *this;
}

bool _Rect_::UnmapDialogRect( HWND hwndDlg )
{
    if( !hwndDlg ) return false;

    _Rect_ orig = *this;

    _Rect_ rect( 0, 0, 1024, 1024 );
    try { if( !::MapDialogRect( hwndDlg, rect ) ) throw 0; }
    catch(...) { return false; }
    // Note: We must encapsulate #::MapDialogRect()# in exception handling
    // because if the HWND passed was not created using a dialogbox creation
    // function, it could happen, that #::MapDialogRect()# throws exception
    // (or even worse).
    // This can happen especially if the HWND is from another process.
    // But only enacapsulating #::MapDialogRect()# does not work because the
    // compiler will optimize away the try and catch statements (if
    // we #return false# instead of #throw 0#).
    // But throwing an exception ourselves when #::MapDialogRect()# returns
    // FALSE, the compiler cannot optimize away the exception handling.

    double fx = (double) rect.Cx() / 1024.0;
    double fy = (double) rect.Cy() / 1024.0;
    if( fx == 0.0 ) fx = 1.0;
    if( fy == 0.0 ) fy = 1.0;

    left   = (long) privateRect::round( orig.left / fx );
    top    = (long) privateRect::round( orig.top  / fy );
    right  = left + (long) privateRect::round( orig.Cx() / fx );
    bottom = top  + (long) privateRect::round( orig.Cy() / fy );

    return true;
}




}
