/*
    File:               argonRect.h
    Description:        Point, Size and Rectangle classes.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         23.11.2003
    last updated:       03.11.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

/*
    IMPORTANT:
    To be able to link correctly your project, copy the following 2 lines
    of code into *one* of your .cpp files (recommended: near the bottom of "stdafx.cpp").
        #define COMPILE_MULTIMON_STUBS
        #include <multimon.h>
    They will cause the stubs of the "multimon.h" multi monitor support
    to compile.
*/

#ifndef __argonRect_h
#define __argonRect_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #ifndef _WIN32_WCE
    #include <multimon.h>
    #endif
    #include <math.h>
    #include <tchar.h>
#endif

#ifndef _argon_no_warning_disables
    #pragma warning( disable: 4239 )
    // Disables a minor warning of nonstandard extension usage. This warning
    // normally occurs when passing a temporary object (no variable)
    // as parameter to a reference to the class.
    // Since in VC++ this construction *is* allowed as an extension and
    // we often use microsoft specific constellations, it is not harmful
    // at all to also allow using this microsoft specific feature
    // since it increases code writing and readability.

    #pragma warning( disable: 4706 )    // assignment within conditional expression

    #pragma warning( disable: 4100 )    // unreferenced parameter

    // The following warning disabling will prevent the compiler from
    // yawning floating point overflow errors (e.g. helpful when
    // using DirectX).
    #pragma warning( disable: 4056 )    // overflow in floating point constant arithmetic
    #pragma warning( disable: 4756 )    // overflow in constant arithmetic

    #pragma warning( disable: 4127 )    // conditional expression is constant
    // This would occur e.g. in something like  if( sizeof(TCHAR) == sizeof(CHAR) )

    #pragma warning( disable: 4201 )    // nonstandard extension used : nameless struct/union
    // Will be necessary e.g. when using   #include <mmsystem.h>
#endif

namespace ArgonLibrary
{




typedef const RECT *PCRECT;

namespace privateRect
{
    template< class T > void tswap( T& a, T& b ) { T h = a; a = b; b = h; }
    template< class T > T tmin( T a, T b ) { return ( a <= b )? a : b; }
    template< class T > T tmax( T a, T b ) { return ( a >= b )? a : b; }

    inline double round( double x ) { return floor(x + 0.5); }
}

class _Pt_;
class _Size_;
class _Rect_;

//class _Pt_ : public POINT
class _Pt_ : public tagPOINT
{
    public :
        _Pt_() {  } // randomly filled
        _Pt_( int _x, int _y ) { x = _x; y = _y; }
        _Pt_( POINT      _pt ) { * (POINT*) this = _pt; }
        _Pt_( SIZE     _size ) { * (POINT*) this = * (POINT*) &_size; }
        _Pt_( DWORD       dw ) { x = (SHORT) LOWORD(dw); y = (SHORT) HIWORD(dw); }
        _Pt_( const WINDOWPOS *pwp ) { x = pwp ? pwp->x : 0; y = pwp ? pwp->y : 0; }

        _Pt_& Set     ( int _x, int _y ) { x = _x; y = _y; return *this; }
        _Pt_& SetX    ( int _x ) { x = _x; return *this; }
        _Pt_& SetY    ( int _y ) { y = _y; return *this; }
        _Pt_& SetEmpty()         { x = y = 0; return *this; }
        _Pt_& Offset( int xofs, int yofs ) { x += xofs;    y += yofs;    return *this; }
        _Pt_& Offset( POINT  pt )          { x += pt.x;    y += pt.y;    return *this; }
        _Pt_& Offset( SIZE size )          { x += size.cx; y += size.cy; return *this; }
        _Pt_& Scale ( int    xscale, int    yscale ) { x *= xscale;   y *= yscale;   return *this; }
        _Pt_& Scale ( SIZE    scale )                { x *= scale.cx; y *= scale.cy; return *this; }
        _Pt_& ScaleF( double xscale, double yscale ) { x = (LONG)(x * xscale); y = (LONG)(y * yscale); return *this; }
        _Pt_& Resize ( int    size ) { x *= size; y *= size; return *this; }
        _Pt_& ResizeF( double size ) { x = (LONG)privateRect::round(x * size); y = (LONG)privateRect::round(y * size); return *this; }
        _Pt_& Divide ( int    invsize ) { if( invsize != 0   ) { x /= invsize; y /= invsize; } return *this; }
        _Pt_& DivideF( double invsize ) { if( invsize != 0.0 ) { x = (LONG)privateRect::round(x / invsize); y = (LONG)privateRect::round(y / invsize); } return *this; }
        _Pt_& StayInRect      ( RECT rc );
        _Pt_& StayInPseudoRect( RECT rc );

        operator PPOINT() { return this; }

        bool IsZero() const { return( x == 0 && y == 0 ); }

        _Size_&       GetAsSize();
        const _Size_& GetAsSize() const;
        _Rect_ GetEmptyRect () const; // top-left and bottom-right point will be set to 'this' position
        _Rect_ GetOneDotRect() const; // top-left will be set to 'this' position and the rect's size will be 1 pixel in each direction

        _Pt_& GetTL( RECT rc ) { x = rc.left;  y = rc.top;    return *this; }
        _Pt_& GetTR( RECT rc ) { x = rc.right; y = rc.top;    return *this; }
        _Pt_& GetBL( RECT rc ) { x = rc.left;  y = rc.bottom; return *this; }
        _Pt_& GetBR( RECT rc ) { x = rc.right; y = rc.bottom; return *this; }

        _Pt_& Absolute() { x = labs(x); y = labs(y); return *this; }
        _Pt_& Negate  () { x = -x; y = -y; return *this; }
        _Pt_& Flip    () { privateRect::tswap( x, y ); return *this; }

        bool operator ==( POINT pt ) const { return( x == pt.x && y == pt.y ); }
        bool operator !=( POINT pt ) const { return( x != pt.x || y != pt.y ); }
        _Pt_& operator +=( SIZE       size ) { x += size.cx; y += size.cy; return *this; }
        _Pt_& operator -=( SIZE       size ) { x -= size.cx; y -= size.cy; return *this; }
        _Pt_& operator +=( POINT        pt ) { x += pt.x;    y += pt.y;    return *this; }
        _Pt_& operator -=( POINT        pt ) { x -= pt.x;    y -= pt.y;    return *this; }
        _Pt_& operator *=( int       scale ) { x *= scale;   y *= scale;   return *this; }
        _Pt_& operator *=( float     scale ) { x = (LONG)privateRect::round(x * scale); y = (LONG)privateRect::round(y * scale); return *this; }
        _Pt_& operator *=( double    scale ) { x = (LONG)privateRect::round(x * scale); y = (LONG)privateRect::round(y * scale); return *this; }
        _Pt_& operator /=( int    invscale ) { if( invscale != 0    ) { x /= invscale; y /= invscale; } return *this; }
        _Pt_& operator /=( float  invscale ) { if( invscale != 0.0f ) { x = (LONG)privateRect::round(x / invscale); y = (LONG)privateRect::round(y / invscale); } return *this; }
        _Pt_& operator /=( double invscale ) { if( invscale != 0.0  ) { x = (LONG)privateRect::round(x / invscale); y = (LONG)privateRect::round(y / invscale); } return *this; }

        _Pt_ operator +( SIZE       size ) const { return _Pt_( x + size.cx, y + size.cy ); }
        _Pt_ operator -( SIZE       size ) const { return _Pt_( x - size.cx, y - size.cy ); }
        _Pt_ operator -()                  const { return _Pt_( -x, -y ); }
        _Pt_ operator +( POINT        pt ) const { return _Pt_( x + pt.x,  y + pt.y  ); }
        _Pt_ operator *( int       scale ) const { return _Pt_( x * scale, y * scale ); }
        _Pt_ operator *( float     scale ) const { return _Pt_( (int)privateRect::round(x * scale), (int)privateRect::round(y * scale) ); }
        _Pt_ operator *( double    scale ) const { return _Pt_( (int)privateRect::round(x * scale), (int)privateRect::round(y * scale) ); }
        _Pt_ operator /( int    invscale ) const { if( invscale != 0    ) return _Pt_( x / invscale, y / invscale ); else return *this; }
        _Pt_ operator /( float  invscale ) const { if( invscale != 0.0f ) return _Pt_( (int)privateRect::round(x / invscale), (int)privateRect::round(y / invscale) ); else return *this; }
        _Pt_ operator /( double invscale ) const { if( invscale != 0.0  ) return _Pt_( (int)privateRect::round(x / invscale), (int)privateRect::round(y / invscale) ); else return *this; }

        _Size_ operator -( POINT pt ) const;

        _Rect_ operator +( PCRECT prc ) const;
        _Rect_ operator -( PCRECT prc ) const;

        void SetWindowPos ( HWND hwnd ) const; // Sets only a new position for a window (does not change its size)
        _Pt_& GetCursorPos  ()            { ::GetCursorPos( this );         return *this; }
        _Pt_& GetMessagePos ()            { *this = ::GetMessagePos();      return *this; }
        _Pt_& ClientToScreen( HWND hwnd ) { ::ClientToScreen( hwnd, this ); return *this; }
        _Pt_& ScreenToClient( HWND hwnd ) { ::ScreenToClient( hwnd, this ); return *this; }
        _Pt_& MapWindowPoints( HWND hwndFrom, HWND hwndTo ) { ::MapWindowPoints( hwndFrom, hwndTo, this, 1 ); return *this; }

            ///<summary> Maps dialog coordinates to pixels (<b>bBackward</b> == false)
            ///  or pixels to dialog coordinates (<b>bBackward</b> == true). <br/>
            ///  This uses <b>GetDialogBaseUnits()</b> to convert the units, so using the
            ///  default system font and not taking a specific dialog's font into account. </summary>
        _Pt_& MapSystemDialogCoords( bool bBackward = false );
};

// SAR_* values for #option# of #_Size_::SetAspectRatio()# and #_Rect_::SetAspectRatio()#
#define SAR_KEEPWIDTH           1   // the width will not be changed, just the height will be adopted to reflect the desired aspect ratio
#define SAR_KEEPHEIGHT          2   // the height will not be changed, just the width will be adopted to reflect the desired aspect ratio
#define SAR_ENLARGETOSMALLEST   3   // either the width or the height will be enlarged, so that the final rect is the smallest bounding rect around the old rect
#define SAR_SHRINKTOBIGGEST     4   // either the width or the height will be made smaller, so that the final rect is the biggest rect inside the old rect

//class _Size_ : public SIZE
class _Size_ : public tagSIZE
{
    public :
        _Size_() {  } // random filled
        _Size_( int _cx, int _cy ) { cx = _cx; cy = _cy; }
        _Size_( SIZE       _size ) { * (SIZE*) this = _size; }
        _Size_( POINT        _pt ) { * (SIZE*) this = * (SIZE*) &_pt; }
        _Size_( DWORD         dw ) { cx = (SHORT) LOWORD(dw); cy = (SHORT) HIWORD(dw); }
        _Size_( RECT         _rc ) { cx = _rc.right - _rc.left; cy = _rc.bottom - _rc.left; }
        _Size_( const WINDOWPOS *pwp ) { cx = pwp ? pwp->cx : 0; cy = pwp ? pwp->cy : 0; }

        _Size_& Set     ( int _cx, int _cy ) { cx = _cx; cy = _cy; return *this; }
        _Size_& SetCx   ( int _cx ) { cx = _cx; return *this; }
        _Size_& SetCy   ( int _cy ) { cy = _cy; return *this; }
        _Size_& SetEmpty()          { cx = cy = 0; return *this; }
        _Size_& Scale ( int    xscale, int    yscale ) { cx *= xscale;   cy *= yscale;   return *this; }
        _Size_& Scale ( SIZE    scale )                { cx *= scale.cx; cy *= scale.cy; return *this; }
        _Size_& ScaleF( double xscale, double yscale ) { cx = (LONG)privateRect::round(cx * xscale); cy = (LONG)privateRect::round(cy * yscale); return *this; }
        _Size_& Resize ( int    size ) { cx *= size; cy *= size; return *this; }
        _Size_& ResizeF( double size ) { cx = (LONG)privateRect::round(cx * size); cy = (LONG)privateRect::round(cy * size); return *this; }
        _Size_& Divide ( int    invsize ) { if( invsize != 0   ) { cx /= invsize; cy /= invsize; } return *this; }
        _Size_& DivideF( double invsize ) { if( invsize != 0.0 ) { cx = (LONG)privateRect::round(cx / invsize); cy = (LONG)privateRect::round(cy / invsize); } return *this; }

        double GetAspectRatio() const { return cy != 0 ? (double) cx / (double) cy : 0.0; }
        _Size_& SetAspectRatio( SIZE sizOfAspectRatio, DWORD option );
        _Size_& SetAspectRatio( double aspect_ratio, DWORD option );

        operator PSIZE() { return this; }

        bool IsEmpty() const { return( cx <= 0 || cy <= 0 ); }
        bool IsNull () const { return( cx == 0 && cy == 0 ); }

        _Pt_&       GetAsPt()       { return *(_Pt_*)this; }
        const _Pt_& GetAsPt() const { return *(_Pt_*)this; }
        _Rect_ GetOriginRect() const; // top-left will be (0,0) and rect's size will be set to 'this' size
        _Pt_   GetCenter() const { return _Pt_( cx >> 1, cy >> 1 ); }

        _Size_& MaxSize( PSIZE psize2 ) { cx = privateRect::tmax( cx, psize2->cx ); cy = privateRect::tmax( cy, psize2->cy ); return *this; } // get the maximum of the sizes; only allow to make size greater. cx and cy may not be smaller than wanted
        _Size_& MinSize( PSIZE psize2 ) { cx = privateRect::tmin( cx, psize2->cx ); cy = privateRect::tmin( cy, psize2->cy ); return *this; } // get the minimum of the sizes; only allow to make size smaller. cx and cy may not be greater than wanted

        _Size_& Absolute() { cx = labs(cx); cy = labs(cy); return *this; }
        _Size_& Negate  () { cx = -cx; cy = -cy; return *this; }
        _Size_& Flip    () { privateRect::tswap( cx, cy ); return *this; }

        bool operator ==( SIZE size ) const { return( cx == size.cx && cy == size.cy ); }
        bool operator !=( SIZE size ) const { return( cx != size.cx || cy != size.cy ); }
        bool operator < ( SIZE size ) const { return( cx <  size.cx && cy <  size.cy ); } // only if both extents are smaller -> then the complete size is smaller
        bool operator > ( SIZE size ) const { return( cx >  size.cx && cy >  size.cy ); } // only if both extents are greater -> then the complete size is greater
        bool operator <=( SIZE size ) const { return( cx <= size.cx || cy <= size.cy ); } // if one extent is smaller or equal -> the complete size is smaller or equal
        bool operator >=( SIZE size ) const { return( cx >= size.cx || cy >= size.cy ); } // if one extent is greater or equal -> the complete size is greater or equal
        _Size_& operator +=( SIZE       size ) { cx += size.cx; cy += size.cy; return *this; }
        _Size_& operator -=( SIZE       size ) { cx -= size.cx; cy -= size.cy; return *this; }
        _Size_& operator *=( int       scale ) { cx *= scale;   cy *= scale;   return *this; }
        _Size_& operator *=( float     scale ) { cx = (LONG)privateRect::round(cx * scale); cy = (LONG)privateRect::round(cy * scale); return *this; }
        _Size_& operator *=( double    scale ) { cx = (LONG)privateRect::round(cx * scale); cy = (LONG)privateRect::round(cy * scale); return *this; }
        _Size_& operator /=( int    invscale ) { if( invscale != 0    ) { cx /= invscale; cy /= invscale; } return *this; }
        _Size_& operator /=( float  invscale ) { if( invscale != 0.0f ) { cx = (LONG)privateRect::round(cx / invscale); cy = (LONG)privateRect::round(cy / invscale); } return *this; }
        _Size_& operator /=( double invscale ) { if( invscale != 0.0  ) { cx = (LONG)privateRect::round(cx / invscale); cy = (LONG)privateRect::round(cy / invscale); } return *this; }

        _Size_ operator +( SIZE       size ) const { return _Size_( cx + size.cx, cy + size.cy ); }
        _Size_ operator -( SIZE       size ) const { return _Size_( cx - size.cx, cy - size.cy ); }
        _Size_ operator -()                  const { return _Size_( -cx, -cy ); }
        _Size_ operator *( int       scale ) const { return _Size_( cx * scale, cy * scale ); }
        _Size_ operator *( float     scale ) const { return _Size_( (int)privateRect::round(cx * scale), (int)privateRect::round(cy * scale) ); }
        _Size_ operator *( double    scale ) const { return _Size_( (int)privateRect::round(cx * scale), (int)privateRect::round(cy * scale) ); }
        _Size_ operator /( int    invscale ) const { if( invscale != 0    ) return _Size_( cx / invscale, cy / invscale ); else return *this; }
        _Size_ operator /( float  invscale ) const { if( invscale != 0.0f ) return _Size_( (int)privateRect::round(cx / invscale), (int)privateRect::round(cy / invscale) ); else return *this; }
        _Size_ operator /( double invscale ) const { if( invscale != 0.0  ) return _Size_( (int)privateRect::round(cx / invscale), (int)privateRect::round(cy / invscale) ); else return *this; }

        _Pt_ operator +( POINT pt ) const { return _Pt_( cx + pt.x, cy + pt.y ); }
        _Pt_ operator -( POINT pt ) const { return _Pt_( cx - pt.x, cy - pt.y ); }

        _Rect_ operator +( PCRECT prc ) const;

        _Size_& GetClientSize( HWND hwnd ) { RECT rc; ::GetClientRect(hwnd, &rc); cx = rc.right - rc.left; cy = rc.bottom - rc.top; return *this; }
        _Size_& GetWindowSize( HWND hwnd ) { RECT rc; ::GetWindowRect(hwnd, &rc); cx = rc.right - rc.left; cy = rc.bottom - rc.top; return *this; }
        void SetWindowSize ( HWND hwnd ) const; // Sets only a new size for a window (does not change its position)

            ///<summary> Maps dialog coordinates to pixels (<b>bBackward</b> == false)
            ///  or pixels to dialog coordinates (<b>bBackward</b> == true). <br/>
            ///  This uses <b>GetDialogBaseUnits()</b> to convert the units, so using the
            ///  default system font and not taking a specific dialog's font into account. </summary>
        _Size_& MapSystemDialogCoords( bool bBackward = false );

        // GetSystemMetrics made easier
        _Size_& GetSystemmetrics( int nIndexCX, int nIndexCY )
        { cx = ::GetSystemMetrics( nIndexCX ); cy = ::GetSystemMetrics( nIndexCY ); return *this; }
        _Size_& GetSysMtrxBorder       () { return GetSystemmetrics( SM_CXBORDER       , SM_CYBORDER        ); }
        #ifdef SM_CXCURSOR
        _Size_& GetSysMtrxCursor       () { return GetSystemmetrics( SM_CXCURSOR       , SM_CYCURSOR        ); }
        #endif
        _Size_& GetSysMtrxDlgFrame     () { return GetSystemmetrics( SM_CXDLGFRAME     , SM_CYDLGFRAME      ); }
        _Size_& GetSysMtrxDoubleClk    () { return GetSystemmetrics( SM_CXDOUBLECLK    , SM_CYDOUBLECLK     ); }
        #ifdef SM_CXDRAG
        _Size_& GetSysMtrxDrag         () { return GetSystemmetrics( SM_CXDRAG         , SM_CYDRAG          ); }
        #endif
        _Size_& GetSysMtrxEdge         () { return GetSystemmetrics( SM_CXEDGE         , SM_CYEDGE          ); }
        _Size_& GetSysMtrxFixedFrame   () { return GetSystemmetrics( SM_CXFIXEDFRAME   , SM_CYFIXEDFRAME    ); }
        #ifdef SM_CXFRAME
        _Size_& GetSysMtrxFrame        () { return GetSystemmetrics( SM_CXFRAME        , SM_CYFRAME         ); }
        #endif
        #ifdef SM_CXFULLSCREEN
        _Size_& GetSysMtrxFullscreen   () { return GetSystemmetrics( SM_CXFULLSCREEN   , SM_CYFULLSCREEN    ); }
        #endif
        _Size_& GetSysMtrxHScroll      () { return GetSystemmetrics( SM_CXHSCROLL      , SM_CYHSCROLL       ); }
        _Size_& GetSysMtrxIcon         () { return GetSystemmetrics( SM_CXICON         , SM_CYICON          ); }
        _Size_& GetSysMtrxIconSpacing  () { return GetSystemmetrics( SM_CXICONSPACING  , SM_CYICONSPACING   ); }
        #ifdef SM_CXMAXIMIZED
        _Size_& GetSysMtrxMaximized    () { return GetSystemmetrics( SM_CXMAXIMIZED    , SM_CYMAXIMIZED     ); }
        #endif
        #ifdef SM_CXMAXTRACK
        _Size_& GetSysMtrxMaxTrack     () { return GetSystemmetrics( SM_CXMAXTRACK     , SM_CYMAXTRACK      ); }
        #endif
        #ifdef SM_CXMENUCHECK
        _Size_& GetSysMtrxMenuCheck    () { return GetSystemmetrics( SM_CXMENUCHECK    , SM_CYMENUCHECK     ); }
        #endif
        #ifdef SM_CXMENUSIZE
        _Size_& GetSysMtrxMenuSize     () { return GetSystemmetrics( SM_CXMENUSIZE     , SM_CYMENUSIZE      ); }
        #endif
        #ifdef SM_CXMIN
        _Size_& GetSysMtrxMin          () { return GetSystemmetrics( SM_CXMIN          , SM_CYMIN           ); }
        #endif
        #ifdef SM_CXMINIMIZED
        _Size_& GetSysMtrxMinimized    () { return GetSystemmetrics( SM_CXMINIMIZED    , SM_CYMINIMIZED     ); }
        #endif
        #ifdef SM_CXMINSPACING
        _Size_& GetSysMtrxMinSpacing   () { return GetSystemmetrics( SM_CXMINSPACING   , SM_CYMINSPACING    ); }
        #endif
        #ifdef SM_CXMINTRACK
        _Size_& GetSysMtrxMinTrack     () { return GetSystemmetrics( SM_CXMINTRACK     , SM_CYMINTRACK      ); }
        #endif
        _Size_& GetSysMtrxScreen       () { return GetSystemmetrics( SM_CXSCREEN       , SM_CYSCREEN        ); }
        #ifdef SM_CXSIZE
        _Size_& GetSysMtrxSize         () { return GetSystemmetrics( SM_CXSIZE         , SM_CYSIZE          ); }
        #endif
        #ifdef SM_CXSIZEFRAME
        _Size_& GetSysMtrxSizeFrame    () { return GetSystemmetrics( SM_CXSIZEFRAME    , SM_CYSIZEFRAME     ); }
        #endif
        _Size_& GetSysMtrxSmIcon       () { return GetSystemmetrics( SM_CXSMICON       , SM_CYSMICON        ); }
        #ifdef SM_CXSMSIZE
        _Size_& GetSysMtrxSmSize       () { return GetSystemmetrics( SM_CXSMSIZE       , SM_CYSMSIZE        ); }
        #endif
        _Size_& GetSysMtrxVirtualScreen() { return GetSystemmetrics( SM_CXVIRTUALSCREEN, SM_CYVIRTUALSCREEN ); }
        _Size_& GetSysMtrxVScroll      () { return GetSystemmetrics( SM_CXVSCROLL      , SM_CYVSCROLL       ); }
        // and additional GetSystemMetrics helper methods
        #ifdef SM_CXMENUSIZE
        _Size_& GetSysMtrxMenubarButtonSize   () { return GetSysMtrxMenuSize(); }
        #endif
        #ifdef SM_CXSIZE
        _Size_& GetSysMtrxCaptionButton       () { return GetSysMtrxSize(); }
        _Size_& GetSysMtrxTitlebarButton      () { return GetSysMtrxSize(); }
        #endif
        #ifdef SM_CXSMSIZE
        _Size_& GetSysMtrxSmallCaptionButton  () { return GetSysMtrxSmSize(); }
        _Size_& GetSysMtrxSmallTitlebarButton () { return GetSysMtrxSmSize(); }
        #endif
        #ifdef SM_CXFRAME
        _Size_& GetSysMtrxThickFrame          () { return GetSysMtrxFrame(); }
        #endif
        #ifdef SM_CXHTHUMB
        _Size_& GetSysMtrxThumb               () { return GetSystemmetrics( SM_CXHTHUMB, SM_CYVTHUMB ); }
        #endif
        _Size_& GetSysMtrxScrollbarThicknesses() { return GetSystemmetrics( SM_CXVSCROLL , SM_CYHSCROLL ); }

        bool    GetBitmapSize   ( HBITMAP hbm, BITMAP *pbm = NULL );
        _Size_& GetBitmapSizeRet( HBITMAP hbm, BITMAP *pbm = NULL ) { GetBitmapSize( hbm, pbm ); return *this; }
        bool    GetIconSize     ( HICON hicon, ICONINFO *pii = NULL );
        _Size_& GetIconSizeRet  ( HICON hicon, ICONINFO *pii = NULL ) { GetIconSize( hicon, pii ); return *this; }

        bool    GetTextPixelSize         ( PCTSTR pstrText, HFONT hfont, HDC hdc = NULL, UINT uDrawtextFormat = DT_CALCRECT );
        _Size_& GetTextPixelSizeRet      ( PCTSTR pstrText, HFONT hfont, HDC hdc = NULL, UINT uDrawtextFormat = DT_CALCRECT ) { GetTextPixelSize( pstrText, hfont, hdc, uDrawtextFormat ); return *this; }
        bool    GetFontPixelSize         ( HFONT hfont, HDC hdc = NULL ) { return GetTextPixelSize( _T("A"), hfont, hdc ); }
        _Size_& GetFontPixelSizeRet      ( HFONT hfont, HDC hdc = NULL ) { GetFontPixelSize( hfont, hdc ); return *this; }
        bool    GetWindowFontPixelSize   ( HWND hwnd );
        _Size_& GetWindowFontPixelSizeRet( HWND hwnd ) { GetWindowFontPixelSize( hwnd ); return *this; }
};

class _Range_
{
    public :
        LONG    m_start;
        LONG    m_length;

    public :
        _Range_() {  } // random filled
        _Range_( int _start, int _length ) { m_start = _start; m_length = _length; }
        _Range_( const _Range_& _range ) { m_start = _range.m_start; m_length = _range.m_length; }

        _Range_& SetStartLength( int _start, int _length ) { m_start = _start; m_length = _length; return *this; }
        _Range_& SetStartEnd   ( int _start, int _end )    { m_start = _start; m_length = _end - _start; return *this; }
        _Range_& SetStart      ( int _start )              { m_length -= _start - m_start; m_start = _start; return *this; }
        _Range_& SetEnd        ( int _end )                { m_length = _end - m_start; return *this; }
        _Range_& SetEmpty      ()                          { m_start = m_length = 0; return *this; }

        bool IsEmpty() const { return m_length <= 0; }
        bool IsNull () const { return m_start == 0 && m_length == 0; }
        bool PosInRange( int pos ) const { return pos >= m_start && pos < m_start + m_length; }

        _Range_& Offset ( int d )        { m_start += d; return *this; }
        _Range_& Inflate( int d )        { m_start -= d; m_length += d + d; return *this; }
        _Range_& Inflate( int s, int e ) { m_start -= s; m_length += s + e; return *this; }
        _Range_& Deflate( int d )        { m_start += d; m_length -= d + d; return *this; }
        _Range_& Deflate( int s, int e ) { m_start += s; m_length -= s + e; return *this; }

        _Range_& MoveStartTo ( int pos ) { m_start = pos; return *this; }
        _Range_& MoveCenterTo( int pos ) { m_start = pos - (m_length >> 1); return *this; }
        _Range_& MoveEndTo   ( int pos ) { m_start = pos - m_length; return *this; }

        bool     IntersectRange       ( const _Range_& range2 );
        _Range_& IntersectRangeRet    ( const _Range_& range2 ) { IntersectRange( range2 ); return *this; }
        bool     IntersectRangeEmul   ( const _Range_& range2 ) { return IntersectRange( range2 ); }
        _Range_& IntersectRangeEmulRet( const _Range_& range2 ) { IntersectRangeEmul( range2 ); return *this; }
        bool     UnionRange           ( const _Range_& range2 );
        _Range_& UnionRangeRet        ( const _Range_& range2 ) { UnionRange( range2 ); return *this; }
        bool     UnionRangeEmul       ( const _Range_& range2 );
        _Range_& UnionRangeEmulRet    ( const _Range_& range2 ) { UnionRangeEmul( range2 ); return *this; }

        _Range_& operator = ( const _Range_& _range ) { m_start = _range.m_start; m_length = _range.m_length; return *this; }
        bool operator ==( const _Range_& _range ) const { return m_start == _range.m_start && m_length == _range.m_length; }
        bool operator !=( const _Range_& _range ) const { return m_start != _range.m_start || m_length != _range.m_length; }
        _Range_& operator +=( int ofs ) { m_start += ofs; return *this; }
        _Range_& operator -=( int ofs ) { m_start -= ofs; return *this; }
        _Range_& operator &=( const _Range_& _range ) { return IntersectRangeEmulRet( _range ); }
        _Range_& operator |=( const _Range_& _range ) { return UnionRangeEmulRet    ( _range ); }

        int GetStart () const { return m_start; }
        int GetEnd   () const { return m_start + m_length; }
        int GetLength() const { return m_length; }
        int GetCenter() const { return m_start + (m_length >> 1); }
};

// MRTF_* values for 'flags' of '_Rect_::MoveRectToFrame()'
#define MRTF_LEFT               0x0001
#define MRTF_CENTER             0x0003
#define MRTF_RIGHT              0x0002
#define MRTF_TOP                0x0004
#define MRTF_VCENTER            0x000c
#define MRTF_BOTTOM             0x0008
#define MRTF_TOPLEFT            ( MRTF_TOP     | MRTF_LEFT   )
#define MRTF_TOPCENTER          ( MRTF_TOP     | MRTF_CENTER )
#define MRTF_TOPRIGHT           ( MRTF_TOP     | MRTF_RIGHT  )
#define MRTF_VCENTERLEFT        ( MRTF_VCENTER | MRTF_LEFT   )
#define MRTF_MIDDLE             ( MRTF_VCENTER | MRTF_CENTER )
#define MRTF_VCENTERRIGHT       ( MRTF_VCENTER | MRTF_RIGHT  )
#define MRTF_BOTTOMLEFT         ( MRTF_BOTTOM  | MRTF_LEFT   )
#define MRTF_BOTTOMCENTER       ( MRTF_BOTTOM  | MRTF_CENTER )
#define MRTF_BOTTOMRIGHT        ( MRTF_BOTTOM  | MRTF_RIGHT  )
#define MRTF_HORZ_MASK          ( MRTF_LEFT | MRTF_CENTER  | MRTF_RIGHT  )
#define MRTF_VERT_MASK          ( MRTF_TOP  | MRTF_VCENTER | MRTF_BOTTOM )

// RSIFF_* values for 'flags' of '_Rect_::RectStayInFrame()'
#define RSIFF_X                 0x0001
#define RSIFF_Y                 0x0002
#define RSIFF_DEFAULT           ( RSIFF_X | RSIFF_Y )

//class _Rect_ : public RECT
class _Rect_ : public tagRECT
{
    public :
        _Rect_() {  } // random filled
        _Rect_( int l, int t, int r, int b ) { left = l; top = t; right = r; bottom = b; }
        _Rect_( const RECT&  rc ) { * (RECT*) this = rc; }
        _Rect_( PCRECT      prc ) { if( prc ) * (RECT*) this = *prc; }
        _Rect_( SIZE       size ) { left = top = 0; right = size.cx; bottom = size.cy; }
        _Rect_( POINT topleft, SIZE         size ) { right = ( left = topleft.x ) + size.cx; bottom = ( top = topleft.y ) + size.cy; }
        _Rect_( POINT topleft, POINT bottomright ) { left = topleft.x; top = topleft.y; right = bottomright.x; bottom = bottomright.y; }
        _Rect_( const WINDOWPOS *pwp ) { left = pwp ? pwp->x : 0; top = pwp ? pwp->cy : 0; right = left + (pwp ? pwp->cx : 0); bottom = top + (pwp ? pwp->cy : 0); }

        int Cx() const { return( right - left ); }
        int Cy() const { return( bottom - top ); }
        int Width () const { return( right - left ); }
        int Height() const { return( bottom - top ); }
        _Rect_& SetCx( int cx ) { right = left + cx; return *this; }
        _Rect_& SetCy( int cy ) { bottom = top + cy; return *this; }
        _Rect_& SetWidth ( int cx ) { right = left + cx; return *this; }
        _Rect_& SetHeight( int cy ) { bottom = top + cy; return *this; }
        _Rect_& SetSize( int cx, int cy ) { right = left + cx; bottom = top + cy; return *this; } // same meaning as 'SetSizeTL()'
        _Rect_& SetSize( SIZE      size ) { return SetSizeTL( size ); }                           // same meaning as 'SetSizeTL()'
        _Rect_& SetSizeTL( int cx, int cy ) { right = left   + cx; bottom = top + cy; return *this; } // the Top-Left corner stays in place
        _Rect_& SetSizeTR( int cx, int cy ) { left  = right  - cx; bottom = top + cy; return *this; } // the Top-Right corner stays in place
        _Rect_& SetSizeBL( int cx, int cy ) { right = left   + cx; top = bottom - cy; return *this; } // the Bottom-Left corner stays in place
        _Rect_& SetSizeBR( int cx, int cy ) { left  = right  - cx; top = bottom - cy; return *this; } // the Bottom-Right corner stays in place
        _Rect_& SetSizeTC( int cx, int cy ) { right = ( left = Xm() - (cx>>1) ) + cx; bottom = top + cy; return *this; } // the Top-horzCenter corner stays in place
        _Rect_& SetSizeBC( int cx, int cy ) { right = ( left = Xm() - (cx>>1) ) + cx; top = bottom - cy; return *this; } // the Bottom-horzCenter corner stays in place
        _Rect_& SetSizeCL( int cx, int cy ) { right = left   + cx; bottom = ( top = Ym() - (cy>>1) ) + cy; return *this; } // the vertCenter-Left corner stays in place
        _Rect_& SetSizeCR( int cx, int cy ) { left  = right  - cx; bottom = ( top = Ym() - (cy>>1) ) + cy; return *this; } // the vertCenter-Right corner stays in place
        _Rect_& SetSizeCC( int cx, int cy ) { right = ( left = Xm() - (cx>>1) ) + cx; bottom = ( top = Ym() - (cy>>1) ) + cy; return *this; } // the Center of the rect (vertCenter-horzCenter) stays in place
        _Rect_& SetSizeTL( SIZE size )      { return SetSizeTL( size.cx, size.cy ); } // the Top-Left corner stays in place
        _Rect_& SetSizeTR( SIZE size )      { return SetSizeTR( size.cx, size.cy ); } // the Top-Right corner stays in place
        _Rect_& SetSizeBL( SIZE size )      { return SetSizeBL( size.cx, size.cy ); } // the Bottom-Left corner stays in place
        _Rect_& SetSizeBR( SIZE size )      { return SetSizeBR( size.cx, size.cy ); } // the Bottom-Right corner stays in place
        _Rect_& SetSizeTC( SIZE size )      { return SetSizeTC( size.cx, size.cy ); } // the Top-horzCenter corner stays in place
        _Rect_& SetSizeBC( SIZE size )      { return SetSizeBC( size.cx, size.cy ); } // the Bottom-horzCenter corner stays in place
        _Rect_& SetSizeCL( SIZE size )      { return SetSizeCL( size.cx, size.cy ); } // the vertCenter-Left corner stays in place
        _Rect_& SetSizeCR( SIZE size )      { return SetSizeCR( size.cx, size.cy ); } // the vertCenter-Right corner stays in place
        _Rect_& SetSizeCC( SIZE size )      { return SetSizeCC( size.cx, size.cy ); } // the Center of the rect (vertCenter-horzCenter) stays in place
        _Rect_& LimitCx     ( int cxmin, int cxmax ) { LimitCxMax( cxmax ); LimitCxMin( cxmin ); return *this; }
        _Rect_& LimitCxMin  ( int cxmin )            { if( right - left < cxmin ) right = left + cxmin; return *this; } // width may not be smaller than 'cxmin'
        _Rect_& LimitCxMax  ( int cxmax )            { if( right - left > cxmax ) right = left + cxmax; return *this; } // width may not be greater than 'cxmax'
        _Rect_& LimitCy     ( int cymin, int cymax ) { LimitCyMax( cymax ); LimitCyMin( cymin ); return *this; }
        _Rect_& LimitCyMin  ( int cymin )            { if( bottom - top < cymin ) bottom = top + cymin; return *this; } // height may not be smaller than 'cymin'
        _Rect_& LimitCyMax  ( int cymax )            { if( bottom - top > cymax ) bottom = top + cymax; return *this; } // height may not be greater than 'cymax'
        _Rect_& LimitSizeMin( int cxmin, int cymin ) { LimitCxMin( cxmin      ); LimitCyMin( cymin      ); return *this; } // width and height may not be smaller than wanted
        _Rect_& LimitSizeMin( SIZE sizemin )         { LimitCxMin( sizemin.cx ); LimitCyMin( sizemin.cy ); return *this; } // width and height may not be smaller than wanted
        _Rect_& LimitSizeMax( int cxmax, int cymax ) { LimitCxMax( cxmax      ); LimitCyMax( cymax      ); return *this; } // width and height may not be greater than wanted
        _Rect_& LimitSizeMax( SIZE sizemax )         { LimitCxMax( sizemax.cx ); LimitCyMax( sizemax.cy ); return *this; } // width and height may not be greater than wanted
        _Rect_& MaxSize     ( SIZE size2 )           { return LimitSizeMin( size2 ); } // get the maximum of the sizes; only allow to make size greater. width and height may not be smaller than wanted
        _Rect_& MinSize     ( SIZE size2 )           { return LimitSizeMax( size2 ); } // get the minimum of the sizes; only allow to make size smaller. width and height may not be greater than wanted
        _Size_ Size() const { return _Size_( right - left, bottom - top ); }
        _Pt_& TopLeft          ()       { return *(_Pt_*) ((POINT*)this);    }
        _Pt_& BottomRight      ()       { return *(_Pt_*)(((POINT*)this)+1); }
        const _Pt_& TopLeft    () const { return *(_Pt_*) ((POINT*)this);    }
        const _Pt_& BottomRight() const { return *(_Pt_*)(((POINT*)this)+1); }
        _Pt_ GetTopLeft     () const { return _Pt_( left,               top                ); }
        _Pt_ GetTopCenter   () const { return _Pt_( (left+right) / 2,   top                ); }
        _Pt_ GetTopRight    () const { return _Pt_( right,              top                ); }
        _Pt_ GetVCenterRight() const { return _Pt_( right,              (top + bottom) / 2 ); }
        _Pt_ GetBottomRight () const { return _Pt_( right,              bottom             ); }
        _Pt_ GetBottomCenter() const { return _Pt_( (left + right) / 2, bottom             ); }
        _Pt_ GetBottomLeft  () const { return _Pt_( left,               bottom             ); }
        _Pt_ GetVCenterLeft () const { return _Pt_( left,               (top + bottom) / 2 ); }
        _Pt_ GetCenter      () const { return _Pt_( (left + right) / 2, (top + bottom) / 2 ); }
        int Xm() const { return( (left + right) >> 1 ); }
        int Ym() const { return( (top + bottom) >> 1 ); }
        _Rect_&      SwapLeftRight()            { SwapLeftRight( (PRECT)this ); return *this; }
        _Rect_&      SwapTopBottom()            { SwapTopBottom( (PRECT)this ); return *this; }
        static PRECT SwapLeftRight( PRECT prc ) { privateRect::tswap( prc->left, prc->right ); return prc; }
        static PRECT SwapTopBottom( PRECT prc ) { privateRect::tswap( prc->top, prc->bottom ); return prc; }
        _Range_ GetHorzRange() const { return _Range_().SetStartEnd( left, right ); }
        _Range_ GetVertRange() const { return _Range_().SetStartEnd( top, bottom ); }
        void SetHorzRange( const _Range_& range ) { right  = ( left = range.m_start ) + range.m_length; }
        void SetVertRange( const _Range_& range ) { bottom = ( top  = range.m_start ) + range.m_length; }
        double GetAspectRatio() const { int _cy = Cy(); return _cy != 0 ? (double) Cx() / (double) _cy : 0.0; }

        operator PRECT ()       { return this; }
        operator PCRECT() const { return this; }

        bool IsRectEmpty() const { return( right <= left || bottom <= top ); }
        bool IsRectNull () const { return( left == 0 && right == 0 && top == 0 && bottom == 0 ); }
        bool PtInRect( POINT pt ) const { return( pt.x >= left && pt.x < right && pt.y >= top && pt.y < bottom ); }

        _Rect_& SetRect     ( int x1, int y1,   int x2, int y2 ) { left = x1; top = y1; right = x2; bottom = y2; return *this; }
        _Rect_& SetRect     ( POINT topleft, SIZE         size ) { right = ( left = topleft.x ) + size.cx; bottom = ( top = topleft.y ) + size.cy; return *this; }
        _Rect_& SetRect     ( POINT topleft, POINT bottomright ) { left = topleft.x; top = topleft.y; right = bottomright.x; bottom = bottomright.y; return *this; }
        _Rect_& SetRectEmpty()                                   { memset( this, 0, sizeof(RECT) ); return *this; }
        _Rect_& CopyRect ( PCRECT prc )       { if( prc ) *(RECT*)this = *prc; return *this; }
        bool    EqualRect( PCRECT prc ) const { return( memcmp( this, prc, sizeof(RECT) ) == 0 ); }
        _Rect_& CopyHorizontal( PCRECT prc ) { if( prc ) { left = prc->left; right  = prc->right;  } return *this; } // copies the left and right values
        _Rect_& CopyVertical  ( PCRECT prc ) { if( prc ) { top  = prc->top;  bottom = prc->bottom; } return *this; } // copies the top and bottom values

        _Rect_& InflateRect( int x, int y ) { left -= x; top -= y; right += x; bottom += y; return *this; }
        _Rect_& InflateRect( SIZE    size ) { left -= size.cx; top -= size.cy; right += size.cx; bottom += size.cy; return *this; }
        _Rect_& InflateRect( PCRECT   prc ) { left -= prc->left; top -= prc->top; right += prc->right; bottom += prc->bottom; return *this; }
        _Rect_& InflateRect( int l, int t, int r, int b ) { left -= l; top -= t; right += r; bottom += b; return *this; }
        _Rect_& DeflateRect( int x, int y ) { left += x; top += y; right -= x; bottom -= y; return *this; }
        _Rect_& DeflateRect( SIZE    size ) { left += size.cx; top += size.cy; right -= size.cx; bottom -= size.cy; return *this; }
        _Rect_& DeflateRect( PCRECT   prc ) { left += prc->left; top += prc->top; right -= prc->right; bottom -= prc->bottom; return *this; }
        _Rect_& DeflateRect( int l, int t, int r, int b ) { left += l; top += t; right -= r; bottom -= b; return *this; }

        _Rect_& NormalizeRect();

        _Rect_& OffsetRect( int x, int y ) { left += x; top += y; right += x; bottom += y; return *this; }
        _Rect_& OffsetRect( POINT     pt ) { left += pt.x; top += pt.y; right += pt.x; bottom += pt.y; return *this; }
        _Rect_& OffsetRect( SIZE    size ) { left += size.cx; top += size.cy; right += size.cx; bottom += size.cy; return *this; }
        _Rect_& MoveRectTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left = x ) + cx; bottom = ( top = y ) + cy; return *this; } // Moves the rect to the position without resizing (same as 'MoveRectTLTo()')
        _Rect_& MoveRectTo( POINT     pt ) { return MoveRectTo( pt.x, pt.y ); } // Moves the rect to the position without resizing (same as 'MoveRectTLTo()')
        _Rect_& MoveRectToOrigin() { return MoveRectTo( 0, 0 ); }
        _Rect_& MoveRectToFrame( _Rect_ rcFrame, DWORD flags );
        _Rect_& RectStayInFrame( RECT   rcFrame, DWORD flags = RSIFF_DEFAULT );
        _Rect_& MoveRectTLTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left  = x ) + cx; bottom = ( top    = y ) + cy; return *this; } // Moves the rect's Top-Left corner to the position without resizing (same as 'MoveRectTo()')
        _Rect_& MoveRectTRTo( int x, int y ) { int cx = Cx(); int cy = Cy(); left  = ( right = x ) - cx; bottom = ( top    = y ) + cy; return *this; } // Moves the rect's Top-Right corner to the position without resizing
        _Rect_& MoveRectBLTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left  = x ) + cx; top    = ( bottom = y ) - cy; return *this; } // Moves the rect's Bottom-Left corner to the position without resizing
        _Rect_& MoveRectBRTo( int x, int y ) { int cx = Cx(); int cy = Cy(); left  = ( right = x ) - cx; top    = ( bottom = y ) - cy; return *this; } // Moves the rect's Bottom-Right corner to the position without resizing
        _Rect_& MoveRectTCTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left  = x - (cx >> 1) ) + cx; bottom = ( top    = y ) + cy; return *this; } // Moves the rect's Top-horzCenter corner to the position without resizing
        _Rect_& MoveRectBCTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left  = x - (cx >> 1) ) + cx; top    = ( bottom = y ) - cy; return *this; } // Moves the rect's Bottom-horzCenter corner to the position without resizing
        _Rect_& MoveRectCLTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left  = x ) + cx; bottom = ( top = y - (cy >> 1) ) + cy; return *this; } // Moves the rect's vertCenter-Left corner to the position without resizing
        _Rect_& MoveRectCRTo( int x, int y ) { int cx = Cx(); int cy = Cy(); left  = ( right = x ) - cx; bottom = ( top = y - (cy >> 1) ) + cy; return *this; } // Moves the rect's vertCenter-Right corner to the position without resizing
        _Rect_& MoveRectCCTo( int x, int y ) { int cx = Cx(); int cy = Cy(); right = ( left  = x - (cx >> 1) ) + cx; bottom = ( top = y - (cy >> 1) ) + cy; return *this; } // Moves the rect's Center (vertCenter-horzCenter) to the position without resizing
        _Rect_& MoveRectTLTo( POINT pt )     { return MoveRectTLTo( pt.x, pt.y ); } // Moves the rect's Top-Left corner to the position without resizing (same as 'MoveRectTo()')
        _Rect_& MoveRectTRTo( POINT pt )     { return MoveRectTRTo( pt.x, pt.y ); } // Moves the rect's Top-Right corner to the position without resizing
        _Rect_& MoveRectBLTo( POINT pt )     { return MoveRectBLTo( pt.x, pt.y ); } // Moves the rect's Bottom-Left corner to the position without resizing
        _Rect_& MoveRectBRTo( POINT pt )     { return MoveRectBRTo( pt.x, pt.y ); } // Moves the rect's Bottom-Right corner to the position without resizing
        _Rect_& MoveRectTCTo( POINT pt )     { return MoveRectTCTo( pt.x, pt.y ); } // Moves the rect's Top-horzCenter corner to the position without resizing
        _Rect_& MoveRectBCTo( POINT pt )     { return MoveRectBCTo( pt.x, pt.y ); } // Moves the rect's Bottom-horzCenter corner to the position without resizing
        _Rect_& MoveRectCLTo( POINT pt )     { return MoveRectCLTo( pt.x, pt.y ); } // Moves the rect's vertCenter-Left corner to the position without resizing
        _Rect_& MoveRectCRTo( POINT pt )     { return MoveRectCRTo( pt.x, pt.y ); } // Moves the rect's vertCenter-Right corner to the position without resizing
        _Rect_& MoveRectCCTo( POINT pt )     { return MoveRectCCTo( pt.x, pt.y ); } // Moves the rect's Center (vertCenter-horzCenter) to the position without resizing
        _Rect_& ZoomRect ( int    xscale, int    yscale ) { left *= xscale; top *= yscale; right *= xscale; bottom *= yscale; return *this; } // Multiplies each coord of the rect
        _Rect_& ZoomRect ( SIZE    scale )                { return ZoomRect( scale.cx, scale.cy ); } // Multiplies each coord of the rect
        _Rect_& ZoomRectF( double xscale, double yscale ); // Multiplies each coord of the rect
        _Rect_& DivRect ( int    xinvscale, int    yinvscale ) { if( xinvscale != 0 && yinvscale != 0 ) { left /= xinvscale; top /= yinvscale; right /= xinvscale; bottom /= yinvscale; } return *this; } // Divides each coord of the rect
        _Rect_& DivRect ( int     invscale )                   { return DivRect( invscale, invscale ); } // Divides each coord of the rect
        _Rect_& DivRectF( double xinvscale, double yinvscale ); // Divides each coord of the rect
        _Rect_& ZoomSize ( int    xscale, int    yscale ) { right = left + (right - left) * xscale; bottom = top + (bottom - top) * yscale; return *this; } // Multiplies only the size (width and height) of the rect
        _Rect_& ZoomSize ( SIZE    scale )                { return ZoomSize( scale.cx, scale.cy ); } // Multiplies only the size (width and height) of the rect
        _Rect_& ZoomSizeF( double xscale, double yscale ); // Multiplies only the size (width and height) of the rect
        _Rect_& DivSize ( int    xinvscale, int    yinvscale ) { if( xinvscale != 0 && yinvscale != 0 ) { right = left + (right - left) / xinvscale; bottom = top + (bottom - top) / yinvscale; } return *this; } // Divides only the size (width and height) of the rect
        _Rect_& DivSize ( SIZE    divsize )                    { return DivSize( divsize.cx, divsize.cy ); } // Divides only the size (width and height) of the rect
        _Rect_& DivSizeF( double xinvscale, double yinvscale ); // Divides only the size (width and height) of the rect
        _Rect_& SetAspectRatio( SIZE sizOfAspectRatio, DWORD option );
        _Rect_& SetAspectRatio( double aspect_ratio, DWORD option );
        _Rect_ MirrorRect            ( POINT pt ) const;
        _Rect_ MirrorOnHorizontalAxis( int    y ) const;
        _Rect_ MirrorOnVerticalAxis  ( int    x ) const;

        _Rect_& QuarterizeRect( PRECT prcTL, PRECT prcTR, PRECT prcBL, PRECT prcBR );

        static bool AreAttachedRects( PCRECT prc1, PCRECT prc2 );
        bool        AreAttachedRects( PCRECT prc2 ) const { return AreAttachedRects( this, prc2 ); }
        static bool AreHorizontallyAttachedRects( PCRECT prc1, PCRECT prc2 );
        bool        AreHorizontallyAttachedRects( PCRECT prc2 ) const { return AreHorizontallyAttachedRects( this, prc2 ); }
        static bool AreVerticallyAttachedRects( PCRECT prc1, PCRECT prc2 );
        bool        AreVerticallyAttachedRects( PCRECT prc2 ) const { return AreVerticallyAttachedRects( this, prc2 ); }

        bool    IntersectRect       ( PCRECT prc );
        bool    IntersectRect       ( PCRECT prc1, PCRECT prc2 );
        _Rect_& IntersectRectRet    ( PCRECT prc )               { IntersectRect( prc );        return *this; }
        _Rect_& IntersectRectRet    ( PCRECT prc1, PCRECT prc2 ) { IntersectRect( prc1, prc2 ); return *this; }
        bool    IntersectRectEmul   ( PCRECT prc )               { IntersectRectEmulRet( this, prc );  return !IsRectEmpty(); }
        bool    IntersectRectEmul   ( PCRECT prc1, PCRECT prc2 ) { IntersectRectEmulRet( prc1, prc2 ); return !IsRectEmpty(); }
        _Rect_& IntersectRectEmulRet( PCRECT prc )               { IntersectRectEmulRet( this, prc );  return *this; }
        _Rect_& IntersectRectEmulRet( PCRECT prc1, PCRECT prc2 );
        bool    UnionRect           ( PCRECT prc );
        bool    UnionRect           ( PCRECT prc1, PCRECT prc2 );
        _Rect_& UnionRectRet        ( PCRECT prc )               { UnionRect( prc );        return *this; }
        _Rect_& UnionRectRet        ( PCRECT prc1, PCRECT prc2 ) { UnionRect( prc1, prc2 ); return *this; }
        bool    UnionRectEmul       ( PCRECT prc )               { UnionRectEmulRet( this, prc );  return !IsRectEmpty(); }
        bool    UnionRectEmul       ( PCRECT prc1, PCRECT prc2 ) { UnionRectEmulRet( prc1, prc2 ); return !IsRectEmpty(); }
        _Rect_& UnionRectEmulRet    ( PCRECT prc )               { UnionRectEmulRet( this, prc );  return *this; }
        _Rect_& UnionRectEmulRet    ( PCRECT prc1, PCRECT prc2 );
        bool    SubtractRect        ( PCRECT prc1, PCRECT prc2 );

        _Rect_& operator = ( const RECT&  rc ) { *(RECT*)this = rc; return *this; }
        bool    operator ==( const RECT&  rc ) const { return( memcmp( this, &rc, sizeof(RECT) ) == 0 ); }
        bool    operator !=( const RECT&  rc ) const { return( memcmp( this, &rc, sizeof(RECT) ) != 0 ); }
        _Rect_& operator +=( POINT        pt ) { left  += pt.x;    top    += pt.y;    right += pt.x; bottom += pt.y; return *this; }
        _Rect_& operator +=( SIZE       size ) { right += size.cx; bottom += size.cy; return *this; }
        _Rect_& operator +=( PCRECT      prc ) { InflateRect( prc ); return *this; }
        _Rect_& operator -=( POINT        pt ) { left  -= pt.x;    top    -= pt.y;    right -= pt.x; bottom -= pt.y; return *this; }
        _Rect_& operator -=( SIZE       size ) { right -= size.cx; bottom -= size.cy; return *this; }
        _Rect_& operator -=( PCRECT      prc ) { DeflateRect( prc ); return *this; }
        _Rect_& operator &=( const RECT&  rc ) { return IntersectRectEmulRet( &rc ); }
        _Rect_& operator |=( const RECT&  rc ) { return UnionRectEmulRet    ( &rc ); }
        _Rect_& operator *=( int       scale ) { return ZoomRect ( scale, scale ); }
        _Rect_& operator *=( float     scale ) { return ZoomRectF( scale, scale ); }
        _Rect_& operator *=( double    scale ) { return ZoomRectF( scale, scale ); }
        _Rect_& operator /=( int    invscale ) { return DivRect ( invscale, invscale ); }
        _Rect_& operator /=( float  invscale ) { return DivRectF( invscale, invscale ); }
        _Rect_& operator /=( double invscale ) { return DivRectF( invscale, invscale ); }

        _Rect_ operator +( POINT  pt ) const { return _Rect_( left + pt.x, top + pt.y, right + pt.x,    bottom + pt.y    ); }
        _Rect_ operator -( POINT  pt ) const { return _Rect_( left - pt.x, top - pt.y, right - pt.x,    bottom - pt.y    ); }
        _Rect_ operator +( SIZE size ) const { return _Rect_( left,        top,        right + size.cx, bottom + size.cy ); }
        _Rect_ operator -( SIZE size ) const { return _Rect_( left,        top,        right - size.cx, bottom - size.cy ); }
        // IMPORTANT: The 'operator+(SIZE)' and 'operator-(SIZE)' operators
        // in here have a different implmenetation than 'CRect'.
        // 'CRect' will offset the rect itself while '_Rect_' offsets only
        // the rect's bottom right corner.
        _Rect_ operator +( PCRECT      prc ) const { _Rect_ rc( this ); rc.InflateRect( prc ); return rc; }
        _Rect_ operator -( PCRECT      prc ) const { _Rect_ rc( this ); rc.DeflateRect( prc ); return rc; }
        _Rect_ operator &( const RECT& rc2 ) const { _Rect_ rc; rc.IntersectRectEmul( this, &rc2 ); return rc; }
        _Rect_ operator |( const RECT& rc2 ) const { _Rect_ rc; rc.UnionRectEmul    ( this, &rc2 ); return rc; }
        _Rect_ operator *( int       scale ) const { return _Rect_( this ).ZoomRect ( scale,    scale    ); }
        _Rect_ operator *( float     scale ) const { return _Rect_( this ).ZoomRectF( scale,    scale    ); }
        _Rect_ operator *( double    scale ) const { return _Rect_( this ).ZoomRectF( scale,    scale    ); }
        _Rect_ operator /( int    invscale ) const { return _Rect_( this ).DivRect  ( invscale, invscale ); }
        _Rect_ operator /( float  invscale ) const { return _Rect_( this ).DivRectF ( invscale, invscale ); }
        _Rect_ operator /( double invscale ) const { return _Rect_( this ).DivRectF ( invscale, invscale ); }
        _Rect_ MulDiv( int nMultiplier, int nDivisor ) const;

        _Rect_& GetWorkArea();
        _Rect_& GetMonitorWorkAreaFromMon( HMONITOR hmon );
        _Rect_& GetMonitorWorkAreaFromWnd( HWND hwnd )
        { return GetMonitorWorkAreaFromMon( MonitorFromWindow(hwnd, MONITOR_DEFAULTTONEAREST) ); }
        _Rect_& GetPrimaryScreen();
        _Rect_& GetVirtualScreen();
        _Rect_& GetMonitorRectFromMon( HMONITOR hmon );
        _Rect_& GetMonitorRectFromWnd( HWND hwnd )
        { return GetMonitorRectFromMon( MonitorFromWindow(hwnd, MONITOR_DEFAULTTONEAREST) ); }
        _Rect_& GetDesktopRect() { return GetWindowRect( GetDesktopWindow() ); }

        _Rect_& GetClientRect( HWND hwnd ) { if( hwnd ) ::GetClientRect(hwnd, this); return *this; }
        _Rect_& GetWindowRect( HWND hwnd ) { if( hwnd ) ::GetWindowRect(hwnd, this); return *this; }
        bool    SetWindowRect   ( HWND hwnd, HWND hwndInsertAfter, UINT flags ) const; // Performs a SetWindowPos() call.
        void    SetWindowRect   ( HWND hwnd ) const; // Sets only a new position and size for a window
        void    SetWindowTopLeft( HWND hwnd ) const; // Sets only a new position for a window (does not change its size)
        void    SetWindowSize   ( HWND hwnd ) const; // Sets only a new size for a window (does not change its position)
        _Rect_& GetWindowRectInClient( HWND hwnd, HWND hwndOfClient );
        _Rect_& GetDlgItemRect( HWND hwndDlg, int idCtrl )       { return GetWindowRectInClient( GetDlgItem(hwndDlg, idCtrl), hwndDlg ); }
        void    SetDlgItemRect( HWND hwndDlg, int idCtrl ) const { SetWindowRect( GetDlgItem(hwndDlg, idCtrl) ); }
        _Rect_& GetControlRect( HWND hwndCtrl )                  { return GetWindowRectInClient( hwndCtrl, GetParent(hwndCtrl) ); }
        bool    AdjustWindowRectEx( HWND hwnd, bool bIncludeScrollbars = true );
        bool    AdjustWindowRectEx( DWORD dwStyle, bool bMenu, DWORD dwExStyle, bool bIncludeScrollbars = true );
        #ifndef _WIN32_WCE
        bool    AdjustWindowRectEx( DWORD dwStyle, const MENUBARINFO *pmbi, DWORD dwExStyle, bool bIncludeScrollbars = true );
        #endif
        bool    ReverseAdjustWindowRectEx( HWND hwnd, bool bIncludeScrollbars = true );
        bool    ReverseAdjustWindowRectEx( DWORD dwStyle, bool bMenu, DWORD dwExStyle, bool bIncludeScrollbars = true );
        #ifndef _WIN32_WCE
        bool    ReverseAdjustWindowRectEx( DWORD dwStyle, const MENUBARINFO *pmbi, DWORD dwExStyle, bool bIncludeScrollbars = true );
        #endif
        _Rect_& ClientToScreen( HWND hwnd ) { if( hwnd ) ::MapWindowPoints( hwnd, NULL, (PPOINT)this, 2 ); return *this; }
        _Rect_& ScreenToClient( HWND hwnd ) { if( hwnd ) ::MapWindowPoints( NULL, hwnd, (PPOINT)this, 2 ); return *this; }
        _Rect_& MapWindowPoints( HWND hwndFrom, HWND hwndTo ) { ::MapWindowPoints( hwndFrom, hwndTo, (PPOINT)this, 2 ); return *this; }
        HDWP    DeferWindowRect   ( HDWP hdefer, HWND hwnd, HWND hwndInsertAfter, UINT flags ) const; // Performs a DeferWindowPos() call.
        HDWP    DeferWindowRect   ( HDWP hdefer, HWND hwnd ) const; // Defers only a new position and size for a window (using #DeferWindowPos()#)
        HDWP    DeferWindowTopLeft( HDWP hdefer, HWND hwnd ) const; // Defers only a new position for a window (does not change its size) (using #DeferWindowPos()#)
        HDWP    DeferWindowSize   ( HDWP hdefer, HWND hwnd ) const; // Defers only a new size for a window (does not change its position) (using #DeferWindowPos()#)
        static void CenterWindowOverWindow( HWND hwndToBeCentered, HWND hwndReferenceWindow );
        static void CenterWindowOverItsParent( HWND hwndToBeCentered );

            ///<summary> Maps dialog coordinates to pixels (<b>bBackward</b> == false)
            ///  or pixels to dialog coordinates (<b>bBackward</b> == true). <br/>
            ///  This uses <b>GetDialogBaseUnits()</b> to convert the units, so using the
            ///  default system font and not taking a specific dialog's font into account. </summary>
        _Rect_& MapSystemDialogCoords( bool bBackward = false );
        bool    MapDialogRect( HWND hwndDlg ) { return ::MapDialogRect( hwndDlg, this ) != FALSE; }
        _Rect_& MapDialogRectRet( HWND hwndDlg ) { ::MapDialogRect( hwndDlg, this ); return *this; }
        bool    UnmapDialogRect( HWND hwndDlg );
        _Rect_& UnmapDialogRectRet( HWND hwndDlg ) { UnmapDialogRect( hwndDlg ); return *this; }

        int     IntersectClipRect( HDC hdc ) const { return ::IntersectClipRect( hdc, left, top, right, bottom ); }
        int     ExcludeClipRect  ( HDC hdc ) const { return ::ExcludeClipRect  ( hdc, left, top, right, bottom ); }
};




}

#endif
