/*
    File:               argonRectF.cpp
    Description:        Point, Size and Rectangle classes for floating point values.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         03.11.2006
    last updated:       03.11.2006 (Roman Komary)

    Copyright (c) 2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonRectF_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonRectF.h"

namespace ArgonLibrary
{




// Use this method if you want 'this' point to be handled
// as being the corner of another rect.
// Use this method if you want 'this' point to be handled
// as being the corner of another rect.
// The point will stay inside the rect. But the result is that the
// point may also lie on the right or bottom border of the rect.
// This is different to '_Pt_::StayInRect()'.
// NOTE:
// A method '_PtF_::StayInRect()' does not exist since we
// are using float values here.
_PtF_& _PtF_::StayInPseudoRect( RECTF rc )
{
    if(      x < rc.left   ) x = rc.left;
    else if( x > rc.right  ) x = rc.right;
    if(      y < rc.top    ) y = rc.top;
    else if( y > rc.bottom ) y = rc.bottom;
    return *this;
}

_SizeF_& _PtF_::GetAsSize()
{
    return * (_SizeF_*) this;
}

const _SizeF_& _PtF_::GetAsSize() const
{
    return * (_SizeF_*) this;
}

_RectF_ _PtF_::GetEmptyRect() const
{
    return _RectF_( x, y, x, y );
}

_RectF_ _PtF_::GetOneDotRect() const
{
    return _RectF_( x, y, x + 1.0f, y + 1.0f );
}



_SizeF_ _PtF_::operator -( POINTF pt ) const
{
    return _SizeF_( x - pt.x, y - pt.y );
}



_RectF_ _PtF_::operator +( PCRECTF prc ) const
{
    if( !prc ) return GetEmptyRect();
    return _RectF_( x + prc->left, y + prc->top, x + prc->right, y + prc->bottom );
}

_RectF_ _PtF_::operator -( PCRECTF prc ) const
{
    if( !prc ) return GetEmptyRect();
    return _RectF_( prc ).MirrorRect( *this );
}




///<param name="sizOfAspectRatio"> the size from which to take the aspect ratio </param>
///<param name="option"> a SAR_* value </param>
_SizeF_& _SizeF_::SetAspectRatio( SIZEF sizOfAspectRatio, DWORD option )
{
    if( sizOfAspectRatio.cx <= 0.0f || sizOfAspectRatio.cy <= 0.0f )
    {
        if( sizOfAspectRatio.cx <= 0.0f ) cx = 0.0f;
        if( sizOfAspectRatio.cy <= 0.0f ) cy = 0.0f;
        return *this;
    }

    if( cx <= 0.0f || cy <= 0.0f )
        return *this;

    if( option == SAR_KEEPWIDTH )
    {
        cy = (cx * sizOfAspectRatio.cy) / sizOfAspectRatio.cx;
        return *this;
    }

    if( option == SAR_KEEPHEIGHT )
    {
        cx = (cy * sizOfAspectRatio.cx) / sizOfAspectRatio.cy;
        return *this;
    }

    if( option == SAR_ENLARGETOSMALLEST )
    {
        float cy2 = (cx * sizOfAspectRatio.cy) / sizOfAspectRatio.cx;
        if( cy2 >= cy ) cy = cy2;
        else            cx = (cy * sizOfAspectRatio.cx) / sizOfAspectRatio.cy;
        return *this;
    }

    if( option == SAR_SHRINKTOBIGGEST )
    {
        float cy2 = (cx * sizOfAspectRatio.cy) / sizOfAspectRatio.cx;
        if( cy2 <= cy ) cy = cy2;
        else            cx = (cy * sizOfAspectRatio.cx) / sizOfAspectRatio.cy;
        return *this;
    }

    return *this;
}

///<param name="aspect_ratio"> the aspect ratio to set (width divided by height) </param>
///<param name="option"> a SAR_* value </param>
_SizeF_& _SizeF_::SetAspectRatio( double aspect_ratio, DWORD option )
{
    if( aspect_ratio <= 0.0 )
    {
        if( aspect_ratio <= 0.0 ) cx = 0.0f;
        if( aspect_ratio <  0.0 ) cy = 0.0f;
        return *this;
    }

    if( cx <= 0.0f || cy <= 0.0f )
        return *this;

    if( option == SAR_KEEPWIDTH )
    {
        cy = (float) (cx / aspect_ratio);
        return *this;
    }

    if( option == SAR_KEEPHEIGHT )
    {
        cx = (float) (cy * aspect_ratio);
        return *this;
    }

    if( option == SAR_ENLARGETOSMALLEST )
    {
        float cy2 = (float) (cx / aspect_ratio);
        if( cy2 >= cy ) cy = cy2;
        else            cx = (float) (cy * aspect_ratio);
        return *this;
    }

    if( option == SAR_SHRINKTOBIGGEST )
    {
        float cy2 = (float) (cx / aspect_ratio);
        if( cy2 <= cy ) cy = cy2;
        else            cx = (float) (cy * aspect_ratio);
        return *this;
    }

    return *this;
}



_RectF_ _SizeF_::GetOriginRect() const
{
    return _RectF_( 0.0f, 0.0f, cx, cy );
}



_RectF_ _SizeF_::operator +( PCRECTF prc ) const
{
    if( !prc ) return GetOriginRect();
    return _RectF_( prc->left, prc->top, cx + prc->right, cy + prc->bottom );
}




bool _RangeF_::IntersectRange( const _RangeF_& range2 )
{
    float _end = privateRectF::tmin( GetEnd(), range2.GetEnd() );
    m_start  = privateRectF::tmax( m_start,  range2.m_start  );
    m_length = _end - m_start;

    return m_length > 0.0f;
}

bool _RangeF_::UnionRange( const _RangeF_& range2 )
{
    if( range2.IsEmpty() )
        return !IsEmpty();

    float _end = privateRectF::tmax( GetEnd(), range2.GetEnd() );
    m_start  = privateRectF::tmin( m_start,  range2.m_start  );
    m_length = _end - m_start;

    return m_length > 0.0f;
}

///<summary> This is similar to <b>UnionRange()</b>, but
///  with the difference that empty ranges (as input and
///  as output) are handled as if for non-empty ranges,
///  and an empty result range's coords not be set to zero.
///  They will then be the really *calculated* coords. </summary>
bool _RangeF_::UnionRangeEmul( const _RangeF_& range2 )
{
    float _end = privateRectF::tmax( GetEnd(), range2.GetEnd() );
    m_start  = privateRectF::tmin( m_start,  range2.m_start  );
    m_length = _end - m_start;

    return m_length > 0.0f;
}




_RectF_& _RectF_::NormalizeRect()
{
    if( left > right  ) privateRectF::tswap( left, right  );
    if( top  > bottom ) privateRectF::tswap( top,  bottom );
    return *this;
}



// Moves the rect to the frame of 'rcFrame' according to 'flags'
// but always that way, that the rect stays inside of 'rcFrame'
// (if 'rcFrame' is larger than 'this' rect of course).
// The size of 'this' rect will never be changed.
//  : flags = combination of MRTF_* values which indicate which direction to move where to.
//            E.g. if this is only MRTF_LEFT, the x-coord of the rect will be moved to
//             the left side of 'rcFrame' but the y-coord will not be changed.
//  > false = Error
_RectF_& _RectF_::MoveRectToFrame( _RectF_ rcFrame, DWORD flags )
{
    if(      ( flags & MRTF_HORZ_MASK ) == MRTF_LEFT   ) MoveRectTo( rcFrame.left,               top );
    else if( ( flags & MRTF_HORZ_MASK ) == MRTF_CENTER ) MoveRectTo( rcFrame.Xm() - Cx() * 0.5f, top );
    else if( ( flags & MRTF_HORZ_MASK ) == MRTF_RIGHT  ) MoveRectTo( rcFrame.right - Cx(),       top );

    if(      ( flags & MRTF_VERT_MASK ) == MRTF_TOP     ) MoveRectTo( left, rcFrame.top                );
    else if( ( flags & MRTF_VERT_MASK ) == MRTF_VCENTER ) MoveRectTo( left, rcFrame.Ym() - Cy() * 0.5f );
    else if( ( flags & MRTF_VERT_MASK ) == MRTF_BOTTOM  ) MoveRectTo( left, rcFrame.bottom - Cy()      );

    return *this;
}

// Moves the rect that way that it is *completely* inside of 'rcFrame'.
// Note: It will only be moved if it is outside.
// Note: Only that coordinates will be moved which have been defined in 'flags'
//  : flags = combination of RSIFF_* values
// Note: If the rect is larger than the frame-rect, then it will be position
//  onto the top and/or left coord of the frame-rect.
_RectF_& _RectF_::RectStayInFrame( RECTF rcFrame, DWORD flags )
{
    if( flags & RSIFF_X ) {
        if( right > rcFrame.right ) MoveRectTo( rcFrame.right - Cx(), top );
        if( left  < rcFrame.left  ) MoveRectTo( rcFrame.left,         top );
    }

    if( flags & RSIFF_Y ) {
        if( bottom > rcFrame.bottom ) MoveRectTo( left, rcFrame.bottom - Cy() );
        if( top    < rcFrame.top    ) MoveRectTo( left, rcFrame.top           );
    }

    return *this;
}

///<param name="sizOfAspectRatio"> the size from which to take the aspect ratio </param>
///<param name="option"> a SAR_* value </param>
///<remarks> In any case, the calculated rect will keep its top-left coordinate unchanged. </remarks>
_RectF_& _RectF_::SetAspectRatio( SIZEF sizOfAspectRatio, DWORD option )
{
    _SizeF_ siz = Size();
    siz.SetAspectRatio( sizOfAspectRatio, option );
    SetRect( GetTopLeft(), siz );
    return *this;
}

///<param name="aspect_ratio"> the aspect ratio to set (width divided by height) </param>
///<param name="option"> a SAR_* value </param>
///<remarks> In any case, the calculated rect will keep its top-left coordinate unchanged. </remarks>
_RectF_& _RectF_::SetAspectRatio( double aspect_ratio, DWORD option )
{
    _SizeF_ siz = Size();
    siz.SetAspectRatio( aspect_ratio, option );
    SetRect( GetTopLeft(), siz );
    return *this;
}

// Mirrors the rect on a point 'pt'.
_RectF_ _RectF_::MirrorRect( POINTF pt ) const
{
    _RectF_ rc;
    pt.x += pt.x; pt.y += pt.y; // double the 'pt'
    rc.left   = pt.x - right;
    rc.top    = pt.y - bottom;
    rc.right  = pt.x - left;
    rc.bottom = pt.y - top;
    return rc;
}

// Mirrors the rect on a *horizontal* axes
// whose y-coord is 'y'.
_RectF_ _RectF_::MirrorOnHorizontalAxis( float y ) const
{
    _RectF_ rc;
    y += y; // double the coord
    rc.left   = left;
    rc.top    = y - bottom;
    rc.right  = right;
    rc.bottom = y - top;
    return rc;
}

// Mirrors the rect on a *vertical* axes
// whose x-coord is 'x'.
_RectF_ _RectF_::MirrorOnVerticalAxis( float x ) const
{
    _RectF_ rc;
    x += x; // double the coord
    rc.left   = x - right;
    rc.top    = top;
    rc.right  = x - left;
    rc.bottom = bottom;
    return rc;
}



// The rect will be devided into 4 rects and
// they will be equally sized if possilbe.
//
//      +-------+-------+
//      |       |       |
//      | prcTL | prcTR |
//      |       |       |
//      +-------+-------+
//      |       |       |
//      | prcBL | prcBR |
//      |       |       |
//      +-------+-------+
//
_RectF_& _RectF_::QuarterizeRect( PRECTF prcTL, PRECTF prcTR, PRECTF prcBL, PRECTF prcBR )
{
    prcTL->left  = prcBL->left  = left;
    prcTR->left  = prcBR->left  =
    prcTL->right = prcBL->right = Xm();
    prcTR->right = prcBR->right = right;

    prcTL->top    = prcTR->top    = top;
    prcBL->top    = prcBR->top    =
    prcTL->bottom = prcTR->bottom = Ym();
    prcBL->bottom = prcBR->bottom = bottom;

    return *this;
}



bool _RectF_::AreAttachedRects( PCRECTF prc1, PCRECTF prc2 )
{
    return AreHorizontallyAttachedRects(prc1, prc2) ? true : AreVerticallyAttachedRects(prc1, prc2);
}

bool _RectF_::AreHorizontallyAttachedRects( PCRECTF prc1, PCRECTF prc2 )
{
    if( prc1->top    != prc2->top    ||
        prc1->bottom != prc2->bottom ) return false;
    return prc1->left  == prc2->right ||
           prc1->right == prc2->left;
}

bool _RectF_::AreVerticallyAttachedRects( PCRECTF prc1, PCRECTF prc2 )
{
    if( prc1->left  != prc2->left  ||
        prc1->right != prc2->right ) return false;
    return prc1->top    == prc2->bottom ||
           prc1->bottom == prc2->top;
}



///<summary> This is similar to <b>IntersectRect()</b>, but
///  with the difference that empty rectangles (as input and
///  as output) are handled as if for non-empty rectangles,
///  and an empty result rect's coords not be set to zero.
///  They will then be the really *calculated* coords. </summary>
_RectF_& _RectF_::IntersectRectEmulRet( PCRECTF prc1, PCRECTF prc2 )
{
    _RectF_ rc1 = prc1 ? *prc1 : _RectF_(0.0f, 0.0f, 0.0f, 0.0f);
    _RectF_ rc2 = prc2 ? *prc2 : _RectF_(0.0f, 0.0f, 0.0f, 0.0f);

    left   = privateRectF::tmax( rc1.left,   rc2.left   );
    right  = privateRectF::tmin( rc1.right,  rc2.right  );
    top    = privateRectF::tmax( rc1.top,    rc2.top    );
    bottom = privateRectF::tmin( rc1.bottom, rc2.bottom );

    return *this;
}

///<summary> This is similar to <b>UnionRect()</b>, but
///  with the difference that empty rectangles (as input and
///  as output) are handled as if for non-empty rectangles,
///  and an empty result rect's coords not be set to zero.
///  They will then be the really *calculated* coords. </summary>
_RectF_& _RectF_::UnionRectEmulRet( PCRECTF prc1, PCRECTF prc2 )
{
    _RectF_ rc1 = prc1 ? *prc1 : _RectF_(0.0f, 0.0f, 0.0f, 0.0f);
    _RectF_ rc2 = prc2 ? *prc2 : _RectF_(0.0f, 0.0f, 0.0f, 0.0f);

    left   = privateRectF::tmin( rc1.left,   rc2.left   );
    right  = privateRectF::tmax( rc1.right,  rc2.right  );
    top    = privateRectF::tmin( rc1.top,    rc2.top    );
    bottom = privateRectF::tmax( rc1.bottom, rc2.bottom );

    return *this;
}



_RectF_ _RectF_::MulDiv( double dMultiplier, double dDivisor ) const
{
    return _RectF_( (float) ((left   * dMultiplier) / dDivisor),
                    (float) ((top    * dMultiplier) / dDivisor),
                    (float) ((right  * dMultiplier) / dDivisor),
                    (float) ((bottom * dMultiplier) / dDivisor) );
}




}
