/*
    File:               argonTime.cpp
    Description:        TimeSpan class for easy handling of
                        time values in granularity of nanoseconds.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         23.11.2003
    last updated:       26.09.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonTime_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonTime.h"

namespace ArgonLibrary
{




struct tm *_localtime_( struct tm *ptm, const time_t *ptime )
{
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)

    if( localtime_s( ptm, ptime ) )
        return NULL;

    return ptm;

    #else

    if( !ptm ) return NULL;
    if( !ptime )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    struct tm *ret = localtime( ptime );
    if( !ret )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    *ptm = *ret;
    return ptm;

    #endif
}

struct tm *_gmtime_( struct tm *ptm, const time_t *ptime )
{
    #if defined(__STDC_SECURE_LIB__) || defined(__STDC_LIB_EXT1__)

    if( gmtime_s( ptm, ptime ) )
        return NULL;

    return ptm;

    #else

    if( !ptm ) return NULL;
    if( !ptime )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    struct tm *ret = gmtime( ptime );
    if( !ret )
        { memset( ptm, -1, sizeof(struct tm) ); return NULL; }

    *ptm = *ret;
    return ptm;

    #endif
}




TimeSpan::TimeSpan( struct tm *ptm, INT64 _nanos )
{
    if( !ptm )
        { m_t = 0; m_nanos = 0; return; }

    m_t = mktime( ptm );
    // 'm_t' is -1 on error

    _AdjustMembers(m_t, _nanos);

    m_nanos = (int) _nanos;
}

///<remarks>
///  The values of the parameters are not restricted to their normal ranges.
///  They even may be negative and will be handled correctly.
///</remarks>
TimeSpan::TimeSpan( int year, int month, int day, int hour, int min, int sec, int dst )
{
    struct tm atm;
    atm.tm_sec      = sec;
    atm.tm_min      = min;
    atm.tm_hour     = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday     = day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon      = month - 1;
    // (year >= 1900) must be
    atm.tm_year     = year - 1900;
    atm.tm_isdst    = dst;

    m_t = mktime( &atm );
    // 'm_t' is -1 on error

    m_nanos = 0;
}

///<remarks>
///  The values of the parameters are not restricted to their normal ranges.
///  They even may be negative and will be handled correctly.
///</remarks>
TimeSpan::TimeSpan( int year, int month, int day, int hour, int min, int sec,
                    int millisecs, int microsecs, int nanosecs, int dst )
{
    struct tm atm;
    atm.tm_sec      = sec;
    atm.tm_min      = min;
    atm.tm_hour     = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday     = day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon      = month - 1;
    // (year >= 1900) must be
    atm.tm_year     = year - 1900;
    atm.tm_isdst    = dst;

    m_t = mktime( &atm );
    // 'm_t' is -1 on error

    INT64 _nanos = (INT64) millisecs * 1000000 + (INT64) microsecs * 1000 + (INT64) nanosecs;
    _AdjustMembers(m_t, _nanos);

    m_nanos = (int) _nanos;
}

///<remarks>
///  The values of the parameters are not restricted to their normal ranges.
///  They even may be negative and will be handled correctly.
///</remarks>
TimeSpan::TimeSpan( long julianday, int hour, int min, int sec, int dst )
{
    WORD year, month, day;
    getGregorianDate( julianday, year, month, day );

    struct tm atm;
    atm.tm_sec      = sec;
    atm.tm_min      = min;
    atm.tm_hour     = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday     = (int) day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon      = (int) month - 1;
    // (year >= 1900) must be
    atm.tm_year     = (int) year - 1900;
    atm.tm_isdst    = dst;

    m_t = mktime( &atm );
    // 'm_t' is -1 on error

    m_nanos = 0;
}

///<remarks>
///  The values of the parameters are not restricted to their normal ranges.
///  They even may be negative and will be handled correctly.
///</remarks>
TimeSpan::TimeSpan( long julianday, int hour, int min, int sec,
                    int millisecs, int microsecs, int nanosecs, int dst )
{
    WORD year, month, day;
    getGregorianDate( julianday, year, month, day );

    struct tm atm;
    atm.tm_sec      = sec;
    atm.tm_min      = min;
    atm.tm_hour     = hour;
    // (day >= 1 && day <= 31) must be
    atm.tm_mday     = (int) day;
    // (month >= 1 && month <= 12) must be
    atm.tm_mon      = (int) month - 1;
    // (year >= 1900) must be
    atm.tm_year     = (int) year - 1900;
    atm.tm_isdst    = dst;

    m_t = mktime( &atm );
    // 'm_t' is -1 on error

    INT64 _nanos = (INT64) millisecs * 1000000 + (INT64) microsecs * 1000 + (INT64) nanosecs;
    _AdjustMembers(m_t, _nanos);

    m_nanos = (int) _nanos;
}

TimeSpan::TimeSpan( WORD wDosDate, WORD wDosTime, int dst )
{
    struct tm atm;
    atm.tm_sec      = (wDosTime & ~0xFFE0) << 1;
    atm.tm_min      = (wDosTime & ~0xF800) >> 5;
    atm.tm_hour     = wDosTime >> 11;
    atm.tm_mday     = wDosDate & ~0xFFE0;
    atm.tm_mon      = ((wDosDate & ~0xFE00) >> 5) - 1;
    atm.tm_year     = (wDosDate >> 9) + 80;
    atm.tm_isdst    = dst;

    m_t = mktime( &atm );
    // 'm_t' is -1 on error

    m_nanos = 0;
}

TimeSpan::TimeSpan( const SYSTEMTIME& syst, int dst )
{
    if( syst.wYear < 1970 ) {
        m_t     = 0;
        m_nanos = 0;
    } else {
        TimeSpan _t( (int) syst.wYear, (int) syst.wMonth,  (int) syst.wDay,
                     (int) syst.wHour, (int) syst.wMinute, (int) syst.wSecond,
                     dst );
        m_t     = _t.m_t;
        m_t    +=   ((int) (UINT) syst.wMilliseconds) / 1000;
        m_nanos = ( ((int) (UINT) syst.wMilliseconds) % 1000 ) * 1000000;
    }
}

TimeSpan::TimeSpan( const FILETIME& filet, int dst )
{
    // first convert file time (UTC time) to local time
    FILETIME localt;
    if( !::FileTimeToLocalFileTime( &filet, &localt ) )
        { m_t = 0; m_nanos = 0; return; }

    // then convert that local time to system time
    SYSTEMTIME syst;
    if( !::FileTimeToSystemTime( &localt, &syst ) )
        { m_t = 0; m_nanos = 0; return; }

    // then convert the system time to a time_t (C-runtime local time)
    TimeSpan _t( syst, dst );
    m_t     = _t.m_t;
    m_nanos = 0;
}



///<summary> Returns current local system time </summary>
TimeSpan TimeSpan::getCurrentTime()
{
    // Note: '::GetLocalTime()' returns the same
    // datetime as 'time()', but also includes milliseconds.
    SYSTEMTIME st;
    ::GetLocalTime( &st );
    return TimeSpan( st );
    //return time( NULL );
}

///<summary> Returns current non-local system time </summary>
TimeSpan TimeSpan::getCurrentTimeUTC()
{
    SYSTEMTIME st;
    ::GetSystemTime( &st );
    return TimeSpan( st );
}

struct tm *TimeSpan::getGmtTm( struct tm *ptm ) const
{
    if( !ptm ) {
        static struct tm s_tm;
        return _gmtime_( &s_tm, &m_t );
    } else
        return _gmtime_( ptm, &m_t );
}

struct tm *TimeSpan::getLocalTm( struct tm *ptm ) const
{
    if( !ptm ) {
        static struct tm s_tm;
        return _localtime_( &s_tm, &m_t );
    } else
        return _localtime_( ptm, &m_t );
}

///<summary> Gets the *local* time in a SYSTEMTIME struct. </summary>
bool TimeSpan::getAsSystemTime( SYSTEMTIME& syst ) const
{
    struct tm _tm;
    struct tm *ptm = getLocalTm( &_tm );
    if( !ptm ) return false;

    syst.wYear         = (WORD) (1900 + ptm->tm_year);
    syst.wMonth        = (WORD) (1 + ptm->tm_mon);
    syst.wDayOfWeek    = (WORD) ptm->tm_wday;
    syst.wDay          = (WORD) ptm->tm_mday;
    syst.wHour         = (WORD) ptm->tm_hour;
    syst.wMinute       = (WORD) ptm->tm_min;
    syst.wSecond       = (WORD) ptm->tm_sec;
    syst.wMilliseconds = (WORD) (m_nanos / 1000000);

    return true;
}

///<summary> Gets the *non-local* time in a SYSTEMTIME struct. </summary>
bool TimeSpan::getAsSystemTimeUTC( SYSTEMTIME& syst ) const
{
    struct tm _tm;
    struct tm *ptm = getGmtTm( &_tm );
    if( !ptm ) return false;

    syst.wYear         = (WORD) (1900 + ptm->tm_year);
    syst.wMonth        = (WORD) (1 + ptm->tm_mon);
    syst.wDayOfWeek    = (WORD) ptm->tm_wday;
    syst.wDay          = (WORD) ptm->tm_mday;
    syst.wHour         = (WORD) ptm->tm_hour;
    syst.wMinute       = (WORD) ptm->tm_min;
    syst.wSecond       = (WORD) ptm->tm_sec;
    syst.wMilliseconds = (WORD) (m_nanos / 1000000);

    return true;
}

///<summary> Gets the *non-local* time in a FILETIME struct. </summary>
bool TimeSpan::getAsFileTime( FILETIME& filet ) const
{
    SYSTEMTIME syst;
    if( !getAsSystemTime( syst ) ) return false;
    // convert system time to local file time
    FILETIME localt;
    if( !::SystemTimeToFileTime( &syst, &localt ) ) return false;
    // convert local file time to UTC file time
    if( !::LocalFileTimeToFileTime( &localt, &filet ) ) return false;
    return true;
}



long TimeSpan::getJulianDay() const
{
    struct tm _tm;
    struct tm *ptm = getLocalTm( &_tm );
    if( !ptm ) return 0;
    return getJulianDay( (WORD) (ptm->tm_year + 1900),
                         (WORD) (ptm->tm_mon  +    1),
                         (WORD) (ptm->tm_mday       ) );
}

long TimeSpan::getJulianDay( WORD wYear, WORD wMonth, WORD wDay )
{
    long y = wYear;
    long m = wMonth;
    long d = wDay;
    if( m > 2L ) m = m - 3L;
    else {       m = m + 9L;  y--; }
    long c  = y / 100;
    long ya = y - 100L * c;
    long j  = (146097L * c) / 4 + (1461L * ya) / 4 + (153L * m + 2L) / 5 + d + 1721119L;
    return j;
}

long TimeSpan::getJulianDay( const SYSTEMTIME& syst )
{
    return getJulianDay( syst.wYear, syst.wMonth, syst.wDay );
}

void TimeSpan::setJulianDay( long nJD, SYSTEMTIME& syst )
{
    getGregorianDate( nJD, syst.wYear, syst.wMonth, syst.wDay );
}

void TimeSpan::getGregorianDate( long nJD, WORD& wYear, WORD& wMonth, WORD& wDay )
{
    long j = nJD - 1721119L;
    long y = (4L * j - 1L) / 146097;
    j = 4L * j - 1L - 146097L * y;
    long d = j / 4;
    j = (4L * d + 3L) / 1461;
    d = 4L * d + 3L - 1461L * j;
    d = (d + 4L) / 4;
    long m = (5L * d - 3L) / 153;
    d = 5L * d - 3L - 153L * m;
    d = (d + 5L) / 5;
    y = 100L * y + j;
    if( m < 10L ) m = m + 3L;
    else {        m = m - 9L;  y++; }
    wYear  = (WORD) y;
    wMonth = (WORD) m;
    wDay   = (WORD) d;
}



// Seconds can range from year 1900 to 2036 (in DWORD range).
UINT64 TimeSpan::getSecondsSinceJD( long nJDRef )
{
    UINT64 n = (UINT64) getJulianDay() - (UINT64) nJDRef;
    struct tm _tm;
    struct tm *ptm = getLocalTm( &_tm );
    if( !ptm ) return 0;
    n = ( n * 24 ) + (UINT64) ptm->tm_hour;
    n = ( n * 60 ) + (UINT64) ptm->tm_min;
    n = ( n * 60 ) + (UINT64) ptm->tm_sec;
    return n;
}

// Seconds can range from year 1900 to 2036 (in DWORD range).
TimeSpan TimeSpan::fromSecondsSinceJD( UINT64 secs, long nJDRef )
{
    UINT64 n = secs;
    int s = (int) (n % 60);     n /= 60;
    int m = (int) (n % 60);     n /= 60;
    int h = (int) (n % 24);     n /= 24;
    long jd = (long) n + nJDRef;
    //INT day_of_week = (jd + 1) % 7;   // 0 = sunday, 1 = monday, ...
    TimeSpan t( jd, h, m, s );
    return t;
}



void TimeSpan::_AdjustMembers( time_t& _t, int& _nanos )
{
    if( _nanos >= 0 ) {
        _t     += (time_t) (_nanos / 1000000000);
        _nanos  =          (_nanos % 1000000000);
    } else {
        _t     -= (time_t)     ((-_nanos - 1) / 1000000000 + 1);
        _nanos  = 1000000000 - ((-_nanos - 1) % 1000000000 + 1);
    }
}

void TimeSpan::_AdjustMembers( time_t& _t, INT64& _nanos )
{
    if( _nanos >= 0 ) {
        _t     += (time_t) (_nanos / 1000000000);
        _nanos  =          (_nanos % 1000000000);
    } else {
        _t     -= (time_t)     ((-_nanos - 1) / 1000000000 + 1);
        _nanos  = 1000000000 - ((-_nanos - 1) % 1000000000 + 1);
    }
}




}
