/*
    File:               argonTime.h
    Description:        TimeSpan class for easy handling of
                        time values in granularity of nanoseconds.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         23.11.2003
    last updated:       26.09.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonTime_h
#define __argonTime_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include <math.h>
    #include <time.h>
    // Note: If your compiler yields problems with the #tm# structure,
    // try including <time.h> in your "stdafx.h" file.
#endif

namespace ArgonLibrary
{




struct tm *_localtime_( struct tm *ptm, const time_t *ptime );
struct tm *_gmtime_	  ( struct tm *ptm, const time_t *ptime );




// Note: 'dst' means Daylight Saving Time
//        >0   is in effect
//        ==0  is not in effect
//        <0   unknown (The C run-time library assumes the
//             United Statess rules for implementing the
//             calculation of Daylight Saving Time).

///<summary>
///  Class <b>TimeSpan</b> supports handling of time values in the range
///  usual to <b>time_t</b> time values, but additionally supports units
///  down to nanoseconds.
///</summary>

class TimeSpan
{
    private :
        time_t m_t;     // seconds since 1.1.1970  (this is the local time)
        int    m_nanos; // nano-seconds (always in range [0 - 1000000000[ nanoseconds)

    public :
        TimeSpan() { m_t = 0; m_nanos = 0; }
        // Note: The following ctors are used to specify a point of time.
        // If you need to create a TimeSpan object acting as span of time,
        // use the #TimeSpan::fromSpan()# or related static methods instead.
        //TimeSpan( time_t _t ) { m_t = _t; m_nanos = 0; }
        TimeSpan( double _t ) { m_t = (time_t) floor(_t); m_nanos = (int) ((_t - floor(_t)) * 1000000000.0); }
        TimeSpan( time_t _t, INT64 _nanos ) { _AdjustMembers(_t, _nanos); m_t = _t; m_nanos = (int) _nanos; }
        TimeSpan( struct tm *ptm, INT64 _nanos );
        TimeSpan( int year, int month, int day, int hour, int min, int sec, int dst = -1 );
        TimeSpan( int year, int month, int day, int hour, int min, int sec,
                  int millisecs, int microsecs, int nanosecs, int dst = -1 );
        TimeSpan( long julianday, int hour, int min, int sec, int dst = -1 );
        TimeSpan( long julianday, int hour, int min, int sec,
                  int millisecs, int microsecs, int nanosecs, int dst = -1 );
        TimeSpan( WORD wDosDate, WORD wDosTime, int dst );  // set 'dst' to -1 for default
        TimeSpan( const TimeSpan& _t ) { m_t = _t.m_t; m_nanos = _t.m_nanos; }
        TimeSpan( const SYSTEMTIME& syst, int dst = -1 );
        TimeSpan( const FILETIME&  filet, int dst = -1 );

        static TimeSpan getCurrentTime    ();
        static TimeSpan getCurrentTimeUTC ();
        struct tm      *getGmtTm          ( struct tm *ptm = NULL ) const;
        struct tm      *getLocalTm        ( struct tm *ptm = NULL ) const;
        bool            getAsSystemTime   ( SYSTEMTIME& syst ) const;
        bool            getAsSystemTimeUTC( SYSTEMTIME& syst ) const;
        bool            getAsFileTime     ( FILETIME& filet ) const;

        operator time_t   () const { return m_t;     }
        time_t   get_Time () const { return m_t;     }
        int      get_Nanos() const { return m_nanos; }

        // Time-stamp values (if this TimeSpan object describes a point of time)
        int   getYear      () const { struct tm _tm; return getLocalTm(&_tm)->tm_year + 1900; }
            ///<summary> month of year (1 = Jan) </summary>
        int   getMonth     () const { struct tm _tm; return getLocalTm(&_tm)->tm_mon + 1; }
            ///<summary> day of month </summary>
        int   getDay       () const { struct tm _tm; return getLocalTm(&_tm)->tm_mday; }
        int   getHour      () const { struct tm _tm; return getLocalTm(&_tm)->tm_hour; }
        int   getMinute    () const { struct tm _tm; return getLocalTm(&_tm)->tm_min; }
        int   getSecond    () const { struct tm _tm; return getLocalTm(&_tm)->tm_sec; }
        int   getMilliSec  () const { return m_nanos / 1000000; }
        DWORD getFractional() const { return GetFractionalFromNanos( m_nanos ); } // this will return a DWORD where the whole possible range of nanoseconds (0 - 999999999) is stretched to DWORD range (0 - 0xffffffff)
            ///<summary> 0 = Sun, 1 = Mon, ..., 6 = Sat </summary>
        int   getDayOfWeek () const { struct tm _tm; return getLocalTm(&_tm)->tm_wday; }
            ///<summary> 1 = Mon, ..., 6 = Sat, 7 = Sun </summary>
        int   getDayOfWeek2() const { struct tm _tm; return (getLocalTm(&_tm)->tm_wday + 6) % 7 + 1; }

        // Time-range values (if this TimeSpan object describes a duration/length/span of time)
        INT64 getSpanDays        () const { return (INT64) (m_t / (24 * 3600L)); }
        INT64 getSpanHours       () const { return (INT64) (getSpanTotalHours()   - getSpanDays() * 24); }
        INT64 getSpanMinutes     () const { return (INT64) (getSpanTotalMinutes() - getSpanTotalHours() * 60); }
        INT64 getSpanSeconds     () const { return (INT64) (getSpanTotalSeconds() - getSpanTotalMinutes() * 60); }
        INT64 getSpanTotalHours  () const { return m_t / 3600; }
        INT64 getSpanTotalMinutes() const { return m_t / 60; }
        INT64 getSpanTotalSeconds() const { return m_t; }
        static TimeSpan fromSpanDays   ( int days  ) { return TimeSpan( (time_t) (days * 24 * 60 * 60), 0 ); }
        static TimeSpan fromSpanHours  ( int hours ) { return TimeSpan( (time_t) (hours * 60 * 60), 0 ); }
        static TimeSpan fromSpanMinutes( int mins  ) { return TimeSpan( (time_t) (mins * 60), 0 ); }
        static TimeSpan fromSpanSeconds( int secs  ) { return TimeSpan( (time_t) secs, 0 ); }
        static TimeSpan fromSpan( int weeks, int days, int hours = 0, int mins = 0, int secs = 0, int millisecs = 0, int microsecs = 0, int nanosecs = 0 )
            { return TimeSpan( (time_t) ((weeks * 7 + days) * 24 * 60 * 60 + hours * 60 * 60 + mins * 60 + secs), (INT64) millisecs * 1000000 + (INT64) microsecs * 1000 + (INT64) nanosecs ); }

        double getAsDouble() const { return (double) m_t + (double) m_nanos * 0.000000001; }

        TimeSpan&  operator = ( const TimeSpan& _t ) { m_t = _t.m_t; m_nanos = _t.m_nanos; return *this; }
        TimeSpan&  operator = ( time_t   _t ) { m_t = _t; m_nanos = 0; return *this; }
        TimeSpan   operator - ( TimeSpan _t ) const { return TimeSpan( m_t - _t.m_t, m_nanos - _t.m_nanos ); }
        TimeSpan   operator + ( TimeSpan _t ) const { return TimeSpan( m_t + _t.m_t, m_nanos + _t.m_nanos ); }
        TimeSpan&  operator +=( TimeSpan _t ) { m_t += _t.m_t; m_nanos += _t.m_nanos; _AdjustMembers(); return *this; }
        TimeSpan&  operator -=( TimeSpan _t ) { m_t -= _t.m_t; m_nanos -= _t.m_nanos; _AdjustMembers(); return *this; }
        bool       operator ==( TimeSpan _t ) const { return m_t == _t.m_t && m_nanos == _t.m_nanos; }
        bool       operator !=( TimeSpan _t ) const { return m_t != _t.m_t || m_nanos != _t.m_nanos; }
        bool       operator < ( TimeSpan _t ) const { return m_t <  _t.m_t || ( m_t == _t.m_t && m_nanos <  _t.m_nanos ); }
        bool       operator > ( TimeSpan _t ) const { return m_t >  _t.m_t || ( m_t == _t.m_t && m_nanos >  _t.m_nanos ); }
        bool       operator <=( TimeSpan _t ) const { return m_t <  _t.m_t || ( m_t == _t.m_t && m_nanos <= _t.m_nanos ); }
        bool       operator >=( TimeSpan _t ) const { return m_t >  _t.m_t || ( m_t == _t.m_t && m_nanos >= _t.m_nanos ); }

        TimeSpan& TruncateToPositive() { if( m_t < 0 ) { m_t = 0; m_nanos = 0; }  return *this; } // if negative -> set to zero

        #define Diff_1stJan1970_1stJan1900  2208988800  // Diff between time() [1.1.1970 UCT] and time server [1.1.1900 UCT]
                                                        // This is ( JulianDay_1stJan1970 - JulianDay_1stJan1900 ) * 24 * 60 * 60

        #define JulianDay_1stJan1900        2415021
        #define JulianDay_1stJan1970        2440588
        #define JulianDay_1stJan1980        2444240
        long getJulianDay() const;
        static long getJulianDay( WORD wYear, WORD wMonth, WORD wDay );
        static long getJulianDay( const SYSTEMTIME& syst );
        static void setJulianDay( long nJD, SYSTEMTIME& syst );
        static void getGregorianDate( long nJD, WORD& wYear, WORD& wMonth, WORD& wDay );

        UINT64 getSecondsSinceJD  ( long nJDRef );
        UINT64 getSecondsSince1900() { return getSecondsSinceJD( JulianDay_1stJan1900 ); }
        static TimeSpan fromSecondsSinceJD  ( UINT64 secs, long nJDRef );
        static TimeSpan fromSecondsSince1900( UINT64 secs ) { return fromSecondsSinceJD( secs, JulianDay_1stJan1900 ); }

        // Note: The following fractional handling is useful for NTP or SNTP time formats.
        // Fractional means, the nanoseconds (range [0 - 999999999]) to be stretched to
        // an unsiged DWORD range [0 - 0xFFFFFFFF].
        static DWORD GetFractionalFromNanos( int  nanos ) { return (DWORD) ceil ( ((double) nanos * ((double) 0x100000000u)) * 0.000000001  ); }
        static int   GetNanosFromFractional( DWORD frac ) { return (int)   floor( ((double) frac  / ((double) 0x100000000u)) * 1000000000.0 ); }
        static TimeSpan fromTimeFractional( time_t _t, DWORD fractional ) { return TimeSpan( _t, GetNanosFromFractional(fractional) ); }

    protected :
        void        _AdjustMembers() { _AdjustMembers( m_t, m_nanos ); }
        static void _AdjustMembers( time_t& _t, int&   _nanos );
        static void _AdjustMembers( time_t& _t, INT64& _nanos );
};




}

#endif
