/*
    File:               argonUser32Dll.cpp
    Description:        Extensions and emulations of parts of the
                        system's USER32.DLL
                        useful mainly to support system features that
                        are not or may be not present on all available
                        operating system versions out there.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         9.11.2003
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonUser32Dll_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonUser32Dll.h"

#ifndef _argon_no_includes
    #include "argonLocking.h"
#endif

namespace ArgonLibrary
{




BOOL WINAPI argonUpdateLayeredWindow(
    HWND hwnd,             // handle to layered window
    HDC hdcDst,            // handle to screen DC
    POINT *pptDst,         // new screen position
    SIZE *psize,           // new size of the layered window
    HDC hdcSrc,            // handle to surface DC
    POINT *pptSrc,         // layer position
    COLORREF crKey,        // color key
    BLENDFUNCTION *pblend, // blend function
    DWORD dwFlags          // options
)
{
    User32Dll::LoadAPI();
    if(   !User32Dll::s_pfnUpdateLayeredWindow ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return User32Dll::s_pfnUpdateLayeredWindow( hwnd, hdcDst, pptDst, psize,
                                                hdcSrc, pptSrc, crKey, pblend, dwFlags);
}

BOOL WINAPI argonSetLayeredWindowAttributes(
    HWND hwnd,             // handle to the layered window
    COLORREF crKey,        // specifies the color key
    BYTE byAlpha,          // value for the blend function
    DWORD dwFlags          // action
)
{
    User32Dll::LoadAPI();
    if(   !User32Dll::s_pfnSetLayeredWindowAttributes ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return User32Dll::s_pfnSetLayeredWindowAttributes( hwnd, crKey, byAlpha, dwFlags );
}

BOOL WINAPI argonGetLayeredWindowAttributes(
    HWND hwnd,
    COLORREF *pcrKey,
    BYTE *pbyAlpha,
    DWORD *pdwFlags
)
{
    User32Dll::LoadAPI();
    if(   !User32Dll::s_pfnGetLayeredWindowAttributes ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return FALSE; }
    return User32Dll::s_pfnGetLayeredWindowAttributes( hwnd, pcrKey, pbyAlpha, pdwFlags );
}

DWORD WINAPI argonGetGuiResources(
    IN HANDLE hProcess,
    IN DWORD uiFlags
)
{
    User32Dll::LoadAPI();
    if(   !User32Dll::s_pfnGetGuiResources ) { SetLastError(ERROR_CALL_NOT_IMPLEMENTED); return 0; }
    return User32Dll::s_pfnGetGuiResources( hProcess, uiFlags );
}



BOOL SetLayeredWindowExstyle( HWND hwnd, BOOL bSet, DWORD dwSwitchAlso )
{
    SetLastError( ERROR_CALL_NOT_IMPLEMENTED );

    if( !hwnd ) return FALSE;
    if( !IsWindow( hwnd ) ) return FALSE;

    if( !User32Dll::IsLayeredWindowsSupported() ) return FALSE;

    SetWindowLongPtr( hwnd, GWL_EXSTYLE,
                      bSet ? (GetWindowLongPtr(hwnd, GWL_EXSTYLE) |  (WS_EX_LAYERED | dwSwitchAlso))
                           : (GetWindowLongPtr(hwnd, GWL_EXSTYLE) & ~(WS_EX_LAYERED | dwSwitchAlso)) );

    SetLastError( ERROR_SUCCESS );

    return TRUE;
}




User32Dll::Unloader           User32Dll::s_Unloader;
HMODULE                       User32Dll::s_hModule                       = NULL;
TfnUpdateLayeredWindow        User32Dll::s_pfnUpdateLayeredWindow        = NULL;
TfnSetLayeredWindowAttributes User32Dll::s_pfnSetLayeredWindowAttributes = NULL;
TfnGetLayeredWindowAttributes User32Dll::s_pfnGetLayeredWindowAttributes = NULL;
TfnGetGuiResources            User32Dll::s_pfnGetGuiResources            = NULL;
bool                          User32Dll::s_bAutoUnload                   = true;

User32Dll::User32Dll()
{
}

// Remarks: We don't need any object of #User32Dll#.
//  We just need its static members.



bool User32Dll::IsLayeredWindowsSupported()
{
    LoadAPI();
    return s_pfnUpdateLayeredWindow != NULL;
    // IMPORTANT: The "USER32.DLL" is always loadable because it contains windows functions
    //  for already very early Windows versions. So we must check separately if the
    //  layered-windows-functions are included in the dll or not.
}



void User32Dll::LoadAPI()
{
    if( s_hModule ) return;

    LockModuleWide lmw;

    if( s_hModule ) return;

    Nullify();

    s_hModule = ::LoadLibrary( _T(USER32DLL_NAME) );

    if( s_hModule )
    {
        s_pfnUpdateLayeredWindow        = (TfnUpdateLayeredWindow       ) GetProcAddress( s_hModule, "UpdateLayeredWindow"        );
        s_pfnSetLayeredWindowAttributes = (TfnSetLayeredWindowAttributes) GetProcAddress( s_hModule, "SetLayeredWindowAttributes" );
        s_pfnGetLayeredWindowAttributes = (TfnGetLayeredWindowAttributes) GetProcAddress( s_hModule, "GetLayeredWindowAttributes" );
        s_pfnGetGuiResources            = (TfnGetGuiResources           ) GetProcAddress( s_hModule, "GetGuiResources"            );
    }
}

void User32Dll::UnloadAPI()
{
    LockModuleWide lmw;

    if( s_hModule )
        ::FreeLibrary( s_hModule );

    Nullify();
}



void User32Dll::Nullify()
{
    s_pfnUpdateLayeredWindow        = NULL;
    s_pfnSetLayeredWindowAttributes = NULL;
    s_pfnGetLayeredWindowAttributes = NULL;
    s_pfnGetGuiResources            = NULL;
    s_hModule = NULL;
}




}
