/*
    File:               argonVersion.cpp
    Description:        Application/Project Version Information.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         5.10.2003
    last updated:       27.09.2006 (Roman Komary)

    Copyright (c) 2003-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonVersion_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonVersion.h"

#ifndef _argon_no_liblinks
    #pragma comment( lib, "version" )
#endif

namespace ArgonLibrary
{




Version::Version()
{
    Ctor();
    Load( Get_Module_Instance() );
}

Version::Version( HMODULE hModule )
{
    Ctor();
    Load( hModule );
}



///<remarks> This method will *always* return a valid pointer. </remarks>
PCSTR Version::GetA( PCSTR pstrVersionString ) const
{
    if( !pstrVersionString || !m_pFileVI || !m_bLoaded )
        return "";

    char str[128];
    int len = _sprintf_( str, "\\StringFileInfo\\%04x%04x\\",
                         m_SelectedLanguage.m_wLanguage, m_SelectedLanguage.m_wCodePage );
    int verlen = (int) _strlen_( pstrVersionString );
    if( len + verlen >= (int) _countof( str ) )
        return "";
    _strcat_( str, pstrVersionString );

    PSTR pstrResult = NULL;
    UINT lenResult = 0;

    if( !VerQueryValueA( m_pFileVI, str, (PVOID*) &pstrResult, &lenResult ) )
        return "";
    if( lenResult == 0 || !pstrResult )
        return "";

    return pstrResult;
}

///<remarks> This method will *always* return a valid pointer. </remarks>
PCWSTR Version::GetW( PCWSTR pstrVersionString ) const
{
    if( !pstrVersionString || !m_pFileVI || !m_bLoaded )
        return L"";

    WCHAR str[128];
    int len = _sprintf_( str, L"\\StringFileInfo\\%04x%04x\\",
                         m_SelectedLanguage.m_wLanguage, m_SelectedLanguage.m_wCodePage );
    int verlen = (int) _strlen_( pstrVersionString );
    if( len + verlen >= (int) _countof( str ) )
        return L"";
    _strcat_( str, pstrVersionString );

    PWSTR pstrResult = NULL;
    UINT lenResult = 0;

    if( !VerQueryValueW( m_pFileVI, str, (PVOID*) &pstrResult, &lenResult ) )
        return L"";
    if( lenResult == 0 || !pstrResult )
        return L"";

    return pstrResult;
}

///<summary>
///  Similar to the <b>Get()</b> methods, except that if the
///  retrieved value is a version information string (in the form e.g.
///  1, 0, 0, 1 or 1.0.0.1 with any number values), the last
///  value, which is interpreted as the build number, will be returned
///  as integer.
///</summary>
int Version::GetBuildNr( PCTSTR pstrVersionString ) const
{
    PCTSTR p = Get( pstrVersionString );
    int len = (int) _strlen_( p );

    int i = len;

    // Skip non-number chars after the last number.
    while( i > 0 ? !_isdigit_( p[i - 1] ) != 0 : false )
        --i;

    // Find start of last number.
    while( i > 0 ? _isdigit_( p[i - 1] ) != 0 : false )
        --i;

    return _stoi_( &p[i] );
}



///<summary>
///  <b>Load()</b> loads version information from the specified
///  module and selects the best language according to the current thread's locale.
///</summary>
///<param name="hModule"> Module for where to load version information from,
///  or NULL if it should be loaded from the current executable. </param>
///<returns> True on success, false on error. </returns>
bool Version::Load( HMODULE hModule )
{
    if( m_bLoaded && hModule == m_hModule )
        return true;

    Unload();

    m_hModule = hModule;

    // Get module filename from where to load version info
    WCHAR strApp[MAX_PATH + 1];
    if( !GetModuleFileNameW( m_hModule, strApp, _countof( strApp ) ) )
        { Unload(); return false; }

    // Get size of version info
    DWORD dwHandle = 0;
    DWORD versize = GetFileVersionInfoSizeW( strApp, &dwHandle );
    if( versize == 0 )
        { Unload(); return false; }

    // alloc buffer
    m_pFileVI = (PBYTE) malloc( versize );
    if( !m_pFileVI )
        { Unload(); return false; }
    m_cFileVI = (int) versize;

    // now get version info
    if( !GetFileVersionInfoW( strApp, dwHandle, versize, m_pFileVI ) )
        { Unload(); return false; }

    m_bLoaded = true;

    if( !SelectLanguageThreadLocale() )
        { Unload(); return false; }

    return true;
}

void Version::Unload()
{
    m_hModule = NULL;
    if( m_pFileVI )
        { free( m_pFileVI ); m_pFileVI = NULL; }
    m_cFileVI = 0;
    m_bLoaded = false;
    m_SelectedLanguage = LangAndCodepage();
}



///<summary>
///  <b>SelectLanguage()</b> choose the best match of the
///  available languages and codepages in the version information
///  for your desired language.
///</summary>
///<param name="lang"> desired language and codepage. Use 0xffff
///  for the codepage if you don't care about it. </param>
///<returns> True on success, false on error. </returns>
bool Version::SelectLanguage( LangAndCodepage lang )
{
    if( !m_pFileVI || !m_bLoaded )
        return false;

    // Get the contained languages of the version info
    LangAndCodepage *pTranslate = NULL;
    UINT translen = 0;
    if( !VerQueryValueW( m_pFileVI, L"\\VarFileInfo\\Translation", (PVOID*) &pTranslate, &translen ) )
        return false;
    translen /= (UINT) sizeof(LangAndCodepage);
    if( !pTranslate || translen == 0 )
        return false;

    LANGID want_lang_id      = (LANGID) lang.m_wLanguage; // the combined form with PRIMARYLANDID and SUBLANGID
    LANGID want_prim_lang_id = PRIMARYLANGID( want_lang_id );
    //LANGID want_sub_lang_id  = SUBLANGID( want_lang_id );

    int iLang;

    // First we try to find a match for PRIMARYLANGID and SUBLANGID and Codepage.

    if( lang.m_wCodePage != 0xffff )
        for( iLang = 0 ; iLang < (int) translen ; ++iLang )
        {
            LANGID lang_id = (LANGID) pTranslate[iLang].m_wLanguage;

            if( lang_id == want_lang_id && lang.m_wCodePage == pTranslate[iLang].m_wCodePage )
                { m_SelectedLanguage = pTranslate[iLang]; return true; }
        }

    // Now we try to find a match for PRIMARYLANGID and SUBLANGID.

    for( iLang = 0 ; iLang < (int) translen ; ++iLang )
    {
        LANGID lang_id = (LANGID) pTranslate[iLang].m_wLanguage;

        if( lang_id == want_lang_id )
            { m_SelectedLanguage = pTranslate[iLang]; return true; }
    }

    // Now we try to find a match for PRIMARYLANGID only.

    for( iLang = 0 ; iLang < (int) translen ; ++iLang )
    {
        LANGID lang_id      = (LANGID) pTranslate[iLang].m_wLanguage;
        LANGID prim_lang_id = PRIMARYLANGID( lang_id );

        if( prim_lang_id == want_prim_lang_id )
            { m_SelectedLanguage = pTranslate[iLang]; return true; }
    }

    // Now we try to find a match for LANG_ENGLISH as
    //  primary language id.

    for( iLang = 0 ; iLang < (int) translen ; ++iLang )
    {
        LANGID lang_id      = (LANGID) pTranslate[iLang].m_wLanguage;
        LANGID prim_lang_id = PRIMARYLANGID( lang_id );

        if( prim_lang_id == LANG_ENGLISH )
            { m_SelectedLanguage = pTranslate[iLang]; return true; }
    }

    // Now we try to find a match for LANG_NEUTRAL as
    //  primary language id.

    for( iLang = 0 ; iLang < (int) translen ; ++iLang )
    {
        LANGID lang_id      = (LANGID) pTranslate[iLang].m_wLanguage;
        LANGID prim_lang_id = PRIMARYLANGID( lang_id );

        if( prim_lang_id == LANG_NEUTRAL )
            { m_SelectedLanguage = pTranslate[iLang]; return true; }
    }

    // Now we get the first language we can find.

    m_SelectedLanguage = *pTranslate;

    return true;
}




}
