/*
    File:               argonWindowDeferrer.cpp
    Description:        This class allows easy managing and handling
                        of deferred window positioning, inclusive
                        for different parent windows.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         20.11.2005
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#define __argonWindowDeferrer_cpp
#define __an_ArgonLibrary_cpp       // helpful flag, if you need to make conditional compilations e.g. in "stdafx.h"

#include "stdafx.h"
#include "argonWindowDeferrer.h"

namespace ArgonLibrary
{




bool WindowSiblingDeferrer::Begin( int numInitialWindows )
{
    if( m_hDwp ) return false;

    m_hDwp = ::BeginDeferWindowPos( numInitialWindows );
    m_hwndParent = NULL;

    return m_hDwp != NULL;
}

bool WindowSiblingDeferrer::Deferwindowpos( HWND hwnd, HWND hwndInsertAfter, int x, int y, int cx, int cy, UINT swpFlags )
{
    if( !hwnd ) return false;

    if( m_hDwp )
    {
        m_hDwp = ::DeferWindowPos( m_hDwp, hwnd, hwndInsertAfter, x, y, cx, cy, swpFlags );

        if( m_hDwp ) { if( !m_hwndParent ) m_hwndParent = GetParent( hwnd ); }
        else m_hwndParent = NULL;
    }

    if( !m_hDwp )
        return ::SetWindowPos( hwnd, hwndInsertAfter, x, y, cx, cy, swpFlags ) != FALSE;

    return m_hDwp != NULL;
}

bool WindowSiblingDeferrer::DeferWindowRect( HWND hwnd, RECT rect )
{
    return Deferwindowpos( hwnd, HWND_TOP, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top,
                           SWP_NOZORDER | SWP_NOACTIVATE );
}

bool WindowSiblingDeferrer::DeferWindowTopLeft( HWND hwnd, int x, int y )
{
    return Deferwindowpos( hwnd, HWND_TOP, x, y, 0, 0,
                           SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}

bool WindowSiblingDeferrer::DeferWindowSize( HWND hwnd, int cx, int cy )
{
    return Deferwindowpos( hwnd, HWND_TOP, 0, 0, cx, cy,
                           SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );
}

bool WindowSiblingDeferrer::End()
{
    if( !m_hDwp ) return false;

    bool ok = ::EndDeferWindowPos( m_hDwp ) != FALSE;

    if( ok )
        { m_hDwp = NULL; m_hwndParent = NULL; }

    return ok;
}




bool WindowDeferrer::Begin( int numInitialWindowsPerSiblingDeferrer )
{
    if( m_bDeferring ) return false;

    m_numInitialWindowsPerSiblingDeferrer = numInitialWindowsPerSiblingDeferrer;
    m_bDeferring = true;

    onBegin();

    return true;
}

bool WindowDeferrer::Deferwindowpos( HWND hwnd, HWND hwndInsertAfter, int x, int y, int cx, int cy, UINT swpFlags )
{
    if( !m_bDeferring || !hwnd ) return false;

    HWND hwndParent = GetParent( hwnd );
    WindowSiblingDeferrer *p = FindSiblingDeferrer( hwndParent );

    if( p )
        return p->Deferwindowpos( hwnd, hwndInsertAfter, x, y, cx, cy, swpFlags );

    p = new WindowSiblingDeferrer;
    if( !p ) return false;

    m_llDwps.Add( p );

    if( !p->Begin( m_numInitialWindowsPerSiblingDeferrer ) )
        { delete p; return false; }

    if( !p->Deferwindowpos( hwnd, hwndInsertAfter, x, y, cx, cy, swpFlags ) )
        { delete p; return false; }

    return true;
}

bool WindowDeferrer::DeferWindowRect( HWND hwnd, RECT rect )
{
    return Deferwindowpos( hwnd, HWND_TOP, rect.left, rect.top, rect.right - rect.left, rect.bottom - rect.top,
                           SWP_NOZORDER | SWP_NOACTIVATE );
}

bool WindowDeferrer::DeferWindowTopLeft( HWND hwnd, int x, int y )
{
    return Deferwindowpos( hwnd, HWND_TOP, x, y, 0, 0,
                           SWP_NOSIZE | SWP_NOZORDER | SWP_NOACTIVATE );
}

bool WindowDeferrer::DeferWindowSize( HWND hwnd, int cx, int cy )
{
    return Deferwindowpos( hwnd, HWND_TOP, 0, 0, cx, cy,
                           SWP_NOMOVE | SWP_NOZORDER | SWP_NOACTIVATE );
}

bool WindowDeferrer::End()
{
    if( !m_bDeferring ) return false;
    if( m_bInEndDefer ) return true;

    m_bInEndDefer = true;
    onEnding();

    bool ok = true;

    while( m_llDwps.m_p1 )
    {
        if( !m_llDwps.m_p1->End() ) ok = false;
        delete m_llDwps.m_p1;
    }

    onEnded();
    m_bDeferring = false;
    m_bInEndDefer = false;

    return ok;
}



WindowSiblingDeferrer *WindowDeferrer::FindSiblingDeferrer( HWND hwndParent )
{
    for( WindowSiblingDeferrer *p = m_llDwps.m_p1 ; p ; p = m_llDwps.next(p) )
        if( p->m_hwndParent == hwndParent )
            return p;

    return NULL;
}




}
