/*
    File:               argonWindowDeferrer.h
    Description:        This class allows easy managing and handling
                        of deferred window positioning, inclusive
                        for different parent windows.
                        This file is part of the ArgonLibrary package.

    Author:             Roman Komary
    created on:         20.11.2005
    last updated:       07.02.2006 (Roman Komary)

    Copyright (c) 2005-2006, Roman Komary.

    Distributed under the Boost Software License, Version 1.0. (See
    accompanying file LICENSE_1_0.ArgonLibrary.txt or copy at
    http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef __argonWindowDeferrer_h
#define __argonWindowDeferrer_h

#ifndef __ArgonLibrary__
#define __ArgonLibrary__            // helpful flag, if you need to make conditional compilations somewhere in your sourcecode
#endif

#ifndef _argon_no_includes
    #include <stddef.h>
    #include "argonLinkedList.h"
#endif

namespace ArgonLibrary
{




class WindowDeferrer;

///<summary>
///  Class <b>WindowSiblingDeferrer</b> can defer windows
///  which all have the same parent window.
///  You must ensure yourself that all windows have the same parent.
///</summary>
class WindowSiblingDeferrer
{
    friend class WindowDeferrer;

    protected :
        HDWP    m_hDwp;
        HWND    m_hwndParent;

    public :
        WindowSiblingDeferrer() { m_hDwp = NULL; m_hwndParent = NULL; }
        ~WindowSiblingDeferrer() { End(); }
    protected :
        WindowSiblingDeferrer( const WindowSiblingDeferrer& src );

    public :
        bool Begin( int numInitialWindows = 16 );
        bool Deferwindowpos    ( HWND hwnd, HWND hwndInsertAfter, int x, int y, int cx, int cy, UINT swpFlags );
        bool DeferWindowRect   ( HWND hwnd, RECT rect );
        bool DeferWindowTopLeft( HWND hwnd, POINT pt ) { return DeferWindowTopLeft( hwnd, pt.x, pt.y ); }
        bool DeferWindowTopLeft( HWND hwnd, int x, int y );
        bool DeferWindowSize   ( HWND hwnd, SIZE siz ) { return DeferWindowSize( hwnd, siz.cx, siz.cy ); }
        bool DeferWindowSize   ( HWND hwnd, int cx, int cy );
        bool End();

        HDWP getHDWP() const { return m_hDwp; }
        HWND getParent() const { return m_hwndParent; }

        bool deferring() const { return m_hDwp != NULL; }

    protected :
        Node<WindowSiblingDeferrer> m_node;
};

///<summary>
///  Class <b>WindowDeferrer</b> can be used to defer windows
///  of even different parent windows. For each group of windows
///  who have the same parent, a <b>WindowSiblingDeferrer</b>
///  object will be used.
///</summary>
class WindowDeferrer
{
    protected :
        LinkedList<WindowSiblingDeferrer> m_llDwps;

        bool    m_bDeferring;
        int     m_numInitialWindowsPerSiblingDeferrer;
        bool    m_bInEndDefer;

    public :
        WindowDeferrer() : m_llDwps( offsetof(WindowSiblingDeferrer, m_node) )
            { m_bDeferring = false; m_numInitialWindowsPerSiblingDeferrer = 0; m_bInEndDefer = false; }
        virtual ~WindowDeferrer() { End(); }
    protected :
        WindowDeferrer( const WindowDeferrer& src );

    public :
        bool Begin( int numInitialWindowsPerSiblingDeferrer = 16 );
        bool Deferwindowpos    ( HWND hwnd, HWND hwndInsertAfter, int x, int y, int cx, int cy, UINT swpFlags );
        bool DeferWindowRect   ( HWND hwnd, RECT rect );
        bool DeferWindowTopLeft( HWND hwnd, POINT pt ) { return DeferWindowTopLeft( hwnd, pt.x, pt.y ); }
        bool DeferWindowTopLeft( HWND hwnd, int x, int y );
        bool DeferWindowSize   ( HWND hwnd, SIZE siz ) { return DeferWindowSize( hwnd, siz.cx, siz.cy ); }
        bool DeferWindowSize   ( HWND hwnd, int cx, int cy );
        bool End();

        bool deferring() const { return m_bDeferring; }

    protected :
        WindowSiblingDeferrer *FindSiblingDeferrer( HWND hwndParent );

    protected :
        virtual void onBegin () {  }
        virtual void onEnding() {  }
        virtual void onEnded () {  }
};




}

#endif
